/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <getopt.h>

#include "ClicApplication.hpp"
#include "nmm/utils/gdparse/IClicValueStream.hpp"
#include "nmm/utils/gdparse/GDParse.hpp"

#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/serialize/StdOValueStream.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/graph/GraphBuilder2.hpp"
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/ISeekable.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/base/resourcemgr/IResourceCollector.hpp"
#include "nmm/utils/utils.hpp"

#include "clic.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <cstdlib>
#include <cctype>
#include <sstream>
#include <bitset>
#include <set>
#include <iostream>


extern unsigned int lineno;
extern char *yytext;
extern list<NodeConfiguration> node_configuration;

namespace NMM {
  ClicApplication::ClicApplication(int argc, char** argv)
    :  NamedObject("Clic-Application"), 
       __app(0), 
       __graph_url(), 
       __composite(0),
       __seekable(0),
       __playback(0),
       __playback_location(),
       __display(0),
       __display_location(),
       __av_sync(0),
       __synchronizer(0),
       __curr_input_file(0), 
       __sync(true), 
       __prefetching(false), 
       __resource(false),
       __eof_counter(0),
       __sink_nodes(0), 
       __fullscreen(false), 
       __delay(0),
       __exit_flag(false),
       __infinite_loop(true),
       __condition(__cond_mutex), 
       __timeout(0),
       __timer(*this), 
       __enable_registry(false), 
       __port(ClientRegistry::default_port),
       __paused(false),
       __quit(false),
       __x_listener(0)
  {
    /* Set message stream to cerr */
    setMessageStream(&cerr, NamedObject::ALL_LEVELS);
    setWarningStream(&cerr, NamedObject::ALL_LEVELS);
    setMessageColor(NamedObject::BLUE_COLOR, NamedObject::ALL_LEVELS);


    /* First parse the arguments to get all filenames and options */
    parseArguments(argc, argv);

    /* Begin> Create an NMMAplication */
    __app = ProxyApplication::getApplication(argc, argv);

    if (__enable_registry) {
      MESSAGE_STREAM("Starting server registry on port " << __port << endl);
      __app->enableServerRegistry(__port);
    }

    if( __graph_url.empty() ) {
      /* request the graph */
      __app->getRegistry().requestGraph(__gd);
      setupGraph();
    }
    else {
      buildGraph();
    }

    if (__timeout > 0) {
      __timer.start(__timeout);
    }
  }

  ClicApplication::~ClicApplication()
  {
    /* Unregister stdin */
    SNetReactor::getInstance().unregisterSocket(0);
    
    /* release the node inside the graph */
    try {
      if( __graph_url.empty() ) {
	/* If synchronization was enabled, disconnect the syncronizer */
	if (__sync) {
	  __gd.disconnectSinkSynchronizer();
	}
	
	/* print out resource statistics */
	if(__resource) {
	  cerr << "saving resource-usage data..." << endl;
	  
	  SequenceValue<INode>& inode_list = __gd.getINodeList();
	  for (size_t i = 0; i< inode_list.size(); ++i) {
	    __resource_mgr->updateNodeInfo(inode_list[i]);
	  }
	  printResourceStatistics();
	}
	
	/* release the graph and all nodes */
	__app->getRegistry().releaseGraph(__gd);
      }
      else {
	destroyGraph();
      }
    } catch (const Exception& e ) {
      cerr << e << endl;
    }
    delete __x_listener;
  }
  
  Result ClicApplication::keyReleased(string& key) {
    if(key == "KEY_f") { // 'f' to toggle fullscreen
      toggleFullscreen();
      return SUCCESS;
    } else if(key == "KEY_Right") {
      seekForward();
      return SUCCESS;
    } else if(key == "KEY_Left") {
      seekBackward();
      return SUCCESS;
    } else if(key == "KEY_q") {
      quit();
      return SUCCESS;
    } else if(key == "KEY_space") {
      togglePause();
      return SUCCESS;
    } else {
      cerr << "ClicApplication: Unsupported key pressed: " << key << endl;
      return FAILURE;
    }
  }

  void ClicApplication::seekForward() {
    if(__seekable) {
      if(__seekable->hasTimeSeek()) {
        __seekable->seekTime(Interval(10, 0));
      } else {
        __seekable->seekPercent(Rational(3,100));
      }
    } else {
      cerr << "Source not seekable" << endl;
    }
  }
  
  void ClicApplication::seekBackward() {
    if(__seekable) {
      if(__seekable->hasTimeSeek()) {
        __seekable->seekTime(Interval(-10, 0));
      } else {
        __seekable->seekPercent(Rational(-3,100));
      }
    } else {
      cerr << "Source not seekable" << endl;
    }
  }
  
  void ClicApplication::pause() {
    if(__synchronizer) {
      __synchronizer->pause();
    } else {
      cerr << "No synchronizer to pause" << endl;
    }
  }
  
  void ClicApplication::play() {
    if(__synchronizer) {
      __synchronizer->wakeup();
    } else {
      cerr << "No synchronizer to wakeup" << endl;
    }
  }
  
  void ClicApplication::togglePause() {
    __paused = !__paused;
    if(__paused) {
      pause();
    } else {
      play();
    }
  }
  
  void ClicApplication::toggleFullscreen() {
    if(!__video_display.get()) {
      cerr << "ClicApplication: No VideoDisplay set!" << endl;
    }
    
    if(__video_display->hasFullscreen()) {
      __fullscreen = !__video_display->getFullscreen();
      __video_display->setFullscreen(__fullscreen);
    } else {
      cerr << "ClicApplication: Fullscreen not supported with current display!" << endl;
    }
  }
  
  void ClicApplication::quit() {
    MutexGuard mg(__cond_mutex);
    __quit = true;
    __condition.notify();
  }
  
  void ClicApplication::parseArguments(int argc, char** argv) {
    int verbose = 0;

    const char *opt_string = "lvrsPhi:V:A:u:o:d:t:p:S";
    struct option options [] = {
      {"input-file",  1,NULL,'i'},
      {"url",         1,NULL,'u'},
      {"audio-output",1,NULL,'A'},
      {"video-output",1,NULL,'V'},
      {"output-file", 1,NULL,'o'},
      {"verbose",     0,NULL,'v'},
      {"resource",    0,NULL,'r'},
      {"sync",        0,NULL,'s'},
      {"prefetching", 0,NULL,'P'},
      {"delay",       1,NULL,'d'},
      {"timeout",     1,NULL,'t'},
      {"loop",        0,NULL,'l'},
      {"port",        1,NULL,'p'},
      {"sever-registry",    0,NULL,'S'},
      {"help",        0,NULL,'h'},
      {NULL,          0,NULL,0}
    };
    int ch,long_option_index;

    while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
      // which option did we get?
      switch (ch) {
        case 'v':
          verbose++;
          break;
        case 'i':
	  __input_files.push_back(getAbsolutePath(optarg));
	    break;
        case 'd':
          __delay = atoi(optarg);
          break;
        case 'o':
          __output_file = optarg;
          break;
        case 'h':
          usage(argv[0]);
          exit(0);
        case 's':
          WARNING_STREAM("Disable synchronizer" << endl);
          __sync = false;
          break;
        case 'P':
          __prefetching = true;
          break;
        case 'r':
          __resource = true;
          break;
        case 'u':
          __graph_url = optarg;
          break;
        case 'A':
          __playback_location = optarg;
          break;
        case 'V':
          __display_location = optarg;
          break;
        case 'l':
          WARNING_STREAM("Disable loop" << endl);
          __infinite_loop = false;
          break;
        case 't':
          __timeout = atoi(optarg);
          break;
        case 'S':
          __enable_registry = true;
          break;
        case 'p' :
          __port = atoi(optarg);
          break;
        default:
          usage(argv[0]);
          throw Exception("Invalid option");
        }
    }

    /* handle *Stream output */
    if (verbose < 1) {
      NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 2) {
      NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 3) {
      NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 4) {
      NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
    }
 
    if( __graph_url.empty() ) { // if no graph url was given -> use .gd file
      // check for filename
      if ( argc - optind < 1 ) {
	usage(argv[0]);
	string nmm_dir = NMM::getNMMDir();
	string error_string ;
	if (!nmm_dir.empty())
	  error_string = string( "Hint : Check directory ") + nmm_dir + string( "share/nmm/gd or directory ") + nmm_dir + string( "apps/clic/gd for graph descriptions.");
	
	throw Exception(string("No filename for a graph description set\n") + error_string);
      }
      else {
	__gd_file = argv[optind];
	/* create graph description from a file */
	getGraphfromGD(__gd_file,__gd);
      }
    }
    else { // some graph url was given -> do not use  .gd file
      if ( !(argc - optind < 1) ) {
	usage(argv[0]);
	throw Exception("Specifying url for graph builder and setting .gd file is not supported \n");
      }
    }
  }

  void ClicApplication::waitUntilKeyPressed(const char c) {
    MESSAGE_STREAM("Enter '" << c << "', to exit the application" << endl);
    MESSAGE_STREAM("Enter 'p', to stop current graph" << endl);
    MESSAGE_STREAM("Enter 's', to start a graph again" << endl);
    MESSAGE_STREAM("Enter 'n', to switch to next input file" << endl);
    if(__resource) {
      MESSAGE_STREAM("Enter 'r', to print resource manager infos" << endl);
    }

    char tmp = 0x0;

    while (1) {
      /* Begin> Lock the mutex, register stdin and wait until data arrived */
      {
        MutexGuard mg(__cond_mutex);
        SNetReactor::getInstance().registerSocket(0, this);
        __condition.wait();
        if(__quit) {
          break;
        }
        /* End< */
        
        /* if exit_flag is set leave the loop */
        if ( __exit_flag ) {
          return;
        }

        cin >> tmp;
      }

      if (tmp == 0x0)
        return;


      if (tmp == c) {
        cerr << "Exit... " << endl;
        // graph description is freed in desctructor
        break;
      }
      if (tmp == 'p' ) {
        __gd.stopGraph();
        cerr << "Stop current graph. Enter 's' to start it again " << endl;
      }
      if (tmp == 's' ) {
        __gd.startGraph();
        cerr << "Start current graph." << endl;
      }
      if (tmp == 'r' ) {
        if (__resource) {
          printCPUInfo();
          printResourceStatistics();
        }
      }
      if (tmp == 'n' ) {
	if(__input_files.size() > 0) {
	  __curr_input_file++;
	  if(__curr_input_file >= __input_files.size() ) {
	    cerr << "reset input file to 0 " << endl;
	    __curr_input_file = 0;
	  }
	  cerr << "Switch to input file " << (__curr_input_file) << ": " << __input_files[__curr_input_file] << endl;
	  /* Begin> handle input files: */
	  SequenceValue<INode>& inode_list = __gd.getINodeList();
	  for (unsigned int i = 0; i< inode_list.size(); ++i) {
	    if(__gd.getInEdgeList(i).size() == 0) {
	      IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
	      if (i_handler.get()) {
		inode_list[i]->reachInitialized();
		i_handler->setFilename(__input_files[__curr_input_file]);
		inode_list[i]->reachStarted();
		break; // currently only works for a single source
	      }
	    }
	  }
	  /* End< */
	}
      }
    }
  }

  void ClicApplication::nextInputFile() {
    if(__input_files.size() > 0) {
      __curr_input_file++;
      if(__curr_input_file >= __input_files.size()) {
        cerr << "All files played" << endl;
        /* If we are not running in inifinte loop, set the exit flag to true and notify the main thread */
        if (!__infinite_loop) {
          __exit_flag = true;
          MutexGuard guard(__cond_mutex);
          __condition.notify();
        }
        /* End< */
        return;
      }
      cerr << "Switch to input file " << (__curr_input_file) << ": " << __input_files[__curr_input_file] << endl;
      /* Begin> handle input files: */
      SequenceValue<INode>& inode_list = __gd.getINodeList();
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
	if(__gd.getInEdgeList(i).size() == 0) { // if (inode_list[i]->getType() == INode::SOURCE) {
	  cerr << "found SOURCE" << endl;
          IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
          if (i_handler.get()) {
            inode_list[i]->reachInitialized();
            i_handler->setFilename(__input_files[__curr_input_file]);
            inode_list[i]->reachStarted();
	    break; // currently only works for a single source
          }
        }
      }
      /* End< */
    }
  }

  void ClicApplication::printResourceStatistics()
  {
    cerr << endl;
    SequenceValue<INode>& inode_list = __gd.getINodeList();
    for (size_t i = 0; i< inode_list.size(); ++i) {
      IResourceCollector_var collector(inode_list[i]->getParentObject()->getInterface<IResourceCollector>());

      cerr << inode_list[i]->getParentObject()->getInterface<INamed>()->getName() << " : " << endl
	   << "    UsedCPUTime              : " << collector->getUsedCPUTime() << endl
	   << "    relative CPU-Usage       : " << collector->getCPUUsage(true) << " %" << endl;

    }
    cerr << endl;
  }

  void ClicApplication::printCPUInfo()
  {
    if(__resource_mgr.get()) {
      string cpu_type;
      unsigned int cpu_speed,cpu_num;
      float cpu_free;
      __resource_mgr->getCPUInfo(cpu_type, cpu_speed, cpu_num, cpu_free);

      cerr << UserInfo().getHostname() << " : " << cpu_num << " x " << cpu_type << ",  "
	   << "speed " << cpu_speed << " bogomips, " << cpu_free << " % free" << endl;
    }
  }

  void ClicApplication::setupGraph() {
    /* Begin> handle input files: */
    if ( __input_files.size() > 0 ) {
      cerr << "Switch to input file " << (__curr_input_file) << ": " << __input_files[__curr_input_file] << endl;
      SequenceValue<INode>& inode_list = __gd.getINodeList();
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
	if(__gd.getInEdgeList(i).size() == 0) {
	  IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
	  if (i_handler.get()) {
	    i_handler->setFilename(__input_files[__curr_input_file]);
	    break; // currently only works for a single source
	  }
	}
      }
    }
    /* End< */
    /* Begin> handle output files: */
    SequenceValue<INode>& inode_list = __gd.getINodeList();
    for (unsigned int i = 0; i< inode_list.size(); ++i) {
      if(__gd.getOutEdgeList(i).size() == 0) { 
      
	++__sink_nodes;
	inode_list[i]->getParentObject()->registerEventListener(ITrack::endTrack_event, 
								new TEDObject0<ClicApplication, Result (ClicApplication:: * )(), Result >
								(this, &ClicApplication::eofHandler));
	inode_list[i]->getParentObject()->registerEventListener(ITrack::finished_event, 
								new TEDObject0<ClicApplication, Result (ClicApplication:: * )(), Result >
								(this, &ClicApplication::finishedHandler));
      }
    }
    if (!__output_file.empty()) {
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
	if(__gd.getOutEdgeList(i).size() == 0) { 
	  ISinkFileHandler_var o_handler(inode_list[i]->getParentObject()->getInterface<ISinkFileHandler>());
	  if (o_handler.get()) {
	    o_handler->setFilename(__output_file);
	    break; // currently only works for a single sink
	  }		    
	}
      }
    }
    /* End< */
    
    /* Begin> Configure the nodes according to its event-description and store the events in the NodeDescription */
    for (list<NodeConfiguration>::iterator it = node_configuration.begin(); it != node_configuration.end(); ++it) {
      if (!it->getEventList().empty()) {
	for (list<NCElement>::iterator it1 = it->getEventList().begin(); it1 != it->getEventList().end(); ++it1) {
	  /* create an ClicValueStream */
	  IClicValueStream iv(it1->event_key, it1->argument_list);

	  /* get the inode */
	  INode* inode = __gd.getINode(*it->getNodeDescription());
	  /* request the event from the node */
	  Event* event= inode->getParentObject()->matchEventPrototype( iv.getEventKey() );

	  if (!event) {
	    throw Exception ("Unknown event with key " + iv.getEventKey() + " in node " + it->getNodeDescription()->getNodeName() );
	  }

	  /* read the parameters from the string */
	  iv.readEvent(*event);

	  /* and add the event to the NodeDescription */
	  __gd.getNodeDescription(it->getNodeDescription()->getInternalID()).addNodeEvent(*event, it1->state);
	  /* delete the old event */
	  delete event;
	}
      }
    }
    /* End< */

    if (__sync) {
      /* connect the synchronizer if synchronization is enabled */
      __gd.connectSinkSynchronizer();      
    }

    if (__prefetching) {
      /* connect the synchronizer if synchronization is enabled */
      __gd.setPrefetchingEnabled(true);      
    }

    /* enable resource measurements */
    if(__resource) {
      __resource_mgr.reset(__app->getRegistry().getRegistryList()[RegistryName(UserInfo().getHostname(), ClientRegistry::default_port)].server_registry->getResourceManager());

      string cpu_descr;
      __resource_mgr->getCPUDescription(cpu_descr);
      cerr << endl << "running on " << cpu_descr << endl << endl;

      SequenceValue<INode>& inode_list = __gd.getINodeList();
      for (size_t i = 0; i< inode_list.size(); ++i) {
        inode_list[i]->getParentObject()->getInterface<IResourceCollector>()->enableResourceProbing();
      }
      cerr << "Enabled resource probing for all nodes." << endl;

    }

    __paused = false;

    /* realize the graph */
    __gd.realizeGraph();

    /* activate the graph */
    __gd.activateGraph();
    /* and finaly start it */
    __gd.startGraph();

    /* Begin> Find display node */
    for (unsigned int i = 0; i< inode_list.size(); ++i) {
      if (inode_list[i]->getType() == INode::SINK) {
	/* set video display */
	__video_display.reset(inode_list[i]->getParentObject()->getInterface<IVideoDisplay>());
	if (!__video_display.get()) {
	  continue;
	}
	
        /* enable display events*/
        enableDisplayEvents();
        
	/* found a display */
	break;
      }
    }
    /* End< */

  }
  
  void ClicApplication::buildGraph() {
    string demux_audio_jack_tag = "audio"; // e.g. set this to audio0 or audio1 to explicitly choose audio track for demux nodes
    string demux_video_jack_tag = "video";

    try {
      NodeDescription __playback_nd("PlaybackNode");
      if( !__playback_location.empty() ) {
	__playback_nd.setLocation(__playback_location);
      }
      NodeDescription __display_nd("XDisplayNode");
      if( !__display_location.empty() ) {
	__display_nd.setLocation(__display_location);
      }
    
      // setup graph builder
      GraphBuilder2 gb;
      if(!gb.setURL(__graph_url)) {
	throw Exception("Invalid URL given");
      }
        
      if(__sync) {
	// create synchronizer
	__av_sync = new MultiAudioVideoSynchronizer();
        __synchronizer = __av_sync->getCheckedInterface<IMultiAudioVideoSynchronizer>();
	gb.setMultiAudioVideoSynchronizer(__synchronizer);
        __paused = false;
      }
      gb.setAudioSink(__playback_nd);
      gb.setVideoSink(__display_nd);
      gb.setDemuxAudioJackTag(demux_audio_jack_tag);
      gb.setDemuxVideoJackTag(demux_video_jack_tag);
      if(__prefetching) {
	gb.setAudioVideoPrefetchingEnabled(__prefetching);
      }

      __composite = gb.createGraph(*__app);
      __composite->reachStarted();
      if(__display) {
        __display->reachStarted();
      }
      
      /* get interface for seeking */
      __seekable = __composite->getInterface<ISeekable>();
      
      /* get video display interface */
      __video_display.reset(__composite->getInterface<IVideoDisplay>());
      if(__video_display.get()) {
        enableDisplayEvents();
      }
    }
    catch (const Exception& e) {
      cerr << e << endl;
    }
    catch(...) {
      cerr << "something went wrong..." << endl;
    }
  } 

  void ClicApplication::destroyGraph() {
    if(__seekable) {
      delete __seekable;
      __seekable = 0;
    }
    
    disableDisplayEvents();
    if(__video_display.get()) {
      __video_display.reset(NULL);
    }    
    
    if(__composite) {
      if(__display && __display->isStarted()) {
        __display->stop();
      }
      __composite->reachActivated();
      __composite->flush();
      if(__display && __display->isActivated()) {
        __display->flush();
      }
      __composite->reachConstructed();
      if(__display) {
        __display->reachConstructed();
      }
      delete __composite;
      __composite = 0;

      if(__display) {
        __app->getRegistry().releaseNode(*__display);
        __display = NULL;
      }
    }
    
    if(__sync) {
      delete __synchronizer;
      __synchronizer = 0;
      delete __av_sync; 
      __av_sync = 0;
    }
  }
  
  void ClicApplication::enableDisplayEvents() {
    if(__video_display.get() && __composite) {
      /* enable display events */
      IDisplayEventSender_var display_event_sender(__video_display->getParentObject()->getInterface<IDisplayEventSender>());
      if (display_event_sender.get()) {
        DEBUG_STREAM( "ENABLE DISPLAY EVENTS!!!!" << endl);
        display_event_sender->enableDisplayEvents();
        
        __x_listener = new TEDObject1<ClicApplication, string>(this, &ClicApplication::keyReleased);
        /* register an event listener */
        __composite->
          registerEventListener(IDisplayEvents::keyReleased_event, __x_listener);
      }
    }
  }
  
  void ClicApplication::disableDisplayEvents() {
    if(__video_display.get()) {
      /* enable display events */
      IDisplayEventSender_var display_event_sender(__video_display->getParentObject()->getInterface<IDisplayEventSender>());
      if (display_event_sender.get()) {
        /* unregister event listener */
        __composite->removeEventListener(__x_listener);
        
        display_event_sender->disableDisplayEvents();        
      }
    }
  }
  
  void ClicApplication::usage(const string& name) {
    cerr << "Usage (for .gd files):" << endl
         << "  " << name << " [options] <file with graph description>" << endl
         << endl
         << "Options:" << endl
         << "  -i : Try to set an input file to all source nodes which support IFilename" << endl
         << "  -o : Try to set an output file to all sink nodes support IFilename" << endl
         << "  -v : increase verbositiy (may be given multiple times)" << endl
         << "  -r : Enable resource probing" << endl
         << "  -l : Exit clic after playing all files " << endl
         << "  -t : Sets a timeout in seconds after that clic automatically exit" << endl
         << "  -s : Disable synchronization " << endl
         << "  -P : Enable prefetching " << endl
         << "  -S : Enable access to the local server registry from other applications " << endl
         << "  -u <url> : Use graph builder to play back given URL" << endl
         << "  -p : Sets a specific port for the server registry " << endl
         << "  -h : Show this help" << endl
         << endl;
    cerr << "Usage (for graph builder):" << endl
         << "  " << name << " [options] -u <url>" << endl
         << "  -u <url> : Use graph builder to play back given URL" << endl
         << endl
         << "Options:" << endl
         << "  -v : increase verbositiy (may be given multiple times)" << endl
         << "  -t <seconds> : Sets a timeout in seconds after that clic automatically exit" << endl
         << "  -s : Disable synchronization " << endl
         << "  -S : Enable access to the local server registry from other applications " << endl
         << "  -p <port> : Sets a specific port for the server registry " << endl
         << "  -A <hostname> : Sets host for possible audio output" << endl
         << "  -V <hostname> : Sets host for possible video output" << endl
         << "  -h : Show this help" << endl
         << endl;
  }

  Result ClicApplication::eofHandler() {
    ++__eof_counter;
    if (__eof_counter == __sink_nodes) {
      if( __input_files.size() > 0 ) {
        MESSAGE_STREAM( "Received \"end_of_file\" Event for input file " << __input_files[__curr_input_file] << endl);

        /* if a delay was defined and it was not the last input file, wait.*/
        if (__delay && !(__curr_input_file >= __input_files.size() - 1 ) )
          sleep(__delay);

        nextInputFile();
      }
      __eof_counter = 0;
    }

    return SUCCESS;
  }

  Result ClicApplication::finishedHandler() {
    cerr << endl ;
    MESSAGE_STREAM( "Received \"finished\" Event for all input files. "<< endl);
    return SUCCESS;
  }

  NetHandler::NetResult ClicApplication::receiveEvent(int, short, short ) {
    /* Begin> Notify the main thread to read data from stdin */
    __cond_mutex.lock();
    __condition.notify();
    __cond_mutex.unlock();
    /* End< */
    return NetHandler::REMOVE;
  }

  void ClicApplication::timeout(Timer* ) {
    __cond_mutex.lock();
    __exit_flag = true;
    __condition.notify();
    __cond_mutex.unlock();
  }
}



