/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        David Maass <dama@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "grab.hpp"

// avoid forward declarations
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"

// used interfaces
#include "nmm/interfaces/device/cdda/ICDDA.hpp"
#include "nmm/interfaces/device/cdda/ICDDB.hpp"
#include "nmm/interfaces/general/progress/IProgressListener.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/base/IThreadedNode.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"

#include "nmm/base/connect.hpp"    // we want do connect :-)
#include "nmm/utils/utils.hpp"     // for string + int
#include "nmm/utils/DirReader.hpp" // for makeDir()
#include <fstream>                 // for writing playlist
#include <iostream>

// we want to be in:
namespace NMM {

  GrabState::GrabState(string id, MMBoxApplication* _app)
    : MMBoxState(id, _app),
      total_tracks(0),
      actual_track_name("unknown track"),
      current_progress(0),
      startTrack_listener(this, &GrabState::startTrack),
      setProgress_listener(this, &GrabState::setProgress),
      finished_listener(this, &GrabState::finished),
      cdda(NULL),
      enc(NULL),
      write(NULL),
      iprogress(NULL),
      text(NULL),
      plane(NULL),
      border(NULL),
      scroll(NULL),
      header(NULL),
      win(NULL)
  {
    // we initialize ALL members explicit and reasonable via colon
    // after the super class:
    // some variables (e.g. Mutexes, ThreadConditions or EDObjects)
    // must be initialized like this, so to be uniform, we initialize
    // all members like this!
    // Attention with reordering warnings from gcc!!!!
  }

  GrabState::~GrabState() {
    // nothing to do here, all allocated resourced are freed in doKill!
  }


  /******************** MMBoxState interface ********************/
  Result GrabState::doCreate(string) {
    GLOBAL_DEBUG_STREAM("GrabState::doCreate()" << endl);

    // init the background
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph(background)) {
        GLOBAL_ERROR_STREAM("Could not create default Graph!" << endl);
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        GLOBAL_ERROR_STREAM("Could not start default Graph!" << endl);
        return FAILURE;
      }
    }

    // set up osd as early as possible, to avoid empty windows
    createWindow();

    // read config values
    string mp3_root_path;
    if (!app->getMMBoxConfig()->getValue("rootpath", mp3_root_path)) {
      mp3_root_path="/";
    }
    string dest_dir;
    if (!app->getMMBoxConfig()->getValue("grabdestinationdirectory", dest_dir)) {
      dest_dir="grabbed CDs";
    }
    // now we can compute the outputdir
    string outputdir = mp3_root_path + "/" + dest_dir;

    string grabcd;
    if (!app->getMMBoxConfig()->getValue("grabcd", grabcd)) {
      grabcd = string("ogg");
    }
    bool grab_ogg;
    string encoder_name;
    string fileext;
    if (grabcd == string("ogg")) {
      grab_ogg = true; 
      encoder_name = string("VorbisEncodeNode");
      fileext = string("ogg");
    } else {
      grab_ogg = false;
      encoder_name = string("MPEGAudioEncodeNode");
      fileext = string("mp3");
    }

    bool create_playlist;
    if (!app->getMMBoxConfig()->getValue("grabcreateplaylist", create_playlist)) {
      create_playlist = true;
    }
    bool do_cddb_query;
    if (!app->getMMBoxConfig()->getValue("cddb", do_cddb_query)) {
      do_cddb_query = true;
    }
    string cddb_server;
    if (!app->getMMBoxConfig()->getValue("cddbserver", cddb_server)) {
      cddb_server = "freedb.freedb.org";
    }
    int bitrate;
    if (!app->getMMBoxConfig()->getValue("grabmp3bitrate", bitrate)) {
      bitrate = 128;
    }
    // we compute the quality for ogg from the bitrate
    float quality;
    switch (bitrate) {
    case 64:
      quality = 0.2;
      break;
    case 128:
      quality = 0.4;
      break;
    case 192:
      quality = 0.6;
      break;
    case 256:
      quality = 0.8;
      break;
    }

    if (false == DirReader::makeDir(outputdir)) {
      GLOBAL_ERROR_STREAM("Failed to mkdir \"" + outputdir + "\"" << endl);
      cleanup();
      return FAILURE;
    }

    try {
      /*Begin> Create requests for needed CDDANode */
      NodeDescription cdda_req;
      // we request a Node, capable of reading AudioCDs
      cdda_req.addInterface(ICDDA::ICDDA_type);
      /*End< */

      /*Begin> Get the ClientRegistry and request Nodes */
      ClientRegistry& registry = app->getNMMApplication()->getRegistry();
      {
        RegistryLock lock(registry);
     
        list<Response> cdda_response = registry.initRequest(cdda_req);
        list<Response> enc_response =
          // request Encoder by name
          registry.initRequest(NodeDescription(encoder_name));
        list<Response> write_response =
          // request WriteNode by name
          registry.initRequest(NodeDescription("GenericWriteNode"));

        if (cdda_response.empty() ||
            enc_response.empty() ||
            write_response.empty() ) {
          GLOBAL_ERROR_STREAM("Size of cdda_response " << cdda_response.size() << endl);
          GLOBAL_ERROR_STREAM("Size of enc_response " << enc_response.size() << endl);
          GLOBAL_ERROR_STREAM("Size of write_response " << write_response.size() << endl);
          // a detailed Exception
          throw Exception(string("A required node is not available\n") +
                          string("Size of cdda_response ") + cdda_response.size() + string("\n") +
                          string("Size of enc_response ") + enc_response.size() + string("\n") +
                          string("Size of write_response ") + write_response.size());
        }

        cdda = registry.requestNode(cdda_response.front());
        enc = registry.requestNode(enc_response.front());
        write = registry.requestNode(write_response.front());
      }
      /* End< */
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("CdState::create(): requesting failed!" << endl);
      GLOBAL_ERROR_STREAM(e << endl);

      // cleanup allocated stuff
      cleanup();
      app->getOsdManager()->displayMessage("ERROR", "Cannot start CDGrabber", "", "OK");
      changeState("OSDInputState");
      app->getOsdManager()->update();
      return SUCCESS;
    } catch (...) {
      cerr << "Something strang went wrong while requesting in GrabState::doCreate()" << endl;
      cleanup();
      return FAILURE;
    }

    //configure the nodes
    ICDDA_var icdda(cdda->getParentObject()->getCheckedInterface<ICDDA>());
    icdda->setRepeat(false);

    ICDDB_var icddb(cdda->getParentObject()->getInterface<ICDDB>());

    if (icddb.get()) {
      // we might got a Source not capable of CDDB
      icddb->setCDDBServer(cddb_server,8880);
      icddb->setCDDB(do_cddb_query);
    }

    // store interfaces for later use
    iprogress = cdda->getParentObject()->getInterface<IProgress>();

    ISinkFileHandler_var isink(write->getParentObject()->getCheckedInterface<ISinkFileHandler>());
    isink->setFileExtension(fileext);
    isink->setWorkingDirectory(outputdir);

    // just now for watching if someone does not comply with our
    // convenience
    if (!app->getOsdNode()->isStarted()) {
      cerr << "OSDNode has unexpected state: " << app->getOsdNode()->getState() << endl;
    }
  
    // initialize & connect the nodes
    try {
      cdda->init();
      enc->init();
      write->init();

      cdda->initOutput();
      connect(cdda, enc);
      cdda->activate();

      enc->initOutput();
      const Format* enc_format(enc->getWorkingOutputProperty()->getDefaultFormat());
      // compute encoder output format from its default on plus values
      // read from config
      Format encode_format(enc_format->getFormatType());
      encode_format.addIntParamValue(Format::sample_rate_param, enc_format->getIntValue(Format::sample_rate_param));
      encode_format.addIntParamValue(Format::channels_param, enc_format->getIntValue(Format::channels_param));
      if (grab_ogg) {
        encode_format.addFloatParamValue(Format::quality_param, quality);
      } else {
        encode_format.addIntParamValue(Format::layer_param, 3);
        // why does the VorbisEncodeNode not have this param?
        encode_format.addIntParamValue(Format::audio_codec_param, enc_format->getIntValue(Format::audio_codec_param));
        encode_format.addIntParamValue(Format::bitrate_param, bitrate * 1000);
        encode_format.addIntParamValue(Format::bitpersample_param, enc_format->getIntValue(Format::bitpersample_param));
      }
      connect(enc, write, &encode_format);
      enc->activate();

      write->initOutput();
      write->activate();
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("GrabState::doCreate(): connecting failed!" << endl);
      GLOBAL_ERROR_STREAM(e << endl);

      // cleanup allocated stuff
      cleanup();
      app->getOsdManager()->displayMessage("Error", "Cannot start CDgrabber.", "", "OK");
      changeState("OSDInputState");
      app->getOsdManager()->update();
      return SUCCESS;
    } catch (...) {
      cerr << "Something strang went wrong while connecting in GrabState::doCreate()" << endl;
      cleanup();
      return FAILURE;
    }

    // registering these Events directly at the writer avoids conflicts
    // with other apps listening on startTrack or finished from their
    // graphs
    write->getParentObject()->registerEventListener(ITrack::startTrack_event, &startTrack_listener);
    write->getParentObject()->registerEventListener(ITrack::finished_event, &finished_listener);
    if (iprogress != NULL) {
      iprogress->sendProgressInformation(true);
      write->getParentObject()->registerEventListener(IProgressListener::setProgress_event, &setProgress_listener);
    }

    // fill up osd, ...
    // gather AudioCD infos
    string disc_artist("unknown artist");
    string disc_title("unknown title");
    if (icddb.get()) {
      disc_artist = icddb->getDiscArtist();
      disc_title = icddb->getDiscTitle();
      actual_track_name = icddb->getTrackName(1);
    } else {
      actual_track_name = string("Track 1");
    }

    // recompute the output dir based on AudioCD infos
    outputdir = (mp3_root_path +
                 "/" + dest_dir + "/" +
                 disc_artist + " - " + disc_title);
    // try to create the needed directory
    if (! DirReader::makeDir(outputdir)) {
      GLOBAL_ERROR_STREAM("Failed to mkdir \"" + outputdir + "\"" << endl);
      cleanup();
      return FAILURE;
    }
    // set the new working directory for the writer
    isink->setWorkingDirectory(outputdir);

    // eventually create the playlist
    ofstream playlist;
    if (create_playlist) {
      // this makeDir should not fail
      DirReader::makeDir(outputdir + "/playlists");
      string playlistname = outputdir + "/playlists/" + disc_artist + " - " + disc_title + ".m3u";
      playlist.open(playlistname.c_str(), ios::trunc | ios::out);
    }

    // fill text view and playlist
    total_tracks = icdda->getTotalTracks();
    text->deleteAllLines();
    header->setText(disc_artist + " - " + disc_title);
    for(unsigned int i=1; i <= total_tracks; i++) {
      string current_track(string("Track ") + i);
      if (icddb.get()) {
        current_track = icddb->getTrackName(i);
      }
      text->addLine(current_track);
      if (create_playlist) {
        playlist << "../" << current_track << "."<<fileext << endl;
      }
    }
    text->selectLine(1);

    // we are finished with the playlist
    if (create_playlist) {
      playlist.close();
    }

    win->update();
    app->getOsdManager()->update();
    app->getOsdManager()->displayProgressBar(actual_track_name);
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->update();
  
    write->start();
    enc->start();
    cdda->start();

    // nice the threads!
    IThreadedNode_var ithread1(cdda->getParentObject()->getCheckedInterface<IThreadedNode>());
    ithread1->lowerPriority(15);

    IThreadedNode_var ithread2(enc->getParentObject()->getCheckedInterface<IThreadedNode>());
    ithread2->lowerPriority(15);

    IThreadedNode_var ithread3(write->getParentObject()->getCheckedInterface<IThreadedNode>());
    ithread3->lowerPriority(15);

    GLOBAL_DEBUG_STREAM("grabState::doCreate() == SUCCESS" << endl);

    return SUCCESS;
  }

  Result GrabState::doKill() {
    GLOBAL_DEBUG_STREAM("CdState::doKill()" << endl);

    if (!cdda) {
      // when we do not have a CDDANode here, we must have had
      // problems and have displayed a error-message, so we only have
      // to disconnect the default graph
      GLOBAL_DEBUG_STREAM("Recovering from Error-Message!" << endl);
    } else {
      app->getOsdManager()->hideProgressBar();
      app->getOsdManager()->hideTime();
      app->getOsdManager()->hideSymbol();

      write->getParentObject()->removeEventListener(&startTrack_listener);
      write->getParentObject()->removeEventListener(&finished_listener);
      if (iprogress != NULL) {
        write->getParentObject()->removeEventListener(&setProgress_listener);
        iprogress->sendProgressInformation(false);
      }
    }

    // in all cases: cleanup & disconnect the default graph
    cleanup();
    app->disconnectDefaultGraph();

    GLOBAL_DEBUG_STREAM("GrabState::doKill() == SUCCESS" << endl);

    return SUCCESS;
  }

  Result GrabState::doInitialize(string) {
    GLOBAL_DEBUG_STREAM("GrabState::doInitialize()" << endl);

    if (!cdda) {
      // when we do not have a CDDANode here, we must have had
      // problems and have displayed a error-message, so we have
      // nothing to do init as we will receive a message (with
      // receiveMessage()) and will deactive ourselve then!
      GLOBAL_DEBUG_STREAM("Displaying Error-Message, doing nothing!" << endl);
      return SUCCESS;
    }

    // init the background
    if (isBackgroundFree()) {
      if (FAILURE == app->createDefaultGraph( background )) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }

    win->setVisible(true);
    win->update();
    app->getOsdManager()->displayProgressBar(actual_track_name);
    app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
    app->getOsdManager()->setProgress(current_progress.getNumerator());
    app->getOsdManager()->updateProgress();
    app->getOsdManager()->update();

    GLOBAL_DEBUG_STREAM("GrabState::doInitialize() == SUCCESS" << endl);

    return SUCCESS;
  }

  Result GrabState::doUninitialize() {
    GLOBAL_DEBUG_STREAM("CdState::doUninitialize()" << endl);

    if (!cdda) {
      // when we do not have a CDDANode here, we must have had
      // problems and display a error-message, so we have
      // nothing to do here!
      GLOBAL_DEBUG_STREAM("Displaying Error-Message, doing nothing!" << endl);
      return SUCCESS;
    }


    app->getOsdManager()->hideProgressBar();
    win->setVisible(false);

    if (isBackgroundFree()) { 
      // we go in the background so:
      app->disconnectVideoGraph();
    }

    GLOBAL_DEBUG_STREAM("GrabState::doUninitialize() == SUCCESS" << endl);

    return SUCCESS;
  }

  /* Switch to last state */
  Result GrabState::Back() {
    GLOBAL_DEBUG_STREAM("GrabState::Back()" << endl);
    GLOBAL_DEBUG_STREAM("Switch to stateid: " << lastid << endl);

    return goToLastState();
  }

  Result GrabState::receiveMessage(string message) {
    GLOBAL_DEBUG_STREAM("GrabState::receiveMessage(" << message << ")" << endl);

    app->getOsdManager()->deleteMessage();

    deactivate();

    goToLastState();

    GLOBAL_DEBUG_STREAM("GrabState::receiveMessage(" << message << ") == SUCCESS" << endl);

    return SUCCESS;
  }
  /******************** MMBoxState interface ********************/


  /******************** EventListener methods ********************/
  Result GrabState::setProgress(u_int64_t& val, u_int64_t& max) {
    current_progress = Rational(val, max);
    GLOBAL_DEBUG_STREAM("GrabState::setProgress(" << current_progress << ")" << endl);

    if (hasFocus()) {
      // no osd updates if we don't have the focus
      app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
      app->getOsdManager()->setProgress(current_progress.getNumerator());
      app->getOsdManager()->updateProgress();
      app->getOsdManager()->update();
    }

    return SUCCESS;
  }

  Result GrabState::startTrack(string& name) {
    GLOBAL_DEBUG_STREAM("CdState::startTrack(" << name << ")" << endl);

    actual_track_name = name;
    // when a new track is started we can reset the progress
    current_progress = Rational(0);

    for (unsigned int i = 0; i <= text->getNumLines(); i++) {
      if (actual_track_name == text->getLineContent(i+1)) {
        text->selectLine(i+1);
        break;
      }
      if (i == text->getNumLines()) {
        cerr << "Error in current-line-computation!" << endl;
      }
    }

    if (hasFocus()) {
      win->update();
      // no osd updates if we don't have the focus
      app->getOsdManager()->displayProgressBar(actual_track_name);
      app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
      app->getOsdManager()->setProgress(current_progress.getNumerator());
      app->getOsdManager()->updateProgress();
      app->getOsdManager()->update();
      app->syncReset();
    }

    return SUCCESS;
  }

  Result GrabState::finished() {
    GLOBAL_DEBUG_STREAM("CdState::finished()" << endl);

    // we are finished, set Progress to 100%
    current_progress = Rational(1);

    actual_track_name = string("Grabbing finished.");
    text->addLine(actual_track_name);
    text->selectLine(total_tracks+1);
    if (hasFocus()) {
      win->update();
      app->getOsdManager()->displayProgressBar(actual_track_name);
      app->getOsdManager()->setMaxProgress(current_progress.getDenominator());
      app->getOsdManager()->setProgress(current_progress.getNumerator());
      app->getOsdManager()->updateProgress();
      app->getOsdManager()->update();
    }

    return SUCCESS;
  }
  /******************** EventListener methods ********************/


  /******************** private helpers ********************/
  void GrabState::cleanup() {
    GLOBAL_DEBUG_STREAM("GrabState::cleanup()" << endl);

    deleteWindow();

    if (enc) {
      enc->reachOutputInitialized();
      if (enc->isOutputConnected()) {
        write->reachInitialized();
        disconnect_output(enc);
      }
      enc->reachInitialized();
    }

    if (cdda && cdda->isOutputConnected()) {
      cdda->reachOutputInitialized();
      disconnect_output(cdda);
    }

    if (cdda) {
      cdda->reachConstructed();
      app->getNMMApplication()->getRegistry().releaseNode(*cdda);
      delete cdda;
      cdda = NULL;    

      if (iprogress) {
        delete iprogress;
        iprogress = NULL;
      }
    }

    if (enc) {
      enc->reachConstructed();
      app->getNMMApplication()->getRegistry().releaseNode(*enc);
      delete enc;
      enc = NULL;    
    }

    if (write) {
      write->reachConstructed();
      app->getNMMApplication()->getRegistry().releaseNode(*write);
      delete write;
      write = NULL;
    }

    GLOBAL_DEBUG_STREAM("GrabState::cleanup() return" << endl);
  }

  void GrabState::createWindow() {
    GLOBAL_DEBUG_STREAM("GrabState::createWindow()" << endl);

    deleteWindow();

    text = app->getOSDWidgetFactory().createTextView();
    text->setFont("medium");
    text->setSelectionColor(0xff0080ff);
  
    scroll = app->getOSDWidgetFactory().createScrollDecorator(*text);
  
    plane = app->getOSDWidgetFactory().createPlaneDecorator(*scroll);
    plane->setPlaneColor( 0x80ff0000 );
  
    border = app->getOSDWidgetFactory().createBorderDecorator(*plane);
  
    header = app->getOSDWidgetFactory().createHeaderDecorator(*border);
    header->setFont("medium");
    header->setText("Accessing CDRom...");

    win = app->getOsdManager()->createWindow();
    win->setDimension(620, 350);
    win->setPosition(50, 50);
    win->setContent(*header);
  
    win->update();
    app->getOsdManager()->update();

    GLOBAL_DEBUG_STREAM("GrabState::createWindow() return" << endl);
  }

  void GrabState::deleteWindow() {
    GLOBAL_DEBUG_STREAM("GrabState::deleteWindow()" << endl);

    if (win) {
      app->getOsdManager()->deleteWindow(*win);
      delete win;
      win = NULL;
    }
  
    if (text) {
      app->getOsdManager()->deleteWidget(*text);
      delete text;
      text = NULL;
    }

    if (plane) {
      app->getOsdManager()->deleteWidget(*plane);
      delete plane;
      plane = NULL;
    }

    if (border) {
      app->getOsdManager()->deleteWidget(*border);
      delete border;
      border = NULL;
    }

    if (scroll) {
      app->getOsdManager()->deleteWidget(*scroll);
      delete scroll;
      scroll = NULL;
    }

    if (header) {
      app->getOsdManager()->deleteWidget(*header);
      delete header;
      header = NULL;
    }

    GLOBAL_DEBUG_STREAM("GrabState::deleteWindow() return" << endl);
  }
  /******************** private helpers ********************/

} // namespace NMM
