/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "SlideShowState.hpp"

#include "nmm/base/connect.hpp"
#include "nmm/utils/utils.hpp"

#include "nmm/interfaces/base/registry/IPluginManager.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/interfaces/video/image/ICreateImage.hpp"

#include <iostream>

#ifdef USE_MAGICK

  static bool isImageFormat(const string& extension)
  {
    string extensions[] = {"jpg", "png", "bmp", "gif", "tga", "ppm"};

    for (unsigned int i = 0; i < sizeof(extensions)/sizeof(string); i++) {
      if (extension == extensions[i]) {
	return true;
      }
    }

    return false;
  }

#endif

namespace NMM {

SlideShowState::SlideShowState( string id, MMBoxApplication* _app)
  : MMBoxState( id, _app ), read_node(0), last_video_node(0) {

  string general_rootpath;
  if (!app->getMMBoxConfig()->getValue("rootpath", general_rootpath)) {
    general_rootpath="/";
  }
  if (!app->getMMBoxConfig()->getValue("slideshowdirectory", rootpath)) {
    rootpath="Pictures";
  }
  rootpath = general_rootpath + "/" + rootpath;

  if (!app->getMMBoxConfig()->getValue("slideshowscaler", use_scaler)) {
    use_scaler = true;
  }
}

SlideShowState::~SlideShowState()
{
}

Result SlideShowState::doCreate( string ) {
  if( isBackgroundFree() ) {

    // NMM-operations may throw exception:
    // catch them, output them, clean up and return FAILURE
    try {

      /* stop and disconnect first. won't hurt (I hope!!) */
      app->disconnectVideoGraph();

      /* Default Graph Node creation */
      graph_descr.clear();

#ifdef USE_MAGICK
      NodeDescription read_nd("MagickReadNode");
#else
      NodeDescription read_nd("PNGReadNode");
#endif

      NodeDescription converter_nd("RGBtoYV12ConverterNode");
      NodeDescription scaler_nd("VideoScalerNode");

      graph_descr.addEdge(read_nd, converter_nd);
      if(use_scaler) {
	graph_descr.addEdge(converter_nd, scaler_nd);
      }

      // get the ClientRegistry and request nodes 
      ClientRegistry& registry = app->getNMMApplication()->getRegistry();
      registry.requestGraph(graph_descr);

      read_node = graph_descr.getINode(read_nd);
      
      if(use_scaler) {
	last_video_node = graph_descr.getINode(scaler_nd);
	}
      else {
	last_video_node = graph_descr.getINode(converter_nd);
      }

      // get interface for setting files from magick node
      IFileHandler_var file_source(read_node->getParentObject()->getCheckedInterface<IFileHandler>());

      // helper objects for accessing directories and files
      dirreader.setRootDir(rootpath);
      dirreader.changeDir( "." );
      dir = dirreader.getDirEntries();

      current_index = 0;
      string filename = loadNextImage(+1);
      if(filename == "") {
	GLOBAL_ERROR_STREAM("No supported image formats at all in " << rootpath << endl);
	throw Exception("No supported image formats at all in " + rootpath);
      }
      file_source->setFilename(filename);

      // realizeGraph connects nodes and activates nodes

      graph_descr.realizeGraph();
      
      // connect last node to osd of mmbox
      last_video_node->reachOutputInitialized();
      app->getOsdNode()->reachInitialized();
      if(use_scaler) {
        Format scaler_format("video/raw");
	scaler_format.addIntParamValue(Format::x_resolution_param, 720);
	scaler_format.addIntParamValue(Format::y_resolution_param, 576);
	scaler_format.addIntParamValue(Format::ratio_x_param, 720);
	scaler_format.addIntParamValue(Format::ratio_y_param, 576);
	connect(last_video_node, app->getOsdNode(), "stream1", &scaler_format);
      }
      else {
	connect(last_video_node, app->getOsdNode(), "stream1");
      }
      last_video_node->reachActivated();

      // pictures are static background, so the osd node has to be set to pause modus
      app->getOsdManager()->setPause(true);

      // start own graph and nodes from mmbox (osd and display)
      graph_descr.startGraph();
      app->getOsdNode()->reachStarted();
      app->getDisplayNode()->reachStarted();
      
      return( SUCCESS );

    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("SlideShowState::doCreate()" << endl);
      GLOBAL_ERROR_STREAM(e << endl);

      return FAILURE;
    } catch (...) {
      GLOBAL_ERROR_STREAM("Something strange went wrong in SlideShowState::doCreate()" << endl);
      app->getNMMApplication()->getRegistry().releaseGraph(graph_descr);

      // no idea, how to clean up, if this happens!
      return FAILURE;
    }
  }
  else {
    GLOBAL_ERROR_STREAM("Background is not free, cannot connect" << endl);
    return FAILURE;
  }
}

Result SlideShowState::doKill() {
  try {
    // disconnect last node from osd of mmbox
    app->getOsdNode()->reachInitialized();
    last_video_node->reachOutputInitialized();
    disconnect_output(last_video_node);

    // release graph description
    app->getNMMApplication()->getRegistry().releaseGraph(graph_descr);
    return SUCCESS;
  } 
  catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("Exception in SlideShowState::doKill" << endl);
    GLOBAL_ERROR_STREAM(e << endl);
    
    if (e.getParent()) {
      GLOBAL_ERROR_STREAM("Parent of Exception: " << e.getParent()->getName() << endl);
    }
  }
  catch (...) {
    GLOBAL_ERROR_STREAM("Something went wrong in SlideShowState::doKill" << endl);
  }
  return FAILURE;
}

Result SlideShowState::doInitialize( string ) {
  app->getOsdManager()->update();
  
  return( SUCCESS );
}

Result SlideShowState::doUninitialize() {

  return( SUCCESS );
}

Result SlideShowState::Back() {
  return( goToLastState() );
}
  
string SlideShowState::loadNextImage(int jump_by) 
{
  current_index += jump_by;
  string filename = "";
  for (unsigned int i=0; i < dir.size(); i++,current_index += jump_by ) {
    if(current_index >= static_cast<int>(dir.size())) {
      current_index = 0;
    }
    if(current_index < 0) {
      current_index = dir.size() - 1;
    }
    if ( !dir[current_index].isdir ) { 
      string suffix(dirreader.getSuffix(dir[current_index].name));

#ifdef USE_MAGICK
      if (isImageFormat(suffix)) {
	filename = string(rootpath + "/" + dir[current_index].name);
	GLOBAL_DEBUG_STREAM("Display image " << filename << endl);
	break;
      }
#else
      if(suffix == "png") {
	filename = string(rootpath + "/" + dir[current_index].name);
	GLOBAL_DEBUG_STREAM("Display image " << filename << endl);
	break;
      }
#endif
    }
  }

  return filename;
}

Result SlideShowState::Down()
{
  string filename = loadNextImage(+1);
  if(filename != "") {
    read_node->reachInitialized();
    IFileHandler_var file_source(read_node->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    read_node->reachStarted();
    return SUCCESS;
  }
  else {
    return FAILURE;
  }
}
Result SlideShowState::Up(){
  string filename = loadNextImage(-1);
  if(filename != "") {
    read_node->reachInitialized();
    IFileHandler_var file_source(read_node->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    read_node->reachStarted();
    return SUCCESS;
  }
  else {
    return FAILURE;
  }
}

}
