/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "taskmgr.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ServiceProxy.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"

#include <stdio.h>

namespace NMM {

  TaskMgrState::TaskMgrState( string id, MMBoxApplication* _app ) : 
    MMBoxState( id, _app ),
    manager(app->getOsdManager()),
    factory(manager->getParentObject()->getCheckedInterface<IWidgetFactory>()),
    task_list(NULL) {
    SReferenceManager::getInstance().registerService("mmboxtaskmgr", this);

  }

  TaskMgrState::~TaskMgrState() {
    if (factory) {
      delete factory;
      factory = NULL;
    }
  
    if (task_list) {
      delete task_list;
      task_list = NULL;
    }
  }

  Result TaskMgrState::receiveMessage( string ) {
    manager->deleteMessage();

    deactivate();
    goToLastState();

    return SUCCESS;
  }

  Result TaskMgrState::doCreate( string ) {
    GLOBAL_DEBUG_STREAM( "Template: initialize: " << stateid << endl );
    
    if ( isBackgroundFree() ) {
      if (FAILURE == app->createDefaultGraph( background )) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }
  
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Creating Window" << endl);
    createWindow();
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Created Window" << endl);
  
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Making list" << endl);
    activeTaskToList();
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Made list" << endl);
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Selecting line" << endl);
    if(task_list->getTextWidget()->getNumLines() > 0)
      task_list->getTextWidget()->selectLine( 1 );
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Selected line" << endl);
  
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Showing list" << endl);
    task_list->show();
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Showed list" << endl);
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Updating list" << endl);
    task_list->update();
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Updated list" << endl);
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Updating manager" << endl);

    manager->update();
    GLOBAL_DEBUG_STREAM("TaskMgrState::doCreate(): Updated manager" << endl);
  
    GLOBAL_DEBUG_STREAM( "Template: initialized: " << stateid << endl );

    manager->setColorBarText("Destroy", "Switch", "", "");

    unsigned int lines = task_list->getTextWidget()->getNumLines();
    if ( lines <= 0 ) {
      manager->hideColorBar();
    }else{
      manager->displayColorBar();
    }
    manager->update();

    return( SUCCESS );
  }

  Result TaskMgrState::doKill() {
    GLOBAL_DEBUG_STREAM( "Template: kill: " << stateid << endl );
 
    manager->hideColorBar();
    manager->update();

    deleteWindow();

    if ( isBackgroundFree() ) app->disconnectVideoGraph();
    return SUCCESS;
  }

  Result TaskMgrState::doInitialize( string ) {
    if ( isBackgroundFree() ) {
      if (FAILURE == app->createDefaultGraph( background )) {
        return FAILURE;
      }

      if (FAILURE == app->startDefaultGraph()) {
        return FAILURE;
      }
    }
  
    task_list->deleteAllLines();
    activeTaskToList();
    task_list->show();
    task_list->update();

    manager->setColorBarText("Destroy", "Switch", "", "");
    unsigned int lines = task_list->getTextWidget()->getNumLines();
    if ( lines <= 0 ) {
      manager->hideColorBar();
    }else{
      manager->displayColorBar();
    }

    manager->update();
  
    return( SUCCESS );
  }

  Result TaskMgrState::doUninitialize() {

    task_list->hide();

    manager->hideColorBar();
    manager->update();
    
    if ( isBackgroundFree() ) app->disconnectVideoGraph();
    return( SUCCESS );
  }

  void TaskMgrState::createWindow() {
    GLOBAL_DEBUG_STREAM("TaskMgrState::createWindow()" << endl);
    deleteWindow();
  

    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new ListUnit()" << endl);
    task_list = new ListUnit( manager, factory, "medium" );
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new setHeader()" << endl);
    task_list->setHeader( "Active Tasks" );

    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new setDimension()" << endl);
    task_list->setDimension(560, 416);
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new setPosition()" << endl);
    task_list->setPosition(80, 50);
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new setColors()" << endl);
    task_list->setColors(0x80ff0000,
                         0x80ff0000,
                         0x800000ff,
                         0x800000ff,
                         0x800000ff);
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new selectWindow()" << endl);
    task_list->selectWindow();
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new stask_list->update()" << endl);
    task_list->update();
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new show()" << endl);
    task_list->show();
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow(): new manager->update()" << endl);

    manager->setColorBarText("Destroy", "Switch", "", "");
    manager->displayColorBar();

    manager->update();
    GLOBAL_DEBUG_STREAM("TaskMgrState::createWindow() finished" << endl);
  }

  void TaskMgrState::deleteWindow() {
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow()" << endl);
    if ( task_list ) {
      delete task_list;
      task_list = NULL;
    }
    GLOBAL_DEBUG_STREAM("TaskMgrState::deleteWindow() finished" << endl);
  }

  void TaskMgrState::activeTaskToList() {
    const MMBoxApplication::mmbox_state_list_t* states;
    MMBoxApplication::mmbox_state_list_t::const_iterator i;

    states = app->getActiveStates();
    for( i=states->begin(); i!=states->end(); i++ ) {
      MMBoxState* state = *i;
      if ( state->getID() != getID() ) {
        task_list->getTextWidget()->addLine( state->getID() );
      }
    }
  }

  /* Switch to last state */
  Result TaskMgrState::Back() {
    deactivate();

    return( goToLastState() );
  }


  Result TaskMgrState::Up() {

    task_list->getTextWidget()->selectUp();
    task_list->update();

    unsigned int sel = task_list->getTextWidget()->getSelectedLine();
    string id = task_list->getTextWidget()->getLineContent( sel );

    if ( id == getID() ) {
      manager->hideColorBar();
    }else{
      manager->displayColorBar();
    }


    manager->update();

    return(SUCCESS);
  }

  Result TaskMgrState::Down() {

    task_list->getTextWidget()->selectDown();
    task_list->update();

    unsigned int sel = task_list->getTextWidget()->getSelectedLine();
    string id = task_list->getTextWidget()->getLineContent( sel );

    if ( id == getID() ) {
      manager->hideColorBar();
    }else{
      manager->displayColorBar();
    }

    manager->update();

    return(SUCCESS);
  }

  Result TaskMgrState::colorKeyGreen() {
    unsigned int sel = task_list->getTextWidget()->getSelectedLine();
    string id = task_list->getTextWidget()->getLineContent( sel );

    /* test, if we do not switch to taskmanager itself */
    if (id != getID()) {
      string state = id.substr(0, id.find_first_of(" "));

      int pos = id.find("(") + 1;

      string host;

      if (pos) {
	host = id.substr(pos, id.length() - pos - 1);
      }
      else {
      	host = "";
      }

      changeState(state, host);

    }

    return(SUCCESS);
  }

  Result TaskMgrState::colorKeyRed() {
    unsigned int sel = task_list->getTextWidget()->getSelectedLine();
    if(sel > 0){

      string id = task_list->getTextWidget()->getLineContent( sel );
      
      /* test, if we do not switch to taskmanager itself */
      if ( id != getID() ) {
        app->getState(id)->kill();
        app->reInitActualState();
      }
    }

    return(SUCCESS);
  }

  void TaskMgrState::getTaskList(list<string > & tasks) {
    tasks.clear();

    const MMBoxApplication::mmbox_state_list_t* states;
    MMBoxApplication::mmbox_state_list_t::const_iterator i;
  
    states = app->getActiveStates();

    // iterate through all active states and add their IDs to the list
    for (i = states->begin(); i!=states->end(); i++) {
      MMBoxState* state = *i;
      tasks.push_back(state->getID());
    }
  }
}

