/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/* C includes */
#include <getopt.h>
#include <stdlib.h>
#include <unistd.h> //for sleep()
#include <time.h>

/* C++ includes */
#include <iostream>
#include <fstream>
#include <string>
#include <cstdio>
#include <iomanip>

/* NMM includes */
// include connect
#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

// include used interfaces
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/device/IDevice.hpp"
#include "nmm/interfaces/device/cdda/ICDDA.hpp"
#include "nmm/interfaces/device/cdda/ICDDB.hpp"
#include "nmm/interfaces/file/ISinkFileHandler.hpp"

// makedir
#include "nmm/utils/DirReader.hpp"

using namespace NMM;

/* Listener class, catch events from the nodes */
class Listener2 {
private:
  bool is_finished;
  string track_name;

public:
  Listener2() { 
    is_finished = false;
    track_name = "";
  }
  Result finished() { is_finished = true; return SUCCESS; }
  bool getfinishedflag() { return is_finished; }
  Result startTrack(const string& name) { 

    track_name = name;
    return SUCCESS; 
  }
  string getTrackName() { 
    return track_name;
  }
};


void usage(void) {
  cerr << "USAGE:" << endl
       << "  grab2mp3 [options]" << endl
       << endl
       << "OPTIONS:" << endl
       << " -c --cddb     : use CDDB server to name tracks" << endl
       << " -p --playlist : create playlist according to the track order" << endl
       << " -g --ogg      : grab to OggVorbis" <<endl
       << " -w --wav      : grab to WAV" <<endl
       << " -b --bitrate  : set the bitrate for the mp3 file" << endl
       << " -q --quality  : set the quality for Ogg-File -1.0=low,0.4=medium,1=high" << endl
       << " -o --output   : set the target directory to write the files to" << endl
       << "                 if not specified: take \"<Artist> - <CD-Title>\"" << endl
       << " -v --verbose  : increase verbosity (may be given multiple times)" << endl
       << " -d --device   : set cdrom-device [/dev/cdrom]" << endl
       << " -n --number   : enable TrackNumbers in Filename" << endl
       << " -h --help     : show this help" << endl
       << endl;
  exit(1);
}



int main(int argc, char *argv[]) {
  /* command line parameter - see getopt(3) */
  const char *opt_string = "cpgwb:q:o::d:vnh";
  struct option options [] = {
        {"cddb",     0, NULL, 'c'},
        {"playlist", 0, NULL, 'p'},
        {"ogg",      0, NULL, 'g'},
        {"wav",      0, NULL, 'w'},
        {"bitrate",  1, NULL, 'b'},
        {"quality",  1, NULL, 'q'},
        {"output",   2, NULL, 'o'},
        {"device",   1, NULL, 'd'},
        {"verbose",  0, NULL, 'v'},
        {"number",   0, NULL, 'n'},
        {"help",     0, NULL, 'h'},
        {NULL,       0, NULL, 0  }
  };
  int verbose = 0;
  bool cddb = false;
  bool create_playlist = false;
  bool grab2vorbis = false;
  bool grab2wav = false;
  int bitrate = 128;
  float quality = 0.4;
  string output = ".";
  string device("/dev/cdrom");
  Listener2* l = new Listener2();
  bool number = false;


  /* --- Parameter --------------------------------------------------- */
  int c,long_option_index;
  string bitrates = "32 40 48 56 64 80 96 112 128 160 192 224 256 320";

  while((c=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
    // which option did we get?
    switch (c) {
    case 'g':
      if (grab2wav) {
        cerr << "Cannot grab both to WAV and OggVorbis, disabling grabing to WAV!" << endl;
        grab2wav = false;
      }
      grab2vorbis = true;
      break;
    case 'w':
      if (grab2vorbis) {
        cerr << "Cannot grab both to OggVorbis and WAV, disabling grabing to OggVorbis!" << endl;
        grab2vorbis = false;
      }
      grab2wav = true;
      break;
    case 'c':
      cddb = true;
      break;
    case 'p':
      create_playlist = true;
      break;
    case 'b':
      bitrate = atoi(optarg);
      break;
    case 'q':
      quality = atoi(optarg);
      (quality < -1) ? quality=-1 : 
        (quality > 1) ? quality=1 : 
          quality=quality;
      break;
    case 'o':
      if (optarg) {
	output = string(optarg);
	while (output.substr(output.size()-1, output.size()-1) == "/") {
	  output = output.substr(0, output.size()-1);
	}
      } else {
	output = string("");
      }
      break;
    case 'd':
      device = optarg;
      break;
    case 'h':
      usage();
      break;
    case 'v':
      verbose = true;
      break;
    case 'n':
      number = true;
      break;
    default:
      usage();
    }
  }
  
  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  GLOBAL_DEBUG_STREAM("bitrate  :" << bitrate << endl);
  GLOBAL_DEBUG_STREAM("quality  :" << quality << endl);
  GLOBAL_DEBUG_STREAM("output   :" << ((output == string("")) ? string("<Artist - Title>") : output) << endl);
  
  
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);

  /* Allocate _var variables for the INode's*/
  INode_var cdda;
  INode_var enc;
  INode_var write;

  try {
    /* --- CDDA Node --------------------------------------------------- */
    NodeDescription cdda_req;
    cdda_req.setNodeName("CDDANode");

    /* --- Converter Node ------------------------------------------ */
    NodeDescription enc_req;
    if (grab2vorbis)
      enc_req.setNodeName("VorbisEncodeNode");
    else
      enc_req.setNodeName("MPEGAudioEncodeNode");

    /* --- Generic Write Node ---------------------------------------------- */
    NodeDescription write_req;
    if (grab2wav) {
      write_req.setNodeName("WavWriteNode");
    } else {
      write_req.setNodeName("GenericWriteNode");
    }

    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    registry.initTransaction();

    list<Response> cdda_resp = registry.initRequest(cdda_req);
    list<Response> enc_resp = registry.initRequest(enc_req);
    list<Response> write_resp = registry.initRequest(write_req);


    if (cdda_resp.empty() || (!grab2wav && enc_resp.empty()) || write_resp.empty() ) {
      GLOBAL_ERROR_STREAM("Size of cdda response " << cdda_resp.size() << endl);
      if (!grab2wav) {
        GLOBAL_ERROR_STREAM("Size of enc response " << enc_resp.size() << endl);
      }
      GLOBAL_ERROR_STREAM("Size of write response " << write_resp.size() << endl);
      throw Exception("A required node is not available");
    }

    /* Begin> request the nodes */
    cdda.reset(registry.requestNode(cdda_resp.front()));
    if (!grab2wav) {
      enc.reset(registry.requestNode(enc_resp.front()));
    }
    write.reset(registry.requestNode(write_resp.front()));
    registry.releaseTransaction();
    /* End< */

    /* initialize and configure the nodes */
    /* --- CDDA Node --------------------------------------------------- */
    IDevice_var idevice(cdda->getParentObject()->getInterface<IDevice>());
    idevice->setDevice(device);

    ICDDB_var icddb(cdda->getParentObject()->getInterface<ICDDB>());
    if (cddb) {
      icddb->setCDDBServer("freedb.freedb.org", 8880);
      icddb->setCDDB(true);
    }

    ICDDA_var icdda(cdda->getParentObject()->getInterface<ICDDA>());
    icdda->setRepeat(false);
    icdda->numberTracks(number);

    cdda->init();
    cdda->initOutput();

    /* get and react on the table of contents */
    int total_tracks = icdda->getTotalTracks();
    string disc_artist = icddb->getDiscArtist();
    string disc_title = icddb->getDiscTitle();
    string fileext;
    if (grab2wav) {
      fileext = string("wav");
    } else if (grab2vorbis) {
      fileext = string("ogg");
    } else {
      fileext = string("mp3");
    }
  
    if (string("") == output) {
      output = disc_artist + string(" - ") + disc_title;
    }
    if (FAILURE == DirReader::makeDir(output)) {
      throw Exception("Failed to mkdir \"" + output + "\"");
    }

    if (create_playlist) {
      DirReader::makeDir(output + "/playlists");
      string playlistname = output + "/playlists/" + disc_artist + " - " + disc_title + ".m3u";
      ofstream playlist;
      playlist.open(playlistname.c_str(), ios::trunc | ios::out);

      for(int i=1; i<=total_tracks; i++) {
	string name = icddb->getTrackName(i);
        playlist << "../" << name << "." << fileext << endl;
      }
    
      playlist.close();
    }

    /* --- MP3 Converter Node ------------------------------------------ */
    if (enc.get()) {
      enc->init();
    }

    /* --- Generic Write Node ---------------------------------------------- */
    write->init();

    ISinkFileHandler_var isink(write->getParentObject()->getInterface<ISinkFileHandler>());
    isink->setFileExtension(fileext);
    isink->setWorkingDirectory(output);
  
    write->
      getParentObject()->
      registerEventListener(ITrack::finished_event, 
			    new TEDObject0<Listener2>(l, &Listener2::finished));
    write->
      getParentObject()->
      registerEventListener(ITrack::startTrack_event,
			    new TEDObject1<Listener2, string, Result(Listener2::*)(const string&)>(l, &Listener2::startTrack));
  
  
    /* --- Connections ------------------------------------------------ */
    if (grab2wav) {
      connect(cdda,write);
    } else {
      connect(cdda, enc);
    }
    cdda->activate();
    
    if (!grab2wav) {
      enc->initOutput();
      Format* encode_format;
      if (grab2vorbis) {
        const Format* enc_format = enc->getWorkingOutputProperty()->getDefaultFormat();
        encode_format = new Format(Format::audio_x_ogg_type);
        encode_format->addIntParamValue(Format::audio_codec_param, enc_format->getIntValue(Format::audio_codec_param));
        encode_format->addIntParamValue(Format::sample_rate_param, enc_format->getIntValue(Format::sample_rate_param));
        encode_format->addIntParamValue(Format::channels_param, enc_format->getIntValue(Format::channels_param));
        encode_format->addFloatParamValue(Format::quality_param, quality);
      } else {
        const Format* enc_format = enc->getWorkingOutputProperty()->getDefaultFormat();
        encode_format = new Format("audio/mpeg");
        encode_format->addIntParamValue(Format::layer_param, 3);
        encode_format->addIntParamValue(Format::sample_rate_param, enc_format->getIntValue(Format::sample_rate_param));
        encode_format->addIntParamValue(Format::channels_param, enc_format->getIntValue(Format::channels_param));
        encode_format->addIntParamValue(Format::bitpersample_param, enc_format->getIntValue(Format::bitpersample_param));
        encode_format->addIntParamValue(Format::audio_codec_param, enc_format->getIntValue(Format::audio_codec_param));
        encode_format->addIntParamValue(Format::bitrate_param, bitrate * 1000);
      }
      connect(enc, write, encode_format);
      enc->activate();
    }
    
    write->initOutput();
    write->activate();
  
    /* --- Start ----------------------------------------------------- */
    cdda->start();
    if (enc.get()) {
      enc->start();
    }
    write->start();
  
    string track_name = "";
    string tmp_name = "";

    /* --- dance baby ------------------------------------------------- */
    while(!(l->getfinishedflag())) {
      tmp_name = l->getTrackName();
      if (tmp_name != track_name) {
	track_name = tmp_name;
	cout << "\'" << track_name << "\' -> \'" << output << "/" << track_name << "." << fileext << "\'" << endl;
      }
      sleep(1); 
    }  
  
    /* --- clean up --------------------------------------------------- */
    cdda->stop();
    if (enc.get()) {
      enc->stop();
    }
    write->stop();
  
    cdda->flush();
    if (enc.get()) {
      enc->flush();
    }
    write->flush();
  
    cdda->deactivate();
    if (enc.get()) {
      enc->deactivate();
    }
    write->deactivate();
  
    cdda->deinitOutput();
    if (enc.get()) {
      enc->deinitOutput();
    }
    write->deinitOutput();
  
    cdda->deinit();
    if (enc.get()) {
      enc->deinit();
    }
    write->deinit();
  
    /*Release the reserved nodes*/
    registry.releaseNode(*cdda);
    if (enc.get()) {
      registry.releaseNode(*enc);
    }
    registry.releaseNode(*write);

    GLOBAL_DEBUG_STREAM("Done!" << endl);
  
  }
  // catch NMM-exception and print it
  catch(Exception e) {
    /* release the tranaction */
    app->getRegistry().releaseTransaction();

    /*Release the reserved nodes*/
    if (cdda.get())
      app->getRegistry().releaseNode(*cdda);
    if (enc.get())
      app->getRegistry().releaseNode(*enc);
    if (write.get())
      app->getRegistry().releaseNode(*write);

    cerr << e << endl;
    return 1;
  }
  // any other exception
  catch(...) {
    /* release the tranaction */
    app->getRegistry().releaseTransaction();

    /*Release the reserved nodes*/
    if (cdda.get())
      app->getRegistry().releaseNode(*cdda);
    if (enc.get())
      app->getRegistry().releaseNode(*enc);
    if (write.get())
      app->getRegistry().releaseNode(*write);

    cerr << "Something went wrong..." << endl;
    return 1;
  }

  return 0;
} /* main */


