/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

// C includes
#include <getopt.h>

#include <iostream>

#include <iostream>

// general includes
#include "nmm/base/connect.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/utils/proc/CPUInfo.hpp"
#include "nmm/utils/utils.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/resourcemgr/IResourceCollector.hpp"
#include "nmm/interfaces/video/IScaler.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/ISeekable.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"

#include "nmm/interfaces/base/IBufferManager.hpp"

// sync
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

// console producer
#include "nmm/misc/console/ConsoleProducer.hpp"

using namespace NMM;


/**
 * Class for receiving events.
 */
class AppEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  AppEventReceiver() {
    dispatcher = new EventDispatcher();
  }

  virtual ~AppEventReceiver() {
    delete dispatcher;
  }

  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }

  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }

  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }

  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }

  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  IVideoDisplay* video_display;
  bool fullscreen;
  bool change_fullscreen;
  IMultiAudioVideoSynchronizer* sync;
  ISeekable* iseek;
  bool pause;
  CPUInfo* free_cpu_monitor;
  list<string>* audio_streams;
  string current_audio_stream;
  string car_audio_stream;
  INode* demux;
  INode* audio;
  list<pair<string, IResourceCollector*> > resmgrs;

  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }

  void reconnectDemuxAudio(const string& old_jack_tag, const string& new_jack_tag) {
    if (demux && audio) {
      try {
        demux->stop();
        audio->stop();

        audio->flush();
      
        demux->deactivate();
        audio->deactivate();

        audio->deinitOutput();

        disconnect_output(demux, old_jack_tag);
        connect(demux, new_jack_tag, audio);
      
        audio->initOutput();
      
        audio->activate();
        demux->activate();

        demux->start();
        audio->start();
      } catch (const Exception& e) {
        cerr << "Error while reconnecting demux' audio stream: " << endl << e << endl;
      } catch (...) {
        cerr << "Something strange went wrong while reconnecting demux' audio stream!" << endl;
      }
    } else {
      cerr << "Demux or Audio Node not set!" << endl;
    }
  }

public:
  EventCatcher() 
    : exit(false),
      exit_mutex(),
      exit_cv(exit_mutex),
      video_display(NULL),
      fullscreen(false),
      change_fullscreen(false),
      sync(NULL),
      iseek(NULL),
      pause(false),
      free_cpu_monitor(NULL),
      audio_streams(NULL),
      current_audio_stream(""),
      car_audio_stream(""),
      demux(NULL),
      audio(NULL),
      resmgrs() {}

  Result endTrack() {
    cout << "End of File" << endl;
    signalExit();
    return SUCCESS;
  }

  void setVideoDisplay(IVideoDisplay* value) {
    video_display = value;
    if (video_display) {
      fullscreen = video_display->getFullscreen() && video_display->hasFullscreen();
    }
  }

  void setSeek(ISeekable* seeker) {
    iseek=seeker;
  }

  void setSynchronizer(IMultiAudioVideoSynchronizer* value) {
    sync = value;
  }
  
  void setCPUMonitor(CPUInfo* value) {
    free_cpu_monitor = value;
  }
  
  void setAudioStreams(list<string>& streams) {
    audio_streams = &streams;
  }
  
  void setCurrentAudioStream(const string& stream) {
    current_audio_stream = stream;
  }
  
  void setCarAudioStream(const string& stream) {
    car_audio_stream = stream;
  }
  
  void setDemuxAudioConnection(INode* node1, INode* node2) {
    demux = node1;
    audio = node2;
  }
  
  void addResourceCollector(IResourceCollector* res) {
    if (res) {
      INamed_var resname(res->getParentObject()->getInterface<INamed>());
      string name;
      if (resname.get()) {
        name = resname->getName();
      } else {
        name = "noname";
      }
      resmgrs.push_back(pair<string, IResourceCollector*>(name, res));
    }
  }

  Result keyReleased(string& key) {

    // if paused, any key should wakeup
    if (pause && sync) {
      pause = false;
      sync->wakeup();
      return SUCCESS;
    }

    if (key == "KEY_q") { // 'q' to exit
      signalExit();

    } else if (key == "KEY_f") { // 'f' to toggle fullscreen
      if (!video_display) {
	cerr << "EventHandler: No VideoDisplay set!" << endl;
	return FAILURE;
      }
      if (video_display->hasFullscreen()) {
	fullscreen = !fullscreen;
	// do switching to fullscreen from application thread
	change_fullscreen = true;
	exit_cv.notifyAll();
      } else {
	cerr << "Fullscreen not supported with current display!" << endl;
      }

    } else if (key == "KEY_Left") { //seek backward
      if (!iseek) {
	cerr<<"no seekinterface set"<<endl;
	return FAILURE;
      }
      iseek->seekPercent(Rational(-7,100));
    } else if (key=="KEY_Right") { //seek forward
      if (!iseek) {
	cerr<<"no seekinterface set"<<endl;
	return FAILURE;
      }
      iseek->seekPercent(Rational(5,100));      

    } else if ((key == "KEY_space")) { // <space> to pause, any key wakes up, see above
      if (!sync) {
	cerr << "EventHandler: No Synchronizer set!" << endl;
	return FAILURE;
      }
      pause = true;
      sync->pause();

    } else if (( key == "KEY_c")) { // 'c' for cpu monitor query
      if (!free_cpu_monitor) {
	cerr << "EventHandler: No CPU monitor set!" << endl;
	return FAILURE;
      }
      for (uint i=0; i<free_cpu_monitor->getCPUNum(); i++) {
	cout << "CPU " << i << ": " << free_cpu_monitor->getFreeCPU(i) << " % free" << endl;
      }
      cout << free_cpu_monitor->getTotalFreeCPU() << " % total" << endl;
      cout << "Name" << ":\t\t" 
           << "CPUUsage" << "\t"
           << "MaxReceivedBufferSize" << "\t"
           << "UsedCPUTime" << "\t\t" 
           << "UsedTime" << "\t"
           << endl;
      for (list<pair<string, IResourceCollector*> >::iterator res = resmgrs.begin();
           res != resmgrs.end();
           res++) {
        if (res->first.length() < 14) {
          cout << res->first << ":\t";
        } else {
          cout << res->first.substr(0, 13) << ".:\t";
        }
        cout << res->second->getCPUUsage() << "\t\t"
             << res->second->getMaxReceivedBufferSize() << "\t\t\t"
             << res->second->getUsedCPUTime() << "\t" 
             << res->second->getUsedTime() << "\t"
             << endl;
      }
    

    } else if (( key == "KEY_n")) { // 'n' for next audio_stream
      if (audio_streams) {
        string old_stream = current_audio_stream;
        for (list<string>::iterator i = audio_streams->begin();
             i != audio_streams->end();
             i++) {
          if (current_audio_stream == *i) {
            i++;
            if ((i != audio_streams->end()) && (*i == car_audio_stream)) {
              i++;
            }
            if (i == audio_streams->end()) {
              i = audio_streams->begin();
              if (*i == car_audio_stream) {
                i++;
              }
            }
            if (i != audio_streams->end()) {
              current_audio_stream = *i;
            }
            break;
          }
        }
        if (old_stream != current_audio_stream) {
          reconnectDemuxAudio(old_stream, current_audio_stream);
          cerr << "Switched audio from " << old_stream << " to " << current_audio_stream << "." << endl;
        }
      } else {
        cerr << "Audio streams not set!" << endl;
      }

    } else if (( key == "KEY_p")) { // 'p' for nexprevious audio_stream
      cerr << "WARNING: 'p' no longer pauses, but selects previous audio stream!" << endl;
      if (audio_streams) {
        string old_stream = current_audio_stream;
        cerr << "Current audio stream: " << current_audio_stream << endl;
        for (list<string>::reverse_iterator i = audio_streams->rbegin();
             i != audio_streams->rend();
             i++) {
          if (current_audio_stream == *i) {
            i++;
            if (i == audio_streams->rend()) {
              i = audio_streams->rbegin();
            }
            current_audio_stream = *i;
            break;
          }
        }
        if (old_stream != current_audio_stream) {
          cerr << "Current audio stream: " << current_audio_stream << endl;
          reconnectDemuxAudio(old_stream, current_audio_stream);
        }
      } else {
        cerr << "Audio streams not set!" << endl;
      }

    } else {
      cerr << "Unsupported key pressed: " << key << endl;
    }

    return SUCCESS;
  }

  void handleEvents() {
    while (!exit) {
      exit_mutex.lock();
      exit_cv.wait();
      if (change_fullscreen) {
	video_display->setFullscreen(fullscreen);
	change_fullscreen = false;
      }
      exit_mutex.unlock();
    }
  }

};


int usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    filename:             mpeg ps file" << endl
       << "    -r --remote-audio     Use remote audio sink on host [localhost]" << endl
       << "                          registry must be specified with: reghost[:regport]" << endl
       << "    -p --auto-aspect      Automatically correct aspect ratio" << endl
       << "    -m --mga              Use Matrox device" << endl
       << "    -f --fullscreen       Fullscreen mode" << endl
       << "    -t --audiostream      Select audio stream number" << endl
       << "    -i --videostream      Select video stream number" << endl
       << "    -a --ac3passthrough   send ac3 data directly to soundcard" << endl
       << "    -s --silent           Disable audio output" << endl
       << "    -n --no-video         Disable video output" << endl
       << "    -W --width            set video output width in pixels" << endl
       << "    -H --height           set video output height in pixels" << endl
       << "    -b --filtering        use bilinear filtering in software scaler (broken)" << endl
       << "    -d --deinterlace      Deinterlace the image" << endl
       << "    -y --no-sync          Disable Synchronisation" << endl
       << "                          (once: inter stream only, twice: completely)" << endl
       << "    -B --no-buffermanager Disable usage of BufferManager from DisplayNode" << endl
       << "    -A --car-cinema       Enable \"car-cinema\" mode with specified host" << endl
       << "                          registry must be specified with: reghost[:regport]" << endl
       << "    -c --cpu              Monitor CPU usage" << endl
       << "    -v --verbose:         increase verbosity (may be given multiple times)" << endl
       << "    -h --help:            Shows this help and exits." << endl;
  return 0;
}


int main(int argc, char ** argv) {
  /* Global options */
  bool remote_audio      = false;
  string remote_audio_host = "localhost";
  int remote_audio_port  = 0;
  int verbose            = 0;
  bool ac3_passthrough   = false;
  int use_ac3            = false;
  int use_mga            = false;
  int audio_track        = -1;
  string audio_stream    = "";
  int video_track        = -1;
  string video_stream    = "";
  bool fullscreen        = false;
  bool no_audio          = false;
  bool no_video          = false;
  bool cpu_monitor       = false;
  int out_width          = -1;
  int out_height         = -1;
  bool do_deinterlace    = false;
  bool do_scale          = false;
  bool filtering         = false;
  bool auto_aspect       = false;
  string filename;
  list<string> demux_audio_streams;
  bool mpeg_video_only = false;
  bool no_sync           = false;
  bool no_intra_sync     = false;
  bool no_buffermanager  = false;
  bool car_cinema       = false;
  string car_cinema_host = "localhost";
  int car_cinema_port    = 0;
  string car_cinema_stream = "";

  const char* optstring = "pmft:i:asnW:r:H:A:cbdyBvh";
  struct option longopts[] = {
    {"remote-audio",     1, NULL, 'r'},
    {"mga",              0, NULL, 'm'},
    {"auto-aspect",      0, NULL, 'p'},
    {"fullscreen",       0, NULL, 'f'},
    {"audiostream",      1, NULL, 't'},
    {"videostream",      1, NULL, 'i'},
    {"ac3passthrough",   0, NULL, 'a'},
    {"silent",           0, NULL, 's'},
    {"no-video",         0, NULL, 'n'},
    {"width",            1, NULL, 'W'},
    {"height",           1, NULL, 'H'},
    {"cpu",              0, NULL, 'c'},
    {"filtering",        0, NULL, 'b'},
    {"deinterlace",      0, NULL, 'd'},
    {"no-sync",          0, NULL, 'y'},
    {"no-buffermanager", 0, NULL, 'B'},
    {"car-cinema",      1, NULL, 'A'},
    {"verbose",          0, NULL, 'v'},
    {"help",             0, NULL, 'h'},
    {NULL,               0, NULL, 0  } /* last row must be zero */
  };
  
  int ch, optionindex;
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'p':
      auto_aspect = true;
      break;
    case 'm':
      use_mga = true;
      break;
    case 'f':
      fullscreen = true;
      break;
    case 't':
      if (optarg) {
        audio_track = atoi(optarg);
      }
      break;
    case 'i':
      if (optarg) {
        video_track = atoi(optarg);
      }
      break;
    case 'r':
      remote_audio = true;
      if(optarg) {
        string remote_registry = string(optarg);
        string::size_type colon_pos = remote_registry.find_first_of(':');
        if (colon_pos == string::npos) {
          remote_audio_host = remote_registry;
          remote_audio_port = ClientRegistry::default_port;
        } else {
          remote_audio_host = remote_registry.substr(0, colon_pos);
          remote_audio_port = atoi(remote_registry.substr(colon_pos+1, string::npos).c_str());
        }
        if (!remote_audio_port) {
          return usage(argv[0]);
        }
      }
      break;
    case 'a':
      ac3_passthrough = true;
      break;
    case 's':
      no_audio = true;
      break;
    case 'n':
      no_video = true;
      break;
    case 'W':
      if (optarg) {
        out_width = atoi(optarg);
      }
      break;
    case 'H':
      if (optarg) {
        out_height = atoi(optarg);
      }
      break;
    case 'c':
      cpu_monitor = true;
      break;
    case 'v':
      verbose++;
      break;
    case 'b':
      filtering = true;
      break;
    case 'd':
      do_deinterlace = true;
      break;
    case 'y':
      if (no_sync) {
        no_intra_sync = true;
      } else {
        no_sync = true;
      }
      break;
    case 'B':
      no_buffermanager = true;
      break;
    case 'A':
      car_cinema = true;
      if(optarg) {
        string remote_registry = string(optarg);
        string::size_type colon_pos = remote_registry.find_first_of(':');
        if (colon_pos == string::npos) {
          car_cinema_host = remote_registry;
          car_cinema_port = ClientRegistry::default_port;
        } else {
          car_cinema_host = remote_registry.substr(0, colon_pos);
          car_cinema_port = atoi(remote_registry.substr(colon_pos+1, string::npos).c_str());
        }
        if (!car_cinema_port) {
          return usage(argv[0]);
        }
      }
      break;
    default: //includes 'h'
      return usage(argv[0]);
    }
  }
  
  if (argc <= optind) { // no file specified
    return usage(argv[0]);
  } else {
    filename = string(argv[optind]);
  }
  
  if (-1 == audio_track) { // no audio track selected defaulting to 0
    audio_track = 0;
  }

  if (-1 == video_track) { // no video track selected defaulting to 0
    video_track = 0;
  }

  // maybe turn off debug messages
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  CPUInfo* free_cpu_monitor = NULL; // cpu-monitor on local host

  if (cpu_monitor) {
    free_cpu_monitor = new CPUInfo;
    free_cpu_monitor->setUpdateIntervall(2);
    free_cpu_monitor->start();
  }

  if ((out_width > 0) || (out_height > 0) || auto_aspect) {
    do_scale = true;
  }
  
  NMMApplication* app = 0;
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var readfile;
  INode_var demux;
  INode_var mpg2yuv;
  INode_var ac3_dec;
  INode_var mp3_dec;
  INode_var display;
  INode_var audio_play;
  INode_var scaler;
  INode_var deinterlacer;
  INode_var car_cinema_audio;
  INode_var car_cinema_dec;
  /*End< */

  /* synchronisation stuff */
  ISinkController_var audio_controller;
  ISinkController_var video_controller;
  ISinkController_var car_cinema_audio_controller;
  MultiAudioVideoSynchronizer mav_sync;
  IMultiAudioVideoSynchronizer_var msync;
  /* synchronisation stuff */

  /* Allocate an event listener */
  EventCatcher ec;
  AppEventReceiver receiver;
  
  int error_level = 0;
  try {
    app = ProxyApplication::getApplication(argc, argv);

    /*Begin> Init requests for nodes */
    NodeDescription readfile_request;
    readfile_request.setNodeName("MPEGReadNode");

    NodeDescription demux_request;
    demux_request.setNodeName("MPEGDemuxNode");
    
    NodeDescription mpg2yuv_request;
    mpg2yuv_request.setNodeName("MPEGVideoDecodeNode");
    
    NodeDescription ac3_dec_request;
    ac3_dec_request.setNodeName("AC3DecodeNode");
    
    NodeDescription mp3_dec_request;
    mp3_dec_request.setNodeName("MPEGAudioDecodeNode");
    
    NodeDescription display_request;
    display_request.setNodeName("XDisplayNode");
    
    NodeDescription audio_play_request;
    audio_play_request.setNodeName("PlaybackNode");
    
    NodeDescription scaler_request;
    scaler_request.setNodeName("VideoScalerNode");
    
    NodeDescription deinterlacer_request;
    deinterlacer_request.setNodeName("YUVDeInterlaceNode");

    NodeDescription car_cinema_audio_request("PlaybackNode");
    NodeDescription car_cinema_dec_request("AC3DecodeNode");
    /*End< */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    if(remote_audio) { 
      cerr << "Adding registry " << remote_audio_host << endl;
      registry.addRegistry(TCPAddress(remote_audio_host, remote_audio_port));
    }
    if (car_cinema) {
      cerr << "Adding registry " << car_cinema_host << endl;
      registry.addRegistry(TCPAddress(car_cinema_host, car_cinema_port));
    }

    {
      RegistryLock lock(registry);
      
      list<Response> readfile_response = registry.initRequest(readfile_request);
      list<Response> demux_response = registry.initRequest(demux_request);
      list<Response> mpg2yuv_response = registry.initRequest(mpg2yuv_request);
      list<Response> ac3_dec_response = registry.initRequest(ac3_dec_request);
      list<Response> mp3_dec_response = registry.initRequest(mp3_dec_request);
      list<Response> display_response = registry.initRequest(display_request);
      list<Response> audio_play_response;
      if(remote_audio) {
	audio_play_request.setLocation(remote_audio_host);
	audio_play_response = registry.initRequest(audio_play_request);
      }
      else {
	audio_play_response = registry.initRequest(audio_play_request);
      }
      list<Response> scaler_response = registry.initRequest(scaler_request);
      list<Response> deinterlacer_response = registry.initRequest(deinterlacer_request);
      list<Response> car_cinema_audio_response;
      list<Response> car_cinema_dec_response;
      if (car_cinema) {
        car_cinema_audio_request.setLocation(car_cinema_host);
        car_cinema_audio_response = registry.initRequest(car_cinema_audio_request);
        car_cinema_dec_request.setLocation(car_cinema_host);
        car_cinema_dec_response = registry.initRequest(car_cinema_dec_request);
      }

      if (readfile_response.empty() ||
          demux_response.empty() ||
          mpg2yuv_response.empty() ||
          ac3_dec_response.empty() ||
          mp3_dec_response.empty() ||
          display_response.empty() ||
          audio_play_response.empty() ||
          scaler_response.empty() ||
          deinterlacer_response.empty() ||
          (car_cinema && (car_cinema_audio_response.empty() || car_cinema_dec_response.empty()))) {
        GLOBAL_ERROR_STREAM("Size of readfile_response " << readfile_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of demux_response " << demux_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mpg2yuv_response " << mpg2yuv_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of ac3_dec_response " << ac3_dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mp3_dec_response " << mp3_dec_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of display_response " << display_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of audio_play_response " << audio_play_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of scaler_response " << scaler_response.size() << endl);
        GLOBAL_ERROR_STREAM("Size of deinterlacer_response " << deinterlacer_response.size() << endl);
        if (car_cinema) {
          GLOBAL_ERROR_STREAM("Size of car_cinema_audio_response " << car_cinema_audio_response.size() << endl);
          GLOBAL_ERROR_STREAM("Size of car_cinema_dec_response " << car_cinema_dec_response.size() << endl);
        }
        throw Exception("A required node is not available");
      }
      
      readfile.reset(registry.requestNode(readfile_response.front()));
      demux.reset(registry.requestNode(demux_response.front()));
      mpg2yuv.reset(registry.requestNode(mpg2yuv_response.front()));
      ac3_dec.reset(registry.requestNode(ac3_dec_response.front()));
      mp3_dec.reset(registry.requestNode(mp3_dec_response.front()));
      display.reset(registry.requestNode(display_response.front()));
      audio_play.reset(registry.requestNode(audio_play_response.front()));
      scaler.reset(registry.requestNode(scaler_response.front()));
      deinterlacer.reset(registry.requestNode(deinterlacer_response.front()));
      if (car_cinema) {
        car_cinema_audio.reset(registry.requestNode(car_cinema_audio_response.front()));
        car_cinema_dec.reset(registry.requestNode(car_cinema_dec_response.front()));
      }
    }
    /*End< */
    
    /*Begin> Initialize and connect Nodes */
    ISeekable_var file_seek(readfile->getParentObject()->getCheckedInterface<ISeekable>());
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    readfile->init();
    
    demux->init();
    
    readfile->initOutput();
    
    if (readfile->getWorkingOutputProperty()->getDefaultFormat()->getFormatType() == Format::video_mpeg_type) {
      mpeg_video_only = true;
      no_audio = true;  // video only implies no audio!
      no_sync = true;   // video only implies no inter-stream-sync!
      cout << "Pure MPEG Video stream. No Audio!" << endl;
      video_stream = string("default");
    } else {
      GLOBAL_DEBUG_STREAM("Connecting readfile to demux..." << endl);
      connect(readfile, demux);
    
      readfile->activate();
      demux->initOutput();
    
      list<string> demux_video_streams;
      list<string> demux_streams = demux->getOutputStreamTags();
      demux_streams.sort();
      cout << "Available Output streams:" << endl;
      for (list<string>::const_iterator i = demux_streams.begin();
           i != demux_streams.end();
           i++) {
        cout << "\t" << *i << endl;
        const Format* format = demux->getWorkingOutputProperty()->getDefaultFormat(*i);
        if (audio_stream.empty()) {
          if (format->getFormatType() == Format::audio_ac3_type) {
            use_ac3 = true;
          }
        }
        if ((format->getFormatType() == Format::audio_ac3_type) || (format->getFormatType() == Format::audio_mpeg_type)) {
          demux_audio_streams.push_back(*i);
        }
        if (format->getFormatType() == Format::video_mpeg_type) {
          demux_video_streams.push_back(*i);
        }
      }

      audio_stream = (use_ac3 ? string("ac3_audio") : string("mpeg_audio")) + audio_track;
      list<string>::iterator stream;
      for (stream = demux_audio_streams.begin();
           stream != demux_audio_streams.end();
           stream++) {
        if (*stream == audio_stream) {
          break;
        }
      }
      if (demux_audio_streams.end() == stream) {
        cerr << "Audio Stream " << audio_stream << " not found. Trying ";
        audio_stream = (!use_ac3 ? string("ac3_audio") : string("mpeg_audio")) + audio_track;
        cerr << audio_stream << "." << endl;
      }
      for (stream = demux_audio_streams.begin();
           stream != demux_audio_streams.end();
           stream++) {
        if (*stream == audio_stream) {
          break;
        }
      }
      if (demux_audio_streams.end() == stream) {
        cerr << "Audio Stream " << audio_stream << " not found. Defaulting to ";
        if (demux_audio_streams.size() != 0) {
          audio_stream = *demux_audio_streams.begin();
        } else {
          audio_stream = string("<no audio>");
          no_audio = true;
        }
        cerr << audio_stream << "." << endl;
      }

      if (demux_audio_streams.size() < 2) {
        cerr << "Could not find a available audio output! Disabling car-cinema!" << endl;
        car_cinema = false;
      }
      if (car_cinema) {
        if (!no_audio) {
          list<string>::iterator car_stream = demux_audio_streams.begin();
          if (*car_stream == audio_stream) {
            car_stream++;
          }
          car_cinema_stream = *car_stream;
        } else {
          car_cinema_stream = *demux_audio_streams.begin();
        }
        cout << "Using " << car_cinema_stream  << " for car-cinema" << endl;
      }

      video_stream = string("mpeg_video") + video_track;
      for (stream = demux_video_streams.begin();
           stream != demux_video_streams.end();
           stream++) {
        if (*stream == video_stream) {
          break;
        }
      }
      if (demux_video_streams.end() == stream) {
        cerr << "Video Stream " << video_stream << " not found. Defaulting to ";
        if (demux_video_streams.size() != 0) {
          video_stream = *demux_video_streams.begin();
        } else {
          video_stream = string("<no video>");
          no_video = true;
        }
        cerr << video_stream << "." << endl;
      }
    }

    if (!no_video) {
      IBiDirQueuedNode_var mpg2yuv_queue(mpg2yuv->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      mpg2yuv_queue->setDownstreamMaxSize(100);
      mpg2yuv->init();
    
      GLOBAL_DEBUG_STREAM("Connecting demux:mpeg_video0 to mpg2yuv..." << endl);
      connect((mpeg_video_only ? readfile : demux), video_stream, mpg2yuv);
    }

    if (!mpeg_video_only) {
      if (!no_audio) {
        IAudioDevice_var audio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
        audio_device->setDevice("/dev/dsp");
      
        if (use_ac3) {
          if (!ac3_passthrough) {
            IBiDirQueuedNode_var ac3_dec_queue(ac3_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
            ac3_dec_queue->setDownstreamMaxSize(100);
            ac3_dec->init();
          
            GLOBAL_DEBUG_STREAM("Connecting demux:" << audio_stream << " to ac3_dec..." << endl);
            connect(demux, audio_stream, ac3_dec);

            if (car_cinema) {
              IBiDirQueuedNode_var ac3_dec_queue(car_cinema_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
              ac3_dec_queue->setDownstreamMaxSize(100);
              car_cinema_dec->init();

              GLOBAL_DEBUG_STREAM("Connecting demux:" << car_cinema_stream << " to car_cinema_dec..." << endl);
              connect(demux, car_cinema_stream, car_cinema_dec);
            }

            demux->activate();
            ac3_dec->initOutput();
          
            GLOBAL_DEBUG_STREAM("Connecting ac3_dec to audio_play..." << endl);
            audio_play->init();
            connect(ac3_dec, audio_play);
          
            ac3_dec->activate();

            if (car_cinema) {
              car_cinema_dec->initOutput();
          
              GLOBAL_DEBUG_STREAM("Connecting car_cinema_dec to car_cinema_audio..." << endl);
              car_cinema_audio->init();
              connect(car_cinema_dec, car_cinema_audio);
          
              car_cinema_dec->activate();
              car_cinema_audio->initOutput();
              car_cinema_audio->activate();
            }
          } else {
            IBiDirQueuedNode_var audio_play_queue(audio_play->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
            audio_play_queue->setDownstreamMaxSize(100);
            audio_play->init();
          
            GLOBAL_DEBUG_STREAM("Connecting demux:ac3_audio0 to audio_play..." << endl);
            connect(demux, audio_stream, audio_play);
            demux->activate();
          }
        } else {
          IBiDirQueuedNode_var mp3_dec_queue(mp3_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
          mp3_dec_queue->setDownstreamMaxSize(100);
          mp3_dec->init();
        
          GLOBAL_DEBUG_STREAM("Connecting demux:" << audio_stream << " to mp3_dec..." << endl);
          connect(demux, audio_stream, mp3_dec);
        
          demux->activate();
          mp3_dec->initOutput();
        
          GLOBAL_DEBUG_STREAM("Connecting mp3_dec to audio_play..." << endl);
          audio_play->init();
          connect(mp3_dec, audio_play);
	
          mp3_dec->activate();
        }
      
        audio_play->initOutput();
        audio_play->activate();
      } else {
        // without audio, intra-stream-sync makes no sence
        no_sync = true;

        if (car_cinema) {
          IBiDirQueuedNode_var ac3_dec_queue(car_cinema_dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
          ac3_dec_queue->setDownstreamMaxSize(100);
          car_cinema_dec->init();
          
          GLOBAL_DEBUG_STREAM("Connecting demux:" << car_cinema_stream << " to car_cinema_dec..." << endl);
          connect(demux, car_cinema_stream, car_cinema_dec);
        }

        demux->activate();

        if (car_cinema) {
          car_cinema_dec->initOutput();
          
          GLOBAL_DEBUG_STREAM("Connecting car_cinema_dec to car_cinema_audio..." << endl);
          car_cinema_audio->init();
          connect(car_cinema_dec, car_cinema_audio);
          
          car_cinema_dec->activate();
          car_cinema_audio->initOutput();
          car_cinema_audio->activate();
        }
      }
    } else {
      readfile->activate();
    }
    
    if (mpg2yuv->isInitialized()) {
      mpg2yuv->initOutput();
    }
    
    INode* video_source = mpg2yuv.get();
    INode* video_buffer_source = mpg2yuv.get();
    Format* video_source_format = NULL;
    
    /* Insert the deinterlacer */
    if (do_deinterlace) {
      deinterlacer->init();
      connect(video_source, deinterlacer.get());
      video_source->activate();
      deinterlacer->initOutput();
      video_source = deinterlacer.get();
    }
    
    /* Insert the scaler */
    if (do_scale) {
      scaler->init();
      connect(video_source, scaler.get());
      
      video_source->activate();
      scaler->initOutput();
      
      Property_var scaler_wout(scaler->getWorkingOutputProperty());
      const Format* scaler_def_format = scaler_wout->getDefaultFormat();
      Format* scaled_video_format = new Format(scaler_def_format->getFormatType());
      int x_resolution = scaler_def_format->getIntValue(Format::x_resolution_param);
      int y_resolution = scaler_def_format->getIntValue(Format::y_resolution_param);
      if (out_width > 0) {
        x_resolution = out_width;
      }
      if (out_height > 0) {
        y_resolution = out_height;
      }
      cerr << "scaling to " << x_resolution << "x" << y_resolution << endl;
      int ratio_x, ratio_y;
      if (auto_aspect) {
        ratio_x = scaler_def_format->getIntValue(Format::ratio_x_param);
        ratio_y = scaler_def_format->getIntValue(Format::ratio_y_param);
      } else {
        ratio_x = x_resolution;
        ratio_y = y_resolution;
      }
      float framerate = scaler_def_format->getFloatValue(Format::framerate_param);
      scaled_video_format->addIntParamValue(Format::x_resolution_param, x_resolution);
      scaled_video_format->addIntParamValue(Format::y_resolution_param, y_resolution);
      scaled_video_format->addIntParamValue(Format::ratio_x_param, ratio_x);
      scaled_video_format->addIntParamValue(Format::ratio_y_param, ratio_y);
      scaled_video_format->addFloatParamValue(Format::framerate_param, framerate);
      scaled_video_format->addStringParamValue(Format::colorspace_param, "yv12");
      scaled_video_format->addStringParamValue(Format::channelorder_param, "yvu");
      scaled_video_format->addStringParamValue(Format::format_param, "planar");
      scaled_video_format->addStringParamValue(Format::endian_param, "little");
      scaled_video_format->addIntParamValue(Format::bitperpixel_param, 12);      
      
      video_source = scaler.get();
      video_buffer_source = scaler.get();
      video_source_format = scaled_video_format;
    }
    
    IVideoDisplay_var video_display;
    if (!no_video) {
      IBiDirQueuedNode_var display_queue(display->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
      display_queue->setDownstreamMaxSize(1);
      display->init();
      video_display.reset(display->getParentObject()->getCheckedInterface<IVideoDisplay>());
      if ( video_display->hasFullscreen() ) {
        video_display->setFullscreen(fullscreen);
      } else {
        cerr << "Fullscreen not supported on current display!" << endl;
      }
      GLOBAL_DEBUG_STREAM("Connecting video source to display..." << endl);
      
      if (video_source_format) {
        connect(video_source, display.get(), video_source_format);
      } else {
        connect(video_source, display.get());
      }
      
      video_source->activate();
      display->initOutput();
      display->activate();
      
      if (!no_buffermanager) {
        IBufferManager_var sink_bm(display->getParentObject()->getCheckedInterface<IBufferManager>());
        IBufferManager_var source_bm(video_buffer_source->getParentObject()->getCheckedInterface<IBufferManager>());
        source_bm->useBufferManager(*sink_bm.get());
      }
    }
    /*End< */
    
    /*Begin> Init synchronization */
    // get the synchronized sinks
    ISynchronizedSink_var audio_sync;
    if (audio_play.get()) {
      audio_sync.reset(audio_play->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    }
    ISynchronizedSink_var video_sync;
    if (display.get()) {
      video_sync.reset(display->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    }
    ISynchronizedSink_var car_cinema_audio_sync;
    if (car_cinema_audio.get()) {
      car_cinema_audio_sync.reset(car_cinema_audio->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    }

    // get the sink controllers
    if (audio_sync.get()) {
      audio_controller.reset(audio_sync->getController());
      if (no_intra_sync) {
        audio_sync->setSynchronized(false);
      } else {
        audio_sync->setSynchronized(true);
      }
    }
    if (!audio_controller.get()) {
      if (!car_cinema) {
        cerr << "Could not get synchronisation controller for audio! Disabling Inter-Stream-Synchronisation!" << endl;
        no_sync = true;
      }
    }
    if (video_sync.get()) {
      video_controller.reset(video_sync->getController());
      // allow dorping 2 frames (same is set from intra-stream-synchronizer.
      // set it here, to be sure it IS set also when no synchronizer is used!
      video_controller->setMaxDiscardingCount(2);
      if (no_intra_sync) {
        video_sync->setSynchronized(false);
      } else {
        video_sync->setSynchronized(true);
      }
    }
    if (!video_controller.get()) {
      cerr << "Could not get synchronisation controller for video! Disabling Inter-Stream-Synchronisation!" << endl;
      no_sync = true;
    }
    if (car_cinema_audio_sync.get()) {
      car_cinema_audio_controller.reset(car_cinema_audio_sync->getController());
      if (no_intra_sync) {
        car_cinema_audio_sync->setSynchronized(false);
      } else {
        car_cinema_audio_sync->setSynchronized(true);
      }
    }

    // get the Inter-Stream-Synchronizer
    if (!no_sync) {
      msync.reset(mav_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());
      // if we do not have one of the controllers, no_sync would be true, so no problem
      if (audio_controller.get()) {
        msync->addAudioSinkController(audio_controller.get());
      }
      msync->addVideoSinkController(video_controller.get());
      if (car_cinema) {
        if (!car_cinema_audio_controller.get()) {
          cerr << "Could not get synchronisation controller for car_cinema_audio! Disabling Inter-Stream-Synchronisation for it!" << endl;
        } else {
          msync->addAudioSinkController(car_cinema_audio_controller.get());
        }
      }
    }
    /*End< */

    /*Begin> CPU monitoring/probing */
    IResourceCollector_var readfile_resmgr(NULL);
    IResourceCollector_var demux_resmgr(NULL);
    IResourceCollector_var mpg2yuv_resmgr(NULL);
    IResourceCollector_var display_resmgr(NULL);
    IResourceCollector_var ac3_dec_resmgr(NULL);
    IResourceCollector_var mp3_dec_resmgr(NULL);
    IResourceCollector_var audio_play_resmgr(NULL);
    IResourceCollector_var scaler_resmgr(NULL);
    IResourceCollector_var deinterlacer_resmgr(NULL);
    IResourceCollector_var car_cinema_audio_resmgr(NULL);
    IResourceCollector_var car_cinema_dec_resmgr(NULL);
    if (cpu_monitor) {
      if (readfile.get() && readfile->isActivated()) {
        readfile_resmgr.reset(readfile->getParentObject()->getInterface<IResourceCollector>());
        readfile_resmgr->enableResourceProbing();
      }
      if (demux.get() && demux->isActivated()) {
        demux_resmgr.reset(demux->getParentObject()->getInterface<IResourceCollector>());
        demux_resmgr->enableResourceProbing();
      }
      if (mpg2yuv.get() && mpg2yuv->isActivated()) {
        mpg2yuv_resmgr.reset(mpg2yuv->getParentObject()->getInterface<IResourceCollector>());
        mpg2yuv_resmgr->enableResourceProbing();
      }
      if (display.get() && display->isActivated()) {
        display_resmgr.reset(display->getParentObject()->getInterface<IResourceCollector>());
        display_resmgr->enableResourceProbing();
      }
      if (ac3_dec.get() && ac3_dec->isActivated()) {
        ac3_dec_resmgr.reset(ac3_dec->getParentObject()->getInterface<IResourceCollector>());
        ac3_dec_resmgr->enableResourceProbing();
      }
      if (mp3_dec.get() && mp3_dec->isActivated()) {
        mp3_dec_resmgr.reset(mp3_dec->getParentObject()->getInterface<IResourceCollector>());
        mp3_dec_resmgr->enableResourceProbing();
      }
      if (audio_play.get() && audio_play->isActivated()) {
        audio_play_resmgr.reset(audio_play->getParentObject()->getInterface<IResourceCollector>());
        audio_play_resmgr->enableResourceProbing();
      }
      if (scaler.get() && scaler->isActivated()) {
        scaler_resmgr.reset(scaler->getParentObject()->getInterface<IResourceCollector>());
        scaler_resmgr->enableResourceProbing();
      }
      if (deinterlacer.get() && deinterlacer->isActivated()) {
        deinterlacer_resmgr.reset(deinterlacer->getParentObject()->getInterface<IResourceCollector>());
        deinterlacer_resmgr->enableResourceProbing();
      }
      if (car_cinema_audio.get() && car_cinema_audio->isActivated()) {
        car_cinema_audio_resmgr.reset(car_cinema_audio->getParentObject()->getInterface<IResourceCollector>());
        car_cinema_audio_resmgr->enableResourceProbing();
      }
      if (car_cinema_dec.get() && car_cinema_dec->isActivated()) {
        car_cinema_dec_resmgr.reset(car_cinema_dec->getParentObject()->getInterface<IResourceCollector>());
        car_cinema_dec_resmgr->enableResourceProbing();
      }
    }
   /*End< */


    /*Begin> Start nodes */
    if (display->isActivated()) {
      display->start();
    }
    if (scaler->isActivated()) {
      scaler->start();
    }
    if (deinterlacer->isActivated()) {
      deinterlacer->start();
    }
    if (mpg2yuv->isActivated()) {
      mpg2yuv->start();
    }
    if (audio_play->isActivated()) {
      audio_play->start();
    }
    if (mp3_dec->isActivated()) {
      mp3_dec->start();
    }
    if (ac3_dec->isActivated()) {
      ac3_dec->start();
    }
    if (demux->isActivated()) {
      demux->start();
    }
    if (readfile->isActivated()) {
      readfile->start();
    }
    if (car_cinema) {
      if (car_cinema_dec->isActivated()) {
        car_cinema_dec->start();
      }
      if (car_cinema_audio->isActivated()) {
        car_cinema_audio->start();
      }
    }
    /*End< */

    /*Begin> Register event listeners */
    if (!no_video) {
      display->getParentObject()->registerEventListener(ITrack::endTrack_event,
                                                        new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
    }
    if (!no_audio) {
      audio_play->getParentObject()->registerEventListener(ITrack::endTrack_event,
                                                           new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
    }

    if (!no_video) {
      display->getParentObject()->
        registerEventListener(IDisplayEvents::keyReleased_event, new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    }

    receiver.getEventDispatcher().registerEventListener(IDisplayEvents::keyReleased_event,
                                                        new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    /*End< */
    
    /* Begin> Add console producer */
    ConsoleProducer console;
    console.readConfig( getNMMResource("XProducer.xml") );
    console.connectTo(&receiver);
    /* End< Add console producer */

    /*Begin> Event loop */
    console.start();
    IDisplayEventSender_var display_event_sender(display->getParentObject()->getInterface<IDisplayEventSender>());
    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->enableDisplayEvents();
    }
    ec.setVideoDisplay(video_display.get());
    ec.setSynchronizer(msync.get());

    if (cpu_monitor) {
      ec.setCPUMonitor(free_cpu_monitor);
      ec.addResourceCollector(readfile_resmgr.get());
      ec.addResourceCollector(demux_resmgr.get());
      ec.addResourceCollector(mpg2yuv_resmgr.get());
      ec.addResourceCollector(display_resmgr.get());
      ec.addResourceCollector(ac3_dec_resmgr.get());
      ec.addResourceCollector(mp3_dec_resmgr.get());
      ec.addResourceCollector(audio_play_resmgr.get());
      ec.addResourceCollector(scaler_resmgr.get());
      ec.addResourceCollector(deinterlacer_resmgr.get());
      ec.addResourceCollector(car_cinema_audio_resmgr.get());
      ec.addResourceCollector(car_cinema_dec_resmgr.get());
    }
    
    ec.setSeek(file_seek.get());
    if (!no_audio) {
      ec.setAudioStreams(demux_audio_streams);
      ec.setCurrentAudioStream(audio_stream);
      ec.setCarAudioStream(car_cinema_stream);
      INode* demux_audio = NULL;
      if (use_ac3) {
        if (!ac3_passthrough) {
          demux_audio = ac3_dec.get();
        } else {
          demux_audio = audio_play.get();
        }
      } else {
        demux_audio = mp3_dec.get();
      }
      ec.setDemuxAudioConnection(demux.get(), demux_audio);
    }

    ec.handleEvents();

    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->disableDisplayEvents();
    }
    console.stop();
    /*End< */

    GLOBAL_DEBUG_STREAM("readfile: " << readfile->isOutputConnected() << endl);
    GLOBAL_DEBUG_STREAM("demux(" << video_stream << "): " << demux->isOutputConnected(video_stream) << endl);
    GLOBAL_DEBUG_STREAM("demux(" << audio_stream << "): " << demux->isOutputConnected(audio_stream) << endl);
    GLOBAL_DEBUG_STREAM("mpg2yuv: " << mpg2yuv->isOutputConnected() << endl);
    GLOBAL_DEBUG_STREAM("ac3_dec: " << ac3_dec->isOutputConnected() << endl);
    GLOBAL_DEBUG_STREAM("mp3_dec: " << mp3_dec->isOutputConnected() << endl);
    GLOBAL_DEBUG_STREAM("display(out): " << display->isOutputConnected() << endl);
    GLOBAL_DEBUG_STREAM("audio_play(in): " << audio_play->isInputConnected() << endl);
    GLOBAL_DEBUG_STREAM("scaler: " << scaler->isOutputConnected() << endl);
    GLOBAL_DEBUG_STREAM("deinterlacer: " << deinterlacer->isOutputConnected() << endl);

  } catch(const Exception& e) {

    cerr << e << endl;
    error_level = 1;

  } catch(...) {

    cerr << "Something strange went wrong" << endl;
    return 2;

  }

  if (cpu_monitor) {
    free_cpu_monitor->stop();
    delete free_cpu_monitor;
  }
  /*Begin> Cleanup */
  try {
    
    /*Begin> Stop nodes */
    if (display.get() && display->isStarted()) {
      display->stop();
    }
    if (scaler.get() && scaler->isStarted()) {
      scaler->stop();
    }
    if (deinterlacer.get() && deinterlacer->isStarted()) {
      deinterlacer->stop();
    }
    if (mpg2yuv.get() && mpg2yuv->isStarted()) {
      mpg2yuv->stop();
    }
    if (audio_play.get() && audio_play->isStarted()) {
      audio_play->stop();
    }
    if (mp3_dec.get() && mp3_dec->isStarted()) {
      mp3_dec->stop();
    }
    if (ac3_dec.get() && ac3_dec->isStarted()) {
      ac3_dec->stop();
    }
    if (demux.get() && demux->isStarted()) {
      demux->stop();
    }
    if (readfile.get() && readfile->isStarted()) {
      readfile->stop();
    }
    if (car_cinema_audio.get() && car_cinema_audio->isStarted()) {
      car_cinema_audio->stop();
    }
    if (car_cinema_dec.get() && car_cinema_dec->isStarted()) {
      car_cinema_dec->stop();
    }
    /*End< */
    
    /*Begin> Flush nodes */
    if (display.get() && display->isActivated()) {
      display->flush();
    }
    if (scaler.get() && scaler->isActivated()) {
      scaler->flush();
    }
    if (deinterlacer.get() && deinterlacer->isActivated()) {
      deinterlacer->flush();
    }
    if (mpg2yuv.get() && mpg2yuv->isActivated()) {
      mpg2yuv->flush();
    }
    if (audio_play.get() && audio_play->isActivated()) {
      audio_play->flush();
    }
    if (mp3_dec.get() && mp3_dec->isActivated()) {
      mp3_dec->flush();
    }
    if (ac3_dec.get() && ac3_dec->isActivated()) {
      ac3_dec->flush();
    }
    if (demux.get() && demux->isActivated()) {
      demux->flush();
    }
    if (readfile.get() && readfile->isActivated()) {
      readfile->flush();
    }
    if (car_cinema_audio.get() && car_cinema_audio->isActivated()) {
      car_cinema_audio->flush();
    }
    if (car_cinema_dec.get() && car_cinema_dec->isActivated()) {
      car_cinema_dec->flush();
    }
    /*End< */

    /*Begin> Disconnect nodes */
    if (display.get() && display->isActivated()) {
      display->deactivate();
    }
    if (display.get() && display->isOutputInitialized()) {
      display->deinitOutput();
    }
    {
      if (display.get() && display->isInputConnected()) {
        if (do_scale) {
          if (scaler.get() && scaler->isActivated()) {
            scaler->deactivate();
          }
        } else if (do_deinterlace) {
          if (deinterlacer.get() && deinterlacer->isActivated()) {
            deinterlacer->deactivate();
          }
        } else {
          if (mpg2yuv.get() && mpg2yuv->isActivated()) {
            mpg2yuv->deactivate();
          }
        }
        disconnect_input(display);
      }
    }
    if (do_scale) {
      if (scaler.get() && scaler->isActivated()) {
        scaler->deactivate();
      }
      if (scaler.get() && scaler->isOutputInitialized()) {
        scaler->deinitOutput();
      }
      {
        if (scaler.get() && scaler->isInputConnected()) {
          if (do_deinterlace) {
            if (deinterlacer.get() && deinterlacer->isActivated()) {
              deinterlacer->deactivate();
            }
          } else {
            if (mpg2yuv.get() && mpg2yuv->isActivated()) {
              mpg2yuv->deactivate();
            }
          }
          disconnect_input(scaler);
        }
      }
    }
    if (do_deinterlace) {
      if (deinterlacer.get() && deinterlacer->isActivated()) {
        deinterlacer->deactivate();
      }
      if (deinterlacer.get() && deinterlacer->isOutputInitialized()) {
        deinterlacer->deinitOutput();
      }
      {
        if (deinterlacer.get() && deinterlacer->isInputConnected()) {
          if (mpg2yuv.get() && mpg2yuv->isActivated()) {
            mpg2yuv->deactivate();
          }
          disconnect_input(deinterlacer);
        }
      }
    }
    if (mpg2yuv.get() && mpg2yuv->isActivated()) {
      mpg2yuv->deactivate();
    }
    if (mpg2yuv.get() && mpg2yuv->isOutputInitialized()) {
      mpg2yuv->deinitOutput();

      if (mpg2yuv.get() && mpg2yuv->isInputConnected()) {
        INode_var pred(mpg2yuv->getPredecessorNode("default"));
        if (pred.get() && pred->isActivated()) {
          pred->deactivate();
        }
        disconnect_input(mpg2yuv);
      }
    }
    if (audio_play.get() && audio_play->isActivated()) {
      audio_play->deactivate();
    }
    if (car_cinema_audio.get() && car_cinema_audio->isActivated()) {
      car_cinema_audio->deactivate();
    }
    if (audio_play.get() && audio_play->isOutputInitialized()) {
      audio_play->deinitOutput();

      if (audio_play.get() && audio_play->isInputConnected()) {
        if (use_ac3) {
          if (ac3_passthrough) {
            if (demux.get() && demux->isActivated()) {
              demux->deactivate();
            }
          } else {
            if (ac3_dec.get() && ac3_dec->isActivated()) {
              ac3_dec->deactivate();
            }
          }
        } else {
          if (mp3_dec.get() && mp3_dec->isActivated()) {
            mp3_dec->deactivate();
          }
        }
        disconnect_input(audio_play);
      }
    }
    if (car_cinema_audio.get() && car_cinema_audio->isOutputInitialized()) {
      car_cinema_audio->deinitOutput();

      if (car_cinema_audio.get() && car_cinema_audio->isInputConnected()) {
        if (car_cinema_dec.get() && car_cinema_dec->isActivated()) {
          car_cinema_dec->deactivate();
        }
        disconnect_input(car_cinema_audio);
      }
    }
    if (car_cinema_dec.get() && car_cinema_dec->isActivated()) {
      car_cinema_dec->deactivate();
    }
    if (car_cinema_dec.get() && car_cinema_dec->isOutputInitialized()) {
      car_cinema_dec->deinitOutput();

      if (car_cinema_dec.get() && car_cinema_dec->isInputConnected()) {
        if (demux.get() && demux->isActivated()) {
          demux->deactivate();
        }
        disconnect_input(car_cinema_dec);
      }
    }
    if (use_ac3) {
      if (!ac3_passthrough) {
        if (ac3_dec.get() && ac3_dec->isActivated()) {
          ac3_dec->deactivate();
        }
        if (ac3_dec.get() && ac3_dec->isOutputInitialized()) {
          ac3_dec->deinitOutput();

          if (ac3_dec.get() && ac3_dec->isInputConnected()) {
            if (demux.get() && demux->isActivated()) {
              demux->deactivate();
            }
            disconnect_input(ac3_dec);
          }
        }
      }
    } else {
      if (mp3_dec.get() && mp3_dec->isActivated()) {
        mp3_dec->deactivate();
      }
      if (mp3_dec.get() && mp3_dec->isOutputInitialized()) {
        mp3_dec->deinitOutput();

        if (mp3_dec.get() && mp3_dec->isInputConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(mp3_dec);
        }
      }
    }
    if (demux.get() && demux->isActivated()) {
      demux->deactivate();
    }
    if (demux.get() && demux->isOutputInitialized()) {
      demux->deinitOutput();

      if (demux.get() && demux->isInputConnected()) {
        if (readfile.get() && readfile->isActivated()) {
          readfile->deactivate();
        }
        disconnect_input(demux);
      }
    }
    if (readfile.get() && readfile->isActivated()) {
      readfile->deactivate();
    }
    if (readfile.get() && readfile->isOutputInitialized()) {
      readfile->deinitOutput();
    }
    /*End< Disconect nodes */

    /*Begin> Deinit nodes */
    if (display.get() && display->isInitialized()) {
      display->deinit();
    }
    if (scaler.get() && scaler->isInitialized()) {
      scaler->deinit();
    }
    if (deinterlacer.get() && deinterlacer->isInitialized()) {
      deinterlacer->deinit();
    }
    if (mpg2yuv.get() && mpg2yuv->isInitialized()) {
      mpg2yuv->deinit();
    }
    if (audio_play.get() && audio_play->isInitialized()) {
      audio_play->deinit();
    }
    if (car_cinema_audio.get() && car_cinema_audio->isInitialized()) {
      car_cinema_audio->deinit();
    }
    if (car_cinema_dec.get() && car_cinema_dec->isInitialized()) {
      car_cinema_dec->deinit();
    }
    if (mp3_dec.get() && mp3_dec->isInitialized()) {
      mp3_dec->deinit();
    }
    if (ac3_dec.get() && ac3_dec->isInitialized()) {
      ac3_dec->deinit();
    }
    if (demux.get() && demux->isInitialized()) {
      demux->deinit();
    }
    if (readfile.get() && readfile->isInitialized()) {
      readfile->deinit();
    }
    /*End< */

    /*Begin> Release nodes */
    if (app != NULL) {
      if (readfile.get()) {
        app->getRegistry().releaseNode(*readfile);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (mpg2yuv.get()) {
        app->getRegistry().releaseNode(*mpg2yuv);
      }
      if (ac3_dec.get()) {
        app->getRegistry().releaseNode(*ac3_dec);
      }
      if (mp3_dec.get()) {
        app->getRegistry().releaseNode(*mp3_dec);
      }
      if (display.get()) {
        app->getRegistry().releaseNode(*display);
      }
      if (deinterlacer.get()) {
        app->getRegistry().releaseNode(*deinterlacer);
      }
      if (scaler.get()) {
        app->getRegistry().releaseNode(*scaler);
      }
      if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
      }
      if (car_cinema_audio.get()) {
        app->getRegistry().releaseNode(*car_cinema_audio);
      }
      if (car_cinema_dec.get()) {
        app->getRegistry().releaseNode(*car_cinema_dec);
      }
    }
    /*End< */

  } catch(const Exception& e) {

    cerr << "Exception during cleanup" << endl;
    cerr << e << endl;

    return 2;

  } catch(...) {

    cerr << "Something strange went wrong during cleanup" << endl;
    return 2;

  }

  return error_level;
}

