/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

// simple mp3 player creating flow graph manually with error handling and access to 'serverregistry'

#include <iostream>

// include NMM 
#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/base/ProxyApplication.hpp"

// include used interfaces
#include "nmm/interfaces/file/IFileHandler.hpp"

// import namespace
using namespace NMM;

int main(int argc, char ** argv)
{
  if(argc != 2) {
    cerr << "Usage:" << endl
	 << "  " << argv[0] << " <mp3file>" << endl
	 << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see http://graphics.cs.uni-sb.de/NMM/current/Docs/messages/index.html
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the NMM application for the example
  NMMApplication* app = 0;

  // the _var variables for the INodes (C++ auto_ptr<>)
  INode_var readfile;
  INode_var decoder;
  INode_var audioplay;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create the descriptions for the GenericReadNode, MPEGAudioDecodeNode, and PlaybackNode
    NodeDescription readfile_descr("GenericReadNode");
    NodeDescription decoder_descr("MPEGAudioDecodeNode");
    NodeDescription audioplay_descr("PlaybackNode");

    // request Nodes: 
    // lock registry with guard, init request for each node, instantiate first result for every request 
    cerr << "request Nodes from registry" << endl;
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);

      list<Response> readfile_response = registry.initRequest(readfile_descr);
      list<Response> decoder_response = registry.initRequest(decoder_descr);
      list<Response> audioplay_response = registry.initRequest(audioplay_descr);

      if (readfile_response.empty() || decoder_response.empty() || audioplay_response.empty() ) {
	cerr << "size of readfile_response  " << readfile_response.size() << endl;
	cerr << "size of decoder_response   " << decoder_response.size() << endl;
	cerr << "size of audioplay_response " << audioplay_response.size() << endl;
	throw Exception("a required node is not available");
      }

      readfile.reset(registry.requestNode(readfile_response.front()));
      decoder.reset(registry.requestNode(decoder_response.front()));
      audioplay.reset(registry.requestNode(audioplay_response.front()));
    } // end of scope for lock -> will release lock

    // set the filename by requesting appropriate interface
    cerr << "set filename to " << argv[1] << endl;
    IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler->setFilename(argv[1]);

    // create a flow graph: readfile -> decoder -> audioplay

    // initialize all three nodes ...
    readfile->init();
    decoder->init();
    audioplay->init();

    // initialize output of source node
    readfile->initOutput();

    // now, output of source node can be connected to initialized input of decoder
    cerr << "connect(readfile, decoder)" << endl;
    connect(readfile, decoder);

    // activate source node so that decoder can initialize its output
    readfile->activate();
    decoder->initOutput();

    // now, output of decoder node can be connected to initialized input of audio sink
    cerr << "connect(decoder, audioplay)" << endl;
    connect(decoder, audioplay);

    // activate decoder node so that audio sink can initialize its output
    decoder->activate();
    audioplay->initOutput();

    // audio sink can be activated now
    audioplay->activate();

    // ... and finally start them!
    cerr << "start() all Nodes" << endl;
    readfile->start();
    decoder->start();
    audioplay->start();
    
    // wait ...
    cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    
    // stop nodes, reset and destroy nodes
    cerr << "stop(), flush(), deactivate(), deinitOutput(), deinit() all Nodes" << endl;
    readfile->stop();
    decoder->stop();
    audioplay->stop();

    readfile->flush();
    decoder->flush();
    audioplay->flush();

    readfile->deactivate();
    decoder->deactivate();
    audioplay->deactivate();

    readfile->deinitOutput();
    decoder->deinitOutput();
    audioplay->deinitOutput();

    readfile->deinit();
    decoder->deinit();
    audioplay->deinit();

    // Release the reserved nodes
    cerr << "release all Nodes" << endl;
    registry.releaseNode(*readfile);
    registry.releaseNode(*decoder);
    registry.releaseNode(*audioplay);

    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    // try to release the reserved nodes
    cerr << "release all Nodes" << endl;
    ClientRegistry& registry = app->getRegistry();
    RegistryLock lock(registry);
    if(readfile.get()) {
      registry.releaseNode(*readfile);
    }
    if(decoder.get()) {
      registry.releaseNode(*decoder);
    }
    if(audioplay.get()) {
      registry.releaseNode(*audioplay);
    }

    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    // try to release the reserved nodes
    cerr << "release all Nodes" << endl;
    ClientRegistry& registry = app->getRegistry();
    RegistryLock lock(registry);
    if(readfile.get()) {
      registry.releaseNode(*readfile);
    }
    if(decoder.get()) {
      registry.releaseNode(*decoder);
    }
    if(audioplay.get()) {
      registry.releaseNode(*audioplay);
    }

    return 1;
  }
}
