/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <getopt.h>
#include <iostream>

#include "nmm/base/connect.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INamed.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/base/IFormatChange.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"
#include "nmm/interfaces/base/IInstream.hpp"
#include "nmm/interfaces/file/IBufferSize.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"

// include the registry
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/utils/utils.hpp"

// console producer
#include "nmm/misc/console/ConsoleProducer.hpp"

// import namespace
using namespace NMM;

/**
 * Class for receiving events.
 */
class ConsoleEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  ConsoleEventReceiver() {
    dispatcher = new EventDispatcher();
  }
  
  virtual ~ConsoleEventReceiver() {
    delete dispatcher;
  }
  
  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};

class AppEventReceiver : public IEventReceiver {
public:
  string name;

  virtual ~AppEventReceiver() { }

  virtual Result receiveEvent(const CEvent& ce, const IEventSender* = 0 )  {
    for (unsigned int i = 0; i < ce.size(); ++i) {
      cerr << "### " << name << " " << ce[i].getKey() << endl;
    }
    return SUCCESS;
  }

  virtual Result receiveEvent(const Event& event, const IEventSender* = 0 )    {
    cerr << "### "  << name << " " << event.getKey() << endl;
    return SUCCESS;
  }

  virtual Result receiveEvent(const CEvent& ce, const IEventSender* = 0 ) const {
    for (unsigned int i = 0; i < ce.size(); ++i) {
      cerr << "### "  << name << " " << ce[i].getKey() << endl;
    }
    return SUCCESS;
  }

  virtual Result receiveEvent(const Event& event, const IEventSender* = 0 )  const {
    cerr << "### "  << name << " " << event.getKey() << endl;
    return SUCCESS;
  }
};

/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  
  INode* readfile;
  INode* audio_play;
  IFileHandler* file_source;
  
  vector<string> files;
  int curr_file_num;
  int max_file_num;
  
  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }

public:
  EventCatcher() 
    : exit(false),
      exit_mutex(),
      exit_cv(exit_mutex),
      readfile(NULL),
      audio_play(NULL),
      file_source(NULL),
      curr_file_num(1),
      max_file_num(-1)
  {}
  
  void setFileReader(INode* readfile, IFileHandler* file_source) {
    this->readfile = readfile;
    this->file_source = file_source;
  }
  
  void setAudioPlay(INode* audio_play) {
    this->audio_play = audio_play;
  }
  
  void addFile(string fname) {
    files.push_back(fname);
    ++max_file_num;
  }
  
  Result setFormat1(Format& )
  {
    cerr << "### NEW FORMAT 1" << endl;
    return SUCCESS;
  }

  Result setFormat2(Format& )
  {
    cerr << "### NEW FORMAT 2" << endl;
    return SUCCESS;
  }

  Result setFormat3(Format& )
  {
    cerr << "### NEW FORMAT 3" << endl;
    return SUCCESS;
  }
  
  Result keyReleased(string& key) {
    if (key == "KEY_q") {
      signalExit();
    } else if (key == "KEY_n") {
      if(curr_file_num > max_file_num) {
        curr_file_num = 0;
      }
      readfile->stop();
      readfile->deactivate();
      readfile->deinitOutput();
      file_source->setFilename(files[curr_file_num++]);
      readfile->initOutput();
      readfile->activate();
      readfile->start();
      cerr << curr_file_num << ": setFilename " << files[curr_file_num-1] << endl;
    } else if (key == "KEY_d") {
      cerr << "Send Instream Messages d" << endl;
      IInstream_var i_instream(readfile->getParentObject()->getCheckedInterface<IInstream>());
      i_instream->sendMessageDownstream(CEvent(Event("DownstreamEvent")));
    } else if (key == "KEY_u") {
      cerr << "Send Instream Messages u" << endl;
      IInstream_var i_instream(audio_play->getParentObject()->getCheckedInterface<IInstream>());
      i_instream->sendMessageUpstream(CEvent(Event("UpstreamEvent")));
    } else {
      GLOBAL_ERROR_STREAM("Unsupported key pressed: " << key << endl);
    }
    
    return SUCCESS;
  }
  
  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }
};


void usage(const char* name) {
  cerr << "Usage:" << endl
       << "  " << name << " [options] <mp3files>" << endl
       << endl
       << "Options:" << endl
       << "  -b : Try to set buffer size of source node to specified value (in Bytes)" << endl
       << "  -v : increase verbositiy (may be given multiple times)" << endl
       << "  -h : Show this help" << endl
       << endl;
}

int main(int argc, char ** argv)
{
  /* Begin> command line parsing */
  string source_node("GenericReadNode");
  int verbose(0);
  int buffer_size(-1);

  const char *opt_string = "b:tvh";
  struct option options [] = {
    {"buffer-size", 1,NULL,'b'},
    {"verbose",     0,NULL,'v'},
    {"help",        0,NULL,'h'},
    {NULL,          0,NULL,0}
  };
  int ch,long_option_index;
  
  while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
    // which option did we get?
    switch (ch) {
    case 'b':
      buffer_size = atoi(optarg);
      if (buffer_size <= 0) {
        cerr << "Buffer size " << buffer_size << " not valid!" << endl;
        buffer_size = -1;
      }
      break;
    case 'v':
      verbose++;
      break;
    case 'h':
    default:
      usage(argv[0]);
      return 1;
    }
  }

  // check for filename
  if ( argc - optind < 1 ) {
    usage(argv[0]);
    return 1;
  }
  
  vector<string> files(argc - optind);
  for (unsigned int i = 0; i < files.size(); i++) {
    files[i] = string(argv[i + optind]);
  }
  
  /* End< command line parsing */

  /* handle *Stream output */
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  /* the NMM application for the example, intialized below */
  NMMApplication* app = 0;

  /* Allocate _var variables for the INode's*/
  INode_var readfile;
  INode_var mp3dec;
  INode_var audio_play;

  /* Allocate an event listener */
  EventCatcher ec;
  AppEventReceiver er1;
  er1.name = "AppEventReceiver 1";
  AppEventReceiver er2;
  er2.name = "AppEventReceiver 2";
  ConsoleEventReceiver cer;
  
  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    app = ProxyApplication::getApplication(argc, argv);
    
    /*Begin> Create the 1. request for the MP3ReadNode*/
    NodeDescription request1;
    request1.setNodeName(source_node);
    /*End< */

    /*Begin> Create the 2. request for the MPEGAudioDecodeNode*/
    NodeDescription request2;
    request2.setNodeName("MPEGAudioDecodeNode");
    /*End< */

    /*Begin> Create the 3. request for the PlaybackNode*/
    NodeDescription request3;
    request3.setNodeName("PlaybackNode");
    /*End< */


    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    registry.initTransaction();

    list<Response> response1 = registry.initRequest(request1);
    list<Response> response2 = registry.initRequest(request2);
    list<Response> response3 = registry.initRequest(request3);


    if (response1.empty() || response2.empty() || response3.empty() ) {
      GLOBAL_ERROR_STREAM("Size of 1. response " << response1.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 2. response " << response2.size() << endl);
      GLOBAL_ERROR_STREAM("Size of 3. response " << response3.size() << endl);
      throw Exception("A required node is not available");
    }

    /* Begin> request the nodes */
    readfile.reset(registry.requestNode(response1.front()));
    mp3dec.reset(registry.requestNode(response2.front()));
    audio_play.reset(registry.requestNode(response3.front()));
    registry.releaseTransaction();
    /* End< */

    /* test for buffer manager */
    IBufferManager_var bm1(readfile->getParentObject()->getCheckedInterface<IBufferManager>());
    IBufferManager_var bm2(mp3dec->getParentObject()->getCheckedInterface<IBufferManager>());
    bm2->useBufferManager(*bm1.get());

    /* Register EventListener */
    /* register TEDObjects for setFormat events */
    
    readfile->getParentObject()->registerEventListener(IFormatChange::setFormat_event, 
						       new TEDObject1<EventCatcher, Format>(&ec, &EventCatcher::setFormat1));
    mp3dec->getParentObject()->registerEventListener(IFormatChange::setFormat_event, 
						     new TEDObject1<EventCatcher, Format>(&ec, &EventCatcher::setFormat2));
    audio_play->getParentObject()->registerEventListener(IFormatChange::setFormat_event, 
							 new TEDObject1<EventCatcher, Format>(&ec, &EventCatcher::setFormat3));
    
    /* register IEventReceiver for upstream, downstream dummy events */
    readfile->getParentObject()->registerEventListener("DownstreamEvent", &er1);
    mp3dec->getParentObject()->registerEventListener("DownstreamEvent", &er1);
    audio_play->getParentObject()->registerEventListener("DownstreamEvent", &er1);

    readfile->getParentObject()->registerEventListener("UpstreamEvent", &er1);
    mp3dec->getParentObject()->registerEventListener("UpstreamEvent", &er1);
    audio_play->getParentObject()->registerEventListener("UpstreamEvent", &er1);

    /* register IEventReceiver for all events */
    // readfile->getParentObject()->registerEventListener(&er2);
    // mp3dec->getParentObject()->registerEventListener(&er2);
    // audio_play->getParentObject()->registerEventListener(&er2);
    
    /* set name for audio_play */
    INamed_var named_audio_play(audio_play->getParentObject()->getCheckedInterface<INamed>());
    named_audio_play->setName("AudioPlay");

    /* set queue size for mp3dec and audio_play*/
    IBiDirQueuedNode_var queued_mp3dec(mp3dec->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    queued_mp3dec->setDownstreamMaxSize(1);
    IBiDirQueuedNode_var queued_audio_play(audio_play->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    queued_audio_play->setDownstreamMaxSize(1);

    /* get and set buffer size of readfile */
    IBufferSize_var bufsz_readfile(readfile->getParentObject()->getInterface<IBufferSize>());
    if (bufsz_readfile.get()) {
      cerr << "Current buffer size of source node is " << bufsz_readfile->getBufferSize() << "." << endl;
      if (buffer_size != -1) {
        bufsz_readfile->setBufferSize(buffer_size);
        cerr << "Set buffer size of source node to " << bufsz_readfile->getBufferSize() << "." << endl;
      }
    } else {
      cerr << "Getting/Setting buffer size of source node not possible!" << endl;
    }
    /* initialize and configure the nodes */
    readfile->init();
    Property_var property(readfile->getOutputProperty());

    IFileHandler_var file_source (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(files[0]);
    cerr << 1 << ": setFilename " << files[0] << endl;

    mp3dec->init();
    audio_play->init();

    readfile->initOutput();
    /* Connect the nodes (this returns a CommunicationChannel*), this automatically
     * sets the transport strategy, so you only need the communication channel in
     * case you want to change something lateron
     */
    connect(readfile, mp3dec);

    readfile->activate();

    mp3dec->initOutput();
    /* and (nearly) the same procedure for the mp3dec and audio_play node 
     * only in this case c_connect is called and the comm.channel is setup
     * manually by choosing the first strategy
     */
    CommunicationChannel_var cc(c_connect(mp3dec, audio_play));
    if(cc.get() != 0) {
      cc->setTransportStrategy(cc->getStrategyList().front());
    }    

    mp3dec->activate();

    audio_play->initOutput();
    audio_play->activate();

    // ... and finally start them
    readfile->start();
    mp3dec->start();
    audio_play->start();
    
    /* Begin> Add console producer */
    ConsoleProducer console;
    console.readConfig( getNMMResource("XProducer.xml") );
    console.connectTo(&cer);
    console.start();
    cer.getEventDispatcher().registerEventListener(IDisplayEvents::keyReleased_event,
                                                   new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    /* End< Add console producer */
    
    /* Begin> Main loop */
    ec.setFileReader(readfile.get(), file_source.get());
    ec.setAudioPlay(audio_play.get());
    for (unsigned int i = 0; i < files.size(); i++) {
      ec.addFile(files[i]);
    }
    ec.handleEvents();
    /* End< Main loop */
    
    // stop nodes, reset and destroy them
    GLOBAL_DEBUG_STREAM("Stopping all Nodes" << endl);
    readfile->stop();
    mp3dec->stop();
    audio_play->stop();

    GLOBAL_DEBUG_STREAM("Flushing all Nodes" << endl);
    readfile->flush();
    mp3dec->flush();
    audio_play->flush();

    GLOBAL_DEBUG_STREAM("Deactivating all Nodes" << endl);
    readfile->deactivate();
    mp3dec->deactivate();
    audio_play->deactivate();

    GLOBAL_DEBUG_STREAM("Deiniting-Output all Nodes" << endl);
    readfile->deinitOutput();
    mp3dec->deinitOutput();
    audio_play->deinitOutput();

    GLOBAL_DEBUG_STREAM("Deinitializing all Nodes" << endl);
    readfile->deinit();
    mp3dec->deinit();
    audio_play->deinit();

    /*Release the reserved nodes*/
    GLOBAL_DEBUG_STREAM("Releasing all Nodes" << endl);
    registry.releaseNode(*readfile);
    registry.releaseNode(*mp3dec);
    registry.releaseNode(*audio_play);
  }
  // catch NMM-exception and print it
  catch(Exception e) {
    /* release the tranaction */
    app->getRegistry().releaseTransaction();

    /*Release the reserved nodes*/
    if (readfile.get())
      app->getRegistry().releaseNode(*readfile);
    if (mp3dec.get())
      app->getRegistry().releaseNode(*mp3dec);
    if (audio_play.get())
      app->getRegistry().releaseNode(*audio_play);
    cerr << e << endl;
    return 1;
  }
  // any other exception
  catch(...) {
    /* release the tranaction */
    app->getRegistry().releaseTransaction();

    if (readfile.get())
      app->getRegistry().releaseNode(*readfile);
    if (mp3dec.get())
      app->getRegistry().releaseNode(*mp3dec);
    if (audio_play.get())
      app->getRegistry().releaseNode(*audio_play);

    cerr << "Something went wrong..." << endl;
    return 1;
  }

  return 0;
}
