/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// general includes
#include "nmm/base/connect.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/ProxyObject.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/general/ICopy.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/audio/visualization/IScope.hpp"

// sync
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <iostream>

using namespace NMM;

bool blur_on = false;
bool verbose = false;
bool spectrum = false;
int blur_step = 10;
string filename;
double framerate = 0;

vector<INode*> nodes;

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}

void startNodes() {
  GLOBAL_DEBUG_STREAM("Starting nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    if (nodes[i]->isActivated()) {
      nodes[i]->start();
    }
  }
}

void stopNodes() {
  GLOBAL_DEBUG_STREAM("Stopping nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    if (nodes[i]->isStarted()) {
      nodes[i]->stop();
    }
  }
}

void flushNodes() {
  GLOBAL_DEBUG_STREAM("Flushing nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    if (nodes[i]->isActivated() || nodes[i]->isStarted()) {
      nodes[i]->flush();
    }
  }
}

void deactivateNodes() {
  GLOBAL_DEBUG_STREAM("Deactivating nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    if (nodes[i]->isActivated()) {
      nodes[i]->deactivate();
    }
  }
}

void deinitOutputNodes() {
  GLOBAL_DEBUG_STREAM("Output-deinitializing nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    if (nodes[i]->isOutputInitialized()) {
      nodes[i]->deinitOutput();
    }
  }
}

void deinitNodes() {
  GLOBAL_DEBUG_STREAM("Deinitializing nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    if (nodes[i]->isInitialized()) {
      nodes[i]->deinit();
    }
  }
}

void releaseNodes(ClientRegistry& registry) {
  GLOBAL_DEBUG_STREAM("Releasing nodes..." << endl);
  for (unsigned int i=0; i<nodes.size(); ++i) {
    registry.releaseNode(*nodes[i]);
    delete nodes[i];
  }
}

void cleanup(ClientRegistry& registry) {
  GLOBAL_DEBUG_STREAM("Cleaning up..." << endl);
  stopNodes();
  flushNodes();
  deactivateNodes();
  deinitOutputNodes();
  deinitNodes();
  releaseNodes(registry);
  
  // just in case
  registry.releaseTransaction();
}

int main(int argc, char ** argv) {
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);
     
  ClientRegistry& registry = app->getRegistry();
  
  bool pause = false;
  
  try {
    
    cout << "Press 'p' + <Enter> to pause, 'p' + <Enter> to continue." << endl;
    cout << "Press 'q' + <Enter> to quit." << endl;
    
    if (argc < 2) {
      GLOBAL_ERROR_STREAM("Not enough arguments..." << endl);
      GLOBAL_ERROR_STREAM("Usage: mp3vis <file> [-blur <step>] [-framerate <rate>]" 
                          << endl);
      exit(1);
    }
  
    if (argc > 6) {
      GLOBAL_ERROR_STREAM("Too many arguments... " << endl);
      GLOBAL_ERROR_STREAM("Usage: mp3vis <file> [-blur <step>] [-framerate <rate>]" 
                          << endl);
      exit(1);
    }
  
    framerate = 25.0;
    blur_step = 10;
    
    if (argc >= 3) {
      for(int i=2; i < argc; i+=1) {
        if (!strcmp(argv[i], "-blur")) {
          blur_on = true;
          GLOBAL_MESSAGE_STREAM("Enable blur effect." << endl);
          blur_step = atoi(argv[++i]);
          if(blur_step > 0) {
            GLOBAL_MESSAGE_STREAM("Blur step is set to " << blur_step << "." << endl);
          } else {
            GLOBAL_MESSAGE_STREAM("Take the default blur step (10)." << endl);
          }
        }
      
        if (!strcmp(argv[i], "-framerate")) {
          framerate = atof(argv[++i]);
          if(framerate > 0) {
            GLOBAL_MESSAGE_STREAM("Framerate is set to " << framerate << "." << endl);
          } else {
            GLOBAL_MESSAGE_STREAM("Take the default framerate (25)." << endl);
            framerate = 25.0;
          }
        }
        
        if (!strcmp(argv[i], "-verbose")) {
          verbose = true;
        }
        
        if (!strcmp(argv[i], "-spectrum")) {
          spectrum = true;
        }
      }
    }
    
    filename = argv[1];
    
    if (!verbose) {
      turnOffAllMessages();
    }
    
    // create requests for nodes
    vector<NodeDescription> request(6);
    request[0].setNodeName("MP3ReadNode");
    request[1].setNodeName("MPEGAudioDecodeNode");
    request[2].setNodeName("CopyNode");
    request[3].setNodeName("PlaybackNode");
    request[4].setNodeName(spectrum ? "SAnalyzerNode" : "ScopeNode");
    request[5].setNodeName("XDisplayNode");
    
    // lock the IRegistry and request nodes
    registry.initTransaction();
    vector< list<Response> > response(request.size());
    for (unsigned int i=0; i<request.size(); ++i) {
      response[i] = registry.initRequest(request[i]);
      if (response[i].empty()) {
        cerr << "Failed to request required node " << request[i].getNodeName() << endl;
        cleanup(registry);
        return 1;
      } else {
        INode* node = registry.requestNode(response[i]);
        nodes.push_back(node);
        INamed_var named(node->getParentObject()->getCheckedInterface<INamed>());
        named->setName(request[i].getNodeName());
      }
    }
    
    // unlock the registry
    registry.releaseTransaction();
    
    // get nodes
    INode* readfile = nodes[0];
    INode* mp3dec = nodes[1];
    INode* copy = nodes[2];
    INode* audio_play = nodes[3];
    INode* scope = nodes[4];
    INode* display = nodes[5];
    
    IBiDirQueuedNode_var queued_audio_play(audio_play->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    queued_audio_play->setDownstreamMaxSize(5);
    IBiDirQueuedNode_var queued_scope(scope->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    queued_scope->setDownstreamMaxSize(5);
    IBiDirQueuedNode_var queued_display(display->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
    queued_display->setDownstreamMaxSize(1);

    // init nodes
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    readfile->init();
    
    mp3dec->init();
    
    ICopy_var icopy(copy->getParentObject()->getCheckedInterface<ICopy>());
    icopy->setNumberOfOutputs(2);
    copy->init();
    
    IAudioDevice_var audio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
    audio_device->setDevice("/dev/dsp");
    audio_play->init();
    
    scope->init();
    IScope_var iscope(scope->getParentObject()->getInterface<IScope>());
    if (iscope.get() != NULL) {
      iscope->setBlurEnabled(blur_on);
      iscope->setBlurStep(blur_step);
    }
    
    display->init();
    
    // setup synchronization
    MultiAudioVideoSynchronizer av_sync;
    IMultiAudioVideoSynchronizer_var sync(av_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());
    
    ISynchronizedSink_var audio_sync(audio_play->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    ISynchronizedSink_var video_sync(display->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    ISinkController_var audio_controller(audio_sync->getController());
    ISinkController_var video_controller(video_sync->getController());
    sync->addAudioSinkController(audio_controller.get());
    sync->addVideoSinkController(video_controller.get());
    
    // output format of visualizer
    Format *video_format = new Format("video/raw");
    video_format->addIntParamValue(Format::x_resolution_param, 400);
    video_format->addIntParamValue(Format::y_resolution_param, 200);
    video_format->addIntParamValue(Format::ratio_x_param, 400);
    video_format->addIntParamValue(Format::ratio_y_param, 200);
    video_format->addFloatParamValue(Format::framerate_param, framerate);
    video_format->addStringParamValue(Format::colorspace_param, "rgb24");
    video_format->addStringParamValue(Format::channelorder_param, "rgb");
    video_format->addStringParamValue(Format::format_param, "packet");
    video_format->addStringParamValue(Format::endian_param, "little");
    video_format->addIntParamValue(Format::bitperpixel_param, 32);
    
    // connect the nodes
    readfile->initOutput();
    
    GLOBAL_DEBUG_STREAM("Trying to connect readfile to mp3dec" << endl);
    connect(readfile, mp3dec);
    
    readfile->activate();
    mp3dec->initOutput();
    
    GLOBAL_DEBUG_STREAM("Trying to connect mp3dec to copy" << endl);
    connect(mp3dec, copy);
    
    mp3dec->activate();
    copy->initOutput();
    
//      Property_var property(copy->getWorkingOutputProperty());
//      const Format* format = property->getDefaultFormat("1");
//      cerr << "begin show format" << endl;
//      if (format == NULL) {
//        cerr << "no format" << endl;
//      } else {
//        cerr << format->getIntValue(Format::channels_param) << endl;
//        cerr << format->getIntValue(Format::bitpersample_param) << endl;
//        cerr << format->getIntValue(Format::sample_rate_param) << endl;
//        cerr << format->getStringValue(Format::encoding_param) << endl;
//        cerr << format->getStringValue(Format::sign_param) << endl;
//        cerr << format->getStringValue(Format::endian_param) << endl;
//      }
//      cerr << "end show format" << endl;
    
    GLOBAL_DEBUG_STREAM("Trying to connect copy:0 to audio_play" << endl);
    connect(copy, "0", audio_play);
    GLOBAL_DEBUG_STREAM("Trying to connect copy:1 to scope" << endl);
    connect(copy, "1", scope);
    
    copy->activate();
    audio_play->initOutput();
    audio_play->activate();
    
    scope->initOutput();
    
    GLOBAL_DEBUG_STREAM("Trying to connect scope to display" << endl);
    connect(scope, display, video_format);
    
    scope->activate();
    display->initOutput();
    display->activate();
    
    startNodes();
    
    while(1) {
      int dummy;
      char c;

      //cerr << "MediaTime: " << display->getMediaTime() << endl;
      c = fgetc(stdin);
      do {dummy = fgetc(stdin);} while (dummy != '\n');
      if ( c == 'q' ) break;
      if ( c == 'p' ) {
        if(!pause) {
          pause = true;
          sync -> pause();
          readfile->stop();
        } else {
          pause = false;
          sync -> wakeup();
          readfile->start();
        }
      }
    }
    
  } catch(Exception e) {
    cerr << e << endl;
    cleanup(registry);
    return 1;
  } catch(...) {
    cerr << "Something went wrong..." << endl;
    cleanup(registry);
    return 1;
  }
  
  cleanup(registry);
  delete app;
  return 0;
}
