/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


// general includes
#include "nmm/base/connect.hpp"
#include "nmm/base/NMMApplication.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ProxyApplication.hpp"
#include "nmm/base/ProxyObject.hpp"
#include "nmm/utils/utils.hpp"

// include used interfaces
#include "nmm/interfaces/base/INode.hpp"
#include "nmm/interfaces/base/IQueued.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/file/ISeekable.hpp"
#include "nmm/interfaces/device/audio/IAudioDevice.hpp"
#include "nmm/interfaces/video/display/IVideoDisplay.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSink.hpp"
#include "nmm/interfaces/video/display/IDisplayEvents.hpp"
#include "nmm/interfaces/video/display/IDisplayEventSender.hpp"
#include "nmm/interfaces/base/IBufferManager.hpp"

// sync
#include "nmm/base/sync/MultiAudioVideoSynchronizer.hpp"

// console producer
#include "nmm/misc/console/ConsoleProducer.hpp"


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>
#include <set>
#include <iostream>

using namespace NMM;


/**
 * Class for receiving events.
 */
class AppEventReceiver : public IEventReceiver {
private:
  EventDispatcher* dispatcher;
  
public:
  AppEventReceiver() {
    dispatcher = new EventDispatcher();
  }
  
  virtual ~AppEventReceiver() {
    delete dispatcher;
  }
  
  EventDispatcher& getEventDispatcher() const {
    return *dispatcher;
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const Event& event, const IEventSender* sender = 0 ) const {
    return doReceiveEvent(event, sender);
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
  virtual Result receiveEvent(const CEvent& event, const IEventSender* sender = 0 ) const {
    Result result = SUCCESS;
    
    for (unsigned int i=0; i<event.size(); ++i) {
      switch (doReceiveEvent(event[i], sender)) {
      case FAILURE:
        result = FAILURE;
        break;
        
      case DELETE:
        result = DELETE;
        break;
        
      default:
        break;
      }
    }
    
    return result;
  }
  
protected:
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) {
    return dispatcher->dispatch(event, sender);
  }
  
  virtual Result doReceiveEvent(const Event& event, const IEventSender* sender = 0) const {
    return dispatcher->dispatch(event, sender);
  }
};


/**
 * Class for handling events.
 * An instance of this class is registered at some nodes as an event listener.
 */
class EventCatcher {
private:
  bool exit;
  ThreadMutex exit_mutex;
  ThreadCondition exit_cv;
  IVideoDisplay* video_display;
  bool fullscreen;
  IMultiAudioVideoSynchronizer* sync;
  ISeekable* iseek;
  bool pause;
  
  void signalExit() {
    MutexGuard em(exit_mutex);
    exit = true;
    exit_cv.notifyAll();
  }
  
public:
  EventCatcher() 
    : exit(false),
      exit_mutex(),
      exit_cv(exit_mutex),
      video_display(NULL),
      fullscreen(false),
      sync(NULL),
      iseek(NULL),
      pause(false)
  {}
  
  Result endTrack() {
    cout << "End of File" << endl;
    signalExit();
    return SUCCESS;
  }

  void setVideoDisplay(IVideoDisplay* value) {
    video_display = value;
    if (video_display) {
      fullscreen = video_display->getFullscreen() && video_display->hasFullscreen();
    }
  }

  void setSeek(ISeekable* seeker) {
    iseek=seeker;
  }

  void setSynchronizer(IMultiAudioVideoSynchronizer* value) {
    sync = value;
  }
  
  Result keyReleased(string& key) {
    
    // if paused, any key should wakeup
    if (pause && sync) {
      pause = false;
      sync->wakeup();
      return SUCCESS;
    }
    
    if (key == "KEY_q") { // 'q' to exit
      signalExit();
      
    } else if (key == "KEY_f") { // 'f' to toggle fullscreen
      if (!video_display) {
	cerr << "EventHandler: No VideoDisplay set!" << endl;
	return FAILURE;
      }
      if (video_display->hasFullscreen()) {
	fullscreen = !fullscreen;
	video_display->setFullscreen(fullscreen);
      } else {
	cerr << "Fullscreen not supported with current display!" << endl;
      }
      
    } else if (key == "KEY_Left") { //seek backward
      if (!iseek) {
	cerr<<"no seekinterface set"<<endl;
	return FAILURE;
      }
      iseek->seekPercent(Rational(-5,100));
    } else if (key=="KEY_Right") { //seek forward
      if (!iseek) {
	cerr<<"no seekinterface set"<<endl;
	return FAILURE;
      }
      iseek->seekPercent(Rational(5,100));      
      
    } else if ((key == "KEY_space")) { // <space> to pause, any key wakes up, see above
      if (!sync) {
	cerr << "EventHandler: No Synchronizer set!" << endl;
	return FAILURE;
      }
      pause = true;
      sync->pause();
      
    } else {
      cerr << "Unsupported key pressed: " << key << endl;
    }
    
    return SUCCESS;
  }

  void handleEvents() {
    exit_mutex.lock();
    while (!exit) {
      exit_cv.wait();
    }
    exit_mutex.unlock();
  }

};


/* Global options */
int verbose            = 0;
bool ac3_passthrough   = false;
int use_ac3            = false;
int use_mga            = false;
int audio_stream       = 0;
int video_stream       = 0;
int fullscreen         = 0;
bool no_audio          = false;
bool no_video          = false;
bool no_sync           = false;
bool dump              = false;
bool use_ffmpeg        = false;
string filename;

void usage( char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] [filename]" << endl
       << endl
       << "    filename:            mpeg ps file" << endl
       << "    -f --fullscreen      Fullscreen mode" << endl
       << "    -t --audiostream     Select audio stream number" << endl
       << "    -i --videostream     Select video stream number" << endl
       << "    -a --ac3passthrough  send ac3 data directly to soundcard" << endl
       << "    -F --ffmpeg          Use ffmpeg as video decoder" << endl
       << "    -s --silent          Disable audio" << endl
       << "    -b --no-video        Disable video output" << endl
       << "    -n --no-sync         Disable synchronization" << endl
       << "    -d --dump:           Dump data instead of playing it" << endl
       << "    -v --verbose:        Verbose mode (lots of messages!)" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  exit(0);
}

void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
}

int main(int argc, char ** argv) {
  const char *optstring = "ft:i:x:y:asbFdnvh";
  struct option longopts[] = {
    {"fullscreen",     0, NULL, 'f'},
    {"audiostream",    1, NULL, 't'},
    {"videostream",    1, NULL, 'i'},
    {"ac3passthrough", 0, NULL, 'a'},
    {"silent",         0, NULL, 's'},
    {"no-video",       0, NULL, 'b'},
    {"ffmpeg",         0, NULL, 'F'},
    {"dump",           0, NULL, 'd'},
    {"no-sync",        0, NULL, 'n'},
    {"verbose",        0, NULL, 'v'},
    {"help",           0, NULL, 'h'},
    {NULL,             0, NULL, 0  } /* last row must be zero */
  };
  
  int ch, optionindex;
  
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
    // which option did we get?
    switch (ch) {
    case 'f':
      fullscreen = true;
      break;
    case 'n':
      no_sync = true;
      break;
    case 't':
      if (optarg) {
	audio_stream = atoi(optarg);
      }
    break;
    case 'i':
      if (optarg) {
	video_stream = atoi(optarg);
      }
      break;
    case 'a':
      ac3_passthrough = true;
      break;
    case 'F':
      use_ffmpeg = true;
      break;
    case 's':
      no_audio = true;
      break;
    case 'b':
      no_video = true;
      break;
    case 'd':
      dump = true;
      break;
    case 'v':
      ++verbose;
      break;
    default: //includes 'h'
      usage(argv[0]);
    }
  }

  if (argc <= optind) { // no file specified
    usage(argv[0]);
  } else {
    filename = string(argv[optind]);
  }
  
  // maybe turn off debug messages
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
  
  /*Begin> Allocate _var variables for the INodes */
  INode_var readfile;
  INode_var demux;
  INode_var ac3_dec;
  INode_var mp3_dec;
  INode_var divx4_dec;
  INode_var display;
  INode_var audio_play;
  /*End< */
  
  /*Begin> synchronisation stuff */
  ISinkController_var audio_controller;
  ISinkController_var video_controller;
  MultiAudioVideoSynchronizer mav_sync;
  IMultiAudioVideoSynchronizer_var msync;
  /*End< synchronisation stuff */
  
  /* Allocate an event listener */
  EventCatcher ec;
  AppEventReceiver receiver;
  
  int error_level = 0;
  NMMApplication* app = NULL;
  try {
    app = ProxyApplication::getApplication(argc, argv);
    
    /*Begin> Init requests for nodes */
    NodeDescription readfile_nd("GenericReadNode" /*"OGMReadNode"*/);
    NodeDescription demux_nd("OGMDemuxNode");
    NodeDescription ac3_dec_nd("AC3DecodeNode");
    NodeDescription mp3_dec_nd("MPEGAudioDecodeNode");
    NodeDescription divx4_dec_nd(use_ffmpeg ? "FFMpegDecodeNode" : "DivX4DecodeNode");
    NodeDescription display_nd("XDisplayNode");
    NodeDescription audio_play_nd("PlaybackNode");
    /*End< Init requests for nodes */
    
    /*Begin> Get the ClientRegistry and request Nodes */
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);
      
      list<Response> readfile_resp = registry.initRequest(readfile_nd);
      list<Response> demux_resp = registry.initRequest(demux_nd);
      list<Response> ac3_dec_resp = registry.initRequest(ac3_dec_nd);
      list<Response> mp3_dec_resp = registry.initRequest(mp3_dec_nd);
      list<Response> divx4_dec_resp = registry.initRequest(divx4_dec_nd);
      list<Response> display_resp = registry.initRequest(display_nd);
      list<Response> audio_play_resp = registry.initRequest(audio_play_nd);
      
      if (readfile_resp.empty() ||
          demux_resp.empty() ||
//           ac3_dec_resp.empty() ||
//           mp3_dec_resp.empty() ||
//           divx4_dec_resp.empty() ||
          (!no_video && display_resp.empty()) ||
          (!no_audio && audio_play_resp.empty())) {
        GLOBAL_ERROR_STREAM("Size of readfile_resp " << readfile_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of demux_resp " << demux_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of ac3_dec_resp " << ac3_dec_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of mp3_dec_resp " << mp3_dec_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of divx4_dec_resp " << divx4_dec_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of display_resp " << display_resp.size() << endl);
        GLOBAL_ERROR_STREAM("Size of audio_play_resp " << audio_play_resp.size() << endl);
        throw Exception("A required node is not available");
      }
      
      readfile.reset(registry.requestNode(readfile_resp.front()));
      demux.reset(registry.requestNode(demux_resp.front()));
      if (!ac3_dec_resp.empty()) {
        ac3_dec.reset(registry.requestNode(ac3_dec_resp.front()));
      }
      if (!mp3_dec_resp.empty()) {
        mp3_dec.reset(registry.requestNode(mp3_dec_resp.front()));
      }
      if (!divx4_dec_resp.empty()) {
        divx4_dec.reset(registry.requestNode(divx4_dec_resp.front()));
      }
      if (!display_resp.empty()) {
        display.reset(registry.requestNode(display_resp.front()));
      }
      if (!audio_play_resp.empty()) {
        audio_play.reset(registry.requestNode(audio_play_resp.front()));
      }
    }
    /*End< */
    
    /*Begin> Initialize and connect Nodes */
    cerr << "=================== begin init/connect nodes ===================" << endl;
    ISeekable_var file_seek(readfile->getParentObject()->getCheckedInterface<ISeekable>());
    IFileHandler_var file_source(readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    file_source->setFilename(filename);
    readfile->init();
    
    demux->init();
    
    readfile->initOutput();
    
    if (!no_audio) {
      IAudioDevice_var audio_device(audio_play->getParentObject()->getCheckedInterface<IAudioDevice>());
      audio_device->setDevice("/dev/dsp");
      audio_play->init();
    }
    
    IVideoDisplay_var video_display;
    if (!no_video) {
      video_display.reset(display->getParentObject()->getCheckedInterface<IVideoDisplay>());
      video_display->setFullscreen(fullscreen);
      display->init();
    }
    
    cerr << "* connect: readfile -> demux" << endl;
    connect(readfile, demux);
    
    readfile->activate();
    
    /*Begin> Analyze stream formats */
    demux->initOutput();
    Property_var demux_wout(demux->getWorkingOutputProperty());
    
    set<string> demux_video_streams;
    set<string> demux_audio_streams;
    list<string> demux_streams = demux->getOutputStreamTags();
    for (list<string>::const_iterator i = demux_streams.begin();
         i != demux_streams.end();
         i++) {
      const Format* format = demux_wout->getDefaultFormat(*i);
      if (format != NULL) {
        cerr << "* stream: " << *i << " is " << format->getFormatType() << endl;
        if (format->getFormatType() == Format::audio_ac3_type) {
          demux_audio_streams.insert(*i);
        } else if (format->getFormatType() == Format::audio_mpeg_type) {
          demux_audio_streams.insert(*i);
        } else if (format->getFormatType() == Format::video_mpeg4_type) {
          demux_video_streams.insert(*i);
        }
      } else {
        cerr << "* stream: " << *i << " has no format!" << endl;
      }
    }
    /*End< Analyze stream formats */
    
    string audio_stream_tag = string("audio") + audio_stream;
    string video_stream_tag = string("video") + video_stream;
    if (!no_video && demux_video_streams.end() == demux_video_streams.find(video_stream_tag)) {
      throw Exception("video stream " + video_stream_tag + " does not exist");
    }
    if (!no_audio && demux_audio_streams.end() == demux_audio_streams.find(audio_stream_tag)) {
      throw Exception("audio stream " + audio_stream_tag + " does not exist");
    }
    
    bool use_ac3 = false;
    if (!no_audio) {
      use_ac3 = (demux_wout->getDefaultFormat(audio_stream_tag)->getFormatType() == Format::audio_ac3_type);
    }
    
    if (!no_video) {
      if (!divx4_dec.get()) {
        throw Exception("No DIVX4 video decoder available");
      }
      
      divx4_dec->init();
      
      cerr << "* connect: demux::" << video_stream_tag << " -> divx4_dec" << endl;
      connect(demux, video_stream_tag, divx4_dec);
    }
    if (!no_audio) {
      if (use_ac3) {
        if (ac3_passthrough) {
          cerr << "* connect: demux::" << audio_stream_tag << " -> audio_play" << endl;
          connect(demux, audio_stream_tag, audio_play);
        } else {
          if (!ac3_dec.get()) {
            throw Exception("No AC3 audio decoder available");
          }
          
          ac3_dec->init();
          
          cerr << "* connect: demux::" << audio_stream_tag << " -> ac3_dec" << endl;
          connect(demux, audio_stream_tag, ac3_dec);
        }
      } else {
        if (!mp3_dec.get()) {
          throw Exception("No MPEG audio decoder available");
        }
        
        mp3_dec->init();
        
        cerr << "* connect: demux::" << audio_stream_tag << " -> mp3_dec" << endl;
        connect(demux, audio_stream_tag, mp3_dec);
      }
    }
    
    demux->activate();
    
    if (!no_audio) {
      if (use_ac3 && !ac3_passthrough) {
        ac3_dec->initOutput();
        
        cerr << "* connect: ac3_dec -> audio_play" << endl;
        connect(ac3_dec, audio_play);
        
        ac3_dec->activate();
      } else if (!use_ac3) {
        mp3_dec->initOutput();
        
        cerr << "* connect: mp3_dec -> audio_play" << endl;
        connect(mp3_dec, audio_play);
        
        mp3_dec->activate();
      }
      
      audio_play->initOutput();
      audio_play->activate();
    }
    
    if (!no_video) {
      divx4_dec->initOutput();
      
      cerr << "* connect: divx4_dec -> display" << endl;
      connect(divx4_dec, display);
      
      divx4_dec->activate();
      display->initOutput();
      display->activate();
    }
    cerr << "=================== end init/connect nodes ===================" << endl;
    /*End< Initialize and connect Nodes */
    
    /*Begin> Setup synchronization */
    cerr << "=================== begin setup synchronization ===================" << endl;
    /* Get synchronized sinks */
    ISynchronizedSink_var audio_sync;
    if (!no_audio && audio_play.get()) {
      audio_sync.reset(audio_play->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    }
    ISynchronizedSink_var video_sync;
    if (!no_video && display.get()) {
      video_sync.reset(display->getParentObject()->getCheckedInterface<ISynchronizedSink>());
    }
    
    /* Get sink controllers and setup intra-stream synchronization */
    ISinkController_var audio_controller;
    if (audio_sync.get()) {
      audio_controller.reset(audio_sync->getController());
      if (no_sync) {
        cerr << "* disable intra-stream sync for audio" << endl;
        audio_sync->setSynchronized(false);
      } else {
        cerr << "* enable intra-stream sync for audio" << endl;
        audio_sync->setSynchronized(true);
      }
    }
    ISinkController_var video_controller;
    if (video_sync.get()) {
      video_controller.reset(video_sync->getController());
      if (no_sync) {
        cerr << "* disable intra-stream sync for video" << endl;
        video_sync->setSynchronized(false);
      } else {
        cerr << "* enable intra-stream sync for video" << endl;
        video_sync->setSynchronized(true);
      }
    }
    
    /* setup inter-stream synchronization */
    if (!no_sync && audio_controller.get() && video_controller.get()) {
      cerr << "* setup inter-stream sync" << endl;
      msync.reset(mav_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());
      msync->addAudioSinkController(audio_controller.get());
      msync->addVideoSinkController(video_controller.get());
    }
    cerr << "=================== end setup synchronization ===================" << endl;
    /*End< Setup synchronization */
    
    /*Begin> Start nodes */
    cerr << "=================== begin start nodes ===================" << endl;
    if (display.get() && display->isActivated()) {
      cerr << "display" << endl;
      display->start();
    }
    if (audio_play.get() && audio_play->isActivated()) {
      cerr << "audio_play" << endl;
      audio_play->start();
    }
    if (ac3_dec.get() && ac3_dec->isActivated()) {
      cerr << "ac3_dec" << endl;
      ac3_dec->start();
    }
    if (mp3_dec.get() && mp3_dec->isActivated()) {
      cerr << "mp3_dec" << endl;
      mp3_dec->start();
    }
    if (divx4_dec.get() && divx4_dec->isActivated()) {
      cerr << "divx4_dec" << endl;
      divx4_dec->start();
    }
    if (demux->isActivated()) {
      cerr << "demux" << endl;
      demux->start();
    }
    if (readfile->isActivated()) {
      cerr << "readfile" << endl;
      readfile->start();
    }
    cerr << "=================== end start nodes ===================" << endl;
    /*End< Start nodes */
    
    /*Begin> Register event listeners */
    if (!no_video) {
      display->getParentObject()->registerEventListener(ITrack::endTrack_event, 
                                                        new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));
    }
    if (!no_audio) {
      audio_play->getParentObject()->registerEventListener(ITrack::endTrack_event, 
                                                           new TEDObject0<EventCatcher>(&ec, &EventCatcher::endTrack));      
    }
    
    if (!no_video) {
      display->getParentObject()->
        registerEventListener(IDisplayEvents::keyReleased_event, new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    }
    
    receiver.getEventDispatcher().registerEventListener(IDisplayEvents::keyReleased_event,
                                                        new TEDObject1<EventCatcher, string>(&ec, &EventCatcher::keyReleased));
    /*End< */
    
    /*Begin> Add console producer */
    ConsoleProducer console;
    console.readConfig( getNMMResource("XProducer.xml") );
    console.connectTo(&receiver);
    /*End< Add console producer */
    
    /*Begin> Event loop */
    console.start();
    IDisplayEventSender_var display_event_sender;
    if (!no_video) {
      display_event_sender.reset(display->getParentObject()->getInterface<IDisplayEventSender>());
      if (display->isStarted() && display_event_sender.get()) {
        display_event_sender->enableDisplayEvents();
      }
    }
    ec.setVideoDisplay(video_display.get());
    ec.setSynchronizer(msync.get());
    ec.setSeek(file_seek.get());
    
    ec.handleEvents();
    
    if (display->isStarted() && display_event_sender.get()) {
      display_event_sender->disableDisplayEvents();
    }
    console.stop();
    /*End< Event loop */
    
  } catch(const Exception& e) {
    
    cerr << e << endl;
    error_level = 1;
    
  } catch(...) {
    
    cerr << "Something strange went wrong" << endl;
    return 2;
    
  }
  
  /*Begin> Cleanup */
  try {
    
    /*Begin> Stop nodes */
    cerr << "=================== begin stop nodes ===================" << endl;
    if (readfile.get() && readfile->isStarted()) {
      readfile->stop();
    }
    if (demux.get() && demux->isStarted()) {
      demux->stop();
    }
    if (ac3_dec.get() && ac3_dec->isStarted()) {
      ac3_dec->stop();
    }
    if (mp3_dec.get() && mp3_dec->isStarted()) {
      mp3_dec->stop();
    }
    if (divx4_dec.get() && divx4_dec->isStarted()) {
      divx4_dec->stop();
    }
    if (audio_play.get() && audio_play->isStarted()) {
      audio_play->stop();
    }
    if (display.get() && display->isStarted()) {
      display->stop();
    }
    cerr << "=================== end stop nodes ===================" << endl; 
    /*End< Stop nodes */
    
    /*Begin> Flush nodes */
    cerr << "=================== begin flush nodes ===================" << endl; 
    if (readfile.get() && readfile->isActivated()) {
      readfile->flush();
    }
    if (demux.get() && demux->isActivated()) {
      demux->flush();
    }
    if (ac3_dec.get() && ac3_dec->isActivated()) {
      ac3_dec->flush();
    }
    if (mp3_dec.get() && mp3_dec->isActivated()) {
      mp3_dec->flush();
    }
    if (divx4_dec.get() && divx4_dec->isActivated()) {
      divx4_dec->flush();
    }
    if (audio_play.get() && audio_play->isActivated()) {
      audio_play->flush();
    }
    if (display.get() && display->isActivated()) {
      display->flush();
    }
    cerr << "=================== end flush nodes ===================" << endl; 
    /*End< Flush nodes */
    
    /*Begin> Disconnect nodes */
    cerr << "=================== begin disconnect/deinit nodes ===================" << endl;
    if (display.get()) {
      if (display->isActivated()) {
        display->deactivate();
      }
      if (display->isOutputInitialized()) {
        display->deinitOutput();
      }
      if (display->isInitialized()) {
        cerr << "display" << endl;
        IInputJack_var display_in_jack(display->getInputStream("default"));
        if (display_in_jack.get() && display_in_jack->isConnected()) {
          if (divx4_dec.get() && divx4_dec->isActivated()) {
            divx4_dec->deactivate();
          }
          disconnect_input(display, "default");
        }
        display->deinit();
      }
    }
    if (audio_play.get()) {
      if (audio_play->isActivated()) {
        audio_play->deactivate();
      }
      if (audio_play->isOutputInitialized()) {
        audio_play->deinitOutput();
      }
      if (audio_play->isInitialized()) {
        cerr << "audio_play" << endl;
        IInputJack_var audio_play_in_jack(audio_play->getInputStream("default"));
        if (audio_play_in_jack.get() && audio_play_in_jack->isConnected()) {
          if (ac3_dec.get() && ac3_dec->isActivated()) {
            ac3_dec->deactivate();
          } else if (mp3_dec.get() && mp3_dec->isActivated()) {
            mp3_dec->deactivate();
          } else if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(audio_play, "default");
        }
        audio_play->deinit();
      }
    }
    if (ac3_dec.get()) {
      if (ac3_dec->isActivated()) {
        ac3_dec->deactivate();
      }
      if (ac3_dec->isOutputInitialized()) {
        ac3_dec->deinitOutput();
      }
      if (ac3_dec->isInitialized()) {
        cerr << "ac3_dec" << endl;
        IInputJack_var ac3_dec_in_jack(ac3_dec->getInputStream("default"));
        if (ac3_dec_in_jack.get() && ac3_dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(ac3_dec, "default");
        }
        ac3_dec->deinit();
      }
    }
    if (mp3_dec.get()) {
      if (mp3_dec->isActivated()) {
        mp3_dec->deactivate();
      }
      if (mp3_dec->isOutputInitialized()) {
        mp3_dec->deinitOutput();
      }
      if (mp3_dec->isInitialized()) {
        cerr << "mp3_dec" << endl;
        IInputJack_var mp3_dec_in_jack(mp3_dec->getInputStream("default"));
        if (mp3_dec_in_jack.get() && mp3_dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(mp3_dec, "default");
        }
        mp3_dec->deinit();
      }
    }
    if (divx4_dec.get()) {
      if (divx4_dec->isActivated()) {
        divx4_dec->deactivate();
      }
      if (divx4_dec->isOutputInitialized()) {
        divx4_dec->deinitOutput();
      }
      if (divx4_dec->isInitialized()) {
        cerr << "divx4_dec" << endl;
        IInputJack_var divx4_dec_in_jack(divx4_dec->getInputStream("default"));
        if (divx4_dec_in_jack.get() && divx4_dec_in_jack->isConnected()) {
          if (demux.get() && demux->isActivated()) {
            demux->deactivate();
          }
          disconnect_input(divx4_dec, "default");
        }
        divx4_dec->deinit();
      }
    }
    if (demux.get()) {
      if (demux->isActivated()) {
        demux->deactivate();
      }
      if (demux->isOutputInitialized()) {
        demux->deinitOutput();
      }
      if (demux->isInitialized()) {
        cerr << "demux" << endl;
        IInputJack_var demux_in_jack(demux->getInputStream("default"));
        if (demux_in_jack.get() && demux_in_jack->isConnected()) {
          if (readfile.get() && readfile->isActivated()) {
            readfile->deactivate();
          }
          disconnect_input(demux, "default");
        }
        demux->deinit();
      }
    }
    if (readfile.get()) {
      if (readfile->isActivated()) {
        readfile->deactivate();
      }
      if (readfile->isOutputInitialized()) {
        readfile->deinitOutput();
      }
      if (readfile->isInitialized()) {
        cerr << "readfile"<< endl;
        readfile->deinit();
      }
    }
    /*End< Disconnect nodes */
    cerr << "=================== end disconnect/deinit nodes ===================" << endl;
    
    /*Begin> Release nodes */
    if (app != NULL) {
      cerr << "=================== begin release nodes ===================" << endl; 
      if (readfile.get()) {
        app->getRegistry().releaseNode(*readfile);
      }
      if (demux.get()) {
        app->getRegistry().releaseNode(*demux);
      }
      if (ac3_dec.get()) {
        app->getRegistry().releaseNode(*ac3_dec);
      }
      if (mp3_dec.get()) {
        app->getRegistry().releaseNode(*mp3_dec);
      }
      if (divx4_dec.get()) {
        app->getRegistry().releaseNode(*divx4_dec);
      }
      if (audio_play.get()) {
        app->getRegistry().releaseNode(*audio_play);
      }
      if (display.get()) {
        app->getRegistry().releaseNode(*display);
      }
      cerr << "=================== end release nodes ===================" << endl; 
    }
    
  } catch(const Exception& e) {
    
    cerr << "Exception during cleanup" << endl;
    cerr << e << endl;
    
    return 2;
    
  } catch(...) {
    
    cerr << "Something strange went wrong during cleanup" << endl;
    return 2;
    
  }
  
  return error_level;
}

