/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_NMMTYPES_HPP
#define NMM_NMMTYPES_HPP NMM_NMMTYPES_HPP

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "nmm/base/serialize/ComplexValue.hpp"

namespace NMM
{
  enum Result { 
    FAILURE = 0, 
    SUCCESS = 1, 
    UNDEF = 2, 
    TIME_OUT = -1, 
    INTERRUPTED = -2, 
    DEFER = -3, 
    BLOCKED = -4, 
    DELETE = -5 
  };

  class IValueStream;
  class OValueStream;

  class NamedObject;

  OValueStream& operator<<(OValueStream& ov, const NMM::Result& t);

  IValueStream& operator>>(IValueStream& iv, NMM::Result& t);

  /**
   * This is the super-class for all Exceptions (or at least for
   * those in nmm/base).
   * Typical use within a NamedObject: throw Exception(this, "something went wrong");
   * Typical use within main: try { ... } catch(const Exception& e) { ... }
   */
  class Exception : public SerializedValue {
  private:
    //! The NamedObject which threw the Exception
    const class NamedObject* parent;
    //! Extra comment of the exception
    TValue<string> comment;

  public:
    /**
     * Default constructor.
     */
    Exception () : SerializedValue("Exception", "NMM")
    {
      parent = 0;
      comment.setValue("Exception");
      insertArgument(&comment);
    }

    /**
     * Constructor, which allows to specify an extra comment
     */
    Exception (const string& s)  : SerializedValue("Exception", "NMM")
    {
      parent = 0;
      comment = "Exception : " + s;
      insertArgument(&comment);
    }

    /**
     * Constructor, which allows to specify the NamedObject which
     * threw the exception and an extra comment
     */
    Exception (const NamedObject* n, const string& s = "") : SerializedValue("Exception", "NMM")
    {
      parent = n;
      comment = "Exception : " + s;
      insertArgument(&comment);
    }

    Exception(const Exception& exception) : SerializedValue(exception)
    {
      *this = exception;
      insertArgument(&comment);
    }

    virtual Exception* copy() const { return new Exception(*this); }

    /**
     * Returns the NamedObject which threw the exceptions
     */
    inline const NamedObject *getParent() const { return parent; }

    /** returns the comment of the exception
      */
    inline const string& getComment()     const { return comment.getRValue();}

    /** sets the comment of the exception
      */
    inline void setComment(const string& s)     { comment.setValue(s); }

    /**
     * Returns the name of the exception and the comment
     */
    friend ostream & operator << (ostream & os, const Exception& e)
    {
      return os << e.getComment();
    }



  };

}

#endif // NMM_NMMTYPES_HPP
