/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_BASEEXCEPTIONS_HPP_
#define _NMM_BASEEXCEPTIONS_HPP_

#include "nmm/NMMTypes.hpp"


namespace NMM
{

  class InputJack;
  class OutputJack;
  class IInputJack;
  class IOutputJack;

  /**
   * This Exception should be thrown if there was an error during state transition
   * more specific, if a call of a 'do'-method failed (e.g. doInit() returns FAILURE)
   */
  class StateTransitionFailedException : public Exception
  {
  public:

    /**
     * Default constructor.
     */
    StateTransitionFailedException () 
      : Exception("StateTransitionFailedExcpetion") { setTypeName("StateTransitionFailedException"); };

    /**
     * Constructor, which allows to specify an extra comment
     */
    StateTransitionFailedException (const string& s) 
      : Exception("StateTransitionFailedException : " + s) {setTypeName("StateTransitionFailedException");};

    /**
     * Constructor, which allows to specify the NamedObject which
     * threw the exception and an extra comment
     */
    StateTransitionFailedException (const NamedObject* n, const string& s = "") 
      : Exception(n, ((s == "") ? "StateTransitionFailedException" : "StateTransitionFailedException : " + s)) {setTypeName("StateTransitionFailedException");};

    StateTransitionFailedException* copy() const { return new StateTransitionFailedException(*this); }
  };

  /**
   * This Exception should be thrown if a required method was not called or completed 
   * without success when changing the state of an object
   */
  class UnsuccessfulRequiredCallException : public Exception
  {
  public:

    /**
     * Default constructor.
     */
    UnsuccessfulRequiredCallException () 
      : Exception("UnsuccessfulRequiredCallExcpetion") { setTypeName("UnsuccessfulRequiredCallException"); };

    /**
     * Constructor, which allows to specify an extra comment
     */
    UnsuccessfulRequiredCallException (const string& s) 
      : Exception("UnsuccessfulRequiredCallException : " + s) {setTypeName("UnsuccessfulRequiredCallException");};

    /**
     * Constructor, which allows to specify the NamedObject which
     * threw the exception and an extra comment
     */
    UnsuccessfulRequiredCallException (const NamedObject* n, const string& s = "") 
      : Exception(n, ((s == "") ? "UnsuccessfulRequiredCallException" : "UnsuccessfulRequiredCallException : " + s)) {setTypeName("UnsuccessfulRequiredCallException");};

    UnsuccessfulRequiredCallException* copy() const { return new UnsuccessfulRequiredCallException(*this); }
  };

  /**
   * This exception will be thrown, if a illegal state transition
   * was performed. This check is implemented in Node.
   */
  class IllegalStateTransitionException : public Exception
  {
  public:

    /**
     * Default constructor.
     */
    IllegalStateTransitionException () 
      : Exception("IllegalStateTransitionExcpetion") { setTypeName("IllegalStateTransitionException");};

    /**
     * Constructor, which allows to specify an extra comment
     */
    IllegalStateTransitionException (const string& s) 
      : Exception("IllegalStateTransitionException : " + s) { setTypeName("IllegalStateTransitionException");};

    /**
     * Constructor, which allows to specify the NamedObject which
     * threw the exception and an extra comment
     */
    IllegalStateTransitionException (const NamedObject* n, const string& s = "") 
      : Exception(n, ((s == "") ? "IllegalStateTransitionException" : "IllegalStateTransitionException : " + s)) { setTypeName("IllegalStateTransitionException");};

    IllegalStateTransitionException* copy() const { return new IllegalStateTransitionException(*this); }
  };

  /**
   * This exception will be thrown, if a operation is performed
   * which is illegal in current state.
   */
  class IllegalStateException : public Exception
  {
  public:

    /**
     * Default constructor.
     */
    IllegalStateException () 
      : Exception("IllegalStateExcpetion") { setTypeName("IllegalStateException"); };

    /**
     * Constructor, which allows to specify an extra comment
     */
    IllegalStateException (const string& s) 
      : Exception("IllegalStateException : " + s) {  setTypeName("IllegalStateException"); };

    /**
     * Constructor, which allows to specify the NamedObject which
     * threw the exception and an extra comment
     */
    IllegalStateException (const NamedObject* n, const string& s = "") 
      : Exception(n, ((s == "") ? "IllegalStateException" : "IllegalStateException : " + s)) { setTypeName("IllegalStateException"); };

    IllegalStateException (const IllegalStateException& exception)
      : Exception(exception) {}

    IllegalStateException* copy() const { return new IllegalStateException(*this); }

  };

   class BindingException : public Exception {
   public:
      BindingException()
        : Exception("BindingException"), __input_jack(0), __output_jack(0) {setTypeName("BindingException");}
      BindingException(const string& s)
        : Exception("BindingException : " + s), __input_jack(0), __output_jack(0) {setTypeName("BindingException");}

      BindingException(const InputJack* i_jack, const OutputJack* o_jack, const string& s = "")
        : Exception("BindingException : " + s), __input_jack(i_jack), __output_jack(o_jack) {setTypeName("BindingException");}

      BindingException* copy() const { return new BindingException(*this); }

      inline const InputJack* getInputJack()    const { return __input_jack;}

      inline const OutputJack* getOutputJack()  const { return __output_jack;}

    private:
      const InputJack*  __input_jack;
      const OutputJack* __output_jack;
   };

   class UnbindingException : public Exception {
   public:
      UnbindingException()
        : Exception("UnbindingException"), __input_jack(0), __output_jack(0) {setTypeName("UnbindingException");}
      UnbindingException(const string& s)
        : Exception("UnbindingException : " + s), __input_jack(0), __output_jack(0) {setTypeName("UnbindingException");}

      UnbindingException(const IInputJack* i_jack, const IOutputJack* o_jack, const string& s = "")
        : Exception("UnbindingException : " + s), __input_jack(i_jack), __output_jack(o_jack) {setTypeName("UnbindingException");}

      UnbindingException* copy() const { return new UnbindingException(*this); }

      inline const IInputJack* getInputJack()    const { return __input_jack;}

      inline const IOutputJack* getOutputJack()  const { return __output_jack;}

    private:
      const IInputJack*  __input_jack;
      const IOutputJack* __output_jack;
   };

  class UnexpectedException : public Exception
  {
  public:

    /**
     * Default constructor.
     */
    UnexpectedException ()
      : Exception("UnexpectedExcpetion") { setTypeName("UnexpectedException"); };

    /**
     * Constructor, which allows to specify an extra comment
     */
    UnexpectedException (const string& s)
      : Exception("UnexpectedException : " + s) {  setTypeName("UnexpectedException"); };

    /**
     * Constructor, which allows to specify the NamedObject which
     * threw the exception and an extra comment
     */
    UnexpectedException (const NamedObject* n, const string& s = "")
      : Exception(n, ((s == "") ? "UnexpectedException" : "UnexpectedException : " + s)) { setTypeName("UnexpectedException"); };

    UnexpectedException (const UnexpectedException& exception)
      : Exception(exception) {}

    UnexpectedException* copy() const { return new UnexpectedException(*this); }

  };

  class UnknownEventException : public Exception
  {
  public:
    UnknownEventException()
      : Exception("UnknownEventException") { setTypeName("UnknownEventException"); }

    UnknownEventException(const string& message)
      : Exception("UnknownEventException : " + message) { setTypeName("UnknownEventException"); }

    UnknownEventException(const NamedObject* n, const string& s)
      : Exception(n, ((s == "") ? "UnknownEventException" : "UnknownEventException : " + s)) { setTypeName("UnknownEventException"); };

    UnknownEventException* copy() const { return new UnknownEventException(*this); }
  };

  
  class WrongDeviceException : public Exception
  {
  public:
    WrongDeviceException(const string& device = "")
      : Exception("WrongDeviceException"), __device(device) { setTypeName("WrongDeviceException"); registerAttributes(); }

    WrongDeviceException(const string& message, const string& device)
      : Exception("WrongDeviceException : " + message), __device(device) { setTypeName("WrongDeviceException"); registerAttributes();}
    
    WrongDeviceException(const NamedObject* n, const string& device)
      : Exception(n, "WrongDeviceException"), __device(device) { setTypeName("WrongDeviceException"); registerAttributes();}
    
    WrongDeviceException(const NamedObject* n, const string& message, const string& device)
      : Exception(n, ((message == "") ? "WrongDeviceException" : "WrongDeviceException : " + message)), __device(device) { setTypeName("WrongDeviceException"); registerAttributes();};

    WrongDeviceException* copy() const { return new WrongDeviceException(*this); }
    const string& getDevice()    const { return __device.getRValue(); }
  private: //method
    void registerAttributes();
  private:
    TValue<string> __device;
  };

  class DeviceNotFoundException : public Exception
  {
  public:
    DeviceNotFoundException(const string& device = "")
      : Exception("DeviceNotFoundException"), __device(device) { setTypeName("DeviceNotFoundException"); registerAttributes(); }

    DeviceNotFoundException(const string& message, const string& device)
      : Exception("DeviceNotFoundException : " + message), __device(device) { setTypeName("DeviceNotFoundException"); registerAttributes();}
    
    DeviceNotFoundException(const NamedObject* n, const string& device)
      : Exception(n, "DeviceNotFoundException"), __device(device) { setTypeName("DeviceNotFoundException"); registerAttributes();};
    
    DeviceNotFoundException(const NamedObject* n, const string& message, const string& device)
      : Exception(n, ((message == "") ? "DeviceNotFoundException" : "DeviceNotFoundException : " + message)), __device(device) { setTypeName("DeviceNotFoundException"); registerAttributes();};

    DeviceNotFoundException* copy() const { return new DeviceNotFoundException(*this); }
    const string& getDevice()       const { return __device.getRValue(); }
  private: //method
    void registerAttributes();
  private:
    TValue<string> __device;
  };
  
  class InterfaceNotFoundException : public Exception
  {
  public:
    InterfaceNotFoundException()
      : Exception("InterfaceNotFoundException") { setTypeName("InterfaceNotFoundException"); }

    InterfaceNotFoundException(const string& message)
      : Exception("InterfaceNotFoundException : " + message) { setTypeName("InterfaceNotFoundException"); }

    InterfaceNotFoundException(const NamedObject* n, const string& s)
      : Exception(n, ((s == "") ? "InterfaceNotFoundException" : "InterfaceNotFoundException : " + s)) { setTypeName("InterfaceNotFoundException");};

    InterfaceNotFoundException* copy() const { return new InterfaceNotFoundException(*this); }
  };


  class InvalidTypeException : public Exception
  {
  public:
    InvalidTypeException()
      : Exception("InvalidTypeException") { setTypeName("InvalidTypeException"); }

    InvalidTypeException(const string& message)
      : Exception("InvalidTypeException : " + message) { setTypeName("InvalidTypeException"); }

    InvalidTypeException(const TypeInfo& expected_type, const TypeInfo& current_type)
      : Exception("InvalidTypeException : "), __expected_type(expected_type), __current_type(current_type)
     { setTypeName("InvalidTypeException"); }

    InvalidTypeException(const NamedObject* n, const string& s)
      : Exception(n, ((s == "") ? "InvalidTypeException" : "InvalidTypeException : " + s)) { setTypeName("InvalidTypeException");};

    InvalidTypeException* copy() const { return new InvalidTypeException(*this); }

    inline    const TypeInfo& getExpectedType() const { return __expected_type; }
    inline    const TypeInfo& getCurrentType()  const { return __current_type; }
    private:
      TypeInfo __expected_type;
      TypeInfo __current_type;
  };
}

#endif
