/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_CONNECTION_MANAGER_HPP
#define NMM_CONNECTION_MANAGER_HPP


#include "nmm/interfaces/base/IConnectionManagerImpl.hpp"
#include "nmm/utils/Singleton.hpp"

#include "TransportStrategy.hpp"
#include "StdNamedObject.hpp"
namespace NMM {
  
  class FormatStrategyMap;
  class ConnectorAddress;
  
  /**
   *  @brief Connection Manager
   *
   *  The %ConnectionManager is responsible for setting up connections for transport strategies that
   *  can't connect to an Acceptor. The actual work is done by %ConnectorAcceptor objects that register
   *  themselves at the %ConnectionManager.
   */
  class ConnectionManager :
    public StdNamedObject,
    public IConnectionManagerImpl
  {
    friend class CreateUsingNew<ConnectionManager>;
  public:
    /**
     *  @name Constructor & Destructor
     */
    //@{
    /**
     *  @brief Virtual destructor
     */
    virtual ~ConnectionManager();
    //@}
    
    /**
     *  @name Methods from NMMObject
     */
    //@{
    /**
     *  @brief Get default address
     *
     *  As a special exception, the %ConnectionManager must not have a %ConnectorAddress.
     *  This limitation is necessary to prevent cyclic %TSAddress structures.
     */
    virtual TSAddress* getDefaultAddress(const string& strategy_type) const;
    //@}
    
    /**
     *  @name IConnectionManager interface
     */
    //@{
    /**
     *  @brief Handle a connection request from a remote host
     *  @param remote_addr Fully specified address of the remote transport strategy
     *  @param local_addr Returns the fully specified address of the local transport strategy
     *  @exception BindingException if no %ConnectorAcceptor was found or the %ConenctorAcceptor throws a BindingException
     *
     *  This method checks if a %ConnectorAcceptor object exists for the requested transport strategy type
     *  and forwards the request to this Connector.
     */
    virtual void handleConnect(const TSAddress* remote_addr, TSAddress*& local_addr) const 
      throw(const BindingException&);
    
    /**
     *  @brief Reliably send event to partner object
     *  @param ts_reference NMM reference of transport strategy which should receive the event
     *  @param e The event
     *  @return The result of the event
     *
     *  This method is used to reliably send a configuration event to a partner object via the %ConnectionManager.
     *  The connection manager pretends that the event was sent via the transport strategy.
     */
    virtual Result sendReliableEvent(const NMMReference& receiver, Event& e) const
      throw(const BindingException&);
    
    /**
     *  @brief Reliably send event to partner strategy
     *  @param receiver NMM Reference of transport strategy which should receive the event
     *  @param e The event
     *  @return The result of the event
     *
     *  This method is used to reliably send a configuration event to a partner strategy via the %ConnectionManager.
     */
    virtual Result sendReliableConfigEvent(const NMMReference& receiver, Event& e) const
      throw(const BindingException&);
    
    /**
     *  @brief Reliably send instream message to partner object
     *  @param receiver NMM Reference of transport strategy which should receive the event
     *  @param message The message
     *  @return The result of the event
     *
     *  This method is used to reliably send an instream to a partner object via the ConnectionManager.
     *  The connection manager pretends that the message was sent via the transport strategy.
     *
     *  This method is only for debugging and should not be used.
     */
    virtual Result sendReliableInstreamMessage(const NMMReference& receiver, CEvent& message) const
      throw(const BindingException&);
    
    /**
     *  @brief Get strategies that support format
     *  @param format Format to check
     *  @return List of transport strategies that support the format
     *
     *  This method returns the list of transport strategies that support the given format.
     */
    virtual list<string> getSupportedStrategies(const Format& format) const;
    //@}
    
    /**
     *  @name Local interface
     */
    //@{
    /**
     *  @brief Request a connection
     *  @param local_addr Fully specified address of the local transport strategy
     *  @param remote_addr Returns the fully specified address of the remote transport strategy
     *  @param remote_icm Interface to remote connection manager
     *  @exception BindingException if the connection could not be set up
     *
     *  This method requests a connection to the remote connection manager given by @c remote_icm. The
     *  request is sent by the %ConnectorAcceptor object which is locally registered for the trasport
     *  strategy type of @c local_addr. A corresponding %ConnectorAcceptor at the remote connection manager
     *  handles the request. If no %ConnectorAcceptor is registered at the local or remote %ConnectionManager
     *  or the connection request fails for some reason, a %BindingException is thrown.
     *
     *  The caller must provide a fully specified address in @c local_addr that can be used by the remote host
     *  to configure its transport strategy. The remote host returns the fully specified address of the partner
     *  transport strategy in @c remote_addr if the connection request was successful.
     */
    void connect(const ConnectorAddress* local_addr, ConnectorAddress*& remote_addr, IConnectionManager* remote_icm) const
      throw(const BindingException&);
    
    /**
     *  @brief Request connection manager interface
     *  @param interface Interface to object on host
     *  @return Connection manager interface, if available, otherwise @a NULL
     *  @exception BindingException if the request fails
     *
     *  This method requests an interface of a connection manager.
     *  To identify the host, an interface from an object on the host is used.
     *  The interface must contain a %ConnectorAddress.
     */
    IConnectionManager* getConnectionManager(const Interface& interface) const
      throw(const BindingException&);    
    //@}
    
    /** @name Configuration
     */
    //@{
    /** Get default strategy.
     * This method returns the default transport strategy type for the given type of connection.
     * The type of connection is the type of communication channel and the binding type.
     * @param channel_type Type of communication channel
     * @param binding_type Binding type
     */
    string getDefaultStrategy(const TypeInfo& channel_type, TSAddress::BindingType binding_type) const;
    
    /** Set default strategy.
     * This method sets the default transport strategy type for the given type of connection.
     * The type of connection is the type of communication channel and the binding type.
     * @param channel_type Type of communication channel
     * @param binding_type Binding type
     & @param strategy_type Strategy type
     */
    void setDefaultStrategy(const TypeInfo& channel_type, TSAddress::BindingType binding_type,
                            const string& strategy_type);
    
    /** Get format strategy map.
     * This method returns the format strategy map of this connection manager.
     */
    inline const FormatStrategyMap& getFormatStrategyMap() const { return *__format_strategy_map; }

    /** Get format strategy map.
     * This method returns the format strategy map of this connection manager.
     */
    inline       FormatStrategyMap& getFormatStrategyMap()       { return *__format_strategy_map; }
    //@}
    
    /** @name Helpers
     */
    //@{
    /** Request service.
     * This method requests an interface of a service. It opens a new connection to the
     * host using the given address and requests the service from the reference manager.
     * @param address Address to use to connect to host
     * @param service name of service
     * @return Interface of requested service
     * @exception BindingException if the request failed
     */
    Interface* getService(const TSAddress& address, const string& service) const
      throw(const BindingException&);

    /** Request service.
     * This method requests an interface of a service. It opens a new connection to the
     * host using the addresses given in the interface and requests the service from the reference manager.
     * @param interface Interface providing addresses of the host
     * @param service name of service
     * @return Interface of requested service
     * @exception BindingException if the request failed
     */
    Interface* getService(const Interface& interface, const string& service) const
      throw(const BindingException&);
    //@}
    
  private:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor.
     */
    ConnectionManager();
    //@}
    
  private:
    typedef pair<TypeInfo, TSAddress::BindingType> ConnectionType;
    typedef map<ConnectionType, string>            StrategyMap;
    typedef StrategyMap::iterator                  StrategyIterator;
    typedef StrategyMap::const_iterator            StrategyConstIterator;
    StrategyMap  __strategy_map;                                            ///< Default strategy map
    
    FormatStrategyMap* __format_strategy_map;
  };
  
  /** Connection manager singleton.
   */
  typedef Singleton<ConnectionManager,
                    CreateUsingNew,
                    DefaultLifetime,
                    MultiThreaded> SConnectionManager;
}

#endif


