/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                    Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef NMM_EVENTDISPATCHER_HPP
#define NMM_EVENTDISPATCHER_HPP

#include "IState.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "EDObject.hpp"

#include "nmm/nmm_hash_map.hpp"

using namespace std;

namespace NMM {

  class IEventReceiver;
  class IEventSender;


  class EventDispatcher : public IState, public NamedObject {
  public:
    /*! \name EventDispatcher's constructors
     * These functions creates or destroy an EventDispatcher
     */
    //@{
    /** Constructor
     */
    EventDispatcher(const char* name = "Dispatcher", int state = -1);

    /** Copy constructor
     */
    EventDispatcher(const EventDispatcher& dispatcher);

    /** Destructor
     */
    virtual ~EventDispatcher();
    //@}


    /*! \name Event registration methods
     * These methods allow to register, remove and access methods as well as EventListener which are handeld by  * the EventDispatcher.
     * An EventListener receives an incoming event after it was successfully processed
     * by the registered method.
     */
    //@{

    /** Adds an EventDispatcher(ED) Object with specific key.
     */
    void registerEvent(const string& s, EDObject* edo);

    /** Removes Event e and returns the corresponding ED-Object if it exist.
     */
    EDObject* removeEvent(const string& e);

    /** Adds an TED-Object as EventListener. Incoming events with corresponding
     * event-keys are forwarded to the TED-Object
     */
    void registerEventListener(const string&, EDObject* );

    /** Adds an IEventReceiver as EventListener. All incomming events are forwarded to the
     * eventhandler. This method is obsolete and you should use an IEventReceiver.
     */
    void registerEventListener(IEventReceiver*);

    /** Adds an IEventReceiver as EventListener. All incomming events are forwarded to the
     * eventhandler. This method is obsolete and you should use an IEventReceiver.
     */
    void registerEventListener(const string& event_key, IEventReceiver*);

    /** Removes an EventListener with a specific key
     */
    void removeEventListener(const string& event_key, IEventReceiver*);

    /** Removes an EventListener
     */
    void removeEventListener(IEventReceiver* receiver);

    /** removes a certain TEDObject
     *  \return list of events the TEDObject listen.
     */
    list<string> removeEventListener(EDObject*);

    /** returns the registered event-keys
     */
    list<string> getEventList()              const;

    /** returns the EDObject with a specific event-key, NULL if no such
     * object exist.
     */
    EDObject* getEDObject(const string& event_key) ;

    /** returns an Event prototype has a substring s as event ket, NULL if no such
      * Event exist
      */
    Event* matchEventPrototype(const string& s);

    /** returns the EDObject with a specific event-key, NULL if no such
     * object exist.
     */
    const EDObject* getEDObject(const string& event_key) const ;

    /** returns true, if Event with key e is registered
     */
    bool hasEvent(const string & e) const;


    /** sets the required bit for an event and two states. If the eventdispatcher is in state from_state
     * and tries to change its state to to_state it checks if an event with key event_key
     */
    void addRequired(const string& event_key, int form_state, int to_state);

    /**
     */
    void addLimited(const string& event_key, int from_sate);

    //@}

    /*! \name Processing and configuration methods
     * These methods can be used to process events and to configure and control the states
     */
    //@{
    /** This method dispatches an event. If any failure occurs during the dispatching process
     * an Exception is created and stored inside the event.
     * \return returns FAILURE if an exception occurs or any other failure, otherwise SUCCESS
     */
    Result dispatch(const Event&, const IEventSender* = 0);

    /**
     * This method processes a composite event inside a node and dispatches all registered
     * methods of the node. 
     * \return FAILURE if one dispatched method returned FAILURE, otherwise SUCCESS
     */
    Result internalProcessCEvent(CEvent* cevent);

    /**
     * This method processes a composite event when leaving a node upstream or downstream:
     * all external listeners are dispatched.
     * \return currently always SUCCESS
     */
    Result externalProcessCEvent(CEvent* cevent);

    /** reset the execution value of state s
     */
    void reset(int state);

    /**
     */
    void reset(const string&, int state);

    /** If all EDObjectes with a mandatory bit for state s are executed successfully 
     * the Eventdispatcher changes the state to s. Otherwise an UnsuccessfulRequiredCallException
     * is thrown.
     */
    void changeState(int state )
      throw (const UnsuccessfulRequiredCallException&);

    /** returns the Eventdispatcher state
     */
    int getState()                           const {return state;}

    /** This methods returns true if the EventDispatcher processes all events with a mandatory bit
     * for the current state.
     */
    void checkMandatoryEvents(int new_state) const
      throw (const UnsuccessfulRequiredCallException&);

    /** This method returns an EventPrototype for unhandeld events.
     * IMORTANT NOTE! The caller of this method is responsible to delete the Events inside the list
     */
    list<Event*> getUnhandledMandatoryEvents(int new_state) const;

    //@}

    void forwardEvent(const Event&, const IEventSender* );
    void externalDispatch(const Event&, const IEventSender* );
    /** returns true if an IEventReceiver is listening to all events
      */
    inline bool hasListener() const { return !__ext_recv_forwarding_list.empty(); }

    /** returns true if an IEventReceiver or TEDObject is listening to a specific event with givent event key
      */
    bool hasSpecificListener(const string&) const;

    /** returns true if an IEventReceiver or TEDObject is listening to any specific event.
      */
    bool hasSpecificListener() const;

    /** returns a list of events all specific TEDObjects and IEventReceiver listen too.
      */
    list<string> getSpecificListener() const;
  private:
    struct string_hash
    {
      size_t operator()(const string& __s) const { return HASH_MAP_STL::__stl_hash_string(__s.c_str()); }
    };

    typedef HASH_MAP_STL::hash_map<string, EDObject*, string_hash >::iterator event_it;
    typedef HASH_MAP_STL::hash_map<string, EDObject*, string_hash >::const_iterator cevent_it;

    typedef HASH_MAP_STL::hash_map<string, list<EDObject*>, string_hash >::iterator EDOMapIt;
    typedef HASH_MAP_STL::hash_map<string, list<EDObject*>, string_hash >::const_iterator CEDOMapIt;

    typedef HASH_MAP_STL::hash_map<string, list<IEventReceiver*>, string_hash >::iterator ReceiverMapIt;
    typedef HASH_MAP_STL::hash_map<string, list<IEventReceiver*>, string_hash >::const_iterator CReceiverMapIt;

    ThreadMutex                                             mutex; 
    ThreadMutex                                             external_mutex;
    ThreadMutex                                             __internal_process_mutex;

    /** stores the EDObjects that are dispatched primarily in the dispatch()-method */
    HASH_MAP_STL::hash_map<string, EDObject*, string_hash >               __int_event_map;
    /** stores the external EDObjects that are dispatched in the externalDispatch()-method 
     *	(called by dispatch() and externalProcessCEvent() 
     */
    HASH_MAP_STL::hash_map<string, list<EDObject*>, string_hash >         __ext_event_map; 
    /** stores the IEventReceiver objects that are dispatched in the externalDispatch()-method (after __ext_event_map)
     *	(called by dispatch() and externalProcessCEvent() 
     */
    HASH_MAP_STL::hash_map<string, list<IEventReceiver*>, string_hash >   __ext_event_recv_map;
    /** stores all IEventReceiver objects that are dispatched in the forwardEvent()-method 
     *	(called by dispatch() and externalProcessCEvent() 
     */
    list<IEventReceiver*>                                   __ext_recv_forwarding_list;
    int                                                     state;
  };


}//end namespace
#endif
