/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_FORMAT_STRATEGY_MAP_HPP
#define NMM_FORMAT_STRATEGY_MAP_HPP

#include "nmm/base/format/Format.hpp"

#include <map>
using std::pair;

#include <list>
using std::list;

namespace NMM {
  
  /** Format strategy map.
   * Maps NMM formats to transport strategy types. Such maps are used to determine which transport
   * strategy types support a given format.
   */
  class FormatStrategyMap {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Drfault constructor.
     */
    FormatStrategyMap();
    
    /** Virtual destructor.
     */
    virtual ~FormatStrategyMap();
    //@}
    
    /** @name Format Map
     */
    //@{
    /** Add supported format.
     * This method declares that the given transport strategy supports the given format by adding
     * an entry to the internal format map.
     * @param strategy_type Strategy type
     * @param format The format to add
     */
    void addSupportedFormat(const string& strategy_type, const Format& format);
    
    /** Get list of supported strategies.
     * This method returns the list of transport strategies that support the given format.
     */
    list<string> getSupportedStrategies(const Format& format) const;

    /** Get list of supported formats.
     * This method returns the list of formats which are supported by the given transport strategy.
     */
    list<Format*> getSupportedFormats(const string& strategy_type);
    
    /** Remove strategy type.
     * This method removes the given strategy type from the map.
     * @param strategy_type Strategy type
     */
    void removeStrategyType(const string& strategy_type);    
    
    /** Check if a combination of format and transport straegy is supported.
     * @param format Format to check
     * @param strategy_type Transport strategy type
     */
    bool isSupported(const Format& format, const string& strategy_type);
    //@}
    
    /** This method displays the map on an output stream.
     */
    void display(ostream& out);
    
  private:
    typedef pair<Format, string>       FormatMapEntry;
    typedef list<FormatMapEntry>       FormatMap;                ///< Type to map a format to a set of strategies
    typedef FormatMap::iterator        FormatMapIterator;
    typedef FormatMap::const_iterator  FormatMapConstIterator;
    
    FormatMap __format_map;                                      ///< The format map
  };
}

#endif


