/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_GENERICBIDIRNODE_HPP_
#define _NMM_GENERICBIDIRNODE_HPP_

#include "QueuedInputJack.hpp"
#include "ThreadedNode.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/memorymgr/IExternalBufferManager.hpp"
#include "nmm/interfaces/base/IInstreamImpl.hpp"
#include "nmm/interfaces/base/IQueuedImpl.hpp"
#include "nmm/interfaces/base/IBufferManagerImpl.hpp"
#include "nmm/interfaces/base/IFilterNode.hpp"
#include "nmm/interfaces/base/IFormatChangeImpl.hpp"
#include "nmm/interfaces/base/sync/ISwitchImpl.hpp"
#include "nmm/base/Usage.hpp"
#include "JackGroup.hpp"

namespace NMM {

  class GenericNode : public ThreadedNode,
		      public IExternalBufferManager,
		      public ISignalReceiver,
		      public IBiDirQueuedNodeImpl,
		      public IInstreamImpl,
		      public IBufferManagerImpl,
		      public IFormatChangeImpl,
		      public ISwitchImpl,
                      public Usage
  {

  private:
    IFilterNode* filter_node; ///< filter_node is unequal null, if this node inherits from GenericFilterNode

    int flush_counter; ///< counts how many doFlush's have returned DELETE (-1) or not DELETE (+1), needed for FlushEvent
    volatile bool working_flag; ///< if this flag is set processBuffer will be called with a 0-pointer
    volatile bool producing_flag; ///< if this flag is set processBuffer will be called with a 0-pointer, if no messages are available
    CEvent* current_event; ///< stores the current event (for deferred events)
    volatile unsigned int event_working_index; ///< stores the index in the current event
    unsigned int pre_fetch_queue_size; 
    /**
     * This struct contains all the information associated with a single
     * input stream. These are the jack to be used, the properties (formats)
     * of the jack, and the queue for incoming buffers.
     */
    class InputStream_t {
    public:
      QueuedInputJack *jack; ///< The actual jack object.
      const Format *format;  ///< The actual format of this stream.
      bool enabled;          ///< If enabled, message will be dequeued
      bool needed;           ///< If true, the stream will not be removed during removeUnneededStreams()

      InputStream_t() : jack(0), format(0), enabled(true), needed(true) {};
      ~InputStream_t() { if(jack) delete jack; if(format) delete format; };
    };

    /**
     * This struct contains all the information associated with a single
     * output stream. These are the jack to be used, the properties (formats)
     * of the jack, and the queue for incoming buffers.
     */
    class OutputStream_t {
    public:
      JackGroup *jack_group;///< The actual jack (group) object.
      const Format *format;  ///< The actual format of this stream.
      bool enabled;          ///< If enabled, message will be dequeued
      bool needed;           ///< If true, the stream will not be removed during removeUnneededStreams()

      OutputStream_t() : jack_group(0), format(0), enabled(true), needed(true) {};
      ~OutputStream_t() { if(jack_group) delete jack_group; if(format) delete format; };
    };

    // needed for map (see STL book)
    struct ltstr {
      bool operator()(const char *s1, const char *s2) const {
        return strcmp(s1, s2) < 0;
      }
    };

    int queue_size; ///< Used to remember the size set in the constructor.
    StreamQueue::Mode queue_mode; ///< Used to remember the queue mode set in the constructor.

    /**
     * This map contains all the input streams. Input streams are stored
     * as InputStream_t structs and accessed via their associated label
     * (the "jack tag"), hence the extra compare function.
     */
    typedef map<const char*, InputStream_t*, ltstr> InputStreamMap_t;
    mutable InputStreamMap_t input_streams; ///< The actual map

    /**
     * This map contains all the output streams. Output streams are stored
     * as OutputStream_t structs and accessed via their associated label
     * (the "jack tag"), hence the extra compare function.
     */
    typedef map<const char*, OutputStream_t*, ltstr> OutputStreamMap_t;
    mutable OutputStreamMap_t output_streams; ///< The actual map

    /** These status variables belong to the output-side of the node */
    int downstr_blocked_queue_count; ///< counts how many downstream connections are blocked
    bool output_queue_not_empty; ///< states if a message is enqueued in some output jack

    /** These status variables belong to the input-side of the node */
    int upstr_blocked_queue_count;  ///< counts how many upstream connections are blocked
    bool input_queue_not_empty; ///< states if a message is enqueued in some input jack

    /** 
     * If a message could be dequeued from inputs, this member holds the input stream from 
     * which the this message was dequeued, otherwise is undefined
     */
    InputStreamMap_t::iterator curr_recv_in_stream;

    /** 
     * If a message could be dequeued from outputs, this member holds the output stream from 
     * which the this message was dequeued, otherwise is undefined
     */
    OutputStreamMap_t::iterator curr_recv_out_stream;
    
    /**
     * Internal method called from run-method to obtain the next message
     * from all enabled outputs in a round-robin manner.
     */ 
    Message *getNextUpstrMessage();

    /**
     * Internal method called from run-method to obtain the next message
     * from all enabled inputs in a round-robin manner.
     */ 
    Message *getNextDownstrMessage();

    /**
     * Sends a message. Upstream messages are sent upstream, 
     * downstream messages are sent downstream. 
     * Buffer-message are sent to the currently set input or output
     * stream. Event-messages are sent to all inputs or outputs.
     */
    virtual Result sendMessage(Message *buffer);

    JackGroup* curr_send_out_stream; ///< Stores a pointer to which output jack the next DOWNSTREAM Message-object will be sent.
    InputJack* curr_send_in_stream;   ///< Stores a pointer to which input jack the next UPSTREAM Message-object will be sent.

    queue<int> signal_queue;          ///< here signals are enqueued before being process in run()
    ThreadMutex signal_queue_mutex;  ///< locks the signal queue
    ThreadCondition signal_queue_cv; ///< cond. variable for waitForSignal()

    /** set to false in constructor, set to true whenever START is handled in run()-method, set to false for STOP */
    bool internal_started_state; 
    ThreadMutex internal_started_mutex;
    ThreadCondition internal_started_cv;

    Message::Direction current_msg_direction; ///< stores the direction of the last dequeued message

    ThreadMutex __internal_statechange_mutex; ///< Mutex to avoid deadlock during state change, that can be caused by block()/unblock()e
  protected://Interface methods
    /**
     * ISwitch interface.
     */
    virtual void switchOutputs();
    virtual void flushOutputs(const Timestamp& ts);
    virtual void setSwitchOutputsEnabled(const bool& enable);
    virtual bool getSwitchOutputsEnabled();
    bool perform_switch; ///< is set when the switch operation should be performed
    bool perform_flush; ///< is set when the flush operation should be performed
    Timestamp perform_flush_ts; ///< the timestamp for the flush operation
    bool switch_enabled; ///< stores whether switching should be performed at all (defaults to false for GenericNode)
    ThreadMutex internal_message_queue_mutex; ///< guards switching operations

    class InternalMessage_t {
    public:
      Message* switch_to_msg;
      Message* switch_from_msg;
      bool reset_switch_to;
      InternalMessage_t() : switch_to_msg(0), switch_from_msg(0), reset_switch_to(false) { };
      InternalMessage_t(Message* to_msg, Message* from_msg, bool reset) 
	: switch_to_msg(to_msg), switch_from_msg(from_msg), reset_switch_to(reset) { };
    };

    typedef queue<InternalMessage_t> InternalMessageQueue_t;
    InternalMessageQueue_t internal_message_queue; ///< queue for internal message (e.g. handover)

    /**
     * IFormatChange interface.
     * Set new input format
     */
    virtual Result setFormat(const Format& format);
    
    inline Result protectedSendMessage(Message* msg) { return sendMessage(msg); };

    /* methods from IBufferManager */
    virtual NMMReference getBufferManager();
    virtual Result useBufferManager(const IBufferManager& manager);
    virtual void releaseBufferManager();

    /**
     * This method returns the successor node of this node.
     * \param jack the name of the output.
     * \return a node or NULL.
     */
    virtual INode*        getSuccessorNode(const string & = "default");

    /**
     * This method returns the predecessor node of this node.
     * \param jack the name of the input.
     * \return a node or NULL.
     */
    virtual INode*        getPredecessorNode(const string & = "default");
    virtual IInputJack*   getInputJack(const NMM::Format &, const string &);
    virtual IInputJack*   getInputStream(const string &);
    virtual IOutputJack*  getOutputJack(const NMM::Format &, const string &);
    virtual IOutputJack*  getUnintializedOutputJack(const Format& format, const string& jack_tag);
    virtual IOutputJack*  getOutputStream(const string &);

    // IBiDirQueued
    /** Method from interface IBiDirQueuedNode */
    virtual unsigned int getDownstreamSize(const string& id = "default") const;
    /** Method from interface IBiDirQueuedNode */
    virtual unsigned int getDownstreamMaxSize(const string& id = "default") const;
    /** Method from interface IBiDirQueuedNode */
    virtual IQueued::Mode getDownstreamMode(const string& id = "default") const;
    /** Method from interface IBiDirQueuedNode */
    virtual Result setDownstreamMaxSize(const unsigned int& size, const string& id = "default"); /* limited(Node::CONSTRUCTED) */
    /** Method from interface IBiDirQueuedNode */
    virtual Result setDownstreamMode(const IQueued::Mode& mode, const string& id = "default"); /* limited(Node::CONSTRUCTED) */
    /** Method from interface IBiDirQueuedNode */
    virtual Result unlimitDownstreamSize() /* limited(INode::STARTED) */ /* in-stream */;
    /** Method from interface IBiDirQueuedNode */
    virtual Result limitDownstreamSize() /* limited(INode::STARTED) */ /* in-stream */ ;
    /** Method from interface IBiDirQueuedNode */
    virtual unsigned int getUpstreamSize(const string& id = "default") const;
    /** Method from interface IBiDirQueuedNode */
    virtual unsigned int getUpstreamMaxSize(const string& id = "default") const;
    /** Method from interface IBiDirQueuedNode */
    virtual IQueued::Mode getUpstreamMode(const string& id = "default") const;
    /** Method from interface IBiDirQueuedNode */
    virtual Result setUpstreamMaxSize(const unsigned int& size, const string& id = "default"); /* limited(Node::CONSTRUCTED) */
    /** Method from interface IBiDirQueuedNode */
    virtual Result setUpstreamMode(const IQueued::Mode& mode, const string& id = "default"); /* limited(Node::CONSTRUCTED) */
    /** Method from interface IBiDirQueuedNode */
    virtual Result setMaxSize(const unsigned int& size, const string& id = "default");
    /** Method from interface IBiDirQueuedNode */
    virtual Result setMode(const IQueued::Mode& mode, const string& id = "default");

    /** Method from interface INode */
    virtual Format* getInputStreamFormat(const string& jack_tag = "default");
    /** Method from interface INode */
    virtual Format* getOutputStreamFormat(const string& jack_tag = "default");
  protected:
    /**
     * Template method that gets called by setFormat (can be implemented in sub-classes)
     */
    virtual Result doSetFormat(const Format&) { return SUCCESS; }

    int dequeueSignal();       ///< dequeues the next Signal (blocking)
    bool isSignalQueueEmpty(); ///< requests the status of the internal signal queue
    Result waitForSignal();    ///< waits for a new signal to arrive (blocking)

    /**
     * inserts an event behind the current position within the cevent
     * This method is called from event-handling methods to insert
     * an event.
     */
    void insertEvent(const Event& e);  

    /**
     * inserts an event behind the current position within the cevent
     * This method is called from event-handling methods to insert
     * an event. The CEvent does the memory management for the given pointer.
     */
    void insertEvent(Event* e);

    /**
     * returns the direction of the last dequeued message
     */
    inline Message::Direction getCurrentMsgDirection() const { return current_msg_direction; }

    /**
     * This template method can be used in plugins to access instream composite events.
     * The method is called before the composite event is processed from the node. 
     */
    virtual Result processCEvent(CEvent* ) { return SUCCESS;}

    /**
     * Sets the working-flag. If this flag is true, processBuffer() is called with 
     * a null-Pointer even if message (i.e. a buffer or an event) could be dequeued.
     * This flag should be set to 'true' in two cases
     * (1) The node receives one buffer as input and produces more than one buffer/event
     *     as output. 
     * (2) The node receives an event and wants to produce one or more buffers/events
     *     in response.
     * Warning: the access to the working-flag is not synchronized.
     */
    void setWorkingFlag(const bool working);

    /**
     * Sets the producing-flag. If this flag is true, processBuffer() is called with 
     * a null-Pointer unless there is a message (i.e. a buffer or an event) which can
     * be dequeued.
     * This flag should mainly be set to 'true' in one case
     * (1) The node is a source node and wants to produce messages all the time
     *     while still being able to process new messages received.
     *     GenericNode has set this flag to false by default.
     *     GenericSourceNode has set this flag to true by default.
     * Warning: the access to the producing-flag is not synchronized.
     */
    void setProducingFlag(const bool working);

    /**
     * This method adds an input stream. The corresponding jack will
     * automatically be constructed. 
     * \param jack_tag the name of the input.
     * \returns SUCCESS if the stream was correctly added, FAILURE otherwise.
     */
    Result addInputStream(const char *jack_tag);

    /**
     * This method adds an output stream. The corresponding jack group will
     * automatically be constructed.
     * \param jack_tag the name of the output.
     * \returns SUCCESS if the stream was correctly added, FAILURE otherwise.
     */
    Result addOutputStream(const char *jack_tag);

    /**
     * This method schedules an input stream to be removed when 
     * removeUnneededInputStreams() is called next time.
     * \param jack_tag the name of the input.
     * \returns SUCCESS if the stream was correctly added, FAILURE otherwise.
     */
    Result removeInputStream(const char *jack_tag);

    /**
     * Removes all output streams with needed == false
     */
    void removeUnneededInputStreams();

    /**
     * This method schedules an output stream to be removed when 
     * removeUnneededOutputStreams() is called next time.
     * \param jack_tag the name of the output.
     * \returns SUCCESS if the stream was correctly added, FAILURE otherwise.
     */
    Result removeOutputStream(const char *jack_tag);

    /**
     * Removes all output streams with needed == false
     */
    void removeUnneededOutputStreams();

    /**
     * This method can be called from processBuffer() or an
     * event handling method to determine from
     * which input the new buffer comes from.
     * The getCurrentMsgDirection() method should be called before
     * to check that the message has direction DOWNSTREAM
     * \return the input/output jack tag of the buffer 
     */
    const char* getCurrentRecvInputStream() const;

    /**
     * This method can be called from processBuffer() or an
     * event handling method to determine from
     * which output the new buffer comes from.
     * The getCurrentMsgDirection() method should be called before
     * to check that the message has direction UPSTREAM
     * \return the input/output jack tag of the buffer 
     */
    const char* getCurrentRecvOutputStream() const;

    /**
     * Use this method to enable or disable input jacks.
     * If a jack is enabled and contains a message, this message will
     * be processed: a buffer will be passed to processBuffer(), an event
     * will be handled. The next time another input will
     * be checked (round-robin). 
     * \param jack_tag the name of the input
     * \param enable the wanted state for the input
     * \return SUCCESS if input was found, else FAILURE
     */
    Result setRecvInputStreamEnabled(const char *jack_tag, bool enable);

    /**
     * This method can be used to check if an input
     * is currently enabled for receiving and dequeueing.
     * \param jack_tag the name of the input
     * \return status of the input
     */
    bool getRecvInputStreamEnabled(const char *jack_tag) const;

    /**
     * Use this method to enable or disable output jacks.
     * If a jack is enabled and contains a message, this message will
     * be processed: a buffer will be passed to processBuffer(), an event
     * will be handled. The next time another input will
     * be checked (round-robin). 
     * \param jack_tag the name of the input
     * \param enable the wanted state for the input
     * \return SUCCESS if input was found, else FAILURE
     */
    Result setRecvOutputStreamEnabled(const char *jack_tag, bool enable);

    /**
     * This method can be used to check if an output
     * is currently enabled for receiving and dequeueing.
     * \param jack_tag the name of the output
     * \return status of the output
     */
    bool getRecvOutputStreamEnabled(const char *jack_tag) const;

    /**
     * Since processBuffer can only return a Message-object, one needs
     * to set the output jack for the next downstream message to be sent. 
     * \param jack_tag name of the output
     * \return SUCCESS if output is valid, FAILURE if not
     */
    Result setCurrentSendOutputStream(const char *jack_tag); 

    /**
     * Get the currently set sending output stream.
     * \return output stream jack tag
     */
    const char* getCurrentSendOutputStream() const;

    /**
     * Since processBuffer can only return a Message-object, one needs
     * to set the input jack for the next upstream message to be sent. 
     * \param jack_tag name of the input
     * \return SUCCESS if output is valid, FAILURE if not
     */
    Result setCurrentSendInputStream(const char *jack_tag); 
    
    /**
     * Get the currently set sending input stream.
     * \return input stream jack tag
     */
    const char* getCurrentSendInputStream() const;

    /**
     * Process one buffers.  
     */
    virtual Message *processBuffer(Buffer *) = 0;

    /**
     * This method returns the list of available input stream tags.
     * \return a list of input stream tags as strings.
     */
    virtual list<string >  getInputStreamTags() const;

    /**
     * This method returns the list of available output stream tags.
     * \return a list of output stream tags as strings.
     */
    virtual list<string >  getOutputStreamTags() const;

    /**
     * This method returns a cloned output jack and fills the given format with
     * the intersection format
     * \param jack_tag the name of the jack
     * \param intersec_format the format to be filled with the intersection format
     * \return the cloned output jack
     */
    virtual IOutputJack* cloneOutputJack(Format*& intersec_format, const string& jack_tag = "default");


    /** This method configures the output format of jack with name jack_tag
      * to given format. If the format is not supported by the OwnWorkingOutputProperty
      * an exception is thrown. 
      */
    virtual void setOutputFormat(const Format&, const string& jack_tag = "default")
      throw (const Exception&);

  public:
    GenericNode(const char *name = "GenericNode",
		     const StreamQueue::Mode queue_mode = StreamQueue::DEFAULT_MODE,
		     const int size = StreamQueue::DEFAULT_SIZE);
		     
    ~GenericNode();
      
    /**
     * This method enables the pre-fetch mode and at the same time sets
     * the number of messages to be received before pre-fetching can be
     * finished. This method has to be called in state CONSTRUCTED.
     */
    void setPrefetchQueueSize(const unsigned int& pre_fetch_queue_size);

    /**
     * Request the number of messages to be prefetched. If zero,
     * prefetching is disabled.
     */
    unsigned int getPrefetchQueueSize();

    /**
     * Blocks until the process of prefetching is finished. This
     * method has to be called in state ACTIVATED.
     */
    void waitPrefetchFinished();

    /**
     * State transition method init() from superclass.
     */
    void init()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * State transition method initOutput() from superclass.
     */
    void initOutput()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * State transition method start() from superclass.
     */
    void start()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * State transition method stop() from superclass.
     */
    void stop()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * State transition method deinitOutput() from superclass.
     */
    void deinitOutput()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * State transition method activate() from superclass.
     */
    void activate()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method internally blocks the node (basically the same as stop, only
     * the state of the node is not changed and the method does not block until the node
     * is really stopped)
     */
    virtual void block();

    /**
     * This method internally unblocks the node (basically the same as start, only
     * the state of the node is not changed)
     */
    virtual void unblock();

    /** 
     * This method is called from successors or predecessors nodes of this node to signal
     * the status of the current send/receiveMessage (e.g. BLOCKED or FREE_QUEUE_SLOT)
     */
    void receiveSignal(Signal event);

    /**
     * The run method from ThreadedNode. The default action of the run
     * method is to call the process() method whenever a Buffer is input.
     * Subclasses that need more control should overwrite this method.
     * This behaviour is both supported and encouraged.
     */
    virtual void run(int num_times = 0);
 

    /**
     * The getInputJack method from Node. 
     */
    virtual const InputJack* getInputJack(const Format *,
                                          const char * jack = "default") const;
    
    /**
     * The getOutputJack method from Node.
     */
    virtual const OutputJack* getOutputJack(const Format* , const char* jack = 0, bool output_initialized = true) const;
    
    /**
     * This method causes the node to perform a flush-operation. 
     * Here, this means flushing all input and output queues. Messages will be released.
     * Additionally, doFlush() will be called. If received instream, only the receiving
     * input or output will be flushed.
     */
    virtual Result flush();

    /**
     * Obtain the format associated with the named input stream.
     * The Format pointer is
     * guaranteed to remain valid over the connectedness of the node.
     *
     * \param jack_tag The tag of the input stream.
     * \returns a pointer to the corresponding format.
     */
    virtual const Format* getInputFormat(const char * jack_tag = "default") const ;

    /**
     * Obtain the format associated with the named output stream.
     * The Format pointer is
     * guaranteed to remain valid over the connectedness of the node.
     *
     * \param jack_tag The tag of the output stream.
     * \returns a pointer to the corresponding format.
     */
    virtual const Format* getOutputFormat(const char * jack_tag = "default") const ;

    /**
     * Currently used for sending instream messages from the application.
     * Should be replaced by some other mechanism.
     */
    virtual InputJack* getOwnInputStream(const string& jack_tag = "default");
    virtual OutputJack* getOwnOutputStream(const string& jack_tag = "default");

    virtual const InputJack* getOwnInputStream(const string&  = "default") const;
    virtual const OutputJack* getOwnOutputStream(const string&  = "default") const;

    virtual Result sendEventUpstr(CEvent* cevent, const char * jack_tag = "default");
    virtual Result sendEventUpstr(CEvent& cevent, const char * jack_tag = "default");
    virtual Result sendEventUpstr( Event&  event, const char * jack_tag = "default");
    virtual Result sendEventDownstr(CEvent* cevent, const char * jack_tag = "default");
    virtual Result sendEventDownstr(CEvent& cevent, const char * jack_tag = "default");
    virtual Result sendEventDownstr(Event&   event, const char * jack_tag = "default");

    /* Forwards incoming instream messages from application to the first jack available*/
    virtual Result sendMessageUpstream(const CEvent& cevent);

    /* Forwards incoming instream messages from application to the first jack available*/
    virtual Result sendMessageDownstream(const CEvent& cevent);

    /** 
     * Gets the status of the working-flag
     */
    bool getWorkingFlag() const;

    /** 
     * Gets the status of the producing-flag
     */
    bool getProducingFlag() const;

    // test for examples/mp3dec/simpleplay
    Result triggerSendMessage(Message *m) { return (this->sendMessage(m)); }
    // test for examples/mp3dec/simpleplay   
    Message *triggerDequeueMessage()        { return getNextDownstrMessage(); } 
    // Test method for scheduling/performance tests
    virtual Message *triggerProcessBuffer(Buffer *buffer) { return processBuffer(buffer); }
    // Test: Get size of queue 
    unsigned int getQueueSize(const char* jack_tag = "default");
    // Test method for scheduling/performance tests 
    virtual Result triggerProcessCEvent(CEvent *e) { return processCEvent(e); }


  /****************************** analyseData and helpers ******************************/
  private:
    /**
     * This class contains all the information needed to cache
     * sent-out messages while analysing.
     */
    class AnalysingCacheEntry {
    public:
      //! the cached message.
      Message& message;
      //! The jack the cached message should be send through.
      list<JackGroup*> jacks;

      AnalysingCacheEntry(Message& _message, list<JackGroup*> _jacks)
        : message(_message), jacks(_jacks) {}
    };

    /**
     * This queue caches buffers returned by processBuffer() while
     * analysing data. They are send to the according OutputJacks at
     * the end of analysis, which store them until the Node is started.
     */
    list<AnalysingCacheEntry>* analysing_cache;

    //! if this flag is set processBuffer should set its output formats
    //  according to the analized data
    bool analysing_flag;

    //! if this flag is set (mostly by processBuffer), the analysing flag
    //  is updated at the end of the current run() iteration
    bool cached_analysing_flag;

    //! the maximum number of iterations allowed in analyseData()
    unsigned int max_analysis_iterations;

    //! if this flag is set the run method should produce data in
    //  ACTIVATED
    bool analyse_producing_flag;

    //! if this flag is set the run method should process data in
    //  ACTIVATED
    bool analyse_processing_flag;

    //! flag to indicate that at least one Buffer was forwarded through
    //  the OutputJack, to which requested data should be send!
    bool analysis_data_forwarded;

    //! indicates, that the predecessing node, from whom we requested
    //  data for "GenericNode::analyseData()", could not provide any
    bool send_data_failed;

    //! Stores a pointer to which output jack a Message must be sent
    //  before we can leave analyse-processing-mode.
    JackGroup* current_analysis_out_stream;

    //! Stores a pointer to which input jack a Message must be sent when
    //  the Node needs more data in analyse-processing-mode.
    InputJack* current_analysis_in_stream;

    //! Stores a iterator (from the input_streams map) to which input
    //  stream the next Message must be sent when the Node needs more
    //  data in analyse-processing-mode.
    InputStreamMap_t::iterator next_analysis_in_stream;

    /**
     * checkes everything needed after run() has successful processed a Buffer.
     * \returns true iff run() should break out of the infinite while-loop.
     */
    bool checkAnalysis();

    /**
     * small helper to send a "INode::sendDataFailed" downstream.
     * \param jack The OutputJack, to send the Event through.
     * \returns true iff the Event could be sent.
     */
    bool transmitSendDataFailed(JackGroup& jack);

    /**
     * small helper to send a "INode::sendData" upstream.
     * \param jack The InputJack, to send the Event through.
     * \returns true iff the Event could be sent.
     */
    bool transmitSendData(InputJack& jack);

    /**
     * small helper to send the next AnalysingCacheEntry to the given
     * Output.
     * \param jack The OutputJack, to send the entry to.
     * \returns true iff there was a entry found.
     */
    bool sendAnalysingCacheEntryTo(JackGroup& jack);

    /**
     * This method computes the next InputJack to send a
     * "INode::sendData" to.
     * \returns a pair of the jack tag and the Pointer of the
     *          InputJack to sent the Event to.
     */
    pair<string, InputJack*> getNextAnalysisInput();

    /**
     * Updates the analysing-flag. If the setAnalysingFlag() tried to
     * set this flag false (mostly from processBuffer() when finished
     * analysing), this method updates the flag realy to false.
     * This method is needed to avoid a iteration of then run() loop
     * with changing analysing-flag. Only at the end of each
     * iteration, the flag is updated to its real value.
     * Warning: the access to the analysing-flag is not synchronized.
     */
    void updateAnalysingFlag();

    /**
     * Sets the analyse-producing-flag. If this flag is true, the run
     * method should produce Buffer (by calling processBuffer()) in
     * ACTIVATED.
     * Warning: the access to the analyse-producing-flag is not
     * synchronized.
     */
    void setAnalyseProducingFlag(const bool analyse_producing);

    /**
     * Sets the analyse-processing-flag. If this flag is true, the run
     * method should process incoming Buffers in ACTIVATED.
     * Warning: the access to the analyse-processing-flag is not
     * synchronized.
     */
    void setAnalyseProcessingFlag(const bool analyse_processing); 

  protected:
    /**
     * This method does all things need, to get some data from this
     * node's predecessor(s) to be analysed by this node's
     * processBuffer method until the processBuffer method signals, that
     * the incoming data is sufficient to be able to set this node's
     * working output property.
     * \param in_jack The jack tag of the InputJack to analyse data from
     * \returns SUCCESS if the analysis succeded, FAILURE otherwise.
     */
    virtual bool analyseData(const string& in_jack = "default");

    /**
     * Sets the analysing-flag. If this flag is true, processBuffer()
     * should analyse the incoming data and set its working output
     * properties and internal variables according to the (format)
     * charakteristics found.
     * Warning: the access to the analysing-flag is not synchronized.
     */
    void setAnalysingFlag(const bool analysing);

    /** 
     * Gets the status of the analysing-flag
     */
    bool getAnalysingFlag() const;

    /**
     * Sets the maximum number of iterations while analysing data.
     * \see analyseData()
     * Warning: the access to the max-analysis-iterations value is not synchronized.
     */
    void setMaxAnalysisIterations(const unsigned int iterations);

    /** 
     * Gets the number of iterations allowed in data analysis
     */
    unsigned int getMaxAnalysisIterations() const;

    /** 
     * Gets the status of the analyse-producing-flag
     */
    bool getAnalyseProducingFlag() const;

    /** 
     * Gets the status of the analyse-processing-flag
     */
    bool getAnalyseProcessingFlag() const;

    // INode Interface
    /**
     * this method signals instream, that a successing node requests
     * some data to do its "GenericNode::analyseData()".
     * \returns DELETE always.
     */
    virtual Result sendData();

    /**
     * this method signals instream, that a predecessing node could
     * not provide some data for "GenericNode::analyseData()"
     * \returns DELETE always.
     */
    virtual Result sendDataFailed();
  /****************************** analyseData and helpers ******************************/
  };
}

#endif

