/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "InstreamChannel.hpp"

#include "nmm/base/ProxyObject.hpp"

#include "nmm/interfaces/base/IInputJack.hpp"
#include "nmm/interfaces/base/IOutputJack.hpp"

#include "CompositeStrategy.hpp"

namespace NMM {

  InstreamChannel::InstreamChannel()
    :  CommunicationChannel(getStaticTypeInfo()),
       __i_strategies_valid(false),
       __strategy_list_valid(false)
  {
    registerAttributes();
  }

  InstreamChannel::InstreamChannel(IInputJack& input_jack, IOutputJack& output_jack)
    throw (const BindingException&)
    :  CommunicationChannel(getStaticTypeInfo()),
       __input_jack(dynamic_cast<IInputJack*>(input_jack.copy())),
       __output_jack(dynamic_cast<IOutputJack*>(output_jack.copy())),
       __i_strategies_valid(false),
       __strategy_list_valid(false)
  {
    registerAttributes();
  }

  InstreamChannel::InstreamChannel(const InstreamChannel& channel) : CommunicationChannel(channel), __i_strategies_valid(false)
  {
    registerAttributes();

    /* Store the IInputJack */
    if (channel.__input_jack.getValue())
      __input_jack.setValue(dynamic_cast<IInputJack*>(channel.__input_jack.getValue()->copy()));

    /* Store the IOutputJack */
    if (channel.__output_jack.getValue())
      __output_jack.setValue(dynamic_cast<IOutputJack*>(channel.__output_jack.getValue()->copy()));
  }

  InstreamChannel::~InstreamChannel() {
    delete __input_jack.getValue();
    delete __output_jack.getValue();
  }

  void InstreamChannel::registerAttributes() {
    insertArgument(&__input_jack);
    insertArgument(&__output_jack);
  }

  list<Interface*>& InstreamChannel::getInterfaceList() {
    if (!__i_strategies_valid)
      updateIStrategies();

    if (!__input_jack_strategy.get())
      return CommunicationChannel::getInterfaceList();

    return __input_jack_strategy->getParentObject()->getInterfaceList();
  }

  list<Interface*>& InstreamChannel::getPartnerInterfaceList() {
    if (!__i_strategies_valid)
      updateIStrategies();

    if (!__output_jack_strategy.get())
      return CommunicationChannel::getInterfaceList();

    return __output_jack_strategy->getParentObject()->getInterfaceList();
  }

  void InstreamChannel::updateIStrategies() {
    /* If no IOutputJack is set return */
    if (!__output_jack.getRValue())
      return;

    /* If no IInputJack is set return */
    if (!__input_jack.getRValue())
      return;


    /* If the TransportStrategy of the IInputJack is not set, request it */
    if (!__input_jack_strategy.get()) {
      __input_jack_strategy.reset(__input_jack->getTransportStrategy());
    }

    /* If the TransportStrategy of the IOutputJack is not set, request it */
    if (!__output_jack_strategy.get()) {
      __output_jack_strategy.reset(__output_jack->getTransportStrategy());
    }

    __i_strategies_valid = true;
  }

  void InstreamChannel::createStrategyList() {
    /* Clear strategy list */
    getStrategyList().clear();
  
    if (!__output_jack.getRValue()) {
      throw BindingException("Invalid output jack" );
    }

    if (!__input_jack.getRValue()) {
      throw BindingException("Invalid input jack" );
    }
  
    /* Begin> Determine the binding type */
    TSAddress::BindingType binding_type;
    if (__output_jack.getRValue()->isLocalTo(*__input_jack.getRValue())) {
      binding_type = TSAddress::LOCAL;
    }
  
    if (__output_jack.getRValue()->isSystemTo(*__input_jack.getRValue())){
      binding_type = TSAddress::SYSTEM;
    }
  
    if (__output_jack.getRValue()->isNetworkTo(*__input_jack.getRValue())){
      binding_type = TSAddress::NETWORK;
    }
    /* End< */
  
    /* Begin> Determine the strategies that can be used, based on binding type */
    switch (binding_type) {
    case TSAddress::LOCAL:
      /* Store the Local addresses, that can be used. */
      if (__output_jack.getRValue()->isLocalTo(*__input_jack.getRValue())) {
        registerStrategyType(TSAddress::LOCAL);
      }
      /* Fall through */
    
    case TSAddress::SYSTEM:
      /* Store the System addresses, that can be used. */
      if (__output_jack.getRValue()->isSystemTo(*__input_jack.getRValue())) {
        registerStrategyType(TSAddress::SYSTEM);
      }
      /* Fall through */
    
    case TSAddress::NETWORK:
      /* Store the Network addresses, that can be used. */
      if (__output_jack.getRValue()->isNetworkTo(*__input_jack.getRValue())) {
        registerStrategyType(TSAddress::NETWORK);
      }
    }
    /* End< */
  
    /* Make strategy list unique */
    getStrategyList().unique();
  
    /* Strategy list is complete */
    finishStrategyList(binding_type);
  }

  void InstreamChannel::registerStrategyType(TSAddress::BindingType binding_type) {
    list<string> input_jack_list  = __input_jack.getRValue()->getSupportedStrategies();
    list<string> output_jack_list = __output_jack.getRValue()->getSupportedStrategies();
  
    /* Begin> Check for default address in output_jack and supported in input_jack */
    for (unsigned int i = 0; i< __output_jack.getRValue()->getAddressList().size(); ++i) {
      if ( __output_jack.getRValue()->getAddressList()[i] ) {
        if (__output_jack.getRValue()->getAddressList()[i]->isBindingType(binding_type)) {
          const string& strategy_type = __output_jack.getRValue()->getAddressList()[i]->getStrategyType();
          if ( find(input_jack_list.begin(),
                    input_jack_list.end(),
                    strategy_type)
               !=   input_jack_list.end()
               )
          {
            addStrategyType(strategy_type);
          }
        }
      }
    }
    /* End */
  
    /* Begin> Check for default address in input_jack and supported in output_jack */
    for (unsigned int i = 0; i< __input_jack.getRValue()->getAddressList().size(); ++i) {
      if (__input_jack.getRValue()->getAddressList()[i]) {
        if (__input_jack.getRValue()->getAddressList()[i]->isBindingType(binding_type)) {
          const string& strategy_type = __input_jack.getRValue()->getAddressList()[i]->getStrategyType();
          if ( find(output_jack_list.begin(),
                    output_jack_list.end(),
                    strategy_type)
               !=   output_jack_list.end()
             )
          {
            addStrategyType(strategy_type);
          }
        }
      }
    }
    /* End< */
  }

  void InstreamChannel::updateAttributes() {
    createStrategyList();
  }
  
  
  TConstructor<InstreamChannel>  __nmm_proxy_channel_prototype ("InstreamChannel", "NMM");
  
  void InstreamChannel::setTransportStrategy(const string& strategy_type) {
    /* First check whether the required transport strategy can be used */
    if ( find( getStrategyList().begin(), getStrategyList().end(), strategy_type ) == getStrategyList().end() )
      throw BindingException("The TransportStrategy " + strategy_type + " can not be used to connect these jacks" );


    TSAddress* address = 0;
    address = __input_jack.getRValue()->getDefaultAddress(strategy_type);
    if (!address) { 
      address = __output_jack.getRValue()->getDefaultAddress(strategy_type);
      if (!address) { 
        throw BindingException("Could not get the default address for strategy " + strategy_type);
      }
      else {
        GLOBAL_DEBUG_STREAM("> input_jack must connect to OutputJack " << endl);
        __input_jack.getRValue()->setTransportStrategy(address);
      }
    } 
    else { 
      GLOBAL_DEBUG_STREAM("> output_jack must connect to InputJack " << endl);
      IOutputJack* tmp_jack = __output_jack.getRValue();
      tmp_jack->setTransportStrategy(address);
    }
  }
  
  void InstreamChannel::setTransportStrategy(const string& buffer_strategy_type, const string& event_strategy_type) {
    GLOBAL_DEBUG_STREAM2("InstreamChannel: Try to connect proxy channel using transport strategies:" << std::endl);
    GLOBAL_DEBUG_STREAM2("  for instream buffers: " << buffer_strategy_type << std::endl);
    GLOBAL_DEBUG_STREAM2("  for instream events : " << event_strategy_type << std::endl);
    
    /* first check whether the required transport strategies can be used */
    if ( find( getStrategyList().begin(), getStrategyList().end(), event_strategy_type ) == getStrategyList().end() )
      throw BindingException("InstreamChannel: The TransportStrategy " + event_strategy_type + " can not be used to connect these jacks" );
    if ( find( getStrategyList().begin(), getStrategyList().end(), buffer_strategy_type ) == getStrategyList().end() )
      throw BindingException("InstreamChannel: The TransportStrategy " + buffer_strategy_type + " can not be used to connect these jacks" );
    
    /* try connection in both directions */
    string comment;
    IJack* local_jack = __input_jack.getRValue();
    IJack* remote_jack = __output_jack.getRValue();
    do {
      /* try connection in the opposite direction */
      swap(local_jack, remote_jack);
      
      if(local_jack == __input_jack.getRValue()) {
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Try to connect input jack to output jack" << std::endl);
      } else {
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Try to connect output jack to input jack" << std::endl);
      }
      
      /* get the default address for CompositeStrategy */
      std::auto_ptr<TSAddress> master_address(remote_jack->getDefaultAddress("CompositeStrategy"));
      if(!master_address.get()) {
        comment = "Failed to get default address for CompositeStrategy";
        GLOBAL_DEBUG_STREAM2("InstreamChannel: " << comment << std::endl);
        continue;
      }
      
      /* get the default addresses of the two sub strategies */
      std::auto_ptr<TSAddress> event_address(remote_jack->getDefaultAddress(event_strategy_type));
      if(!event_address.get()) {
        comment = "Failed to get default address for " + event_strategy_type;
        GLOBAL_DEBUG_STREAM2("InstreamChannel: " << comment << std::endl);
        continue;
      }
      auto_ptr<TSAddress> buffer_address(remote_jack->getDefaultAddress(buffer_strategy_type));
      if(!buffer_address.get()) {
        comment = "Failed to get default address for " + buffer_strategy_type;
        GLOBAL_DEBUG_STREAM2("InstreamChannel: " << comment << std::endl);
        continue;
      }
      
      /* create a CompositeAddress */
      std::auto_ptr<TSAddress> address(new CompositeAddress(master_address->getNMMReference(),
                                                            *master_address, *buffer_address, *event_address));
      
      try {
        /* try to connect */
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Trying to connect" << std::endl);
        local_jack->setTransportStrategy(address.get());
        
        /* success */
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Connection setup successful" << std::endl);
        return;
      } catch(const BindingException& e) {
        GLOBAL_DEBUG_STREAM("InstreamChannel: Connection failed" << std::endl);
        GLOBAL_DEBUG_STREAM("InstreamChannel: " << e << std::endl);
        comment = e.getComment();
      }
    } while(local_jack == __output_jack.getRValue());
    
    /* we failed */
    GLOBAL_DEBUG_STREAM2("InstreamChannel: Connection failed in both directions" << std::endl);
    throw BindingException(comment);
  }
  
  void  InstreamChannel::setValueStream(const IValueStream& iv, const OValueStream& ov) {
    if (!__input_jack.getValue() )
      return;

    if (!__output_jack.getValue() )
      return;

    if (!__input_jack_strategy.get()) {
      __input_jack_strategy.reset(__input_jack->getTransportStrategy());
      if (!__input_jack_strategy.get()) {
        GLOBAL_WARNING_STREAM("No Transportstrategy available. Either the Jack is not connected or its a local binding" << endl);
        return;
      }
    }

    if (!__output_jack_strategy.get()) {
      __output_jack_strategy.reset(__output_jack->getTransportStrategy());
      if (!__input_jack_strategy.get()) {
        GLOBAL_WARNING_STREAM("No Transportstrategy available. Either the Jack is not connected or its a local binding" << endl);
        return;
      }
    }

    __input_jack_strategy->setSendValueStream(iv.getIValueStreamType());
    __output_jack_strategy->setSendValueStream(SValueFactory::getInstance().getOValueStreamPartner( ov.getOValueStreamType() ));
  }

  const list<string>& InstreamChannel::getStrategyList() const {
    if (!__strategy_list_valid) {
      // XXX
      const_cast<InstreamChannel*>(this)->__strategy_list_valid = true;
      const_cast<InstreamChannel*>(this)->createStrategyList();
    }
  
    return CommunicationChannel::getStrategyList();
  }

  list<string>& InstreamChannel::getStrategyList() {
    if (!__strategy_list_valid) {
      __strategy_list_valid = true;
      createStrategyList();
    }
  
    return CommunicationChannel::getStrategyList();
  }
  
  TypeInfo InstreamChannel::getStaticTypeInfo() {
    return TypeInfo("InstreamChannel", "NMM");
  }
  
}
