/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <algorithm>

#include "Interface.hpp"
#include "nmm/base/serialize/ValueStream.hpp"
#include "IReceiver.hpp"
#include "ProxyObject.hpp"
#include "TransportStrategy.hpp"
#include "TSFactory.hpp"
#include "NMMApplication.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"


namespace NMM {

Interface::Interface(NMMObject& receiver, const string& interface_type, const string& name_space)
  : SerializedValue(interface_type, name_space), __parent_object(&receiver), __receiver(0), __parent_reference(receiver.getReference()), __is_local(true)
{
  __parent_user_info.setValue(UserInfo());
  __address_list.useTValue(false);
  __supported_strategies.setValue(STSFactory::getInstance().getStrategyList());
  registerAttributes();
}

Interface::Interface(const string& interface_type, const string& name_space)
  : SerializedValue(interface_type, name_space), __parent_object(0), __receiver(0), __parent_reference(0), __is_local(true)
{
  __address_list.useTValue(false);
  __supported_strategies.setValue(STSFactory::getInstance().getStrategyList());
  registerAttributes();
}

Interface::Interface(const Interface& interface)
  : SerializedValue(interface), IEventSender(interface), __receiver(0), __is_local(true)
{
  __address_list.useTValue(false);
  registerAttributes();
  *this = interface;
}

Interface::~Interface() {
  if (__receiver)
    __receiver->decreaseInterfaceCounter();

  for (unsigned int i = 0; i < __address_list.size(); ++i)
    delete __address_list[i];
}

void Interface::registerAttributes() {
  /* Begin> Register the Attributes */
  insertArgument(&__parent_reference);
  insertArgument(&__parent_user_info);
  insertArgument(&__address_list);
  insertArgument(&__supported_strategies);
  /* End< */
}

bool Interface::isSubType(const string& type) {
  if (find(__interfaces.begin(), __interfaces.end(), type) == __interfaces.end())
    return false;
  return true;
}

void Interface::addInterfaceType(const string& interface) {
  if (find(__interfaces.begin(), __interfaces.end(), interface) == __interfaces.end())
    __interfaces.push_back(interface);
}


const SequenceValue<TSAddress>& Interface::getAddressList() const {
  if (__address_list.empty())
    const_cast<Interface*>(this)->prepareAttributes();

  return __address_list;
}

void Interface::prepareAttributes() {
  if (!__is_local)  //if the interface is not on its local system use the old strategy list
    return;

  __address_list.clear();

  /* Get the supported TransportStrategies */
  list<string> tmp_list = STSFactory::getInstance().getStrategyList();
  TSAddress* ts_address = 0;
  /* Begin> Store the default address for each TransportStrategy */
  for (list<string>::iterator it = tmp_list.begin(); it != tmp_list.end(); ++it) {
    ts_address = getDefaultAddress(*it);
    if (ts_address) {
      /* Set the NMMReference of the NMMObject */
      ts_address->setNMMReference(__parent_reference.getRValue());
      __address_list.push_back(ts_address);
    }
  }
  /*End< */

}

Interface& Interface::operator=(const Interface& interface) {
  if (this == &interface) {
    return *this;
  }
  __interfaces = interface.__interfaces;
  __parent_object      = interface.__parent_object;


  /* Begin> decrease the reference counter if a receiver exist*/
  if (__receiver) {
    __receiver->decreaseInterfaceCounter();
  }
  /* End< */

  __receiver           = interface.__receiver;

  if (__receiver) {
    __receiver->increaseInterfaceCounter();
  }

  __parent_reference   = interface.__parent_reference;
  __parent_user_info   = interface.__parent_user_info;
  __address_list       = interface.__address_list;
  __is_local           = interface.__is_local;
  __supported_strategies = interface.__supported_strategies;
  return *this;                    
}

ProxyObject* Interface::getParentObject()          {
  if (!__receiver)
    requestParentObject();

  return __receiver;
}

const ProxyObject* Interface::getParentObject() const    {
  if (!__receiver)
    requestParentObject();

  return __receiver;
}

ProxyObject* Interface::getEventReceiver()   {
  if (!__receiver)
    requestParentObject();

  return __receiver;
}

const ProxyObject* Interface::getEventReceiver() const  {
  if (!__receiver)
    requestParentObject();

  return __receiver;
}

const NMMReference& Interface::getParentReference() const {
  return __parent_reference.getRValue();
}

void Interface::setEventReceiver(NMMObject& receiver) {
  __parent_object     = &receiver;
  __parent_reference  = receiver.getReference();
}

void Interface::requestParentObject() const {
  __receiver = SReferenceManager::getInstance().requestProxyObject(*this);
  __receiver->increaseInterfaceCounter();
}

TSAddress* Interface::getDefaultAddress(const string& strategy_type) {
  /* If the NMMObject is known, forward the method call*/
  if (__parent_object)
    return __parent_object->getDefaultAddress(strategy_type);

  /* Begin> Search the default address for a specific TransportStrategy */
  for (unsigned int i = 0; i < __address_list.size(); ++i) {
    if (__address_list[i])
      if (__address_list[i]->getStrategyType() == strategy_type)
        return __address_list[i]->copy();
  }
  /* End< */

  return 0;
}

void Interface::updateAttributes() {
  __is_local = false;
}

bool Interface::isLocalTo(const Interface& interface) const {
  return interface.getParentInfo() == getParentInfo();
}

bool Interface::isSystemTo(const Interface& interface) const {
  if (isLocalTo(interface))
    return false;

  if (getParentInfo().getHostname() != interface.getParentInfo().getHostname())
    return false;

  if (getParentInfo().getDomainname() != interface.getParentInfo().getDomainname())
    return false;

  return true;
}

bool Interface::isNetworkTo(const Interface& interface) const {
  if (isLocalTo(interface))
    return false;
  if (isSystemTo(interface))
    return false;

  return true;
}

void Interface::establishBinding()
    throw (const BindingException&)
{
  getParentObject()->establishBinding();
}


Result Interface::sendEvent( const CEvent& ce) {
  if (!__receiver)
    requestParentObject();

  return __receiver->receiveEvent(ce, this);
}

Result Interface::sendEvent( const Event& e)   {
  if (!__receiver)
    requestParentObject();

  return __receiver->receiveEvent(e, this);
}

Result Interface::sendEvent( const CEvent& ce)   const    {
  if (!__receiver)
    requestParentObject();

  return __receiver->receiveEvent(ce, this);
}

Result Interface::sendEvent( const Event& e)     const    {
  if (!__receiver)
    requestParentObject();

  return __receiver->receiveEvent(e, this);
}



template <typename T>
NMM::OValueStream& writePtrList(NMM::OValueStream& ov, const list<T>& v_list, const string type_name) {
  typedef typename list<T>::const_iterator ConstIt; //must be declared as a type
  /* Write the TypInfo */
  ov.beginType("listptr_"+type_name, "std");

  /* Write all elements of the list */
  for (ConstIt it = v_list.begin(); it !=  v_list.end(); ++it) {
    if (*it) {
      ov << **it;
    } else {
      GLOBAL_WARNING_STREAM("Serializing NULL-Ptr in writePtrList." << endl);
      ov.beginType("NULL");
      ov.endType();
    }
  }

  ov.endType();
  return ov;
}


template <class T>
NMM::IValueStream& readPtrList(NMM::IValueStream& iv, list<T>& v_list, const string& type_name) {
  /* Begin> Type checking */
  if (iv.getCurrentType() != "listptr_"+type_name) {
    GLOBAL_ERROR_STREAM("List> Invalid type assignment" << endl);
    GLOBAL_ERROR_STREAM ("List> Current type : " << iv.getCurrentType() << "                Expected type: " << "listptr_"+type_name << endl);
    GLOBAL_ERROR_STREAM ("List> Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: " << "std" << endl);
    throw NMM::InvalidTypeException(iv.getCurrentTypeInfo(), NMM::TypeInfo("listptr_"+type_name, "std"));
  }
  /* End< */

  /* Begin> Access internal attributes */
  if (!iv.goDown()) {
    GLOBAL_DEBUG_STREAM("List seems to be empty" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> While the Stream contains elements, read them */
  while(!iv.isEmpty()) {
    NMM::Value* tmp = NMM::SValueFactory::getInstance().createValue(iv.getCurrentTypeInfo());
    if (!tmp) {
      GLOBAL_ERROR_STREAM("ARRG! In readPtrList" << endl);
      throw NMM::Exception("Could not create Value form " + iv.getCurrentTypeInfo().t_name);
    }

    tmp->createFromStream(iv);

    T list_entry = dynamic_cast<T>(tmp);
    if (!list_entry) {
      delete tmp;
    } else {
      v_list.push_back(list_entry);
    }
  }
  /* End< */

  iv.goUp();

  return iv;
}


NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<NMM::Interface*>& interfaceptr_list) {
  return writePtrList<NMM::Interface*>(ov, interfaceptr_list, "Interface");
}
NMM::IValueStream& operator>>(NMM::IValueStream& ov, list<NMM::Interface*>& interfaceptr_list) {
  return readPtrList<NMM::Interface*>(ov, interfaceptr_list, "Interface");
}

NMM::TValueConstructor<list<NMM::Interface*> >     __nmm_list_interfaceptr_value("listptr_Interface", "std");

}
