/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_JACKGROUP_HPP
#define NMM_JACKGROUP_HPP

#include "Interface.hpp"
#include "QueuedOutputJack.hpp"

namespace NMM {

  class Node;

  class JackGroup {
    
  private:

    size_t curr_dequeue_jack;
    Node* parent;

    class InternalJack_t {
    public:
      QueuedOutputJack* jack;
      bool connected;
      bool switch_to;
      InternalJack_t() : jack(0), connected(true), switch_to(false) {};
      InternalJack_t(QueuedOutputJack* _jack, bool _connected, bool _switch_to) 
	: jack(_jack), connected(_connected), switch_to(_switch_to) {};
      ~InternalJack_t() { };
    };

    typedef vector< InternalJack_t > InternalJacks_t;
    mutable InternalJacks_t jacks;

  public:

    /**
     * The default constructor. A JackGroup needs a parent Node so that
     * newly created Jacks can inherit the parent.
     * \param parent The parent Node.
     */
    JackGroup(Node* parent_node, QueuedOutputJack* jack);

    /**
     * Destructor
     */
    ~JackGroup();

    /**
     * Send a message to all jacks. The message is sent to all jacks in
     * this group by calling their sendMessage method. The message's
     * reference count is increased automatically.
     * \param msg The message to be sent.
     * \return the number of input stream queues which will block 
     *         when calling sendMessage the next time. This count will
     *         be used to wait until all stream queues have at least one
     *         empty slot again.
     */
    int sendMessage(Message* message) const;

    /**
     * \return true if some output was not connected before, but is connected now
     */
    bool checkNewConnections();

    /**
     * Adds a jack to the jack list.
     * \param jack The jack to add.
     */
    Result addJack(const QueuedOutputJack* jack) const;

    /**
     * Removes the given jack.
     * \param jack the jack to be removed
     * \return SUCCESS if jack was found and could be removed, otherwise FAILURE 
     */ 
    Result removeJack(const QueuedOutputJack* jack);

    /**
     * Dequeues the next message of this jack group (round-robin)
     * \return Message* the next message.
     */
    Message* dequeueMessage();

    /**
     * Flushes all jacks in this group
     */
    Result flush();

    /**
     * Returns the master jack
     */
    inline QueuedOutputJack* getMaster() { return jacks[0].jack; }

    /**
     * Checks whether at least one jack within this jack group is connected
     */
    bool isConnected();
    
    /**
     * Sends the first Message to all outputs that were disconnected before
     * and discovered to be connected now (switch_to == true);
     * sends the second given Message to all other outputs (switch_to == false)
     * \return the blocked queue count
     */
    int sendSwitchMessages(Message* switch_to_msg, Message* switch_from_msg);

    /**
     * (Re)sets the switch_to flag to the given value
     * \param status the value
     */
    void resetSwitchTo(bool status = false);

  };
}

#endif // NMM_JACKGROUP_HPP
