/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef NMM_MESSAGE_HPP
#define NMM_MESSAGE_HPP

#include "nmm/base/sync/Timestamp.hpp"
#include "nmm/utils/thread/Thread.hpp"

namespace NMM {


  class Message  : public SerializedValue {
  public:
    enum Type{EVENT = 0, BUFFER, FLUSH_EVENT} ;

    enum Direction{DOWNSTREAM = 0, UPSTREAM, OUTOFBAND, UNDEF };
    /** Constructor
     */
    Message(Type t);
    /** CopyConstructor
     */
    Message(const Message& );
    /** Destructor
     */
    virtual ~Message();

    /**
     * Sets a label for this buffer.
     * Labels can be used in muxed streams (e.g. a stream
     * of audio and video data running over a single jack)
     * to distinguish between the different buffer types
     * (e.g. between video and audio buffers)
     * Use an instream event to set the labels to meaningful
     * values (e.g. "video" is 0, "audio" is 1)
     */
    inline void setLabel(unsigned int l) { label = l; }

    /**
     * Returns the label of a buffer
     */
    inline unsigned int getLabel() const { return label.getValue(); }

    /**
     * Set this to true, if you do not want this Buffer object
     * to get lost. Currently, this only prevents the Buffer object
     * to be discarded from a StreamQueue object. In the future,
     * network transmission, etc. will take care of this bit as well.
     */
    inline void setMandatoryBit(bool m)                  { mandatory.setValue(m); }

    /**
     * Sets the direction where this message should will sent to
     * UPSTREAM or DOWNSTREAM.
     */
    inline void setDirection(Direction d) { direction.setValue(d); }

    /**
     * Sets the buffers timestamp.
     */
    inline void setTimestamp(const Timestamp& t)          { time.setValue(t); time.getRValue().is_set = true; }

    /**
     * The method to release a buffer. The Buffer automatically
     * informs its own Allocator (stored in allocator) that
     * it can be reused as soon as the reference count reaches 0.
     * This method reduces the reference count by one. If the
     * reference count reaches 0, the Buffer should also reset its
     * used byte count.
     */
    void release()                          const;

    /**
     * Increase the reference count. Used when the buffer is
     * multicast.
     * \param by The number to increase the count by.
     */
    void increaseCount (unsigned int by = 1)const;

    /**
     * Returns the status of the mandatory bit.
     */
    inline bool getMandatoryBit()                  const { return mandatory.getRValue(); }

    /**
     * Returns the direction, i.e. UPSTREAM or DOWNSTREAM.
     */
    inline Direction getDirection()                const { return direction.getRValue(); }

    /**
     * Get the current reference count. 0 means that this buffer
     * is currently not in use.
     */
    inline int getCount ()                         const { return ref_count; }

    /** returns the Message type
     */
    inline Type getType()                          const { return type.getRValue(); }

    /**
     * Returns the Messages's timestamp.
     */
    inline const Timestamp& getTimestamp ()        const {return time.getRValue();}

    /** compares the sync_time of two events.
     */
    bool operator<(const Message& e)               const { return getTimestamp().sync_time < e.getTimestamp().sync_time;}

    Message& operator=(const Message& );

    Message* copy() const = 0;
  protected:
    /** This method is called, if ref_count is 0.
     */
    virtual void erase() const = 0;

    /** change Message type
     */
    void setType(Type new_type);

  private:
    void registerValues();
  private:
    TValue<Direction> direction;
    TValue<Type> type;
    TValue<Timestamp> time; ///< The timestamp.
    TValue<bool> mandatory;
    TValue<unsigned int> label;  ///< label of the message for muxed streams

    mutable int ref_count;
    ThreadMutex mutex;
  };

  OValueStream& operator<<(OValueStream& ov, const Message::Type& t);

  IValueStream& operator>>(IValueStream& iv, Message::Type& t);

  OValueStream& operator<<(OValueStream& ov, const Message::Direction& t);

  IValueStream& operator>>(IValueStream& iv, Message::Direction& t);

}
#endif
