/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "NMMObject.hpp"
#include "Node.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"
#include "nmm/interfaces/base/registry/IServerRegistry.hpp"
#include "NMMApplication.hpp"
#include "TSFactory.hpp"
#include "OutOfBandChannel.hpp"
#include "StdNamedObject.hpp"
#include "TransportStrategy.hpp"
#include "nmm/base/serialize/XMLValueStream.hpp"
#include <algorithm>
#include <fstream>

namespace NMM {
  NMMObject::NMMObject()
    : __current_sender(0), __internal_mutex(ThreadMutex::RECURSIVE), __i_registry(0), __ref_count(0)
  {
    /* register the NMMObject and get the NMMReference*/
    __reference = SReferenceManager::getInstance().registerNMMObject(this);

    /* Register the Events */
    registerEvents();
  }

  NMMObject::~NMMObject() {
    /* Unregister the NMMObject */
    SReferenceManager::getInstance().unregisterNMMObject(__reference);

    __internal_mutex.lock();

    /* Begin> Disconnect existing CommunicationChannels */
    for (list<OutOfBandChannel*>::iterator it = __channel_list.begin(); it != __channel_list.end(); ++it) {
      if ((*it)->getTransportStrategy()) {
        /* reset the TransportStrategy */
        (*it)->getTransportStrategy()->setBindingChannel(0);
        /* destroy the TransportStrategy */
        (*it)->getTransportStrategy()->scheduleDestruction();
      }
      delete *it;
    }
    /* End< */
    __channel_list.clear();
    __internal_mutex.unlock();

    /* Delete the interfaces */
    for (list<Interface*>::iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      delete *it;

    /* delete the IRegistry */
    delete __i_registry;
  }

  NMMObject::NMMObject(const NMMObject& nmm_object)
    : IEventReceiver(nmm_object)
  {
    /* Nothing must be done! */
  }

  Result NMMObject::receiveEvent(const Event&  event, const IEventSender* sender) {
    setCurrentSender(const_cast<IEventSender*>(sender));
    return __event_dispatcher.dispatch(event, sender);
  }

  Result NMMObject::receiveEvent(const CEvent&  event, const IEventSender* sender) {
    setCurrentSender(const_cast<IEventSender*>(sender));
    Result result = SUCCESS;
    for (unsigned int i = 0; i< event.size(); ++i) {
      if (__event_dispatcher.dispatch( event[i], sender ) == FAILURE) {
        result = FAILURE;
      }
    }
    return result;
  }

  Result NMMObject::receiveEvent(const Event&  event, const IEventSender* sender) const{
    setCurrentSender(const_cast<IEventSender*>(sender));
    return __event_dispatcher.dispatch(event, sender);
  }

  Result NMMObject::receiveEvent(const CEvent&  event, const IEventSender* sender) const {
    setCurrentSender(const_cast<IEventSender*>(sender));
    Result result = SUCCESS;
    for (unsigned int i = 0; i< event.size(); ++i) {
      if (const_cast<EventDispatcher&>(__event_dispatcher).dispatch( event[i], sender ) == FAILURE) {
        result = FAILURE;
      }
    }
    return result;
  }

  OutOfBandChannel* NMMObject::requestCommunicationChannel() {
    /*Begin> Create a new CommunicationChannel and return it*/
    MutexGuard mg(__internal_mutex); 
    for (list<OutOfBandChannel*>::iterator it = __channel_list.begin(); it != __channel_list.end(); ++it)
    {
      if ((*it)->getTransportStrategy() == 0) {
        return *it;
      }
    }

    OutOfBandChannel* new_channel = new OutOfBandChannel(*this);
    __channel_list.push_back(new_channel);
    return new_channel;
    /*End< */
  }

  void NMMObject::releaseCommunicationChannel(CommunicationChannel* channel) {
    /*Begin> remove Channel channel*/
    __internal_mutex.lock();
    list<OutOfBandChannel*>::iterator it = find(__channel_list.begin(), __channel_list.end(), static_cast<OutOfBandChannel*>(channel));
    if (it != __channel_list.end()) {
      __channel_list.erase(it);
      delete channel;            
    } else {
      GLOBAL_ERROR_STREAM2("The object " << __reference << " is not owner of the channel " << channel<<  endl);
    }
    __internal_mutex.unlock();
    /*End <*/
  }

  Event* NMMObject::getEventPrototype(const string& event_key)    const {
    EDObject* edo = getEventDispatcher().getEDObject(event_key);
    if (edo) {
      return edo->createEvent(event_key);
    }

    GLOBAL_ERROR_STREAM("No prototype for event " << event_key << " available " << endl);
    return 0;
  }

  Event* NMMObject::matchEventPrototype(const string& event_key)    const {
    return getEventDispatcher().matchEventPrototype(event_key);
  }

  Result NMMObject::getReference(NMMReference& reference) const {
    reference = __reference;
    return SUCCESS;
  }

  Result NMMObject::getInterfaceList(list<string>& interface_list) const {
    for (list<Interface*>::const_iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      interface_list.push_back((*it)->getInterfaceType());

    return SUCCESS;
  }

  void NMMObject::registerEvents() {
    getEventDispatcher().registerEvent("NMMObject::getReference",     new TEDObject1<NMMObject, NMMReference, Result(NMMObject::*)(NMMReference&) const>(this, &NMMObject::getReference));
    getEventDispatcher().registerEvent("NMMObject::getInterface",     new TEDObject1<NMMObject, int, Interface* (NMMObject::*)( int ) , Interface* >(this, &NMMObject::getInternalInterface));
    getEventDispatcher().registerEvent("NMMObject::getInterfaceSize", new TEDObject0<NMMObject, int (NMMObject::*)( ) , int >(this, &NMMObject::getInterfaceSize));
    getEventDispatcher().registerEvent("NMMObject::getInterfaceList", new TEDObject0<NMMObject, SequenceValue<Interface> (NMMObject::*)( ) const, SequenceValue<Interface> >(this, &NMMObject::getInternalInterfaceList));
    getEventDispatcher().registerEvent("NMMObject::getUserInfo",      new TEDObject1<NMMObject, UserInfo, Result(NMMObject::*)(UserInfo& ) const >(this, &NMMObject::getUserInfo));
    getEventDispatcher().registerEvent("NMMObject::getStrategyList",  new TEDObject1<NMMObject, list<string>, Result(NMMObject::*)(list<string>& ) const >(this, &NMMObject::getStrategyList));
    getEventDispatcher().registerEvent("NMMObject::getDefaultAddress",  new TEDObject1<NMMObject, string, TSAddress* (NMMObject::*)(const string&) const, TSAddress* >(this, &NMMObject::getDefaultAddress));
    getEventDispatcher().registerEvent("NMMObject::registerEventListener", new TEDObject0<NMMObject, void (NMMObject::*)(), void >(this, &NMMObject::registerEventListener));
    getEventDispatcher().registerEvent("NMMObject::removeEventListener", new TEDObject0<NMMObject, void (NMMObject::*)(), void >(this, &NMMObject::removeEventListener));
    getEventDispatcher().registerEvent("NMMObject::registerSpecificEventListener", new TEDObject1<NMMObject, string, void (NMMObject::*)(const string&), void >(this, &NMMObject::registerSpecificEventListener));
    getEventDispatcher().registerEvent("NMMObject::removeSpecificEventListener", new TEDObject1<NMMObject, string, void (NMMObject::*)(const string&), void >(this, &NMMObject::removeSpecificEventListener));
    getEventDispatcher().registerEvent("NMMObject::connectWithOutOfBandChannel", new TEDObject0<NMMObject, void (NMMObject::*)(), void >(this, &NMMObject::connectWithOutOfBandChannel));
    getEventDispatcher().registerEvent("NMMObject::disconnectFromOutOfBandChannel", new TEDObject1<NMMObject, bool, void (NMMObject::*)(const bool&), void >(this, &NMMObject::disconnectFromOutOfBandChannel));
    getEventDispatcher().registerEvent("NMMObject::getPartnerStrategy",new TEDObject0<NMMObject, ITransportStrategy* (NMMObject::*)(), ITransportStrategy* >(this, &NMMObject::getPartnerStrategy));
    getEventDispatcher().registerEvent("NMMObject::getEventList",new TEDObject0<NMMObject, list<string> (NMMObject::*)()const, list<string> >(this, &NMMObject::getEventList));
    getEventDispatcher().registerEvent("NMMObject::getEventPrototype",new TEDObject1<NMMObject, string, Event* (NMMObject::*)(const string&)const, Event* >(this, &NMMObject::getEventPrototype));
    getEventDispatcher().registerEvent("NMMObject::matchEventPrototype",new TEDObject1<NMMObject, string, Event* (NMMObject::*)(const string&)const, Event* >(this, &NMMObject::matchEventPrototype));
    getEventDispatcher().registerEvent("NMMObject::getRegistry",new TEDObject0<NMMObject, IServerRegistry* (NMMObject::*)(), IServerRegistry* >(this, &NMMObject::getRegistry));
    getEventDispatcher().registerEvent("NMMObject::setRegistry",new TEDObject1<NMMObject, IServerRegistry, void (NMMObject::*)(const IServerRegistry&), void >(this, &NMMObject::setRegistry));
  }

  Result NMMObject::getUserInfo(UserInfo& user_info) const {
    user_info = UserInfo();
    return SUCCESS;
  }

  Result NMMObject::getStrategyList(list<string>& strategy_list) const {
    strategy_list = STSFactory::getInstance().getStrategyList();
    return SUCCESS;
  }

  TSAddress* NMMObject::getDefaultAddress(const string& strategy) const {
    TSAddress* address = STSFactory::getInstance().getDefaultAddress(strategy);
    if (address) {
      TSAddress* tmp_address = address->copy();
      tmp_address->setNMMReference(getReference());
      return tmp_address;
    }

    return 0;
  }

  void NMMObject::registerEventListener(IEventReceiver* receiver) {
    if (receiver)
      __event_dispatcher.registerEventListener(receiver);
  }

  void NMMObject::registerEventListener(const string& event_key, IEventReceiver* receiver) {
    if (receiver)
      __event_dispatcher.registerEventListener(event_key, receiver);
  }

  void NMMObject::registerEventListener(const string& key, EDObject* edo) {
    __event_dispatcher.registerEventListener(key, edo);
  }

  void NMMObject::removeEventListener(IEventReceiver* receiver) {
    __event_dispatcher.removeEventListener(receiver);
  }

  void NMMObject::removeEventListener(const string& event_key, IEventReceiver* receiver) {
    if (receiver)
      __event_dispatcher.removeEventListener(event_key, receiver);
  }

  void NMMObject::removeEventListener(EDObject* ed) {
    __event_dispatcher.removeEventListener(ed);
  }

  void NMMObject::connectWithOutOfBandChannel() {
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy) {
      return;
    }
    OutOfBandChannel* channel = requestCommunicationChannel();
    channel->setTransportStrategy(strategy);
  }

  void NMMObject::disconnectFromOutOfBandChannel(const bool& delete_strategy) {
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy) {
      return;
    }
    releaseCommunicationChannel(strategy->getBindingChannel());
    if(delete_strategy) {
      /* Call unregister to all Transport-Strategies registered to this object */
      removeEventListener(strategy);
      strategy->setBindingChannel(0);
      strategy->scheduleDestruction();
    }
  }

  bool NMMObject::isLocalTo(const Interface& interface)  {
    return interface.getParentInfo() == UserInfo();
  }

  bool NMMObject::isSystemTo(const Interface& interface) {
    if (isLocalTo(interface))
      return false;

    if (UserInfo().getHostname() != interface.getParentInfo().getHostname())
      return false;

    if (UserInfo().getDomainname() != interface.getParentInfo().getDomainname())
      return false;

    return true;
  }

  bool NMMObject::isNetworkTo(const Interface& interface) {
    if (isLocalTo(interface))
      return false;
    if (isSystemTo(interface))
      return false;

    return true;
  }

  SequenceValue<Interface> NMMObject::getInternalInterfaceList() const {
    SequenceValue<Interface> interface_list;
    interface_list.useTValue(false);
    for (list<Interface*>::const_iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      interface_list.push_back((*it)->copy());

    return interface_list;
  }

  int NMMObject::getInterfaceSize() {
    return getInterfaceList().size();
  }

  Interface* NMMObject::getInternalInterface( int i ) {
    if ( getInterfaceList().size() > (unsigned int)(i) ) {
      list<Interface*>::iterator it = getInterfaceList().begin();
      advance(it, i);
      return (*it)->copy();
    }
    return 0;
  }

  ITransportStrategy* NMMObject::getPartnerStrategy() {
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy)
      return 0;

    return strategy->getInterface<ITransportStrategy>();    
  }

  list<string> NMMObject::getEventList() const {
    return getEventDispatcher().getEventList();
  }

  IServerRegistry* NMMObject::getRegistry() {
    if ( __i_registry )
      return dynamic_cast<IServerRegistry*> (__i_registry->copy());

    return 0;
  }

  void NMMObject::setRegistry(const IServerRegistry& registry) {
    delete __i_registry;
    __i_registry = dynamic_cast<IServerRegistry*>(registry.copy());
  }


}

