/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_NODE_HPP_
#define _NMM_NODE_HPP_

#include "nmm/interfaces/base/INodeImpl.hpp"
#include "StdNamedObject.hpp"
#include "ReferenceCount.hpp"

namespace NMM {


  class Jack;
  class InputJack;
  class OutputJack;
  class Property;
  class EventDispatcher;
  class Message;
  class Event;
  class CEvent;
  class EDObject;
  class Format;
  class Value;
  class NodeDescription;

  /**
   * This is the abstract superclass of all nodes. 
   */
  class Node : public INodeImpl, public StdNamedObject, public ReferenceCount {
    friend class ThreadedNode;

  private:
    //! mutex for internal use
    ThreadMutex internal_mutex;
    //! The input-property of the node
    mutable Property* input_property;
    //! The output-property of the node
    mutable Property* output_property;
    //! The input-property of the node
    mutable Property* working_input_property;
    //! The output-property of the node
    mutable Property* working_output_property;
    //! The type of the node
    INode::Type node_type;

    /**
     * This is used to specify parameters that have impact on the cpu-utilisation of the used algorithm,
     * but do not belong to the connection-format.
     * The configuration must be setup before activation of the node
     */
    ComputationConfig* computation_config;
    ComputationConfig* working_computation_config;

  private: //methods
    /** This internal method allocates the InputProperty
      */
    Property* getInternalInputProperty()          const;

    /** This internal method allocates the OutputProperty
      */
    Property* getInternalOutputProperty()         const;

    /** This internal method allocates the WorkingInputProperty
      */
    Property* getInternalWorkingInputProperty()   const;

    /** This internal method allocates the WorkingOutputProperty
      */
    Property* getInternalWorkingOutputProperty()  const;

  protected:

    //! mutex to lock current state
    ThreadMutex state_mutex;

    virtual Result doInit()         { return SUCCESS; }
    virtual Result doDeinit()       { return SUCCESS; }
    virtual Result doInitOutput()   { return SUCCESS; }
    virtual Result doDeinitOutput() { return SUCCESS; }
    virtual Result doActivate()     { return SUCCESS; }
    virtual Result doDeactivate()   { return SUCCESS; }
    virtual Result doStart()        { return SUCCESS; }
    virtual Result doStop()         { return SUCCESS; }
    virtual Result doFlush()        { return SUCCESS; }

    /** returns the own output stream jack
     */
    virtual OutputJack* getOwnOutputStream(const string&  = "default") = 0;

    /** returns the own output stream jack
     */
    virtual const OutputJack* getOwnOutputStream(const string&  = "default") const = 0;

    /** returns the own input stream jack
     */
    virtual InputJack* getOwnInputStream(const string&  = "default") = 0;

    /** returns the own input stream jack
     */
    virtual const InputJack* getOwnInputStream(const string&  = "default") const = 0;

    /** returns the Input-Properties
     */
    virtual Property* getOwnInputProperty()  ;

    /** returns the Output-Properties
     */
    virtual Property* getOwnOutputProperty() ;

    /** returns the Input-Properties
     */
    virtual Property* getOwnWorkingInputProperty()  ;

    /** returns the Output-Properties
     */
    virtual Property* getOwnWorkingOutputProperty() ;

    /** Locks the current state. This method is for internal use only.
     */
    void stateLock()   {state_mutex.lock();  }

    /** Unlocks the current state. This method is for internal use only.
     */
    void stateUnlock() {state_mutex.unlock();}

    /**
     * Sets the type of the Node.
     *  \param type the type of the node.
     */
    virtual void setType(INode::Type type);

    /** resets the execution value for all events and the specified state. Before a state change
      * is possible the node must receive and successfully process all events with a mandatory bit for this state.
      * This method takes only affect on registered methods with mandatory mode for the
      * given state.
      */
    inline void reset(int state)                    { getEventDispatcher().reset(state); }

    /** resets the execution value for the given key and state. Before a state change is
      * possible the node must receive and successfully process an event with the specific key.
      * again. This method takes only affect on registered methods with mandatory mode for the
      * given state.
      */
    inline void reset(const string& key, int state) { getEventDispatcher().reset(key, state); }


  public:

    /**
     * The main constructor. Node constructors should create their
     * (default) jacks, but start no threads.
     * Subclasses should implement constructors with more
     * parameters / functionality as needed.
     */
    Node(const char* name = "Node");

    /** Destructor
     */
    virtual ~Node();

    /**
     * This method will try to reach the speficied state by calling the
     * required state transition methods
     */
    virtual void reachState(const INode::State& state) 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method tries to reach the state CONSTRUCTED
     */
    virtual void reachConstructed() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method tries to reach the state INITIALIZED
     */
    virtual void reachInitialized() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method tries to reach the state OUTPUT_INITIALIZED
     */
    virtual void reachOutputInitialized() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method tries to reach the state ACTIAVTED
     */
    virtual void reachActivated() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method tries to reach the state STARTED
     */
    virtual void reachStarted() 
      throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&);

    /**
     * This method initialize a node with all input and output formats and the node enters state "init".
     */
    virtual void init()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) ;

    /**
     * This method initialize a node and the node enters state "init".
     */
    virtual void initOutput()
      throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) ;

    /**
     * This method activates a node. The internal threads are started
     * and the node enters an "active" state. This does not mean output
     * is generated from any input since the node's play method needs
     * to be called.
     */
    virtual void activate()
      throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&);

    /**
     * This method deactivates a node. 
     */
    virtual void deactivate()
      throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&);

    /*
     * This method causes the node to start generating output
     * or to generate output from input.
     */
    virtual void start()
      throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)  ;

    /**
     * This method causes the node to stop generating output
     * or to no longer generate output from input.
     */
    virtual void stop()
      throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&);

    virtual void deinit()
      throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&);

    virtual void deinitOutput()
      throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&);

    /**
     * This method causes the node to perform a flush-operation.
     */
    virtual Result flush();

    virtual bool isConstructed()       const { return (getState() == INode::CONSTRUCTED ? true : false); }
    virtual bool isInitialized()       const { return (getState() == INode::INITIALIZED ? true : false); }
    virtual bool isOutputInitialized() const { return (getState() == INode::OUTPUT_INITIALIZED ? true : false); }
    virtual bool isActivated()         const { return (getState() == INode::ACTIVATED ? true : false); }
    virtual bool isStarted()           const { return (getState() == INode::STARTED ? true : false); }

    /** This method returns current state
     */
    virtual INode::State getState()               const { return static_cast<INode::State>(getEventDispatcher().getState()); }

    /**
     * This method internally blocks the node (basically the same as stop, only
     * the state of the node is not changed and the method does not block until the node
     * is really stopped)
     */
    virtual void block() = 0;

    /**
     * This method internally unblocks the node (basically the same as start, only
     * the state of the node is not changed)
     */
    virtual void unblock() = 0;

    /**
       This method returns the specified input jack with the given specification in 
       the parameter Format and connected to a certain jack.
       \param format the format with the specifications.
       \param jack the name of the jack.
       \return An InputJack or NULL.
    */
    virtual const InputJack* getInputJack(const Format* , const char*  jack = 0) const = 0;
    virtual IInputJack* getInputJack(const Format&, const string& jack);

    /**
       This method returns the specified output jack with the given specification in
       the parameter Format and connected to a certain jack.
       \param pFormat the format with the specifications.
       \param jack the name of the jack.
       \return An OutputJack or NULL.
    */
    virtual const OutputJack* getOutputJack(const Format* , 
					    const char* jack = 0, 
					    bool output_initialized = true) const = 0;
    virtual IOutputJack* getOutputJack(const Format&, const string&);

    /**
     * Obtain the format associated with the named input stream.
     * The Format pointer is
     * guaranteed to remain valid over the connectedness of the node.
     *
     * \param jack_tag The tag of the input stream.
     * \returns a pointer to the corresponding format.
     */
    virtual const Format* getInputFormat(const char * jack_tag = "default") const = 0;

    /**
     * Obtain the format associated with the named output stream.
     * The Format pointer is
     * guaranteed to remain valid over the connectedness of the node.
     *
     * \param jack_tag The tag of the output stream.
     * \returns a pointer to the corresponding format.
     */
    virtual const Format* getOutputFormat(const char * jack_tag = "default") const = 0;


    /**
       This method returns the cost for the specified jack_tag. This can be a jack or
       a jackgroup, depending on the structure of the node.
       \param jack_tag a jack or jackgroup.
       \return the costs for this jack or jackgroup.
    **/
    virtual float getCost(const string& ) const {return 1;};


    /**
       Returns the type of the Node.
       \return the type of the node.
    **/
    virtual INode::Type getType() const;

    /**
       Returns the internal cost for the node. These costs accrue for the basic
       functionality of the node. The total costs for a node are the sum of the
       internal cost and the cost of the input- and output-jacks.
       return the internal cost of the node.
    **/
    virtual float getInternalCost()  const {return 1;}


    virtual INode*        getSuccessorNode(const string &)                  {return 0;}
    virtual INode*        getPredecessorNode(const string &)                {return 0;}
    virtual IInputJack*   getInputStream(const string &)                    {return 0;}
    virtual IOutputJack*  getOutputStream(const string &)                   {return 0;}

    virtual bool isOutputConnected(const string&) const;
    virtual bool isInputConnected(const string&) const;

    /**
     * Sends a composite event asynchronously upstream.
     *
     * \param cevent the composite event to be sent.
     * \param in_jack the input jack from which this composite 
     *  event is to be sent. This is interesting for multiplexers
     *  with more than one input. If no input jack is specified, the
     *  composite event will be sent to all input jacks.
     * \return SUCCESS or FAILURE
     */
    virtual Result sendEventUpstr(CEvent* cevent, const char*  jack_tag = "default") = 0;
    virtual Result sendEventUpstr(CEvent& cevent, const char*  jack_tag = "default") = 0;
    virtual Result sendEventUpstr(Event&  event, const char*  jack_tag = "default")  = 0;
    virtual Result sendEventDownstr(CEvent* cevent, const char*  jack_tag = "default") = 0;
    virtual Result sendEventDownstr(CEvent& cevent, const char*  jack_tag = "default") = 0;
    virtual Result sendEventDownstr(Event&   event, const char*  jack_tag = "default") = 0;

    /**
     * This method returns the InputProperty of the node
     * \return Property *
     */
    virtual const Property *getInputProperty() const; 
    virtual       Property *getInputProperty();


    /**
     * This method returns the OutputProperty of the node
     * \return Property *
     */
    virtual const Property *getOutputProperty() const;
    virtual       Property *getOutputProperty();


    /**
     * This method returns the InputProperty of the node
     * \return Property *
     */
    virtual const Property *getWorkingInputProperty() const; 
    virtual       Property *getWorkingInputProperty() ;


    /**
     * This method returns the OutputProperty of the node
     * \return Property *
     */
    virtual const Property *getWorkingOutputProperty() const;
    virtual       Property *getWorkingOutputProperty() ;

  protected: //events
    /**
       This method returns the cost for the specified jack_tag. This can be a jack or
       a jackgroup, depending on the structure of the node.
       \param jack_tag a jack or jackgroup.
       \return the costs for this jack or jackgroup.
    **/
    virtual Result getCost(const string& jack_tag, float& cost) const { cost = getCost(jack_tag.c_str()); return SUCCESS;}

    /**
       Returns the internal cost for the node. These costs accrue for the basic
       functionality of the node. The total costs for a node are the sum of the
       internal cost and the cost of the input- and output-jacks.
       return the internal cost of the node.
    **/
    virtual Result getInternalCost(float& cost)  const {cost = getInternalCost(); return SUCCESS;}

        /**
     * Get the comutation configuration of the node
     * \returns a pointer to the corresponding configuration (or 0 if non was set)
     */
    virtual ComputationConfig* getComputationConfig() const ;

    /**
     * Set the comutation configuration to the given one.
     * A range of all possible values should be specified.
     * Any previously assigned configuration will be deleted
     * \param computation_config the new configuration
     */
    virtual void setComputationConfig(const ComputationConfig* computation_config);

    /**
     * Get the working comutation configuration of the node (which is used for the actual connection)
     * \returns a pointer to the corresponding configuration (or 0 if non was set)
     */
    virtual ComputationConfig* getWorkingComputationConfig() const ;

    /**
     * Set the working comutation configuration to the given one.
     * Any previously assigned configuration will be deleted
     * \param computation_config the new configuration
     */
    virtual void setWorkingComputationConfig(const ComputationConfig* computation_config);


  };


}

#endif // _NMM_NODE_HPP_
