/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "OutputJack.hpp"
#include "InputJack.hpp"
#include "Node.hpp"
#include "InstreamChannel.hpp"
#include "nmm/interfaces/base/INodeImpl.hpp"
#include "nmm/interfaces/net/IObjectAcceptor.hpp"
#include "TransportStrategy.hpp"
#include "TSFactory.hpp"
#include "GenericNode.hpp"
#include <iostream>
#include "nmm/base/registry/ReferenceManager.hpp"
#include "ConnectionManager.hpp"
#include "FormatStrategyMap.hpp"

namespace NMM {
  OutputJack::OutputJack(Node* node, const string& jack_tag, const string& name)
  : Jack(node, jack_tag, name), __transport_strategy(0)
  {
    data_jack    = this;
  }

  OutputJack::~OutputJack() {
    if (__transport_strategy) {
      __transport_strategy->scheduleDestruction();
    }
  }

  bool OutputJack::isConnected() const 
  {
    if((getPartner() || data_partner.get()) && getMessageReceiver()) 
      return true; 
    else
      return false; 
  }

  void OutputJack::connectTo(const InputJack* in_jack) const
    throw (const BindingException&)
  {
    if(in_jack == 0) {
      throw BindingException(in_jack, this, "InputJack is 0");
    }
    else {
      if (getPartner() == 0) {
        setPartner(const_cast<InputJack*>(in_jack));
        const_cast<OutputJack*>(this)->setMessageReceiver(const_cast<InputJack*>(in_jack));
        NODE_DEBUG_STREAM2((getParent()), "Valid Partner connected" << endl);
      } else {
        NODE_WARNING_STREAM2((getParent()), "Jack already connected" << endl);
        throw BindingException(in_jack, this, "Outputjack already connected");
      }
    }
  }

  void OutputJack::disconnect()
  {
    setPartner(0);
    setMessageReceiver(0);
    if (__transport_strategy)
      __transport_strategy->scheduleDestruction();

    __transport_strategy = 0;
    data_partner.reset(0);
  }

  void OutputJack::setTransportStrategy(TransportStrategy* strategy) {
    /* Begin> If an old strategy is available, release it. */
    if (__transport_strategy)
      __transport_strategy->scheduleDestruction();
    /* End< */

    /* stores the new strategy */
    __transport_strategy = strategy;

    /* Begin> Handle upstream messages */
    /* If the strategy receives upstream-messages, to the parent*/
    strategy->setMessageReceiver(this);
    /* End< */

    /* Begin> Handle downstream messages. */
    /* If the the jack receives downstream-messages from its strategy they are forwarded to the parent */
    setMessageReceiver(strategy);
    /* Signals from the stratgey, like FREE_QUEUE_SLOT are send to the jacks parent*/
    strategy->setSignalReceiver(this);
    /* End< */

    /* Begin> Configure the TransportStrategie's Buffermanager */
    /* If a TransportStrategy receives a Buffer, it needs a BufferManager to request it. */
    if (getParent()) {
      /* Request the IExternalBuffermanager */
      IExternalBufferManager* buffer_manager = dynamic_cast<IExternalBufferManager*>(getParent());
      if (buffer_manager) {
        /* and tell the strategy to use it */
        strategy->setBufferManager(buffer_manager);
      } else {
        /* Otherwise the TransportStrategy can not receive messages */
        NODE_ERROR_STREAM(getParent(), "Parent implements not the IExternalManager" << endl);
      }
    }
    /* End< */
    
    /* Begin> Configure the TransportStrategy's Format */
    if (getParent()) {
      const Format* format = getParent()->getOutputFormat(getJackTag().c_str());
      if (format) {
        MESSAGE_STREAM("Configure OutputJack with format " << format->getFormatType() << endl);
        strategy->configureForOutputJack(*format);
      }
    }
    /* End< */
  }

  void OutputJack::setTransportStrategy(const TSAddress* address) throw(const BindingException&)
  {
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Begin> Handle local case!
     *        For performance reason, the jack communicates with its jack partner without a
     *        TransportStrategy in a local case.
     */
    /* If the address describes a local binding, its a NMMReference describes an object in this address space*/
    if (address->getBindingType() == TSAddress::LOCAL) {
      /* Request the NMMObject with NMMReference in the given address. The type of this Object must be
       * OutputJack
       */
      InputJack* input_jack = SReferenceManager::getInstance().getObject<InputJack>( address->getNMMReference() );

      /* If the address points not to an OutputJack, throw an Exception */
      if (!input_jack)    {
        throw BindingException("No such InputJack available");
      }

      /* Connect this jack with Call the connectTo method for this jack */
      connectTo(input_jack);
      /* Connect the output_jack with this jack */
      input_jack->connectTo(this);
      /* The output_jack is our partner and receives Messages from out parent */
      setMessageReceiver(input_jack);
      return;
    }
    /* End< */

    /* Begin> Handle distributed case */
    /* 1. Create a TransportStrategy which can connect to the given address */
    TransportStrategy* strategy = STSFactory::getInstance().createStrategy(address->getStrategyType());

    /* If this jack can not create such a TransportStrategy throw a BindingException */
    if (!strategy)
      throw  BindingException("Can not create strategy with type "+ address->getStrategyType());

    /* 2. The strategy connects to the given address */
    strategy->setAddress(*address);

    /* 3. The connection between this object and the other object is used to send instream messagess.
     *    So configure the jack to the new strategy*/
    setTransportStrategy(strategy);

    /* 4. Tell the other jack to use it for instream message communication too.*/
    Event event(IJack::connectToJack_event);
    if ( strategy->receiveReliableEvent(event) != SUCCESS ) {
      /* Begin> Error handling */
      if (event.getException()) {
        NODE_DEBUG_STREAM2( getParent() , *event.getException() << endl);
      } else {
        NODE_DEBUG_STREAM2( (getParent()) , "could not set TS in InputJack" << endl);
      }
      /* End< */
    }
  }

  CommunicationChannel* OutputJack::connectTo(const IInputJack& partner) 
      throw(const BindingException& )
  {
    // setPartner(0); // should not be needed
    data_partner.reset(dynamic_cast<IInputJack*>(partner.copy()));
    IOutputJack_var tmp_jack(getInterface<IOutputJack>());
    return new InstreamChannel(*data_partner.get(), *tmp_jack.get());
  }

  void OutputJack::connectToJack() {
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy)
      throw BindingException();

    setTransportStrategy(strategy);
  }


  ITransportStrategy* OutputJack::getTransportStrategy() {
    if (__transport_strategy)
      return __transport_strategy->getInterface<ITransportStrategy>();

    return 0;
  }

  IInputJack* OutputJack::getDataPartner()
  {
    if (data_partner.get()) {
      return dynamic_cast<IInputJack*>(data_partner.get()->copy()); 
    }
    else {
      return 0;
    }
  }

  TSAddress* OutputJack::getDefaultAddress(const string& strategy_type) const {
    const Format* format = getParent()->getOutputFormat(getJackTag().c_str());
    if (SConnectionManager::getInstance().getFormatStrategyMap().isSupported(*format, strategy_type)) {
      return NMMObject::getDefaultAddress(strategy_type);
    } else {
      return NULL;
    }
  }
}
