/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TransportStrategy.hpp"
#include "nmm/base/registry/ReferenceManager.hpp"
#include "nmm/base/registry/RegistryStructs.hpp"
#include "nmm/base/registry/ClientRegistry.hpp"
#include "CommunicationChannel.hpp"
#include "ProxyObject.hpp"
#include "Message.hpp"
#include "TSFactory.hpp"

namespace NMM {


LocalAddress::LocalAddress(const NMMReference& reference)
  : TSAddress(reference, "LocalStrategy", TypeInfo("LocalAddress", "NMM"), TSAddress::LOCAL)
{
  setHostname( SProcessUserInfo::getInstance().getHostname() );
  setPort( ClientRegistry::default_port );
}



TransportStrategy::TransportStrategy()
    : __message_receiver(0), __event_receiver(0), __signal_receiver(0), __buffer_manager(0), __com_channel(0), __is_strategy(0)
{
  
}

TransportStrategy::~TransportStrategy() {
  if (getBindingChannel())
    getBindingChannel()->setTransportStrategy("NULL");
}

void TransportStrategy::setEventReceiver(IEventReceiver* event_receiver) {
  if (__is_strategy)
    __is_strategy->removeEventSender(this);

  __is_strategy = dynamic_cast<TransportStrategy*>(event_receiver);
  if (__is_strategy)
    __is_strategy->addEventSender(this);

  __event_receiver = event_receiver;
}


ITransportStrategy* TransportStrategy::getPartnerStrategy()             const
{
  TOutValue<ITransportStrategy*> return_value(0);
  Event event("NMMObject::getPartnerStrategy");
  event.setReturnValue(&return_value);
  if (sendReliableEvent(event) != SUCCESS)
    if (event.getException())
      throw *event.getException();

  return return_value.getValue();
}

Result TransportStrategy::sendReliableEvent(Event& e) const {
  return sendEvent(e);
}

Result TransportStrategy::receiveReliableEvent(Event& e, const IEventSender* sender) const {
  return receiveEvent(e, sender);
}

void TransportStrategy::configureForOutputJack(const Format& /*format*/) {
  
}

void TransportStrategy::configureForInputJack(const Format& /*format*/) {
  
}

LocalStrategy::~LocalStrategy() {
  if (__partner_strategy)
    __partner_strategy->setPartnerStrategy(0);
}


void LocalStrategy::setPartnerStrategy(LocalStrategy* partner_strategy)  {
  __partner_strategy  = partner_strategy;
  if (!__partner_strategy) {
    ProxyObject* parent_object = dynamic_cast<ProxyObject*>(getEventReceiver());
    if (parent_object)
      parent_object->disconnectFromObject();
    else
      delete this;
  }
}

Result LocalStrategy::receiveMessage(Message* message, const IMessageSender* sender) {
  if (sender != this)
    if (__partner_strategy) 
      return __partner_strategy->forwardMessage(message);
    else {
      message->release();
      return FAILURE;
    }

  return SUCCESS;
}

Result LocalStrategy::receiveEvent(const CEvent& event, const IEventSender* sender) {
  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
    return NMMObject::receiveEvent(event, sender);

  if (sender != this)
    if (__partner_strategy)
      return __partner_strategy->forwardEvent(event);
    else {
      return FAILURE;
    }

  return SUCCESS;
}

Result LocalStrategy::receiveEvent(const Event& event, const IEventSender* sender) {
  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
    return NMMObject::receiveEvent(event, sender);

  if (sender != this)
    if (__partner_strategy)
      return __partner_strategy->forwardEvent(event);
    else {
      return FAILURE;
    }

  return SUCCESS;
}

Result LocalStrategy::receiveEvent(const CEvent& event, const IEventSender* sender) const {
  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
    return NMMObject::receiveEvent(event, sender);

  if (sender != this)
    if (__partner_strategy)
      return __partner_strategy->forwardEvent(event);
    else {
      return FAILURE;
    }

  return SUCCESS;
}

Result LocalStrategy::receiveEvent(const Event& event, const IEventSender* sender)  const {
  if ( find(getEventSenderList().begin(), getEventSenderList().end(), sender) != getEventSenderList().end())
    return NMMObject::receiveEvent(event, sender);

  if (sender != this)
    if (__partner_strategy)
      return __partner_strategy->forwardEvent(event);
    else {
      return FAILURE;
    }

  return SUCCESS;
}




Result LocalStrategy::sendEvent(const CEvent& event ) {
  if (__partner_strategy)
    return __partner_strategy->forwardEvent(event);
  else
    return FAILURE;
}

Result LocalStrategy::sendEvent(const Event& event) {
  if (__partner_strategy)
    return __partner_strategy->forwardEvent(event);
  else
    return FAILURE;
}

Result LocalStrategy::sendEvent(const CEvent& event) const {
  if (__partner_strategy)
    return __partner_strategy->forwardEvent(event);
  else
    return FAILURE;
}

Result LocalStrategy::sendEvent(const Event& event)  const {
  if (__partner_strategy)
    return __partner_strategy->forwardEvent(event);
  else
    return FAILURE;
}

Result LocalStrategy::sendMessage(Message* message) {
  if (__partner_strategy)
    return __partner_strategy->forwardMessage(message);
  else
    return FAILURE;
}

void LocalStrategy::setAddress(const TSAddress& addr)
  throw (const BindingException&)
{
  const LocalAddress* tmp_address = dynamic_cast<const LocalAddress*>(&addr);
  if (!tmp_address) {
    throw BindingException("Invalid address type");
  } 

  /* Begin> Create a partner strategy and set our object as partner strategy*/
  __partner_strategy = new LocalStrategy();
  __partner_strategy->setPartnerStrategy(this);
  /* End< */

  /* Begin> Configure our partner strategy to the object in the address */
  __partner_strategy->setEventReceiver(SReferenceManager::getInstance().getNMMObject(tmp_address->getNMMReference()));
  __partner_strategy->setMessageReceiver(SReferenceManager::getInstance().getObject<IMessageReceiver>(tmp_address->getNMMReference()));
  /* End< */
}

Result LocalStrategy::forwardMessage(Message* message)   const {
  if (getMessageReceiver())
    return const_cast<IMessageReceiver*>(getMessageReceiver())->receiveMessage(message, this);

  return FAILURE;
}

Result LocalStrategy::forwardEvent(const Event& event)         const {
  if (getEventReceiver())
    return getEventReceiver()->receiveEvent(event, this);

  return FAILURE;
}

Result LocalStrategy::forwardEvent(const CEvent& event)        const {
  if (getEventReceiver())
    return getEventReceiver()->receiveEvent(event, this);

  return FAILURE;
}


void LocalStrategy::setValueStream(const IValueStream& , const OValueStream&) {
  GLOBAL_ERROR_STREAM("LocalStrategy::setValueStream > A LocalStrategy never uses a ValueStream" << endl);
}

void LocalStrategy::setSendValueStream(const string& ) {
  GLOBAL_ERROR_STREAM("LocalStrategy::setSendValueStream > A LocalStrategy never uses a ValueStream" << endl);
}
void LocalStrategy::setReceiveValueStream(const string&){
  GLOBAL_ERROR_STREAM("LocalStrategy::setSendValueStream > A LocalStrategy never uses a ValueStream" << endl);
}


const char LocalStrategy::__strategy_type[] = "LocalStrategy";

TTSFactoryRegistrator< TTSFactoryObject<LocalStrategy> > prototype("LocalStrategy");
TConstructor<LocalAddress> __local_address(TypeInfo("LocalAddress", "NMM"));
}
