/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_R_FLOAT_HPP
#define _NMM_R_FLOAT_HPP

#include "definition.hpp"

namespace NMM {
	
/**
   This class implements a FLOAT range.
*/
  class R_FLOAT 
  {
    
  private:	
    //! The lower bound of the range.
    float m_l_value;
    //! The upper bound of the range.
    float m_h_value;
    //! The lower weight of the range.
    float m_l_weight;
    //! The upper weight of the range.
    float m_h_weight;

  public:
    
    /**
       The default Constructor.
    */
    R_FLOAT() {};
    
    /**
       Constructor with additional parameters wich set directly 
       the member-variables m_l_value and m_h_value.
       Additionally the weighting factor can be specified in the
       parameters l_weight and h_weight.
       \param float lParam: the lower bound of the range.
       \param float hParam: the upper bound of the range.
       \param float l_weight: the lower weight of the range.
       \param float h_weight: the upper weight of the range.
    **/
    R_FLOAT(float lParam, float hParam, float l_weight = DEFAULT_WEIGHT, float h_weight = DEFAULT_WEIGHT);

    /**
       The default Destructor
    **/
    ~R_FLOAT();
    
    /**
       Returns the upper bound of the range.
       \return float: the upper bound of the range.
    **/
    float getHighValue() const;
    
    /**
       Returns the lower bound of the range.
       \return float: the lower bound of the range.
    **/
    float getLowValue() const;
    
    /**
       Returns the lower weight of the range.
       \return float: the lower weight of the range.
    **/
    float getLowWeight() const;

    /**
       Returns the upper weight of the range.
       \return float: the upper weight of the range.
    **/
    float getHighWeight() const;
    
    /**
       Sets the upper bound of the range.
       \param float h_param: the upper bound of the range.
    */
    void  setHighValue(float h_param);
    
    /**
       Sets the lower bound of the range.
       \param float l_param: the lower bound of the range.
    */
    void  setLowValue(float l_param);
    
    /**
       Sets the lower weight of the range.
       \param float l_weight: the lower weight of the range.
    **/
    void setLowWeight(float l_weight);

    /**
       Sets the upper weight of the range.
       \param float h_weight: the upper weight of the range.
    **/
    void setHighWeight(float h_weight);
    
    /**
       Debugging method. Prints out the range.
    */
    void showValue();

    /**
       Returns the standardized weight for the range.
       \return float: the standardized weight.
    **/
    float getStandardizedWeight();

    /**
       Returns if the parameter value is included in the range.
       \param float value: the checked value.
       \return true if the value is included otherwise false.
    **/
    bool isValueIncluded(float value);
    /**
       Returns if the range from lValue to hValue is included in the range.
       \param float lValue: the lower bound of the checked range.
       \param float hvalue: the upper bound of the checked range.
       \return true if the range is included otherwise false.
    **/
    bool isRangeIncluded(float lValue, float hValue);

    /**
       Returns if the range from lValue to hValue has an intersection with the range.
       \param float lValue: the lower bound of the range.
       \param float hValue: the upper bound of the range.
       \return true if the range has an intersection otherwise false.
    **/
    bool isRangeMatching(float lValue, float hValue);

    
    /**
       Returns if the range from lValue to hValue is identical with the current instance.
       \param float lValue: the upper bound of the range.
       \param float hValue: the lower bound of the range.
       \return true if identical, otherwise false.
    **/
    bool isRangeIdentical(float lValue, float hValue) const;
    
    friend bool operator == (const R_FLOAT & lhs, const R_FLOAT & rhs);

    friend ostream & operator << (ostream & os, const R_FLOAT & reference);

  };
}
#endif

