/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_R_IPARAMETER_HPP
#define _NMM_R_IPARAMETER_HPP

#include "Parameter.hpp"

namespace NMM {

/**
   This class stores ranges of INT32 parameters.
   The range itself is implemented by the class R_INT.
**/

  class R_IParameter : public Parameter {
    
  protected:

    //! The STL-List of the range-values.
    r_int32_t_list m_value_list;

    //! Private Variable for browsing the STL-List Value-List
//     mutable R_INT_Iterator m_value_counter;
    mutable list<R_INT *>::const_iterator m_value_counter;

    //! The default value
    int32_t m_default_value;
    
    //! The default range.
    R_INT * m_default_range;
    
    /**
       This function computes the value of the fraction 

         (HighWeight - LowWeight)
         ------------------------
         (HighValue  - LowValue)
       
	 for the specified R_FLOAT range.
       \param R_FLOAT *: a R_FLOAT range.
       \return float: the value of the fraction.
     **/
    virtual float computeFactor(R_INT * param) const;

    /**
       This function checks if there is an intersection between the
       interval (m_lValue - m_hValue) and (lValue - hValue).
       Note: The difference to the inherited function 
             bool isValueIncluded(void *, void *) is that you have
	     to specify both of the interval limits instead of
	     iterating the hole m_value_list!
       \param float m_lValue: the lower bound of the first range.
       \param float m_hValue: the upper bound of the first range.
       \param float lValue: the lower bound of the second range.
       \param float hvalue: the upper bound of the second range.
       \return true if the intersection is not empty, otherwise false.
     **/
    bool isRangeIntersection(int32_t m_lValue, int32_t m_hValue, int32_t lValue, int32_t hValue) const;

  public:
    /**
       The default Constructor.
       \param string parameter_name: the name of the parameter.
    */
    R_IParameter(const string & parameter_name = "");
    
    /**
       Constructor with a parameter pValue that is directly added to the Value-List.
       \param string parameter_name: the name of the parameter.
       \param void * pValue: an R_INT pointer.
    */
    R_IParameter(const string & parameter_name,void * pValue);
    
    /**
       The default Destructor.
    */
    ~R_IParameter();
    
    virtual void addValue(void *pValue);
    
    virtual void addValue(R_INT *value);

    virtual bool isValueIncluded(void *pValue) const;

    virtual bool isValueIncluded(int32_t value) const;
    
    virtual bool isValueIncluded(void *lValue,void *hValue) const;

    virtual bool isValueIncluded(int32_t lValue, int32_t hValue) const;
    
    virtual bool isRangeMatching(int32_t lValue, int32_t hValue) const;
    
    virtual bool isRangeIdentical(int32_t lValue, int32_t hValue) const;
    
    virtual void showParameter() const;
    
    virtual ostream& showParameter(ostream&) const;
    
    virtual ostream& showParameterVerbose(ostream&) const;

    virtual void *nextValue() const;

    virtual Parameter * duplicate() const;

    /**
       This function creates a new parameter from the current
       instance and the argument of the function by computing
       intersections from the intervals.
       \param Parameter * : a parameter.
       \return Parameter * : the new "intersected" parameter.
     **/
    virtual Parameter * getIntersection(const Parameter * reference) const;

    /**
       This function computes the range-value of the parameter value.
       \param value: the value that will be "new-weighted".
       \return float: the new weight.
     **/
    virtual float getWeight(void *pValue) const;

    virtual uint getNumberOfValues() const;

    virtual bool isParameterIdentical(const Parameter *reference) const;

    virtual bool isParameterMatching(const Parameter *reference) const;

    virtual void *getFirstValue() const;
    
    virtual float getStandardizedWeight() const;

    virtual void setDefaultValue(int32_t value);

    virtual int32_t getDefaultValue() const;

    virtual int32_t getLowestValue() const;

    virtual int32_t getHighestValue() const;

    virtual const r_int32_t_list & getRIntValueList() const;

    virtual R_INT * getDefaultRange() const;
    virtual void fromStream(IValueStream& ) ;
    virtual void toStream(OValueStream&)  const ;
  };
}
#endif
