/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_UDP_STRATEGY_HPP
#define NMM_UDP_STRATEGY_HPP

#include "UDPGeneric.hpp"

#include "nmm/interfaces/base/IConnectionManager.hpp"
#include "nmm/base/ObjectAcceptor.hpp"
#include "nmm/base/ConnectorAddress.hpp"
#include "nmm/base/TransportStrategy.hpp"
#include "nmm/utils/net/NetReactor.hpp"

namespace NMM {
  class Node;
  class IValueStream;
  class OValueStream;
  class NetStream;

  /** This class implements an UDP address which stores all information to establish a connection.
   */
  class UDPAddress : public ConnectorAddress {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    UDPAddress (const NMMReference& = 0);

    /** Constructor, with hostname and port of the address
     */
    UDPAddress (const string& name, int port, const NMMReference& = 0);

    /** Copy Constructor
     */
    UDPAddress (const UDPAddress&);

    /** Virtual copy Constructor
     */
    UDPAddress* copy()                          const { return new UDPAddress(*this); }
    //@}

    /** @name Access information
     */
    //@{
    /** This method sets the port
     */
    void setPort(int port);

    /** This method sets the host name
     */
    void setHostname(const string& host);

    /** This method sets the complete INetAddr
     */
    void setINetAddr(const INetAddr& net_addr)        { __net_addr = net_addr; }

    /** This method returns the INetAddr.
     */
    const INetAddr& getINetAddr()               const { return __net_addr; }

    /** This method returns the INetAddr.
     */
    INetAddr& getINetAddr()                     { return __net_addr; }
    //@}
  private:
    INetAddr __net_addr;                              ///< The network address
  };


  template<>
  class TConstructor<UDPAddress> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual UDPAddress*          createValue()          const {return new UDPAddress();}
    virtual TValue<UDPAddress>*  createTValue()         const {return new TValue<UDPAddress>();}

    virtual TValue<TSAddress*>*  createTPointerValue()  const {return new TValue<TSAddress*>();}

    virtual TConstructor<UDPAddress>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  
  /** UDP transport strategy.
   * This class implements a simple UDP transport strategy. This transport strategy can send
   * NMM messages, buffers and events using either a direct UDP NetStream or a PacketStream using
   * a datagram socket. It can use either a ConnectorAddress to set up a connection via a ConnectionManager
   * or a UDPAddress to connect to a partner strategy directly.
   */
  class UDPStrategy
    : public TransportStrategy,
      public UDPGeneric,
      public DispatchedNetHandler,
      public TSAddressOwner,
      protected ThreadedObject
  {
    enum UDPMessage { EVENT_REQUEST = 0, EVENT_RESPONSE , CEVENT_REQUEST, INSTREAM_BUFFER, INSTREAM_MESSAGE, CONFIG_EVENT};
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    UDPStrategy() ;

    /** Destructor
     */
    virtual ~UDPStrategy();

    /** virtual copy constructor
     */
    virtual UDPStrategy* copy() const { return new UDPStrategy(); }

    /** Destroy the UDPStrategy
     */
    virtual void scheduleDestruction();
    //@}

    /** @name Sender & Receiver Interface
     */
    //@{
    /** This method sends an incoming CEvent over a udp connection
     */
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 );

    /** This method sends an incoming Event over a udp connection
     */
    virtual Result receiveEvent(const Event&, const IEventSender* = 0 );

    /** This method sends an incoming CEvent over a udp connection
     */
    virtual Result receiveEvent(const CEvent&, const IEventSender* = 0 ) const;

    /** This method sends an incoming Event over a udp connection
     */
    virtual Result receiveEvent(const Event&, const IEventSender* = 0 )  const;
    
    /** This method sends an incoming Event reliably over a udp connection
     */
    virtual Result receiveReliableEvent(Event&, const IEventSender* = 0 )  const;
    
    /** This method sends an incoming Message over a udp connection
     */
    virtual Result receiveMessage(Message* , const IMessageSender* sender);

    /** This method sends an incoming Message over a udp connection
     */
    virtual Result sendMessage(Message* );

    /** This method sends an incoming CEvent over a udp connection
     */
    virtual Result sendEvent( const CEvent& ce);

    /** This method sends an incoming Event over a udp connection
     */
    virtual Result sendEvent( const Event& e);

    /** This method sends an incoming CEvent over a udp connection
     */
    virtual Result sendEvent( const CEvent& ce)   const;

    /** This method sends an incoming Event over a udp connection
     */
    virtual Result sendEvent( const Event& e)     const;
    
    /** This method sends an incoming Event reliably over a udp connection
     */
    virtual Result sendReliableEvent( Event& e) const;
    //@}


    /** @name Configure the Transport-Strategy
     */
    //@{
    /** This method sets the ValueStreams that should be used
     */
    virtual void setValueStream(const IValueStream& , const OValueStream&);

    /** This method sets the connection address
     */
    virtual void setAddress(const TSAddress& addr)
      throw (const BindingException&);

    /** This method processes internal NMM-Signals from Nodes or jacks
     */
    virtual void   receiveSignal(Signal signal);

    /** This method returns the typename of the Transport Strategy
     */
    virtual string getStrategyType() const { return getStaticStrategyType(); }
    
    /** This method sets the NetStreams that should be used.
     * @param udp_stream NetStrean for unreliable transmission
     */
    virtual void setNetStream(const NetStream& udp_stream);
    //@}

    /** @name DispatchedNetHandler methods
     */
    //@{
    /** This method is called if the event POLLHUP occurs.
     * @param socket The socket on which an event occurs.
     */
    virtual NetResult errorOccured(int socket)       { return closeConnection(socket); }

    /** This method is called if the event POLLINVAL occurs.
     * @param socket The socket on which an event occurs.
     */
    virtual NetResult invalidSocket(int socket)      { return closeConnection(socket); }

    /** This method is called if the event POLLERR occurs.
     * @param socket The socket on which an event occurs.
     */
    virtual NetResult closeConnection(int socket);
    
    /** This method handles incoming data
     */
    virtual NetResult receiveData(int socket);

    /** This method is called when the socket is writable
     */
    virtual NetResult writeData(int socket);
    //@}
    
    
    /** @name Addresses and Connection Parameters
     */
    //@{
    /** This method returns the local UDP address.
     */
    virtual const UDPAddress& getLocalAddress() const throw() {
      return *__local_address;
    }
    
    /** This method returns the remote UDP address.
     */
    const UDPAddress* getRemoteAddress() const {
      return __remote_address;
    }
    //@}
    
    /** @name Packet Stream Options
     */
    //@{
    /** This method sets whether the UDPStrategy should use the PacketStream.
     * The packet stream guarantees consistent, yet unreliable, delivery of NMM messages.
     */
    virtual void setUsePacketStream(const bool& use_packet_stream);
    
    /** This method gets whether the UDPStrategy should use the PacketStream.
     * The packet stream guarantees consistent, yet unreliable, delivery of NMM messages.
     */
    virtual bool getUsePacketStream() const;
    
    /** This method sets the maximum size of a UDP packet.
     */
    virtual void setPacketSize(const ssize_t& size);
    
    /** This method returns the maximum size of a UDP packet.
     */
    virtual ssize_t getPacketSize() const;
    
    /** This method sets the maximum number of partial messages in the receive queue.
     * Packets in the receive queue with a message ID less than the value set by this method are
     * thrown away.
     */
    virtual void setMaxPartialMessages(const unsigned int& max_partial);
    
    /** This method returns the maximum number of partial messages in the receive queue.
     * Packets in the receive queue with a message ID less than the value returned by this method are
     * thrown away.
     */
    virtual unsigned int getMaxPartialMessages() const;
    //@}
    
    /** @name Protocol Statistics
     */
    //@{
    /** This method resets all statistics in the UDPStrategy
     */
    virtual void resetStatistics();
    
    /** This method returns the total amount of bytes sent.
     */
    virtual ssize_t getTotalSentBytes() const;
    
    /** This method returns the total amount of bytes received.
     */
    virtual ssize_t getTotalReceivedBytes() const;
    
    /** This method returns the total number of packets sent
     */
    virtual unsigned int getTotalSentPackets() const;
    
    /** This method returns the total number of packets received
     */
    virtual unsigned int getTotalReceivedPackets() const;
    
    /** This method returns the total number of discarded incoming packets
     */
    virtual unsigned int getTotalDiscardedPackets() const;
    
    /** This method returns the total number of lost incoming packets
     */
    virtual unsigned int getTotalLostPackets() const;
    //@}
    
    /** @name Static type information.
     */
    //@{
    /** This method returns the type name of strategy.
     */
    static inline string getStaticStrategyType() { return __strategy_type; }
    //@}
    
  protected:
    /** @name Internal dispatch methods
     */
    //@{
    /** This method dispatch incoming messgages to the other methods in this section
     *  @param stream The stream on which the message was received
     */
    virtual void dispatchMessage(NetStream* stream);

    /** This method handles an incoming Event.
     */
    void dispatchEventRequest(IValueStream&);

    /** This method handles an Event response.
     */
    void dispatchEventResponse(IValueStream& );

    /** This method handles an incoming CEvent.
     */
    void dispatchCEventRequest(IValueStream&);

    /** This method handles an incoming Config Event.
     */
    void dispatchConfigEvent(IValueStream& );

    /** This method handles an incoming instream message.
     */
    void dispatchInstreamMessage(IValueStream& iv, UDPMessage);

    /** This method handles an incoming upstream message.
     */
    void dispatchUpstreamMessage(IValueStream& iv, UDPMessage);
    //@}

    /** @name Methods to send Data
     */
    //@{
    /** This method sends an response to an received Event or CEvent
     */
    virtual void sendResponse(Result, const Value* , UDPMessage);

    /** This method sends an Event or CEvent
     */
    virtual Result sendValue(const Value& e, UDPMessage)     const;
    
    /** This method reliably sends a config Event
     */
    virtual Result sendReliableConfigEvent(Event& e)     const;
    
    /** This method sends the message in an own thread.
     */
    void internalMessageSend(Message* message);

    /** This method waits until the connection is established
     */
    void waitForConnect() const;

    /** This method creates a local UDP socket
     */
    virtual void createLocalSocket()
      throw (const ConnectionException& );
    
    /** This method connects the local UDP socket to a peer address.
     */
    virtual void connectToPeer()
      throw (const ConnectionException& );
    //@}
    
    /** @name Methods to send Data
     */
    //@{
    /** This thread handles incoming data and is used to send messages
     */
    virtual void run();

    /** Wait for the internal thread.
     */
    void conditionWait();

    /** Notify waiting threads.
     */
    void conditionNotify();
    //@}
    /** @name ValueStreams
     */
    //@{
    /** This method sets the ValueStream that should be used to send data
     */
    virtual void setSendValueStream(const string& );

    /** This method sets the ValueStream that should be used for incoming data
     */
    virtual void setReceiveValueStream(const string&);
    //@}
    
  private:
    /* Begin> Attributes for send/receiveEvent */
    ThreadMutex                   __event_mutex;                ///< Mutex for __event_condition
    ThreadCondition               __event_condition;            ///< Condition to be notified about event response
    ThreadMutex                   __event_sending_mutex;        ///< This mutex ensures that only one event can be send
    mutable Value*                __current_value;              ///< Pointer to the current Value that is send
    Result                        __current_result;             ///< Result for the returning Value
    /* End< */

    /* Begin> Pointers to the used ValueStreams */
    OValueStream*                 __o_stream;                   ///< OValueStream which is used for serialization
    IValueStream*                 __i_stream;                   ///< IValueStream which is used for deserialization
    /* End< */
    
    /* Begin> Pointers to the used NetStreams */
    NetStream*                    __udp_stream;                 ///< UDP net stream */
    bool                          __use_packet_stream;          ///< Whether we are using a packet stream */
    /* End< */
    
    /* Begin> Thread specific attributes */
    bool                          __is_running;                 ///< Threads running flag
    ThreadMutex                   __condition_thread_mutex;     ///< Mutex for Threads condition __thread_condition
    ThreadCondition               __thread_condition;           ///< Thread condition to wait if its nothing to do
    /* End< */

    /* Begin> Message sending attributes */
    Message*                      __send_message;               ///< Temporary pointer to the message that should be send
    ThreadMutex                   __send_message_mutex;         ///< Mutex for save message sending
    /* End< */

    /* Begin> Connection specific attributes */
    UDPAddress*                   __local_address;              ///< Local UDP address
    UDPAddress*                   __remote_address;             ///< Remote UDP address
    IConnectionManager*           __icm;                        ///< Connection manager
    NMMReference                  __remote_ts_ref;              ///< NMM reference of remote transport strategy
    /* End< */

    /* Begin> Attributes for incoming messages from network */
    bool                          __incoming_message;           ///< Flag that something was received on the UDP socket
    bool                          __blocked;                    ///< Flag that is true if the parent returns blocked
    ThreadMutex                   __incoming_message_mutex;     ///< Mutex for the incoming messages
    ThreadMutex                   __internal_thread_mutex;      ///< ThreadMutex to sync the register/unregisterSocket method
    ThreadMutex                   __socket_mutex;               ///< Mutex to protect the socket
    ThreadMutex                   __net_stream_mutex;           ///< Mutex to protect the net stream
    /* End< */
    
    /* Begin> NetReactor event registration */
    int                           __udp_socket_events;          ///< Events the udp socket is currently registered for
    /* End< */
    
    /* Begin> UDP port range */
    static int                    __udp_min_port;               ///< Minimum UDP port
    static int                    __udp_max_port;               ///< Maximum UDP port
    /* End< */
    
  private:
    static const char __strategy_type[];                        ///< Static type name of this strategy
  };

}
#endif
