/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#include "LocalNodeFactory.hpp"
#include "FactoryObject.hpp"

namespace NMM {


LocalNodeFactory::LocalNodeFactory()
  : NodeFactory("LocalNodeFactory",0)
{ }

LocalNodeFactory::~LocalNodeFactory() {
  /* destroy the FacotorObjects */
  for (map<ObjectID, FactoryObject*>::iterator it = __factory_map.begin(); it != __factory_map.end(); ++it)
    delete it->second;
}

void LocalNodeFactory::registerNode(const ObjectID& node_id, const FactoryObject& fo) {
  /* Begin> Check if the ObjectID is already registered */
  map<ObjectID, FactoryObject*>::iterator it = __factory_map.find(node_id);
  if (it != __factory_map.end()) {
    WARNING_STREAM("A FactoryObject for plugin " << fo.getPluginName() << " and Node " << fo.getNodeName() << " is already registered" << endl);
    return;
  }
  /* End< */

  /* insert a copy of the FactoryObject */
  __factory_map.insert(pair<ObjectID, FactoryObject*>( node_id, fo.copy()));
}

void LocalNodeFactory::unregisterNode(const ObjectID& node_id) {
  /* Search the iterator for the given ObjectID */
  map<ObjectID, FactoryObject*>::iterator it = __factory_map.find(node_id);

  /* If no such ObjectID was registered return */
  if (it == __factory_map.end()) {
    WARNING_STREAM("No FactoryObject for plugin "<< node_id <<  endl);
    return;
  }

  /* Delete the FactoryObject */
  delete it->second;

  /* remove the entry from the map */
  __factory_map.erase(it);
}


Node& LocalNodeFactory::createNode (const FactoryTicket& fp)
    throw (const FactoryException& )
{
  /* search the FactoryObject for the Node */
  map<ObjectID, FactoryObject*>::iterator it = __factory_map.find(fp.getObjectID());

  /* If no such FactoryObject is registered, throw an exception */
  if (it == __factory_map.end())
    throw FactoryException("Node FactoryObject available of this Node");

  /* The FactoryTicket must be valid, otherwise an exception is thrown */
  if (!fp.isValid())
    throw FactoryException("FactoryTicket is invalid");

  /* create the node */
  Node* tmp_node = it->second->createNode();

  /* Check if the node is successfully created */
  if (!tmp_node)
    throw FactoryException("Could not create an instance of the Node");

  //Send the configuration event to the node. This event restores the configuration
  //which is used for the node registration.
  tmp_node->receiveEvent(fp.getConfigEvent());

  /*Begin> Create an entry in node_list for the new Node*/
  __node_list.insert(pair<Node*, FactoryObject*>(tmp_node, it->second));
  /*End< */

  /* The FactoryTicket is invalid now */
  setTicketInvalid(fp);

  /* return the node */
  return *tmp_node;
}

void LocalNodeFactory::releaseNode (const Node& node)
    throw (const FactoryException& )
{
  /* seach the FactoryObject for the given node */
  map<Node*, FactoryObject*>::iterator it = __node_list.find(const_cast<Node*>(&node));

  /* if no such FactoryObject available throw an exception */
  if (it == __node_list.end()) {
    throw FactoryException("No such node created");
  }

  /* release the node */
  it->second->releaseNode(const_cast<Node*>(&node));

  /* remove the entry from map */
  __node_list.erase(it);
}


}//end namespace
