/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         Michael Repplinger <repplinger@cs.uni-sb.de>
 *                     Marco Lohse        <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_OBJECTDESCRIPTION_HPP
#define NMM_OBJECTDESCRIPTION_HPP

#include "nmm/base/Event.hpp"
#include "nmm/base/serialize/ComplexValue.hpp"

namespace NMM {

class Interface;
class IValueStream;
class OValueStream;
/** \class ObjectDescription
  * A ObjectDescription describes the capabilities of a object.
  */
class ObjectDescription  : public SerializedValue {
public:
/*! @name Constructor & Destructor
  */
//@{
/** constructor
  */
  ObjectDescription(MemoryMgm = Value::INVALUE);

/** constructor
  */
  explicit ObjectDescription(const string& object_name, MemoryMgm = Value::INVALUE);

/** destructor
  */
  ~ObjectDescription();

/** copy constructor
  */
  ObjectDescription(const ObjectDescription& );

/** virtual copy constructor
  */
  virtual ObjectDescription* copy() const { return new ObjectDescription(*this); }
//@}

/** @name set/addParameter-methods
  * Methods to set the parameter inside this description
  */
//@{

/** sets the name of wich is described in this object
  */
  inline void setObjectName       (const string& object_name)     { __object_name.setValue(object_name); }

/** sets the name of a library that offers the object
  */
  inline void setLibraryName    (const string& library_name)      { __library_name.setValue(library_name); }

/** sets a new interface list of supported interfaces. The old list is overwritten.
  * \sa addInterface
  */
  inline void setInterfaceList(const list<string>& interface_list){ __interface_list.setValue(interface_list); }

/** adds an interface to the description
  */
  void addInterface(const Interface& interface);

/** adds an interface to the description
  */
  void addInterface(const string& interface_type);

/** Sets the additional description based on a key and a corrseponding value(both of type string).
  * This description can be used to describe some additional parameter, like the developer's name
  * or the location of a camera.
  */
  inline void setServiceDescription(const map<string, string>& value_map)  { __service_description.setValue(value_map); }

/** adds an additional description based on key and a corrseponding value(both with type string).
  * An additional description can be used to describe some additional parameter, like the developer's name
  * or the location of a camera.
  */
  inline void addServiceDescription(const string& key, const string& value) { __service_description.getRValue().insert(pair<string, string>(key, value)); }

/** sets the transport strategy that should be used for out of band communication
  */
  inline void setTransportStrategy(const string& transport_strategy)        { __transport_strategy.setValue(transport_strategy); }

/** sets the location of an object
  */
  void setLocation(const string& location);

/** sets the port to access the location of an object
  */
  void setPort(int port);

/** ignores the location string of an object
  */
  void disableLocation();

/** ignores not the location string of an object
  */
  void enableLocation();
//@}

/** @name getParameter-methods
  * Methods to get the parameter inside this description
  */
//@{
/** returns the name of the plugin
  */
  inline const string& getLibraryName()       const   { return __library_name.getRValue(); }

/** return the name of an object inside the plugin
  */
  inline const string& getObjectName()          const   { return __object_name.getRValue(); }

/** This method tests if the given ObjectDescrption nd is a subset of this one
  */
  bool isSubset(const ObjectDescription& nd)    const;

/** returns a list of supported interfaces
  */
  inline const list<string>& getInterfaceList()  const   { return __interface_list.getRValue(); }

/** returns the location of the node
  */
  inline const string&   getLocation()           const   { return __location.getRValue(); }

/** returns the location of the node
  */
  inline       string&   getLocation()                   { return __location.getRValue(); }

/** returns the used port
  */
  inline int   getPort()                         const   { return __port.getRValue(); }
  
/** returns the map with the additional service description.
  */
  inline const map<string, string>& getServiceDescription() const { return __service_description.getRValue(); }

/** returns the map with the additional service description.
  */
  inline       map<string, string>& getServiceDescription()       { return __service_description.getRValue(); }

/** returns the transport strategy that should be used for out of band communication
  */
  inline const string& getTransportStrategy()              const { return __transport_strategy.getRValue(); }

/** returns the transport strategy that should be used for out of band communication
  */
  inline       string& getTransportStrategy()                    { return __transport_strategy.getRValue(); }

/** returns the value of the location-flag
 *@return true if the location-flag is enabled.
 *@return false otherwise
 */
  inline       bool&          isLocationEnabled()          { return __location_flag.getRValue(); }
  inline const bool&          isLocationEnabled()    const { return __location_flag.getRValue(); }
//@}


/** @name Object-Configuration
  * Methods to access the configuration event. Both methods are for internal use only!!
  */
//@{
  inline const CEvent& getConfigEvent()       const   { return __config_event;}
  inline       CEvent& getConfigEvent()               { return __config_event;}
//@}

protected:
  CEvent                          __config_event;         ///< stores the Events that must be send to a object to get this configuration.

private:
/** register all attibutes by the base class SerializedValue
  */
  void registerAttributes();

/** checks if the interface list in the given ObjectDescription is a subset of this Interface list
  */
  bool isInterfaceSubset(const ObjectDescription& ) const;


private:
  TValue<string>                  __object_name;          ///< stores the typename of a object
  TValue<string>                  __library_name;         ///< stores the library which can create the object
  TValue<list<string> >           __interface_list;       ///< stores the interfaces supported by this object
  TValue<map<string, string> >    __service_description;  ///< stores additional description
  TValue<string>                  __location;             ///< The hostname of the system
  TValue<string>                  __transport_strategy;   ///< stores the transport strategy that should be used for out of band communication
  TValue<bool>                    __location_flag;        ///< indicates, wether the location is ignored by the isSubset method or not.
  TValue<int>                     __port;                 ///< stores the used port to request the object
};


} //end namespace
#endif
