/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_PLUGIN_HPP
#define NMM_PLUGIN_HPP

#include <list>
#include <string>
#include <vector>
#include "LocalRegistry.hpp"
#include "LocalNodeFactory.hpp"
#include "FactoryObject.hpp"

/**
 * This macro can be used to register a single plugin.
 */
#define NMM_REGISTER_PLUGIN(node, name) \
static NMM::TPlugin<node> plugin0(name); \
NMM::Plugin* nmm_plugins[] = {&plugin0, 0};

/**
 * This macro can be used to register special plugins like PlaybackNodePlugin etc.
 */
#define NMM_REGISTER_SPECIAL_PLUGIN(plugin) \
NMM::Plugin* nmm_plugins[] = {&plugin, 0};


namespace NMM {
class Registry;
class NodeFactory;
class Exception;

/** \class Plugin
  * This abstract class is used by the PluginManager to load and register a Node with its specific
  * configuration. Each Node uses its own Plugin and register the specific functionality wich can
  * system specific by a registry. Plugins are normaly grouped to a library that can be load at
  * runtime to extend the applications funcionality.
  * \sa PluginManager
  */
class Plugin : private IEventReceiver {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  * \param node_name sets the nodes name
  * \param max_instance sets the maximum instance count. A value of -1 means no limitation.
  */
  Plugin(const string& node_name, int max_instance);

/** Desturctor
  */
  virtual ~Plugin();
//@}

/** @name Plugin Registration
  * These methods register or unregister a specific Plugin
  */
//@{
/** This method initialize the plugin. Normaly this method can be used to check that all needed
  * ressources like libraries can be initalized or are available.
  * \sa nmmFini
  */
  virtual void nmmInit()
    throw (const Exception&)                      = 0;

/** This method deinit the plugin. This method can be used to deinit needed libraries.
  * \sa nmmInit
  */
  virtual void nmmFini()                          = 0;

/** This method is called to register the plugin's nodes by a registry.
  * \sa nmmRegistryFini
  */
  virtual void nmmRegistryInit(Registry& )
    throw (const Exception&)                      = 0;

/** This method removes registered nodes.
  * \sa nmmRegistryInit
  */
  virtual void nmmRegistryFini(Registry& )        = 0;

/** This method is called to register the node by a NodeFactory.
  * \sa nmmFactoryInit
  */
  virtual void nmmFactoryInit(NodeFactory&)
    throw (const Exception&)                      = 0;

/** This method removes the registred nodes from the NodeFactory.
  * \sa nmmFactoryFini
  */
  virtual void nmmFactoryFini(NodeFactory&)       = 0;
//@}


/** @name Plugin information
  * These methods returns some plugin specific information like the plugins version.
  */
//@{
/** This method returns the library name
  */
  inline const string& getLibraryName()                     const { return __library_name;}

/** This method returns the node name
  */
  inline const string& getNodeName()                        const { return __node_name; }

/** This method is used the get the NMM-version which was used to compile the plugin
  */
  inline const string& getNMMVersion()                      const { return __nmm_version; }

/** This method is used the get the Minor number of the NMM-version which was used to compile the plugin
  */
  inline int           getMinorNumber()                     const { return __nmm_minor_nbr;}

/** This method is used the get the Major number of the NMM-version which was used to compile the plugin
  */
  inline int           getMajorNumber()                     const { return __nmm_major_nbr;}

/** This method is used the get the Micro number of the NMM-version which was used to compile the plugin
  */
  inline int           getMicroNumber()                     const { return __nmm_micro_nbr;}

/** Returns the i'th ObjectID
  */
  inline const ObjectID& getObjectID(unsigned int index)    const { return __id_vector[index]; }

/** Returns the ObjectID count
  */
  inline unsigned int getSize()                             const { return __id_vector.size(); }
//@}

/** @name Internal methods
  * These methods are for internal use!!
  */
//@{
/** For each registered Node a ObjectID is stored. This method is for INTERNAL USE!
  */
  inline void addObjectID(const ObjectID& node_id)                    { __id_vector.push_back(node_id); }

/** This method sets the library name and is set by the PluginManager if the library is load.
  * This method is for internal use!
  */
  inline void setLibraryName(const string& library_name)              { __library_name = library_name;}

  inline void setPluginIndex(unsigned int index)                      { __index = index; }

  inline unsigned int getPluginIndex()                       const    { return __index;  }
//@}

protected:

/** @name Developer methods
  * These methods are for the plugin developer and offers extendend functionality to simplify
  * the development of a node specific plugin
  */
//@{
/** This method propare an INode for the registration. It enables event logging to receive
  * configuration events for a node. These events are stored with the node description to
  * restore the node with the registered configuration.
  */
  INode* prepareRegistration(Node& node);


/** This method creates an NodeDescription for a specific Node.
  */
  void createNodeDescription(INode&, NodeDescription& );

/** This method registers the NodeDescription by a registre
  */
  void registerNode(const NodeDescription&, int max_instance);

/** This metod changes the Registry wich is used to register the node.
  * If this pointer is NULL, wich is the default value, the LocalRegistry is used
  */
  inline void setRegistry(LocalRegistry* registry)                { __registry = registry;  }

/** This metod changes the NodeFactory wich is used to create the node.
  * If this pointer is NULL, wich is the default value, the LocalNodeFactory is used
  */
  inline void setNodeFactory(LocalNodeFactory* node_factory)      { __node_factory = node_factory;}

/** returns the current Registry
  */
  inline LocalRegistry*     getRegistry()                         { return __registry; }

/** returns the current NodeFactory
  */
  inline LocalNodeFactory*  getNodeFactory()                      { return __node_factory;}

/** returns the maximum instance count
  */
  inline int getMaxInstanceCount()                       const    { return __max_instance;}
//@}

private:

/** Implementation of class IEventReceiver.
  */
//@{
  virtual Result receiveEvent( const CEvent& ce, const IEventSender* );
  virtual Result receiveEvent( const Event& e  , const IEventSender* );

  virtual Result receiveEvent( const CEvent& ce, const IEventSender* )   const;
  virtual Result receiveEvent( const Event& e  , const IEventSender* )   const;
//@}

/** Disable the event logging
  */
  void disableEventLogging(INode& node);

/** Returns the configuration event
  */
//@{
  inline        list<Event>& getConfigEvents()              { return __config_events; }
  inline  const list<Event>& getConfigEvents()      const   { return __config_events; }
//@}

private:
  string                  __library_name;     //< The library's name
  string                  __nmm_version;      //< The NMM version used to compile the plugin
  int                     __nmm_major_nbr;    //< The NMM major number version used to compile the plugin
  int                     __nmm_minor_nbr;    //< The NMM minor number version used to compile the plugin
  int                     __nmm_micro_nbr;    //< The NMM micro number version used to compile the plugin
  vector<ObjectID>        __id_vector;        //< The vector stores the NodeId's for each possible registration
  mutable list<Event>     __config_events;    //< The configuration events that are send to the node
  LocalRegistry*          __registry;         //< Pointer to the Registry that administrate the node
  LocalNodeFactory*       __node_factory;     //< Pointer to the NodeFactory that can create the node
  string                  __node_name;        //< The node-name 
  int                     __max_instance;     //< The maximum instance count
  unsigned int            __index;            //< Stores the index inside the plugin_array
};


/** \class TPlugin
  * This parametrized class offers a standard implementation of the class Plugin.
  * It can be used for a simple registeration for most nodes. If the node needs no configuration events
  * this class can be used without any extension. If the node can be used for several devices
  * and needs some configuration events for state transition "init" you must reimplement the
  * nmmRegistryInit method. To register the node you can use the registerNode method.
  */
template <typename T >
class TPlugin : public Plugin {
public:
/** @name Constructor & Destructor
  */
//@{
/** Constructor
  */
  TPlugin(const string& node_name, int max_instance = -1)
    : Plugin(node_name, max_instance) { }

/** Destructor
  */
  virtual ~TPlugin() {  }
//@}

/** @name Plugin Registration
  * These methods register or unregister a specific Plugin
  */
//@{
/** This method initialize the plugin. Normaly this method can be used to check that all needed
  * ressources like libraries can be initalized or are available.
  * \sa nmmFini
  */
  virtual void nmmInit()
    throw (const Exception&)  {}

/** This method deinit the plugin. This method can be used to deinit needed libraries.
  * \sa nmmInit
  */
  virtual void nmmFini()      {}

/** This method is called to register the plugin's nodes by a registry.
  * \sa nmmRegistryFini
  */
  virtual void nmmRegistryInit(Registry& )
    throw (const Exception&);

/** This method removes registered nodes.
  * \sa nmmRegistryInit
  */
  virtual void nmmRegistryFini(Registry& );

/** This method is called to register the node by a NodeFactory.
  * \sa nmmFactoryInit
  */
  virtual void nmmFactoryInit(NodeFactory&)
    throw (const Exception&);

/** This method removes the registred nodes from the NodeFactory.
  * \sa nmmFactoryFini
  */
  virtual void nmmFactoryFini(NodeFactory&);
//@}
};


template <typename T>
void TPlugin<T>::nmmRegistryInit(Registry& )
  throw (const Exception& )
{
  /* 1. create a dummy node  */
  T node;

  /* 3. prepare the interface for registration and get the interface that must be used */
  INode_var i_node( prepareRegistration( node ) );

  /* 4. create the NodeDescription */
  NodeDescription description;
  createNodeDescription(*i_node.get(), description);

  /* 5. and finaly register the node */
  registerNode( description, getMaxInstanceCount() );
}

template <typename T>
void TPlugin<T>::nmmRegistryFini(Registry& )
{
  /* If the node was registered by a specific registry use this one */
  if (getRegistry())
    for ( unsigned int i = 0; i< getSize(); ++i )
      getRegistry()->unregisterNode(getObjectID(i));

  /* If no specific registry was used use the default LocalRegistry */
  else
    for ( unsigned int i = 0; i< getSize(); ++i )
      SLocalRegistry::getInstance().unregisterNode(getObjectID(i));
}

template <typename T>
void TPlugin<T>::nmmFactoryInit(NodeFactory& )
   throw (const Exception& )
{
  if (getSize() < 1) {
    throw Exception("No ObjectID available for node " + getNodeName());
  }
  for (unsigned int i = 0; i < getSize(); ++i)
    /* If a specific NodeFactory is set use this one */
    if (getNodeFactory())
      getNodeFactory()->registerNode(getObjectID(i), TFactoryObject<T>(getLibraryName(), getNodeName()));

    /* If no specific NodeFactory is set, use the default LocalNodeFactory */
    else
      SLocalNodeFactory::getInstance().registerNode(getObjectID(i), TFactoryObject<T>(getLibraryName(), getNodeName()));
}

template <typename T>
void TPlugin<T>::nmmFactoryFini(NodeFactory& )
{
  for (unsigned int i = 0; i< getSize(); ++i)
    /* If the node is created by a specifc NodeFactory unregister the node by this registry */
    if (getNodeFactory())
      getNodeFactory()->unregisterNode(getObjectID(i));
    /* Otherwise call unregister by the LocalNodeFactory */
    else
      SLocalNodeFactory::getInstance().unregisterNode(getObjectID(i));
}

}//end namespace
#endif

