/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "ReferenceManager.hpp"
#include "nmm/base/ProxyObject.hpp"

namespace NMM {
ReferenceManager::ReferenceManager()
  : NamedObject("ReferenceManager")   {

}

ReferenceManager::~ReferenceManager() {
             
}


NMMReference ReferenceManager::registerNMMObject(NMMObject* object) {
  if (!object)
    return 0;

  /* Use the current value of the internal reference counter and increment it */
  NMMReference reference = __reference_counter++;

  /* Store the NMMObject with the NMMReference */
  __object_map.insert(pair<NMMReference, NMMObject*>(reference, object));

  /* return the reference */
  return reference;
}

void ReferenceManager::unregisterNMMObject(const NMMReference& reference){
  /* remove the NMMReference from the map */
  __object_map.erase(reference);
}

NMMObject* ReferenceManager::getNMMObject(const NMMReference& reference){
  /* search the Iterator */
  InternalIt it = __object_map.find(reference);
  if (it == __object_map.end()) {
    ERROR_STREAM("An object with reference " << reference << " is not registered" << endl);
    /* If the NMMReference is unknown, return 0*/
    return 0;
  }

  /* return the NMMObject */
  return it->second;
}

const  NMMObject* ReferenceManager::getNMMObject(const NMMReference& reference) const{
  /* search the Iterator */
  CInternalIt it = __object_map.find(reference);
  if (it == __object_map.end()) {
    ERROR_STREAM("An object with reference " << reference << " is not registered" << endl);
    /* If the NMMReference is unknown, return 0*/
    return 0;
  }

  /* return the NMMObject */
  return it->second;
}


ProxyObject* ReferenceManager::requestProxyObject(const Interface& interface) {
  ProxyIt tmp_it = __proxy_map.end();

  /* Begin> Check if the ProxyObject already exist, or must be created */

  /* Check if the UserInfo of the interface is already known */
  for (ProxyIt it = __proxy_map.begin(); it != __proxy_map.end(); ++it) {
    if (it->user_info == interface.getParentInfo()) {
      tmp_it = it;
      break;
    }
  }

  /* If the UserInfo is unknown, no ProxyObject can exist, so we must create it*/
  if ( tmp_it == __proxy_map.end() ) {
    return allocateProxyObject(interface);
  }

  /* If the UserInfo exist, check if the ProxyObject to the NMMObject already exist */
  map<NMMReference, ProxyObject*>::iterator map_it = tmp_it->proxy_map.find(interface.getParentReference());
  if (map_it == tmp_it->proxy_map.end()) {
    return allocateProxyObject(interface);
  }
  /* End< */

  return map_it->second;
}

ProxyObject* ReferenceManager::allocateProxyObject(const Interface& interface)
{
  DEBUG_STREAM("Allocate Proxy object  " << endl);
  /* create Proxy and enable auto delete */
  ProxyObject* proxy_object = new ProxyObject(interface, true);

  /* Begin> Search if an InternalStruct for the UserInfo already exist in the __proxy_map*/
  ProxyIt tmp_it = __proxy_map.end();
  for (ProxyIt it = __proxy_map.begin(); it != __proxy_map.end(); ++it) {
    if (it->user_info == interface.getParentInfo()) {
      tmp_it = it;
      break;
    }
  }
  /* End< */

  /* Begin> If no InternalStruct exist, create it */
  if ( tmp_it == __proxy_map.end() ) {
    __proxy_map.push_front(InternalProxyStruct(interface.getParentInfo()));
    tmp_it = __proxy_map.begin();
  }
  /* End< */

  /* Insert the ProxyObject */
  tmp_it->proxy_map.insert(pair<NMMReference, ProxyObject*>(interface.getParentReference(), proxy_object));

  /* Get the default address for the first TransportStrategy from the Interface */
  const TSAddress* address = const_cast<Interface&>(interface).getDefaultAddress(proxy_object->getOutOfBandChannel().getStrategyList().front());

  /* If no address available we can not connect to the object and an exception is thrown */
  if (!address) {
    throw BindingException("No address available\n");
  }

  /* connect to the NMMObject */
  proxy_object->connectToObject(*address);

  /* return the ProxyObject */
  return proxy_object;
}

void ReferenceManager::registerService(const string& service_name, NMMObject* service) {
  ServiceIt it = __service_map.find(service_name);
  if (it != __service_map.end()) {
    ERROR_STREAM("A Service with name is already registered" << endl);
  }
  else
    __service_map.insert(pair<string, NMMObject*>(service_name, service));
}

NMMObject* ReferenceManager::getService(const string& service_name) {
  ServiceIt it = __service_map.find(service_name);
  if (it != __service_map.end())
    return it->second;
  return 0;
}

list<string> ReferenceManager::getServiceList() const {
  list<string> service_list;
  for (CServiceIt it = __service_map.begin(); it != __service_map.end(); ++it)
    service_list.push_back(it->first);

  return service_list;
}

void ReferenceManager::removeService(const string& service) {
  __service_map.erase(service);
}


void ReferenceManager::unregisterProxyObject( ProxyObject& proxy_object, const UserInfo& user_info ) {
  ProxyIt tmp_it = __proxy_map.end();

  /* Begin> Search the position in the list for given UserInfo */
  for (ProxyIt it = __proxy_map.begin(); it != __proxy_map.end(); ++it) {
    if (it->user_info == user_info) {
      tmp_it = it;
      break;
    }
  }
  /* End< */

  if (tmp_it == __proxy_map.end()) {
    WARNING_STREAM("Could not unregister Proxy. No such System known" << endl);
    return;
  }

  /* remove the ProxyObject */
  tmp_it->proxy_map.erase(proxy_object.getParentReference());

  /* If the List is empty, remove the InternalStruct */
  if (tmp_it->proxy_map.empty())
    __proxy_map.erase(tmp_it);
}

/* Initialize the reference counter with 1, because it is the first address we can use. 0 is reserved */
NMMReference ReferenceManager::__reference_counter = 1;


}//end namespace



