
/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *                    NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "RMConfigManager.hpp"

#include <sstream>
#include <fstream>

#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>

#include "nmm/utils/utils.hpp"


namespace NMM {

RMConfigManager::RMConfigManager() : NamedObject("RMConfigManager") {
  net_file   = "network";
  mesh_file  = "meshes";
  stringstream name;
  name << getenv("HOME") << "/.nmm/" << getHostname() << "/";;
  path = name.str();
  name.clear();
  testDirs();

  config.clear();
  net_config.clear();

  interface_n = 0;
}

bool RMConfigManager::save() {
    stringstream name;
    name << path << mesh_file;
    filename = name.str();
    DEBUG_STREAM2("saving meshes (" << filename << ")" << endl);
#if(__GNUC__ >=3)
    ofstream rm_info(filename.c_str(),ios_base::trunc);//,0666);
#else
    ofstream rm_info(filename.c_str(),ios::trunc,0666);
#endif
    if (!rm_info) {
      ERROR_STREAM2("Unabel to open " << filename.c_str() << " for writing" << endl);
      return 0;
    }
    rm_info << "NMM_RM_1.1" << endl;

    map<string, map<string, string > >::iterator node_iterator;

    for (node_iterator = config.begin(); node_iterator != config.end(); ++node_iterator) { // write all nodes
      rm_info << "n\n" << node_iterator->first << endl;
      map< string, string>::iterator format_iterator;
      for (format_iterator = node_iterator->second.begin(); format_iterator != node_iterator->second.end(); ++format_iterator) { // and every cpu
        rm_info << "c\n" << format_iterator->first << endl;
        rm_info << format_iterator->second << endl;
      }
    }
    rm_info.close();

    stringstream netname;
    netname << path << net_file;
    filename = netname.str();
    DEBUG_STREAM2("saving network (" << filename << ")" << endl);
#if(__GNUC__ >=3)
    ofstream net_info(filename.c_str(),ios_base::trunc);//,0666);
#else
    ofstream net_info(filename.c_str(),ios::trunc,0666);
#endif
    map<string, pair<float,uint> >::iterator net_iterator;
    for(net_iterator = net_config.begin(); net_iterator != net_config.end(); ++net_iterator) {
      net_info << net_iterator->first << endl;
      net_info << net_iterator->second.first << endl;
      net_info << net_iterator->second.second << endl;
    }
    DEBUG_STREAM2( "saved configuration file" << endl );

    return 1;
}

bool RMConfigManager::load() {
    stringstream name;
    name << path << mesh_file;
    filename = name.str();
    DEBUG_STREAM2("loading meshes (" << filename << ")" << endl);
    fstream rm_info(filename.c_str(),ios::in|ios::binary);
   // rm_info.setf(ios::skipws);
    string buffer;
    if (!rm_info) {
      ERROR_STREAM2("unable to open meshes (" << filename << ")" << endl);
      ERROR_STREAM2("will be generated when saving" << endl);
    } else {
      bool version_ok = true;
      getline(rm_info,buffer);
      if (buffer == "NMM_RM_1.1") {
        DEBUG_STREAM2("configuration file found" << endl);
        buffer = "";
      } else {
        ERROR_STREAM2("version mismatch" << endl);
        version_ok = false;
      }

      if (version_ok) {
        bool read_error = false;
        string node,type,mesh;

        getline(rm_info, buffer);
        while ( ! (rm_info.eof() || read_error) ) {
          if (buffer == "n" ) {
            getline(rm_info, node);
            DEBUG_STREAM2("reading info for node " << node << endl);
            getline(rm_info, buffer);

            while (buffer == "c") {
              buffer = "";
              getline(rm_info, type);
              DEBUG_STREAM2("reading info for node " << node << " and formattype " << type << " on host " << getHostname() << endl);
              getline(rm_info,mesh);
              getline(rm_info, buffer);

              config[node][type] = mesh;
            }
            for (string::iterator i = buffer.begin(); i!= buffer.end(); ++i) {
              rm_info.putback(*i);
            }
            getline(rm_info, buffer);
          } else {
            read_error = true;
          }
        }
      }
      rm_info.close();
    }

    stringstream netname;
    netname << path << net_file;
    filename = netname.str();
    DEBUG_STREAM2("loading network (" << filename << ")" << endl);
    fstream net_info(filename.c_str(),ios::in|ios::binary);
    if (! net_info) {
      ERROR_STREAM2("unable to open network config (" << filename << ")" << endl);
      return 0;
    }
    string host;
    getline(net_info, host);
    while ( ! net_info.eof() ) {
      float speed;
      uint  interf;
      net_info >> speed;
      net_info >> interf;
      if (! host.empty()) {
        DEBUG_STREAM2("RMConfigManager: network: " << host << " on interface " << interf << "  BW " << speed << endl );
        net_config[host] = pair<float,uint>((speed *1024),interf);
        interface_speeds[interf].push_back( (speed *1024) );
        if (interface_n < interf) {
          interface_n = interf;
        }
      }
      net_info >> host;
    }
    net_info.close();

    return true;
}

void RMConfigManager::testDirs() {
  stringstream name;
  name << getenv("HOME") << "/.nmm";
  if (opendir(name.str().c_str()) == NULL) {
   if (errno == ENOENT) {
     DEBUG_STREAM2("Directory "<< name.str() << " does not exist." << endl);
     DEBUG_STREAM2("Creating directory "<< name.str() << endl);
     if ( mkdir(name.str().c_str(),0777) != 0) {
       DEBUG_STREAM2("Unable to create directory "<< name.str() << endl);
     }
   }
  }
  name << "/" << getHostname();
  if (opendir(name.str().c_str()) == NULL) {
    if (errno == ENOENT) {
      DEBUG_STREAM2("Directory "<< name.str() << " does not exist." << endl);
      DEBUG_STREAM2("Creating directory "<< name.str() << endl);
      if ( mkdir(name.str().c_str(),0777) != 0) {
        DEBUG_STREAM2("Unable to create directory "<< name.str() << endl);
      }
    }
  }
}

void RMConfigManager::setMesh(string node_name, string format_id, string mesh) {
  config[node_name][format_id] = mesh;
}

string RMConfigManager::getMesh(string node_name, string format_id) {
  if (config.find(node_name) != config.end()) {
    if (config[node_name].find(format_id) != config[node_name].end()) {
      return config[node_name][format_id];
    }
  }

  return string("");
}

list<pair<string,string> > RMConfigManager::getMeshes(string node_name) {
  list<pair<string,string> >* ret_list = new list<pair<string,string> >;

  map<string, map<string, string > >::iterator node_iterator;

  node_iterator = config.find(node_name);
  if (node_iterator == config.end()) {
    WARNING_STREAM2("no mesh found for node " << node_name << " on host " << getHostname() << endl);
    return *ret_list;
  }

  map< string, string>::iterator format_iterator;
  for (format_iterator = node_iterator->second.begin(); format_iterator != node_iterator->second.end(); ++format_iterator) { // and every cpu
    ret_list->push_back(pair<string,string>(format_iterator->first, format_iterator->second));
  }

  return *ret_list;
}

float RMConfigManager::getBandwidth(string host) {
  if (net_config.find(host) != net_config.end()) {
    return net_config[host].first;
  }

  ERROR_STREAM2("Bitrate to host " << host << " unknown" << endl);

  return 0.0;
}

void RMConfigManager::setBandwidth(string host, float speed, uint interf) {
  net_config[host] = pair<float,uint>(speed, interf) ;
}

void RMConfigManager::setInterfaces(uint interf_n) {
  interface_n = interf_n;
}

uint RMConfigManager::getInterfaces() {
  return interface_n;
}

float RMConfigManager::getMaxBW(uint interf) {

  if (max_speeds.find(interf) != max_speeds.end() ) {
    return max_speeds[interf];
  }

  list<float> speeds = interface_speeds[interf];
  float max_speed = 0.0F;
  for (list<float>::iterator i = speeds.begin(); i != speeds.end(); ++i) {
    if ( *i > max_speed ) {
      max_speed = *i;
    }
  }
  max_speeds[interf] = max_speed;

  return max_speed;
}


/*

void RMConfigManager::setCPU(string cpu) {
  cpu_info = cpu;
}

string RMConfigManager::getCPU() {
  return cpu_info;
}
*/
}

