/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "RMMesh.hpp"
#include <fstream>
#include <iostream>

namespace NMM {

    RMMesh::RMMesh(): NMM::SerializedValue("RMMesh","NMM"),  NamedObject("RMMesh") {
      registerValues();
      mesh_dimension = 0;
      start = 0;
      dimension_order = list<string>();
      vertex_index = list<RMVertex>();
    }

    RMMesh::RMMesh(const list<string>& order) : NMM::SerializedValue("RMMesh","NMM"), NamedObject("RMMesh") {
      registerValues();
      mesh_dimension = order.size();
      start = 0;
      dimension_order = list<string>();
      for (list<string>::const_iterator i = order.begin(); i != order.end(); ++i) {
        dimension_order.getRValue().push_back(*i);
      }
      vertex_index = list<RMVertex>();
    }

    RMMesh::RMMesh(const RMMesh& rm) : NMM::SerializedValue(rm), NamedObject("RMMesh")  {
      registerValues();
      *this = rm;
    }


    RMMesh::~RMMesh() {
      deleteAll();
      dimension_order.getRValue().clear();
      // clear the vertex_index
      vertex_index.getRValue().clear();
    }

    void RMMesh::deleteAll() {
      start->deleteVertices();
    }


    void RMMesh::setPointers(RMIndexVertex* index, RMVertex* vertex, list<string>::iterator act_dim, bool dummy) {
      DEBUG_STREAM2("setPointers : dummy?" << dummy << endl);
      if ( ! index->isLeaf() ) {
        DEBUG_STREAM2("set pointers: is at " << *act_dim << " (" << vertex->value(*act_dim) << ")" << endl);
        // take lower and higher indices and check all other dimensions for identical values

        RMIndexVertex* lower;
        RMIndexVertex* higher;
        QualityParameter lower_value, higher_value;
        lower_value = higher_value = vertex->value(*act_dim);

        bool higher_found = false , lower_found = false;

        do {
          if ((! lower_found) && lower_value) {
            QualityParameter next;
            lower  = index->getLowerIndex(lower_value, next);
            lower_value = next;
          } else {
            lower = 0;
          }

          if ((! higher_found) && higher_value) {
            QualityParameter next;
            higher = index->getHigherIndex(higher_value, next);
            higher_value = next;
          } else {
            higher = 0;
          }

          list<string>::iterator dim_pos = act_dim;
          bool leaf = false; // flag to indicate leaf-level

          // go down lower tree and check for identical values
          if (lower) {
            while ( (! leaf) && (lower) ) {
              dim_pos++;
              DEBUG_STREAM2("set pointers: lower value found, going down to " << *dim_pos << endl);
              if (lower->isLeaf()) {
                DEBUG_STREAM2("set pointers: is at leaf level. setting up pointers" << endl);
                leaf = true;
                RMVertex* lower_vertex = lower->getVertex(vertex->value(*dim_pos));
                if (lower_vertex != 0) {
                  lower_found = true;
                  vertex->setLower(*act_dim,lower_vertex);
                  if (!dummy) { // reverse direction and slope is not setup for dummy's
                    slopes.addSlope(*act_dim, Range(lower_vertex->value(*act_dim),vertex->value(*act_dim))  , slopes.computeSlope(*act_dim, lower_vertex, vertex));
                    lower_vertex->setHigher(*act_dim,vertex);
                  }
                }
              } else {
                lower = lower->getNextIndex(vertex->value(*dim_pos));
              }
            }
          }
          if ( (! lower) && (! lower_found) ) {
            DEBUG_STREAM2( "set pointers: no lower entry for dimension " << *act_dim << endl);
            vertex->setLower(*act_dim,0);
          }
          // go down higher tree and check for identical values
          dim_pos = act_dim;
          leaf = false;
          if (higher) {
            while ( (! leaf) && (higher) ) {
              dim_pos++;
              DEBUG_STREAM2( "set pointers: higher value found, going down to " << *dim_pos << endl);
              if (higher->isLeaf()) {
                DEBUG_STREAM2("set pointers: is at leaf level. setting up pointers" << endl);
                leaf = true;
                RMVertex* higher_vertex = higher->getVertex(vertex->value(*dim_pos));
                if (higher_vertex != 0) {
                  higher_found = true;
                  vertex->setHigher(*act_dim,higher_vertex);
                  if (!dummy) { // reverse direction and slope is not setup for dummy's
                    slopes.addSlope(*act_dim, Range(vertex->value(*act_dim), higher_vertex->value(*act_dim)) , slopes.computeSlope(*act_dim, vertex , higher_vertex));
                    higher_vertex->setLower(*act_dim,vertex);
                  }
                }
              } else {
                higher = higher->getNextIndex(vertex->value(*dim_pos));
              }
            }
          }
          if ( (! higher) && (! higher_found) ) {
            DEBUG_STREAM2( "set pointers: no higher entry for dimension " << *act_dim << endl);
            vertex->setHigher(*act_dim,0);
          }

        } while ( (higher && ( ! higher_found)) || (lower && ( ! lower_found )) );

        // call recursive with next index
        list<string>::iterator next_pos = act_dim;
        RMIndexVertex* next_index = index->getNextIndex(vertex->value(*act_dim));
        if (next_index) {
          setPointers(index->getNextIndex(vertex->value(*act_dim)) , vertex, ++next_pos, dummy);
        }
      } else {
        // on a leaf index seek nodes that differ only in this dimension
        RMVertex* lower  = index->getLowerVertex(vertex->value(*act_dim));
        RMVertex* higher = index->getHigherVertex(vertex->value(*act_dim));
        if (lower) {
          DEBUG_STREAM2("set pointers: lower leaf found" << endl);

          vertex->setLower(*act_dim,lower);
          if (!dummy) {
            slopes.addSlope(*act_dim, Range(lower->value(*act_dim),vertex->value(*act_dim))  , slopes.computeSlope(*act_dim, lower,vertex));
            lower->setHigher(*act_dim,vertex);
          }
        }
        if (higher) {
          DEBUG_STREAM2("set pointers: higher leaf found" << endl);

          vertex->setHigher(*act_dim,higher);
          if (!dummy) {
            slopes.addSlope(*act_dim, Range(vertex->value(*act_dim), higher->value(*act_dim)) , slopes.computeSlope(*act_dim, vertex , higher));
            higher->setLower(*act_dim,vertex);
          }
        }
      }
    }


    Result RMMesh::addVertex(const Format* description, const float cpu_usage, const float bandwidth, bool dummy ) {

      if (dummy) {
        DEBUG_STREAM2("addVertex: adding dummy vertex" << endl);
      } else {
        DEBUG_STREAM2("addVertex: adding vertex" << endl);
      }
      //// parse format
      const parameter_list desc_dim = description->getParameterList();
      typedef list<Parameter*>::const_iterator PLI;

      RMVertex* new_vertex;
      RMVertex* a_new_vertex = new RMVertex(dimension_order.getRValue(), dummy);

      if (a_new_vertex->dimensionality() != mesh_dimension.getRValue()) {
        ERROR_STREAM2("addVertex: RM dimension mismatch (unable to add " << (*description) << " to a mesh of dimension " << mesh_dimension.getRValue() << endl);
        return FAILURE;
      }

      if (!dummy) {
          vertex_index.getRValue().push_back(*a_new_vertex);
          new_vertex = &(vertex_index.getRValue().back());
          delete a_new_vertex;
      }  else {
          new_vertex = a_new_vertex;
      }

      // set the format values
      for (PLI i = desc_dim.begin(); i != desc_dim.end(); ++i) {
        string param_name = (*i)->getParameterName();
        Parameter::PType param_type = (*i)->getParameterType();
          switch (param_type) {
            case Parameter::INT32:
              new_vertex->setValue(param_name, QualityParameter(static_cast<int>(description->getIntValue( param_name ))) );
              break;
            case Parameter::FLOAT:
              new_vertex->setValue(param_name, QualityParameter(description->getFloatValue( param_name )) );
              break;
            case Parameter::STRING:
              new_vertex->setValue(param_name, QualityParameter(description->getStringValue( param_name )) );
              break;
            default:
              break;
          }
      }

      // set the cpu-usage and bandwidth for the specified format
      new_vertex->setProcessingTime(cpu_usage);
      new_vertex->setBandwidth(bandwidth);

      //// add the vertex to the index
      addVertex(*new_vertex);

      if (start != 0) {
        DEBUG_STREAM2( "addVertex: setting up pointers" << endl);
        // vertex lookup to set the lower/higher pointers of the vertex correctly for every dimension
        setPointers(start, new_vertex, dimension_order.getRValue().begin(), dummy);
        DEBUG_STREAM2( "addVertex: added vertex " << *new_vertex << endl);
      }
      return SUCCESS;
    }

    Result RMMesh::addVertex(RMVertex& vertex) {
      RMVertex* new_vertex = &vertex;

      //// add the vertex to the index
      if (start == 0) { // first vertex in mesh ?
        // the mesh is empty so we don't need to care about existing index vertices

        DEBUG_STREAM2( "addVertex: adding to empty mesh " << endl);
        // add indices
        list<string>::iterator dim_i = dimension_order.getRValue().begin();
        start = new RMIndexVertex( *dim_i );   // add first index
        DEBUG_STREAM2( "addVertex: first mesh index is at dimension " << *dim_i << endl);
        RMIndexVertex* previous_index = start;
        string previous_dim = *dim_i;
        RMIndexVertex* v_index;
        while (++dim_i != dimension_order.getRValue().end()) { // add index for every dimension
          v_index = new RMIndexVertex( *dim_i );
          DEBUG_STREAM2( "addVertex: next mesh index is at dimension " << *dim_i << endl);
          previous_index->addIndex(new_vertex->value(previous_dim),v_index);
          previous_index = v_index;
          previous_dim =  *dim_i;
        }
        // the last index is a leaf
        v_index->setLeaf();

        DEBUG_STREAM2("addVertex: add leaf: dim "<< previous_dim << " , value " << new_vertex->value(previous_dim) << endl);
        v_index->addLeaf(new_vertex->value(previous_dim),new_vertex);
        DEBUG_STREAM2( "addVertex: added vertex " << *new_vertex << endl);
        // for the first vertex we don't need to set the lower/higher pointers

        return SUCCESS;
      } else {
        // the mesh is not empty
        DEBUG_STREAM2("addVertex: adding to existing mesh" << endl);
        list<string>::iterator test_dim = dimension_order.getRValue().begin();
        RMIndexVertex* v_index = start;

        while (test_dim != dimension_order.getRValue().end()) { // check all dimensions
          DEBUG_STREAM2("addVertex: testing vertex-dimension " << *test_dim << endl);
          if ( v_index->name() != *test_dim ) {
            DEBUG_STREAM2( "addVertex: RM dimension mismatch ( " << v_index->name() << " !=  " << *test_dim << " )" << endl);
            delete new_vertex;
            return FAILURE;
          } else {
            QualityParameter act_value = new_vertex->value(*test_dim);
            if (v_index->isIncluded(act_value)) {
              DEBUG_STREAM2( "addVertex: value "<< act_value <<" found" << endl);
              // dimension-index is existing
              if (v_index->isLeaf()) {
                WARNING_STREAM2( "addVertex: vertex already existing (" << v_index << ") new value = " << new_vertex->getProcessingTime() << endl);
                return FAILURE;
              } else {
                // no leaf -> going one level down
                v_index = v_index->getNextIndex(act_value);
              }
            } else {
              DEBUG_STREAM2( "addVertex: new value" << endl);
              // value not included -> add Index/Leaf
              if (v_index->isLeaf()) { // leaf
                v_index->addLeaf(act_value, new_vertex);
              } else {
                list<string>::iterator next_dim = test_dim;
                next_dim++;
                if (next_dim == dimension_order.getRValue().end()) {
                  v_index->setLeaf();
                  v_index->addLeaf(act_value, new_vertex);
                } else {
                  RMIndexVertex* next_index = new RMIndexVertex(*next_dim);
                  v_index->addIndex(act_value, next_index);
                  v_index = next_index;
                }
              }
            }
          }
          test_dim++;
        } // end dimension check
      }
      return SUCCESS;
    }

    Result RMMesh::eraseDummyVertex(RMVertex* vertex) {
      if (vertex->getID() != 0) {
        ERROR_STREAM2("eraseDummyVertex: only dummy vertices can be erased " << *vertex << endl);
        return FAILURE;
      }

      RMIndexVertex* index;
      RMIndexVertex* next_index = start;

      list<RMIndexVertex*> index_vertices_to_be_erased;
      RMIndexVertex* last_to_remove = 0;
      QualityParameter value_to_remove = 0;

//      list<string> dims = vertex->getDimensions();

      for (list<string>::iterator i = dimension_order.getRValue().begin(); i != dimension_order.getRValue().end(); ++i) {
        index = next_index;
        if (index->name() != *i) {
          ERROR_STREAM2("eraseDummyVertex: dimension mismatch " << index->name() << " != " << *i << endl);
          return FAILURE;
        }
        // if the index has only this entry it can eventually be deleted (all lower dimensions must be checked)

        DEBUG_STREAM2("eraseDummyVertex: checking " << index->name() << " for " << vertex->value(*i) << endl);
        bool single_value = false;
        if (index->isLeaf()) {
          if ((index->getHigherVertex(vertex->value(*i)) == 0) && (index->getLowerVertex(vertex->value(*i)) == 0))
            single_value = true;
        } else {
          QualityParameter dummy;
          if ((index->getHigherIndex(vertex->value(*i)) == 0) && (index->getLowerIndex(vertex->value(*i)) == 0))
            single_value = true;
        }
        if (single_value) {
          index_vertices_to_be_erased.push_back(index);
        } else {
          last_to_remove = index;
          value_to_remove = vertex->value(*i);
          index_vertices_to_be_erased.clear(); // the index contains more than this vertex, so it (and all upper indices) can not be deleted
        }
        if ( ! index->isLeaf()) {  // set next_index if not on leaf level
          next_index = index->getNextIndex(vertex->value(*i));
        } else {
          // erase the dummy vertex
          DEBUG_STREAM2("eraseDummyVertex: deleting dummy vertex" << *(index->getVertex(vertex->value(*i))) << endl);
          delete index->getVertex(vertex->value(*i));
          break;
        }
      }

      // remove unneeded indices (list of indices to delete?)
      for (list<RMIndexVertex*>::reverse_iterator i = index_vertices_to_be_erased.rbegin(); i != index_vertices_to_be_erased.rend(); ++i) {
        DEBUG_STREAM2("eraseDummyVertex: deleting index " << (*i)->name() << endl);
        delete *i;
      }
      if (last_to_remove) {
        DEBUG_STREAM2("removing index entry "<< value_to_remove << " on index " << last_to_remove->name() << endl);
        last_to_remove->eraseIndex(value_to_remove);
      }

      return SUCCESS;
    }


    map<string,QualityParameter>* RMMesh::formatToMap(const Format& description) {

      map<string,QualityParameter>* vertex_dims = new map<string,QualityParameter>;

      const parameter_list desc_dim = description.getParameterList();

      typedef list<Parameter*>::const_iterator PLI;

      // the format-parameters are vertex-dimensions
      for (PLI i = desc_dim.begin(); i != desc_dim.end(); ++i) {
        Parameter::PType param_type = (*i)->getParameterType();
        string param_name = (*i)->getParameterName();
        switch (param_type) {
          case Parameter::INT32:
            (*vertex_dims)[param_name] = QualityParameter(static_cast<int>(description.getIntValue( param_name )));
            break;
          case Parameter::FLOAT:
            (*vertex_dims)[param_name] = QualityParameter(description.getFloatValue( param_name ));
            break;
          case Parameter::STRING:
            (*vertex_dims)[param_name] = QualityParameter(description.getStringValue( param_name ));
            break;
          default:
            break;
        }
      }

      return vertex_dims;
    }


    RMVertex* RMMesh::getVertex(const Format* description) {
      // test for dimension mismatch
      if (description->getNumberOfValues() != mesh_dimension.getRValue()) {
        ERROR_STREAM2("dimension mismatch" << endl);
        return 0;
      }

      // seek for the vertex and return a pointer to it

      DEBUG_STREAM2("Looking for matching vertex" << endl);

      // convert the format description to a map of string, float pairs
      map<string,QualityParameter>* vertex_dims = formatToMap(*description);
      list<string>::const_iterator dim_i = dimension_order.getRValue().begin();

      RMIndexVertex* act_index = start; // start with the first index
      while ( (act_index != 0) && (! act_index->isLeaf()) ) {
         act_index = act_index->getNextIndex((*vertex_dims)[*dim_i]);
         dim_i++;
      }
      RMVertex* return_vertex;
      if (act_index != 0) {
        return_vertex = act_index->getVertex((*vertex_dims)[*dim_i]);
      } else {
        return_vertex = 0;
      }
      delete vertex_dims;

      return return_vertex;
    }


    float RMMesh::getBandwidth(Format* description) {
      // test for dimension mismatch
      if (description->getNumberOfValues() != mesh_dimension.getRValue()) {
        ERROR_STREAM2("dimension mismatch" << endl);
        return 0.0F;
      }

      // look if there is already a vertex with the specified format
      RMVertex* ret_vert = getVertex(description);
      if (ret_vert) {
        return ret_vert->getBandwidth();
      } else {
        DEBUG_STREAM2("getUsage: no matching vertex found" << endl);
      }
      return 0.0F;

      // estiumation

    }


    float RMMesh::getUsage(Format* description) {
      // test for dimension mismatch
      if (description->getNumberOfValues() != mesh_dimension.getRValue()) {
        ERROR_STREAM2("dimension mismatch" << endl);
        return 0.0F;
      }

      if (vertex_index.getRValue().empty()) {
        WARNING_STREAM2("empty mesh" << endl);
        return 0.0F;
      }

      // look if there is already a vertex with the specified format
      RMVertex* ret_vert = getVertex(description);
      if (ret_vert) {
        return ret_vert->getProcessingTime();
      } else {
        DEBUG_STREAM2("getUsage: no matching vertex found" << endl);
      }

      // if not we need to estimate the cpu usage
      // this is done by linear interpolation

      // insert a dummy vertex to find a format that nearly matches
      addVertex(description,0.0F,true);

      // look for vertices that differ only in one dimension
      RMVertex* vertex = getVertex(description); // get the vertex we just inserted
//      map<string,QualityParameter>* dims = formatToMap(*description); // parse format

//      for (map<string,QualityParameter>::iterator i = dims->begin(); i != dims->end(); ++i) {
      for (list<string>::iterator i = dimension_order.getRValue().begin(); i != dimension_order.getRValue().end(); ++i) {
        RMVertex* lower = vertex->lower(*i);
        RMVertex* higher = vertex->higher(*i);
        if (lower && higher) { //closest match
          DEBUG_STREAM2("getUsage: found a dimension with closest values" << endl);
          float x_to_estimate = vertex->value(*i);
          QualityParameter x1 = lower->value(*i);
          float y1 = lower->getProcessingTime();
          QualityParameter x2 = higher->value(*i);
          float y2 = higher->getProcessingTime();
          DEBUG_STREAM2("getUsage: closest values:"<< x1 << " = " << y1 << ",  '" << x_to_estimate << "'  ," << x2 << " = " << y2 << endl);
          float ret_val = linearEstimate(x_to_estimate,x1,y1,x2,y2);
          DEBUG_STREAM2("getUsage: estimated value: "<< ret_val << endl);

          DEBUG_STREAM2("getUsage: deleting dummy vertex" << endl);
          eraseDummyVertex(vertex);// delete dummy

          return ret_val;
        }
        if (lower) {  // try to extrapolate on lower values
          DEBUG_STREAM2("found lower vertex on dim " << *i << endl);
          higher = lower;
          lower = higher->lower(*i);
          if (lower) {
            DEBUG_STREAM2("getUsage: found a dimension with close values" << endl);
            float x_to_estimate = vertex->value(*i);
            QualityParameter x1 = lower->value(*i);
            float y1 = lower->getProcessingTime();
            QualityParameter x2 = higher->value(*i);
            float y2 = higher->getProcessingTime();

            float ret_val = linearEstimate(x_to_estimate,x1,y1,x2,y2);
            DEBUG_STREAM2("getUsage: estimated value: "<< ret_val << endl);

            DEBUG_STREAM2("getUsage: deleting dummy vertex" << endl);
            eraseDummyVertex(vertex);// delete dummy

            return ret_val;
          } else {
            lower = higher;
            DEBUG_STREAM2("using slope to estimate dim " << *i << endl);
            const Range dimension_range(lower->value(*i), vertex->value(*i));
            float slope = slopes.getSlope(*i,dimension_range);
            float diff = vertex->value(*i) - lower->value(*i);
            float ret_val = lower->getProcessingTime() * (diff  * slope);
            DEBUG_STREAM2("getUsage: estimated value: "<< ret_val << endl);

            DEBUG_STREAM2("getUsage: deleting dummy vertex" << endl);
            eraseDummyVertex(vertex);// delete dummy

            return ret_val;
          }
        }
        if (higher) {  // try to extrapolate on higher values
          DEBUG_STREAM2("found higher vertex on dim " << *i << endl);
          lower = higher;
          higher = lower->higher(*i);
          if (higher) {
            DEBUG_STREAM2("getUsage: found a dimension with close values" << endl);
            float x_to_estimate = vertex->value(*i);
            QualityParameter x1 = lower->value(*i);
            float y1 = lower->getProcessingTime();
            QualityParameter x2 = higher->value(*i);
            float y2 = higher->getProcessingTime();

            float ret_val = linearEstimate(x_to_estimate,x1,y1,x2,y2);
            DEBUG_STREAM2("getUsage: estimated value: "<< ret_val << endl);

            DEBUG_STREAM2("getUsage: deleting dummy vertex" << endl);
            eraseDummyVertex(vertex);// delete dummy

            return ret_val;
          } else {
            //************************************************* untested !?
            higher = lower;
            DEBUG_STREAM2("using slope to estimate dim " << *i << endl);
            const Range dimension_range(vertex->value(*i), higher->value(*i));
            float slope = slopes.getSlope(*i,dimension_range);
            float diff = higher->value(*i) - vertex->value(*i);
            float ret_val = higher->getProcessingTime() * (diff * slope);
            DEBUG_STREAM2("getUsage: estimated value: "<< ret_val << endl);

            DEBUG_STREAM2("getUsage: deleting dummy vertex" << endl);
            eraseDummyVertex(vertex);// delete dummy

            return ret_val;
          }
        }
      }
      eraseDummyVertex(vertex);
//////////////////////////////

      DEBUG_STREAM2("no close match for getUsage()" << endl);

      // no close match (no vertex which differs in only one dimension)

      // we first need to find the first dimension that does not match

      vertex = new RMVertex(*description);
      RMIndexVertex* index_position = findFirstNonMatching(*vertex);
      if (index_position->isLeaf()) {
        DEBUG_STREAM2("RMMesh::getUsage(): found vertex that differs only in last dimension" << endl);
      }
      DEBUG_STREAM2("no further match at "<< index_position->name() << endl);

      //check for string value (no interpolation possible)
      if ( (vertex->value(index_position->name())).getParameterType() == QualityParameter::STRING) {
        WARNING_STREAM2("unable to interpolate between string values" << endl);
        delete vertex;

        //FIXME return vertex with another string value

        return 0.0F;
      }

      // now we need to find a 'close' vertex, so we seek for a higher or lower value

      while (! index_position->isLeaf()) {
        DEBUG_STREAM2("getUsage: seeking at level "<< index_position->name() << endl);

        //FIXME check if there is a matching slope

        // first try to get the exact value
        RMIndexVertex* next = index_position->getNextIndex(vertex->value(index_position->name()));
        if (next == 0) { // then try the lower value
          DEBUG_STREAM("taking lower value on " << index_position->name() << endl);
          next = index_position->getLowerIndex(vertex->value(index_position->name()));
        }
        if (next == 0) { // finally try the higher value
          DEBUG_STREAM("taking higher value on " << index_position->name() << endl);
          next = index_position->getHigherIndex(vertex->value(index_position->name()));
        }
        if (next == 0) {
          ERROR_STREAM2("RM: empty index in mesh!" << endl);
          return 0;
        }
        index_position = next;
      }

      float ret_val = 0;
      RMVertex* from_vertex;
      if ((from_vertex = index_position->getVertex(vertex->value(index_position->name()))) != 0 ) { //identical value
        DEBUG_STREAM2("estimating from: " << *from_vertex << endl);
        ret_val = computeUsage(*from_vertex, *vertex);
      }
      if ((from_vertex = index_position->getLowerVertex(vertex->value(index_position->name()))) != 0 ) { // lower value
        DEBUG_STREAM2("estimating from: " << *from_vertex << endl);
        ret_val = computeUsage(*from_vertex, *vertex);
      } else {
        if ((from_vertex = index_position->getHigherVertex(vertex->value(index_position->name()))) != 0 ) { // higher value
          DEBUG_STREAM2("estimating from: " << *from_vertex << endl);
          ret_val = computeUsage(*from_vertex, *vertex);
        }
      }

      delete vertex;
      return ret_val;
    }


    // this is only called if it is sure that no matching vertex is available
    RMIndexVertex* RMMesh::findFirstNonMatching(RMVertex &vertex) {
       RMIndexVertex* act_index = start;
       RMIndexVertex* last_index = start;
//       list<string> vertex_dims = vertex.getDimensions();

       for (LI i = dimension_order.getRValue().begin(); i != dimension_order.getRValue().end() ; ++i) {
         if (! last_index->isLeaf()) {
           act_index = last_index->getNextIndex(vertex.value(*i));
           if (act_index) {
              last_index = act_index;
           } else {
              DEBUG_STREAM2("found non matching parameter at dimension: "<< *i << endl);
              break;
           }
         } else {
           // the first non matching dimension must be at leaf level
           return last_index;
         }
       }
       return last_index;
    }

    float RMMesh::computeUsage(RMVertex &from, RMVertex &to) {
      DEBUG_STREAM2( "estimation from: " << from << "  to   " << to << endl);
      float act_usage = from.getProcessingTime();
      // multiply the usage value with the slope for every dimension
      list<string> dimensions = from.getDimensions();

      for (LI i = dimensions.begin(); i != dimensions.end(); ++i) {
        QualityParameter from_val = from.value(*i); 
        QualityParameter to_val   = to.value(*i); 
        if (from_val != to_val) {
          QualityParameter diff;
          float slope;
          if (from_val > to_val) {
            diff = from_val - to_val;
            slope = slopes.getSlope(*i,Range(to_val,from_val));
            DEBUG_STREAM2("using slope : " << from_val << " ... " << to_val << "  =  " << slope << endl);
          } else {
            diff = to_val - from_val;
            slope = slopes.getSlope(*i,Range(from_val,to_val));
            DEBUG_STREAM2("using slope : " << from_val << " ... " << to_val << "  =  " << slope << endl);
          }
          act_usage *= (diff * slope);
        }
      }
      return act_usage;
    }

    float RMMesh::linearEstimate(float x_to_estimate, QualityParameter x1, float y1, QualityParameter x2, float y2) {
      float x_a = x1 > x2 ? x2 : x1;   // lower value of x1 x2
      float x_e = x1 > x2 ? x1 : x2;   // higher value of x1 x2
      float y_a = y1 > y2 ? y2 : y1;   // lower value of y1 y2
      float y_e = y1 > y2 ? y1 : y2;   // higher value of y1 y2
      float x_d = x_e - x_a;           // delta x

      float y_d = y_e - y_a;           // delta y

      float x_y_slope = y_d / x_d;

      float y_est;

      bool extrapolation_front;
      bool extrapolation_back;
      if (x_a > x_to_estimate) {
        DEBUG_STREAM2("extrapolating on front" << endl);
        extrapolation_front = true;
      }
      if (x_to_estimate > x_e) {
        DEBUG_STREAM2("extrapolating on back" << endl);
        extrapolation_back = true;
      }

      if (extrapolation_front) {
        y_est = y_a - ((x_a - x_to_estimate) * x_y_slope);
      } else {
        if (extrapolation_back) {
          y_est = y_e + ((x_to_estimate - x_e) * x_y_slope);
        } else {
          y_est = ((x_to_estimate - x_a) * x_y_slope) + y_a;
        }
      }
      return y_est;
    }

    Result RMMesh::savePoints(const string& filename) {
#if(__GNUC__>=3)
      ofstream outfile(filename.c_str(),ios_base::out|ios::trunc);//,0666);
#else 
      ofstream outfile(filename.c_str(),ios::out|ios::trunc, 0666);
#endif
      //  param       value    num_value
      map<string, map<string , uint> > string_trans; // translation of strings to numercal values
      if (outfile) {
        // first write the dimensionality
        outfile << mesh_dimension.getRValue() << endl;
        // then write all vertices to disc
        for (list<RMVertex>::iterator i = vertex_index.getRValue().begin(); i !=  vertex_index.getRValue().end(); ++i) {
          // process dimensions in the same order
          for (list<string>::iterator j = dimension_order.getRValue().begin(); j != dimension_order.getRValue().end(); ++j) {
            if (i->value(*j).getParameterType() == QualityParameter::STRING) {
              // translate string to numercal value
              if (string_trans.find(*j) == string_trans.end()) { // parameter not in map
                string_trans[*j][(i->value(*j)).getStringValue()] = 1; // initial value is 1
              } else {
                if (string_trans[*j].find((i->value(*j)).getStringValue()) == string_trans[*j].end()) { // parameter in map but value not
                  string_trans[*j][(i->value(*j)).getStringValue()] = 100 * string_trans[*j].size();
                }
              }
              // now the parameter and the value are in the map
              outfile << string_trans[*j][(i->value(*j)).getStringValue()] << ':';
            } else {
              outfile << i->value(*j) << ':';
            }
          }
          outfile << i->getProcessingTime() << ',' << i->getBandwidth() << endl;
        }
        return SUCCESS;
      } else {
        return FAILURE;
      }
    }

    void RMMesh::setFormat(const Format& f) {
      mesh_format = f;
    }

    void RMMesh::getFormat(Format& f) {
      f = mesh_format.getRValue();
    }


    void RMMesh::printIndex() {
      for (list<RMVertex>::iterator i = vertex_index.getRValue().begin(); i != vertex_index.getRValue().end(); i++) {
        cerr << *i << endl;
      }
    }
    void RMMesh::printDimensions() {
      for (list<string>::iterator i = dimension_order.getRValue().begin(); i != dimension_order.getRValue().end(); i++) {
        cerr << *i << endl;
      }
      cerr << "#dimensions: " << dimension_order.getRValue().size() << endl;
      cerr << "#mesh-dimensions: " << mesh_dimension.getRValue() << endl;
    }

    void RMMesh::updateAttributes() {
      // copy the vertex_index list to a vector to allow fast access to its elements

      vector<RMVertex*> v_vertex_index = vector<RMVertex*>(vertex_index.getRValue().size()+1);
      uint k = 0;
      for (list<RMVertex>::iterator i = vertex_index.getRValue().begin(); i != vertex_index.getRValue().end(); ++i) {
        v_vertex_index[++k] = &(*i);
      }

      // update the lower/higher-pointers and add to index
      for (list<RMVertex>::iterator i = vertex_index.getRValue().begin(); i != vertex_index.getRValue().end(); ++i) {
        i->update(v_vertex_index);
        addVertex(*i);
      }

    }

    void RMMesh::prepareAttributes() {
      for (list<RMVertex>::iterator i = vertex_index.getRValue().begin(); i != vertex_index.getRValue().end(); ++i) {
        i->prepare();
      }
    }

    void RMMesh::registerValues() {
      insertArgument(&vertex_index);
      insertArgument(&dimension_order);
      insertArgument(&mesh_dimension);
      insertArgument(&mesh_format);
    }

  OValueStream& operator<<(NMM::OValueStream& ov, const map<string,map<Range    ,pair<float,uint> > >& slope_map) {
    return writeMap<string, map<Range    ,pair<float,uint> > >(ov, slope_map, string_type_info, TypeInfo("Slope", "NMM"));
  }

  IValueStream& operator>>(NMM::IValueStream& iv, map<string, map<Range    ,pair<float,uint> > >& slope_map) {
    return readMap<string, map<Range    ,pair<float,uint> > >(iv, slope_map, string_type_info, TypeInfo("Slope", "NMM"));
  }

  OValueStream& operator<<(NMM::OValueStream& ov, const map<Range    ,pair<float,uint> >& range_map) {
    return writeMap<Range    ,pair<float,uint> >(ov, range_map, string_type_info, TypeInfo("RangeMap", "NMM"));
  }

  IValueStream& operator>>(NMM::IValueStream& iv, map<Range    ,pair<float,uint> >& range_map) {
    return readMap<Range    ,pair<float,uint> >(iv, range_map, string_type_info, TypeInfo("RangeMap", "NMM"));
  }

  OValueStream& operator<<(NMM::OValueStream& ov, const pair<float,uint>& slope_pair) {
    return writePair<float,uint>(ov, slope_pair, "Float" , "UINT");
  }

  IValueStream& operator>>(NMM::IValueStream& iv, pair<float,uint>& slope_pair) {
    return readPair<float,uint>(iv, slope_pair, "Float" , "UINT");
  }

}
