/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "Slope.hpp"

using namespace NMM;

float Slope::getSlope(const string& dim, Range start) {
  DEBUG_STREAM2("lookup for slope on dimension: " << dim << "  (" << start.start() << " - " << start.end() << ")" << endl);
  if (slopes.getRValue().find(dim) == slopes.getRValue().end()) {
    ERROR_STREAM2("RM: no slope for dimension " << dim << endl);
    return 0.0F;
  }
  map<Range, pair< float,uint> > dim_slope = slopes.getRValue()[dim];
  map<Range, pair< float,uint> >::iterator mapi;
  if (dim_slope.empty()) {
    ERROR_STREAM2("RM: no slope for dimension " << dim << endl);
    return 0.0F;
  }
  if ((mapi = dim_slope.find(start)) != dim_slope.end()) {
    return (mapi->second.first); // we have a value for exactly this Range
  }
  if ((mapi = dim_slope.find(start.reverse())) != dim_slope.end()) {
    return (1 / mapi->second.first); // we have a value for the reverse Range
  }

  if (start.start().getParameterType() == QualityParameter::STRING) {
    WARNING_STREAM2("No match for string parameter (" << start.start() << " -> " << start.end() << endl );
  }
  DEBUG_STREAM2("RM: no slope for range " << start.start() << "-" << start.end() << " at dimension " << dim << endl);
  // no exact value -> take a close one
  dim_slope[start] = pair<float, uint>(0,0);
  mapi = dim_slope.find(start);
  if (mapi != dim_slope.begin()) { // estimate from preceding element
    mapi--;
  } else { // estimate from following element
    mapi++;
  }
  dim_slope.erase(start);
  DEBUG_STREAM2("RM: taking slope for range " << mapi->first.start() << "-" << mapi->first.end() << " (" << mapi->second.first << ")" << endl);

  QualityParameter org_range = mapi->first.end() - mapi->first.start();
  QualityParameter new_range = start.end() - start.start();
  float return_slope = mapi->second.first / (org_range / new_range);

  return return_slope;
}

float Slope::computeSlope(const string& dim, RMVertex* lower, RMVertex* higher) {
  DEBUG_STREAM2("computing slope between " << *lower << " and " << *higher << endl);
  // check for string QualityParameters
  if  (higher->value(dim).getParameterType() ==  QualityParameter::STRING) {
    return (higher->getProcessingTime() / lower->getProcessingTime());
  }

  return (higher->getProcessingTime() / lower->getProcessingTime()) /  (higher->value(dim) - lower->value(dim));
}


Result Slope::addSlope(const string& dim, Range start, float val) {
  // a slope is added for a dimension

  DEBUG_STREAM2("adding slope ("<< dim << ", from: " << start.start() << " to: " << start.end() << " is " << val << ")" << endl);

  if (slopes.getRValue().find(dim) == slopes.getRValue().end()) {
    DEBUG_STREAM2("creating slope for dimension " << dim << endl);
  }
  if (slopes.getRValue()[dim][start] != pair<float,uint>( 0,0)) { // a slope at this start-point already exists
    pair<float,uint> slope = slopes.getRValue()[dim][start];
    pair<float,uint> new_slope;
    new_slope.first = ((slope.first * slope.second) + val) / (slope.second + 1);
    new_slope.second = (slope.second + 1);
    slopes.getRValue()[dim][start] = new_slope;
  } else { // new slope
    pair<float,uint> new_slope;
    new_slope.first = val;
    new_slope.second = 1;
    slopes.getRValue()[dim][start] = new_slope;
  }

  return SUCCESS;
}

Slope::Slope(const Slope& s) :  NMM::SerializedValue(s) , NamedObject("Slope") {
  registerValues();
  *this = s;
}

TValueConstructor<map<string,  map<Range    ,pair<float,uint> > > > __nmm_map_string_range ("map_string_range", "std");
TConstructor<Range> __nmm_range ("Range", "NMM");
