/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef _NMM_COMPLEXVALUE_HPP_
#define _NMM_COMPLEXVALUE_HPP_

#include "Value.hpp"

#include <string>
#include <list>
#include <vector>

namespace NMM {


class IValueStream;
class OValueStream;

/** This class is base class for all serializable objects.
  * If a class inherits SerializedValue, it can register its attributes, that  should
  * be serialized. This class implements methods that serialize or deserilaize this object
  * with any ValueStream. The attributes must be a subtype of Value. If the attibute is
  * not a Value you can use the TValue-Wrapper class.
  */
class SerializedValue : public Value {
public:
/** @name Constructor & Destructor
  */
//@{
/** Default constructor
  * \param reconstruction_mode This flag specifies if the object should use the ValueFactory to
  *        create attributes if it is deserialized
  * \param memory_mgm Specified the value-type
  */
  SerializedValue(bool reconstruction_mode = false, Value::MemoryMgm memory_mgm = INOUTVALUE);

/** Constructor with type information
  * \param type_name Typename from the inherited object
  * \param ns Namespace from the inherited object
  * \param reconstruction_mode This flag specifies if the object should use the ValueFactory to
  *        create attributes if it is deserialized
  * \param memory_mgm Specified the value-type
  */
  SerializedValue(const string& type_name , const string& ns , bool reconstruction_mode = false, Value::MemoryMgm memory_mgm = INOUTVALUE);

/** Constructor with type information
  * \param ti TypeInfo from the inherited object
  * \param reconstruction_mode This flag specifies if the object should use the ValueFactory to
  *        create attributes if it is deserialized
  * \param memory_mgm Specified the value-type
  */
  SerializedValue(const TypeInfo& ti, bool reconstruction_mode = false, Value::MemoryMgm memory_mgm = INOUTVALUE);

/** Copy constructor
  * \param sv Creates a copy from given object
  */
  SerializedValue(const SerializedValue& sv);
//@}

/** @name Serialization & Deserialization
  */
//@{
/** This methods deserialize an object from given IValueStream and writes the values to the registered
  * values. The registered attributes are desrialized by calling this method too. So the registered values
  * are not allocated again, they are used to store the values in the IValueStream.
  * \param iv IValueStream to read the values
  */
  virtual void assignFromStream(IValueStream& iv);

/** This methods deserialize an object from given IValueStream and writes the values to the registered
  * values. The registered attributes are desrialized by calling this method too. So the registered values
  * allocates memory again.
  * \param iv IValueStream to read the values
  */
  virtual void createFromStream(IValueStream& iv);

/** This methods serialize an object and writes all registered attributes to the given OValueStream.
  * \param ov OValueStream to write registered attributes.
  */
  virtual void toStream(OValueStream& ov)          const;
//@}

/** @name Type information
  */
//@{
/** This method returns the type name of the object as string
  */
  inline  const string& getTypeName()           const { return __type_info.t_name; }

/** This method returns the namespace of the object as string
  */
  inline  const string& getNamespace()          const { return __type_info.t_namespace; }

/** This method returns the typeinformation in an TypeInfo struct
  */
  inline  const TypeInfo& getTypeInfo()         const { return __type_info; }
//@}

/** @name Memory management
  */
//@{
/** This method frowards the method call to all registered values. In-Values are deleted, all
  * other types not.
  */
  virtual void releaseValue()                   const;

/** This method creates a copy of this object by calling the copy constructor. The caller of this
  * method is responsible to destroy the object.
  */
  virtual SerializedValue* copy()               const { return new SerializedValue(*this); }
//@}

/** @name Operator's
  */
//@{
/** Implements the assignment operator
  */
  const SerializedValue& operator=(const SerializedValue&);
//@}


protected:
/** @name Configure object
  */
//@{
/** Inserts a new Argument that should be serialized or deserialized
  */
  inline  void insertArgument(Value* value)           { __arg_array.push_back(value); }

/** Inserts a new Argument that should be serialized or deserialized
  */
  inline  void registerAttribute(Value* value)        { __arg_array.push_back(value); }

/** Sets the type name of the object
  */
  inline  void setTypeName(const string& type_name)   { __type_info.t_name = type_name;}
                                                        
/** Sets the namespace of the object
  */
  inline  void setNamespace(const string& ns )        { __type_info.t_namespace = ns; }

/** Sets the TypeInfo object
  */
  inline  void setTypeInfo(const TypeInfo& ti)        { __type_info = ti; }
//@}

/** @name Access attributes
  */
//@{
/** returns the size of the internal vector that stores the registered attributes
  */
  inline  size_t getSize()                      const { return __arg_array.size(); }

/** returns the i'th value in the array of registered attributess
  */
  inline        Value* getArgument(size_t pos)        { return __arg_array[pos];}

/** returns the i'th value in the array of registered attributess
  */
  inline  const Value* getArgument(size_t pos)  const { return __arg_array[pos];}

/** returns the i'th value in the array of registered attributess
  */
  inline        Value* getAttribute(size_t pos)       { return __arg_array[pos];}

/** returns the i'th value in the array of registered attributess
  */
  inline  const Value* getAttribute(size_t pos) const { return __arg_array[pos];}
//@}


/** @name Update Serialization
  */
//@{
/** This template method is called after every deserialization process and can
  * be used by a base class to be informed about some changes
  * \sa fromStream
  * \sa createFromStream
  */
  virtual void updateAttributes()                               {};

/** This template method is called before every serialization process and can
  * be used by a base class to be informed about a serialization process.
  * If a child class uses some implementation specific strategies like lazy evaluation
  * this method can be used to update the attributes in case of a objectserialization.
  * \sa fromStream
  * \sa createFromStream
  */
  virtual void prepareAttributes()                              {};
//@}

private: //methods
  /* private assignment operator that is used by operator= */
  const SerializedValue& assign(const SerializedValue& sv);

private: //member
  vector<Value*>  __arg_array;                                  ///< stores the registered values
  TypeInfo        __type_info;                                  ///< stored the TypeInfo of this object
  bool            __reconstruction_mode;                        ///< flag if the deserialization should allocate the object
};


template<typename T = Value >
class SequenceValue : public Value , public vector<T*> {
public:
/** @name Constructor & Destructor
  */
//@{
/** Default-Constructor
  */
  SequenceValue(MemoryMgm memory_mgm = INVALUE);

/** Constructor, which accepts a specific size
  */
  SequenceValue(size_t size, MemoryMgm memory_mgm = INOUTVALUE);

/** Copy-Constructor
  */
  SequenceValue(const SequenceValue& );
//@}

/** @name Serialization & Deserialization
  */
//@{
/** This methods deserialize an object from given IValueStream and writes the values to the registered
  * values. The registered attributes are desrialized by calling this method too. So the registered values
  * are not allocated again, they are used to store the values in the IValueStream.
  * \param iv IValueStream to read the values
  */
  virtual void assignFromStream(IValueStream& iv);

/** This methods deserialize an object from given IValueStream and writes the values to the registered
  * values. The registered attributes are desrialized by calling this method too. So the registered values
  * allocates memory again.
  * \param iv IValueStream to read the values
  */
  virtual void createFromStream(IValueStream& iv);

/** This methods serialize an object and writes all registered attributes to the given OValueStream.
  * \param ov OValueStream to write registered attributes.
  */
  virtual void toStream(OValueStream& ov)          const;

/** This mehtod sets a config-flag to specify that a TValue-Wrapper should be used for every
  * values. If this flag is set to false a TValue wrapper is only used if the stored type is
  * not a subtype of Value.
  */
  inline  void useTValue(bool use_tvalue)                   { __use_tvalue = use_tvalue; }
//@}

/** @name Memory management
  */
//@{
/** This method frowards the method call to all registered values. In-Values are deleted, all
  * other types not.
  */
  virtual void releaseValue()                   const;

/** This method creates a copy of this object by calling the copy constructor. The caller of this
  * method is responsible to destroy the object.
  */
  virtual SequenceValue<T>* copy()              const;
//@}

/** @name Type information
  */
//@{
/** This method sets the type_name of this object
  */
  inline void setTypeName(const string& type_name)  {__type_info.t_name = type_name;}

/** This method sets the namespace of this object
  */
  inline void setNamespace(const string& ns)        {__type_info.t_namespace = ns; }

/** This method returns the type name of the object as string
  */
  inline  const string& getTypeName()           const { return __type_info.t_name; }

/** This method returns the namespace of the object as string
  */
  inline  const string& getNamespace()          const { return __type_info.t_namespace; }

/** This method returns the typeinformation in an TypeInfo struct
  */
  inline const TypeInfo& getTypeInfo()  const       {return __type_info;}
//@}

/** @name Operator's
  */
//@{
/** Implements the assignment operator
  */
  const SequenceValue<T>& operator=(const SequenceValue<T>&);
//@}
private:
/** creates a copy from given SequenceValue<T>
  */
  const SequenceValue<T>& assign(const SequenceValue<T>&);

/** This method delets all elements
  */
  void eraseElements();

private: //attributes
  TypeInfo  __type_info;                      ///< stores the TypeInfo
  bool      __use_tvalue;                     ///< define if a TValue-Wrapper is used for Value-Types too
};





/** SequenceValue implementation **/
template<typename T>
SequenceValue<T>::SequenceValue(size_t _size, MemoryMgm memory_mgm )
  : Value(Value::SEQUENCE, memory_mgm), vector<T*>(_size), __type_info("SequenceValue","NMM"), __use_tvalue(true)
{
  /* Initialize the internal array with NULL pointer */
  for (unsigned int i = 0; i < _size; ++i)
    (*this)[i] = NULL;
}

template<typename T>
SequenceValue<T>::SequenceValue(MemoryMgm memory_mgm )
  : Value(Value::SEQUENCE, memory_mgm), __type_info("SequenceValue", "NMM"), __use_tvalue(true)
{
}

template<typename T>
SequenceValue<T>::SequenceValue(const SequenceValue<T>& av)
  : Value(av), vector<T*>(av)
{
  assign(av);
}

template<typename T>
const SequenceValue<T>& SequenceValue<T>::operator=(const SequenceValue<T> & seq){
  /* cleanup old vector */
  eraseElements();

  /* call the operator= from the base class Value */
  Value::operator=(seq);
  /* call the operator= from the base class vector */
  vector<T*>::operator=(seq);

  /* create a copy of values in geven seq */
  return assign(seq);
}

template<typename T>
const SequenceValue<T> & SequenceValue<T>::assign(const SequenceValue<T> & seq){
  /*Begin> create a copy of the new values*/
  __use_tvalue = seq.__use_tvalue;

  for (size_t i = 0; i< seq.size(); ++i)
    if (seq[i])
      (*this)[i] = dynamic_cast<T*>(seq[i]->copy());
    else
      (*this)[i] = 0;
  /*End */
  __type_info = seq.getTypeInfo();
  return *this;
}

template<typename T>
void SequenceValue<T>::eraseElements() {
  /*Begin> release the old values */
  for (size_t i = 0; i< this->size(); ++i)
    if ((*this)[i])
      (*this)[i]->releaseValue();
  /*End< */
  this->clear();
}

template<typename T>
void SequenceValue<T>::assignFromStream(IValueStream& ivs){
  /* Begin> Type checking */
  if ( (ivs.getCurrentType() != getTypeName()) || ( ivs.getCurrentNamespace() != getNamespace())) {
    GLOBAL_ERROR_STREAM ("Serialized Value> Current type : " << ivs.getCurrentType() << "                Expected type: " << getTypeName() << endl);
    GLOBAL_ERROR_STREAM ("Serialized Value> Current namespace : " << ivs.getCurrentNamespace() << "      Expected namespace: " << getNamespace() << endl);
    return;
  }
  /* End< */

  /* Begin> Try to access attributes */
  if (!ivs.goDown()) {
    GLOBAL_DEBUG_STREAM2("Array seems to be empty");
    ivs.nextElement();
    return;
  }
  /* End< */

  /* Should we TValue-Wrapper for all Values (which is need for Events) */
  ivs >> __use_tvalue;

  bool is_pointer = false;

  unsigned int i = 0;
  while(!ivs.isEmpty()) {
    /* Is the Value a Pointer Type ?*/
    ivs >> is_pointer;
    if (i < this->size()) {
      /* If an object already exist, use it*/
      if ((*this)[i]) {
        /* forward the assign fromStream call, if it exist */
        ((*this)[i])->assignFromStream(ivs);
      }
      /* Otherwise, we must create it */
      else {
        Value* tmp_d = 0;
        if (__use_tvalue) {
          if (is_pointer) {
            /* create TValue which manages a pointer to the object */
            tmp_d = SValueFactory::getInstance().createTPointerValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
          } else {
            /* create a TValue for the object */
            tmp_d = SValueFactory::getInstance().createTValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
          }
        } else {
          /* create a simple Value object, without TValue-Wrapper */
          tmp_d = SValueFactory::getInstance().createValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
        }

        /* Begin> Use the created Value with the given ValueStream */
        if (!tmp_d) {
          /* If no such object could be created, skip this element */
          ivs.nextElement(); //skip this element
          /* If no such object could be created, we receive a NULL pointer */
          (*this)[i] = 0;
        } else {
          /* Otherwise call createFromStream */
          tmp_d->createFromStream(ivs);

          /* Perform type check */
          T* t_value = dynamic_cast<T*>(tmp_d);
          if (t_value) {
            /* insert the element */
            (*this)[i] = t_value;
          } else {
            /* If the dynamic_cast fails, no such object with correct type could be created, set a NULL-Ptr */
            (*this)[i] = 0;
            /* delete the tmp_value */
            delete tmp_d;
          }
        }
        /* End< */
      }
    } else {
      Value* tmp = 0;
      if (__use_tvalue) {
        if (is_pointer) {
          /* create TValue wich manages a pointer to the object */
          tmp = SValueFactory::getInstance().createTPointerValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
        } else {
          /* create a TValue for the object */
          tmp = SValueFactory::getInstance().createTValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
        }
      } else {
        /* create a simple Value object, without TValue-Wrapper */
        tmp = SValueFactory::getInstance().createValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
      }
      /* Begin> Use the created Value with the given ValueStream */

      if ( tmp ) {
        /* read the value */
        tmp->createFromStream(ivs);

        T* t_value = dynamic_cast<T*>(tmp);
        if (t_value) {
          /* insert the element */
          push_back(t_value);
        } else {
          /* If the dynamic_cast fails, no such object could be created, insert a NULL-Ptr */
          this->push_back(0);
          /* delete the tmp_value */
          delete tmp;
        }
      } else {
        /* we received a NULL pointer and insert it */
        this->push_back(0);
        ivs.nextElement();
      }
    }
    ++i;
  }

  ivs.goUp();
}


template<typename T>
void SequenceValue<T>::createFromStream(IValueStream& ivs){
  /* Begin> Type checking */
  if ( (ivs.getCurrentType() != getTypeName()) || (ivs.getCurrentNamespace() != getNamespace())) {
    GLOBAL_ERROR_STREAM ("Invalid Type");
    return;
  }
  /* End< */

  /* Begin> Try to access attributes */
  if (!ivs.goDown()) {
    GLOBAL_DEBUG_STREAM2 ("Array seems to be empty");
    ivs.nextElement();
    return;
  }
  /* End< */

  /* Begin> Release the old values and create new one */
  for (unsigned int i = 0; i< this->size(); ++i)
    if ((*this)[i])
      ((*this)[i]->releaseValue());
  /* End <*/

  /* Clear the internal parameters */
  this->clear();

  ivs >> __use_tvalue;

  bool is_pointer = false;

  Value* tmp = 0;
  while(!ivs.isEmpty()) {
    tmp = 0;
    /* Read the TValue-type that should be created */
    ivs >> is_pointer;

    if (__use_tvalue) {
      if (is_pointer) {
        /* create TValue wich manages a pointer to the object */
        tmp = SValueFactory::getInstance().createTPointerValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
      } else {
        /* create a TValue for the object */
        tmp = SValueFactory::getInstance().createTValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
      }
    } else {
      /* create a simple Value object, without TValue-Wrapper */
      tmp = SValueFactory::getInstance().createValue(ivs.getCurrentType(), ivs.getCurrentNamespace());
    }

    /* Begin> Use the created Value with the given ValueStream */
    T* value = dynamic_cast<T*>(tmp);
    if (value) {
      /* If the dynamic_cast was successfull, read the value */
      value->createFromStream(ivs);
      /* Insert the new value */
      push_back(value);
    } else {
      /* If the dynamic_cast fails, insert a NULL Pointer */
      this->push_back(0);
      ivs.nextElement();
      /* delete the tmp value */
      delete tmp;
    }
    /* End< */
  }

  ivs.goUp();
}


template<typename T>
void SequenceValue<T>::toStream(OValueStream& ovs)  const{
  /* Write the TypeInfo */
  ovs.beginType(getTypeName(), getNamespace());

  ovs << __use_tvalue;
  for (unsigned int i = 0; i< this->size(); ++i) {
    if ((*this)[i])  {
      /* Write the pointer type */
      ovs << (*this)[i]->isPointerType();
      /* write the Value */
      (*this)[i]->toStream(ovs);
    } else {
      ovs << false; //No Pointer type
      /* Write the TypeInfo for a NULL-Pointer */
      ovs.beginType("NULL", "std");
      ovs.endType();
    }
  }
  ovs.endType();
}

template<typename T>
SequenceValue<T>* SequenceValue<T>::copy()            const{
  if (getMemoryStrategy() != Value::INVALUE)
    return const_cast<SequenceValue<T>*>(this);

  return new SequenceValue<T>(*this);
}

template<typename T>
void SequenceValue<T>::releaseValue() const {
  if (getMemoryStrategy() == Value::INVALUE) {
    for (unsigned int i = 0; i< this->size(); ++i) {
      if ( (*this)[i] )
        (*this)[i]->releaseValue();
    }
  }
}



/*Additional ValueStream operators. These stl-structs can be used as TValue's */
NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<string>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<string>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<float>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<float>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<double>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<double>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<u_int32_t>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<u_int32_t>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<u_int16_t>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<u_int16_t>& string_list);

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<int32_t>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<int32_t>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<int16_t>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<int16_t>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<long>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<long>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const list<unsigned long>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov,list<unsigned long>& string_list) ;


NMM::OValueStream& operator<<(NMM::OValueStream& ov, const pair<int32_t, int32_t>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, pair<int32_t, int32_t>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const pair<u_int64_t, u_int64_t>& string_list) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, pair<u_int64_t, u_int64_t>& string_list) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<string, string>& string_map) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<string, string>& string_map) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const map<string, int>& string_int_map) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, map<string, int>& string_int_map) ;

NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<bool>& bool_vector) ;
NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<bool>& bool_vector) ;


}//end namespace


template <typename T>
NMM::OValueStream& writeList(NMM::OValueStream& ov, const list<T>& v_list, const string type_name) {
  typedef typename list<T>::const_iterator ConstIt; //must be declared as a type
  /* Write the TypInfo */
  ov.beginType("list_"+type_name, "std");

  /* Write all elements of the list */
  for (ConstIt it = v_list.begin(); it !=  v_list.end(); ++it)
    ov << *it;

  ov.endType();
  return ov;
}


template <class T>
NMM::IValueStream& readList(NMM::IValueStream& iv, list<T>& v_list, const string& type_name) {
  /* Begin> Type checking */
  if (iv.getCurrentType() != "list_"+type_name) {
    GLOBAL_ERROR_STREAM("List> Invalid type assignment" << endl);
    GLOBAL_ERROR_STREAM ("List> Current type : " << iv.getCurrentType() << "                Expected type: " << "list_"+type_name << endl);
    GLOBAL_ERROR_STREAM ("List> Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: " << "std" << endl);

    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> Access internal attributes */
  if (!iv.goDown()) {
    GLOBAL_DEBUG_STREAM("List seems to be empty" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> While the Stream contains elements, read them */
  while(!iv.isEmpty()) {
    T tmp;
    iv >> tmp;
    v_list.push_back(tmp);
  }
  /* End< */

  iv.goUp();

  return iv;
}


template <typename T>
NMM::OValueStream& writeVector(NMM::OValueStream& ov, const vector<T>& v_vector, const string type_name) {
  typedef typename vector<T>::const_iterator ConstIt; //must be declared as a type
  /* Write the TypInfo */
  ov.beginType("vector_"+type_name, "std");


  /* Write the size */
  ov << v_vector.size();

  /* Write all elements of the list */
  for (ConstIt it = v_vector.begin(); it !=  v_vector.end(); ++it)
    ov << *it;

  ov.endType();
  return ov;
}


template <class T>
NMM::IValueStream& readVector(NMM::IValueStream& iv, vector<T>& v_vector, const string& type_name) {
  /* Begin> Type checking */
  if (iv.getCurrentType() != "vector_"+type_name) {
    GLOBAL_ERROR_STREAM("List> Invalid type assignment" << endl);
    GLOBAL_ERROR_STREAM ("List> Current type : " << iv.getCurrentType() << "                Expected type: " << "vector_"+type_name << endl);
    GLOBAL_ERROR_STREAM ("List> Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: " << "std" << endl);

    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> Access internal attributes */
  if (!iv.goDown()) {
    GLOBAL_DEBUG_STREAM("List seems to be empty" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  size_t size;

  /* read the size of the vector */
  iv >> size;

  v_vector.resize(size);

  /* Begin> While the Stream contains elements, read them */
  for (size_t i = 0; i < size; ++i) {
    iv >> v_vector[i];
  }
  /* End< */

  if (!iv.isEmpty()) {
    GLOBAL_ERROR_STREAM("Unhandled elements from vector_" << type_name << " in IValueStream" << endl);
  }

  iv.goUp();

  return iv;
}


NMM::IValueStream& readBoolVector(NMM::IValueStream& iv, vector<bool>& v_vector, const string& type_name);


template< class T, class V >
NMM::OValueStream& writeMap(NMM::OValueStream& ov, const map<T, V>& value, const NMM::TypeInfo& first_type, const NMM::TypeInfo& second_type)
{
  typedef typename map<T, V>::const_iterator ConstIt; //must be declared as a type

  /* Write the TypeInfo */
  ov.beginType("map_"+first_type.t_name+"_"+second_type.t_name, "std");

  /* Begin> First write the key, then the value */
  for (ConstIt it = value.begin(); it != value.end(); ++it) {
    ov << it->first;
    ov << it->second;
  }
  /* End< */

  ov.endType();

  return ov;
}

template< class T, class V >
NMM::IValueStream& readMap(NMM::IValueStream& iv, map<T, V>& value, const NMM::TypeInfo& first_type, const NMM::TypeInfo& second_type)
{
  /* Begin> Type checking */
  if (iv.getCurrentType() != "map_"+first_type.t_name+"_"+second_type.t_name && iv.getCurrentNamespace() != "std") {
    GLOBAL_ERROR_STREAM("Map> Invalid type assignment" << endl);
    GLOBAL_ERROR_STREAM ("Map> Current type : " << iv.getCurrentType() << "                Expected type: " << "pair_"+first_type.t_name+"_"+second_type.t_name << endl);
    GLOBAL_ERROR_STREAM ("Map> Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: " << "std" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> Access internal attributes */
  if (!iv.goDown()) {
    GLOBAL_DEBUG_STREAM("Invalid Map" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> Read the key, then the value */
  while(!iv.isEmpty()) {
    pair<T, V> tmp_pair;
    iv >> tmp_pair.first;
    iv >> tmp_pair.second;
    value.insert(tmp_pair);
  }
  /* End< */

  iv.goUp();

  return iv;
}


template< class T, class V >
NMM::OValueStream& writePair(NMM::OValueStream& ov, const pair<T, V>& value, const string& first_type, const string& second_type) {
  /* Write the TypeInfo */
  ov.beginType("pair_"+first_type+"_"+second_type, "std");
  /* Write the first element */
  ov << value.first;
  /* Write the second elemetn */
  ov << value.second;

  ov.endType();
  return ov;
}

template< class T, class V >
NMM::IValueStream& readPair (NMM::IValueStream& iv, pair<T, V>& value, const string& first_type, const string& second_type) {
  /* Begin> Type checking */
  if (iv.getCurrentType() != "pair_"+first_type+"_"+second_type && iv.getCurrentNamespace() != "std") {
    GLOBAL_ERROR_STREAM("Pair> Invalid type assignment" << endl);
    GLOBAL_ERROR_STREAM ("Pair> Current type : " << iv.getCurrentType() << "                Expected type: " << "pair_"+first_type+"_"+second_type << endl);
    GLOBAL_ERROR_STREAM ("Pair> Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: " << "std" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Begin> Access the internal attributes */
  if (!iv.goDown()) {
    GLOBAL_DEBUG_STREAM("Invalid Pair" << endl);
    iv.nextElement();
    return iv;
  }
  /* End< */

  /* Read the first element */
  iv >> value.first;
  /* Read the second element */
  iv >> value.second;

  iv.goUp();

  return iv;
}



#endif // _NMM_VALUE_HPP_


