/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef NMM_NETSTREAM_HPP
#define NMM_NETSTREAM_HPP

#include "nmm/base/serialize/ValueStream.hpp"

namespace NMM {

  class Sock;

  /**
   *  @brief Network stream
   *
   *  This class is an input and output stream which reads/writes data from/to a network socket.
   *  The actual socket read/write operation is implemented in subclasses.
   */
  class NetStream : public IStream, public OStream {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    NetStream(Sock& sock);
    
    /** Copy constructor
     */
    NetStream(const NetStream& other);
    
    /** Destructor
     */
    virtual ~NetStream();

    /** virtual Copy constructor
     */
    virtual NetStream* copy() const = 0;
    //@}

    /** @name Operator
     */
    //@{
    /** This method writes an u_int16 i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const u_int16_t&)      ;

    /** This method writes an int16 to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const int16_t&   )     ;

    /** This method writes an u_int32 to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const u_int32_t&)      ;

    /** This method writes an int32 to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const int32_t&   )     ;
    
    /** This method writes an u_int64 i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const u_int64_t& )     ;

    /** This method writes an int64 i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const int64_t& )       ;

    /** This method writes a bool i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const bool& )          ;

    /** This method writes an unsigned long i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const unsigned long&)  ;
    
    /** This method writes a long i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const long& )          ;

    /** This method writes a double i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const double& )        ;

    /** This method writes a float i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const float& )         ;

    /** This method writes a char i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const char& )          ;

    /** This method writes a unsigned char i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const unsigned char&)  ;

    /** This method writes a string i to the stream and returns a reference to the stream.
     */
    virtual OStream& operator<<(const string&)         ;
    //@}

    /** @name Write char-array
     */
    //@{
    /** This method writes an char buffer of given size to the stream.
     * \param buffer Poitner to the char array.
     * \param size Size of the buffer.
     */
    virtual void writeBuffer(const char* buffer, unsigned int size);

    /** This method writes an unsigned char buffer of given size to the stream.
     * \param buffer Poitner to the unsigned char array.
     * \param size Size of the buffer.
     */
    virtual void writeBuffer(const unsigned char* buffer, unsigned int size);
    //@}

    /** @name Operator
     */
    //@{
    /** This method reads an u_int16_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(u_int16_t&)       ;

    /** This method reads an int16_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(int16_t&)         ;

    /** This method reads an u_int32_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(u_int32_t&  )     ;

    /** This method reads an int32_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(int32_t&  )       ;
    
    /** This method reads an u_int64_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(u_int64_t&  )     ;

    /** This method reads an int64_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(int64_t&  )       ;

    /** This method reads a bool i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(bool&  )          ;

    /** This method reads an unsigned long i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(unsigned long&)   ;
    
    /** This method reads a long from the stream and returns a reference to the stream.     
     */
    virtual  IStream& operator>>(long& )           ;
    
    /** This method reads a double i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(double& )         ;

    /** This method reads a float i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(float& )          ;

    /** This method reads a char i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(char& )           ;

    /** This method reads an u_int16_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(unsigned char&)   ;

    /** This method reads an u_int16_t i from the stream and returns a reference to the stream.
     */
    virtual  IStream& operator>>(string&)          ;
    //@}

    /** @name Read char-array
     */
    //@{
    /** This method reads size chars and store they in the given buffer.
     * \param buffer The buffer to store the read char's. The minimum size of the array must be size
     * \param size The size of char's that should be read
     */
    virtual void readBuffer(char*, unsigned int size)           ;

    /** This method reads size unsigned chars and store they in the given buffer.
     * \param buffer The buffer to store the read unsigned char's. The minimum size of the array must be size
     * \param size The size of char's that should be read
     */
    virtual void readBuffer(unsigned char* buffer, unsigned int size);
    //@}

    /** @name NetStream specifc mehtods
     */
    //@{
    /** This method returns the Sock that is used to read/write the data
     */
    Sock& getSock() {
      return __sock;
    }

    /** This method returns the Sock that is used to read/write the data
     */
    const Sock& getSock() const {
      return __sock;
    }

    /** This method handles incoming data.
     * \return true if enough data was received so that readData can be called
     */
    virtual bool receivePacket();
    //@}

    /** @name Some method for network statistics
     */
    //@{
    /**
     * This method resets the sent-bytes counter
     */
    inline void resetReadCounter()                { __read_counter = 0; }

    /**
     * This method resets the received-bytes counter
     */
    inline void resetWriteCounter()               { __write_counter = 0; }
    
    /**
     * This method counts bytes that have been received
     */
    inline void countReadBytes(u_int32_t count)   { __read_counter += count; }
    
    /**
     * This method counts bytes that have been written
     */
    inline void countWriteBytes(u_int32_t count)  { __write_counter += count; }
    
    /**
     * This method reads the sent-bytes counter
     * \return number of bytes sent since last reset
     */
    inline u_int32_t getWriteCounter()   { return __write_counter; }

    /**
     * This method returns the received bytes counter
     */
    inline u_int32_t getReadCounter()    { return __read_counter; }
    //@}


  protected:
    /** Internal method to write the data to the network
     */
    virtual void writeValue(const void*, size_t) = 0;

    /** Internal method to read the data from the network
     */
    virtual void readValue(void*, size_t) = 0;

  private:
    ostream*      __stream;           ///< debug stream, to forward incoming data to another stream
    
    Sock&         __sock;             ///< socket to read to/write from
    
    u_int32_t     __write_counter;    ///< counter to count sent bytes
    u_int32_t     __read_counter;     ///< counter to count received bytes
  };
}

#endif





