/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "PacketNetStream.hpp"

#include "nmm/utils/net/Sock.hpp"
#include "nmm/utils/net/SockStream.hpp"

#include <errno.h>       // for errnos
#include <sys/socket.h>  // for inet_ntoa
#include <sys/types.h>
#include <fstream>


namespace NMM {

  PacketHeader::PacketHeader(int message_id, int packet_no, int packet_count, ssize_t size)
    : __message_id(message_id),
      __packet_no(packet_no),
      __packet_count(packet_count),
      __size(size)
  {}
  
  PacketNetStream::PacketNetStream(Sock& sock)
    : NetStream(sock),
      __max_packet_size(8192),
      __max_partial(10),
      __i_packet(__max_packet_size),
      __o_packet(__max_packet_size),
      __stream_socket(false),
      __i_message_pos(0),
      __o_message_pos(0),
      __i_message_id(0),
      __o_message_id(0),
      __total_sent_bytes(0),
      __total_received_bytes(0),
      __total_sent_packets(0),
      __total_received_packets(0),
      __total_discarded_packets(0),
      __total_lost_packets(0)
  {
    if(dynamic_cast<SockStream*>(&getSock())) {
      __stream_socket = true;
    }
  }
  
  PacketNetStream::PacketNetStream(const PacketNetStream& other)
    : NetStream(other),
      __max_packet_size(other.__max_packet_size),
      __max_partial(other.__max_partial),
      __i_packet(__max_packet_size),
      __o_packet(__max_packet_size),
      __stream_socket(false),
      __i_message_pos(0),
      __o_message_pos(0),
      __i_message_id(0),
      __o_message_id(0),
      __total_sent_bytes(0),
      __total_received_bytes(0),
      __total_sent_packets(0),
      __total_received_packets(0),
      __total_discarded_packets(0),
      __total_lost_packets(0)
  {
    if (dynamic_cast<SockStream*>(&getSock())) {
      __stream_socket = true;
    } 
  }
  
  PacketNetStream::~PacketNetStream() {}
  
  PacketNetStream* PacketNetStream::copy() const {
    return new PacketNetStream(*this);
  }
  
  void PacketNetStream::writeValue(const void* value, size_t size) {
    GLOBAL_DEBUG_STREAM2("PacketNetStream: writeValue size=" << size << endl);
    
    /* Check if buffer is large enough, enlarge it if needed */
    if (__o_message_pos + size > __o_message.size()) {
      ssize_t new_size = __o_message.size();
      if (!new_size) {
        new_size = 8;
      }
      while (new_size < static_cast<ssize_t>(__o_message_pos + size)) {
        new_size += new_size;
      }
      __o_message.resize(new_size);
      
      GLOBAL_DEBUG_STREAM2("PacketNetStream: resize output message boffer. size is now " << __o_message.size() << endl);
    }
    
    /* Write data to internal buffer */
    memcpy(&__o_message[__o_message_pos], value, size);
    __o_message_pos += size;
  }
  
  void PacketNetStream::readValue(void* value, size_t size) {
    GLOBAL_DEBUG_STREAM2("PacketNetStream: readValue size=" << size << endl);
    
    /* Check if size exceeds buffer size */
    if (__i_message_pos + size > __i_message.size()) {
      // XXX: fw: throw correct exception!!
      //throw Exception("Read buffer overflow!");
    }
    
    /* Read data from internal buffer */
    memcpy(value, &__i_message[__i_message_pos], size);
    __i_message_pos += size;
  }
  
  bool PacketNetStream::receivePacket() {
    MutexGuard g(__packet_mutex);
    
    /* Read packet from stream */
    ssize_t size = readPacket(&__i_packet[0]);
    if (!size) {
      throw ConnectionClosed(EPIPE);
    }
    
    /* Check if packet is big enough to contain a complete header */
    if (size < static_cast<ssize_t>(sizeof(PacketHeader))) {
      GLOBAL_ERROR_STREAM("PacketNetStream: Received packet is truncated (no valid header)" << endl);
      return false;
    }
    
    /* Analyze new packet header */
    PacketHeader& new_header = *reinterpret_cast<PacketHeader*>(&__i_packet[0]);
    int new_message_id = new_header.getMessageID();
    int new_packet_no = new_header.getPacketNumber();
    ssize_t new_packet_size = new_header.getPacketSize();
    
    /* Verify packet size versus header data */
    ssize_t correct_size = sizeof(PacketHeader) + new_packet_size;
    if (size != correct_size) {
      GLOBAL_ERROR_STREAM("PacketNetStream: Received packet is corrupt (size=" << size
                          << ", should be " << correct_size << ")" << endl);
      return false;
    }
    
    /* From this point on, assume packet is consistent */
    ++__total_received_packets;
    GLOBAL_DEBUG_STREAM2("PacketNetStream: receivePacket"
                         << " message_id=" << new_message_id
                         << " packet_no=" << new_packet_no
                         << " packet_count=" << new_header.getPacketCount()
                         << " size=" << new_header.getPacketSize()
                         << endl);
    
    /* Insert the packet into the queue */
    /* The list is sorted by message ID and packet number */
    list<Packet>::iterator i;
    for (i=__i_packets.begin(); i != __i_packets.end(); ++i) {
      /* Analyze existing packet header */
      Packet& packet = *i;
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      int message_id = header.getMessageID();
      int packet_no = header.getPacketNumber();
      
      if (message_id == new_message_id) {
        if (packet_no == new_packet_no) {
          /* Packet received multipl times. ignore it */
          continue;
        } else if (packet_no < new_packet_no) {
          continue;
        }
      } else if (message_id < new_message_id) {
        continue;
      }
      
      break;
    }
    
    /* Insert before this packet */
    __i_packets.insert(i, __i_packet);
    
    /* Discard old messages */
    discardOldPackets();
    
    /* Check if we have a complete message now */
    return checkForCompleteMessage();
  }
  
  void PacketNetStream::writeData() {
    GLOBAL_DEBUG_STREAM2("PacketNetStream: writeData" << endl);
    
    ssize_t size = __o_message_pos;
    int packet_no = 0;
    int write_pos = 0;
    
    __o_message_pos = 0;
    
    /* Calculate minimum number of packets needed */
    ssize_t max_bytes_per_packet = __max_packet_size - sizeof(PacketHeader);
    int packet_count = (size + max_bytes_per_packet-1) / max_bytes_per_packet;
    
    /* Split the message into packets and send them */
    while (size > 0) {
      /* Calculate number of bytes to put in this packet */
      ssize_t packet_bytes = max_bytes_per_packet;
      if (packet_bytes > size) {
        packet_bytes = size;
      }
      
      /* Write packet header */
      MutexGuard g(__packet_mutex);
      PacketHeader* header = reinterpret_cast<PacketHeader*>(&__o_packet[0]);
      *header = PacketHeader(__o_message_id, packet_no, packet_count, packet_bytes);
      
      GLOBAL_DEBUG_STREAM2("PacketNetStream: sendPacket"
                           << " message_id=" << __o_message_id
                           << " packet_no=" << packet_no
                           << " packet_count=" << packet_count
                           << " size=" << packet_bytes
                           << endl);
      
      /* Write packet data */
      memcpy(&__o_packet[sizeof(PacketHeader)], &__o_message[write_pos], packet_bytes);
      
      write_pos += packet_bytes;
      size -= packet_bytes;
      ++packet_no;
      
      /* Send packet */
      writePacket(&__o_packet[0]);
    }
    
    ++__o_message_id;
  }
  
  void PacketNetStream::readData() {
    int read_pos = 0;
    int packet_no = 0;
    int packet_count;

    GLOBAL_DEBUG_STREAM2("PacketNetStream: readData" << endl);
    
    if (__i_message_id < 0) {
      // XXX
      //throw Exception("Don't have a complete message!");
      GLOBAL_ERROR_STREAM("PacketNetStream::writeData() called, but no complete message available!");
      return;
    }
    
    /* Put all packets of the current input message together */
    /* Assume that the packets of this message are stored in the right order */
    for (list<Packet>::iterator i=__i_packets.begin(); i != __i_packets.end(); ) {
      Packet& packet = *i;
      
      /* Read packet header and check message ID */
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      if (header.getMessageID() != __i_message_id) {
        ++i;
        continue;
      }
      ssize_t packet_bytes = header.getPacketSize();
      packet_count = header.getPacketCount();
      if (packet_no != header.getPacketNumber()) {
        GLOBAL_ERROR_STREAM2("PacketNetStream: readData: WRONG PACKET_NO!!!" << endl);
        // XXX: something's wrong!
        //throw Exception("Packet not stored in correct order! SHOULD NOT HAPPEN!");
      }
      
      GLOBAL_DEBUG_STREAM2("PacketNetStream: add packet " << packet_no << endl);
      
      /* Enlarge message input buffer if necessary */
      if (static_cast<ssize_t>(read_pos + packet_bytes) > static_cast<ssize_t>(__i_message.size())) {
        ssize_t new_size = __i_message.size();
        if (!new_size) {
          new_size = 8;
        }
        while (new_size < static_cast<ssize_t>(read_pos + packet_bytes)) {
          new_size += new_size;
        }
        __i_message.resize(new_size);
        
        GLOBAL_DEBUG_STREAM2("PacketNetStream: resize input message boffer. size is now " << __i_message.size() << endl);
      }
      
      /* Read packet data */
      memcpy(&__i_message[read_pos], &packet[sizeof(PacketHeader)], packet_bytes);
      
      read_pos += packet_bytes;
      ++packet_no;
      
      /* Remove packet from list */
      i = __i_packets.erase(i);
      
      /* Check if done */
      if (packet_no >= packet_count) {
        break;
      }
    }
    
    __i_message_pos = 0;
    ++__i_message_id;
  }
  
  void PacketNetStream::writePacket(const void* packet) {
    const char* data = static_cast<const char*>(packet);
    ssize_t size = static_cast<const PacketHeader*>(packet)->getPacketSize() + sizeof(PacketHeader);
    size = getSock().send(data, size, 0);
    
    __total_sent_bytes += size;
    ++__total_sent_packets;
    
    GLOBAL_DEBUG_STREAM2("PacketNetStream: writePacket raw_size=" << size << endl);
  }
  
  ssize_t PacketNetStream::readPacket(void* packet) {
    ssize_t size;
    if (__stream_socket) {
      getSock().readStream(static_cast<char*>(packet), sizeof(PacketHeader), sizeof(PacketHeader));
      size = static_cast<PacketHeader*>(packet)->getPacketSize() + sizeof(PacketHeader);
      getSock().readStream(static_cast<char*>(packet) + sizeof(PacketHeader), size - sizeof(PacketHeader), size - sizeof(PacketHeader));
    } else {
      size = getSock().recv(static_cast<char*>(packet), __max_packet_size, 0);
      if (!size) {
        return 0;
      }
    }
    
    __total_received_bytes += size;
    
    GLOBAL_DEBUG_STREAM2("PacketNetStream: readPacket raw_size=" << size << endl);
    return size;
  }
  
  bool PacketNetStream::checkForCompleteMessage() {
    int curr_message_id = -1;
    int curr_packet_no = -1;

    GLOBAL_DEBUG_STREAM2("PacketNetStream: check for complete message" << endl);
    
    for (list<Packet>::iterator i=__i_packets.begin(); i != __i_packets.end(); ++i) {
      /* Analyze existing packet header */
      Packet& packet = *i;
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      int message_id = header.getMessageID();
      int packet_count = header.getPacketCount();
      
      if (message_id > curr_message_id) {
        /* New message. Restart the packet counter */
        curr_message_id = message_id;
        curr_packet_no = 0;
      }
      
      if (curr_packet_no == packet_count-1) {
        /* We have all packets of this message */
        __i_message_id = message_id;
        return true;
      }
      
      ++curr_packet_no;
    }
    
    /* No complete message found */
    __i_message_id = -1;
    return false;
  }
  
  void PacketNetStream::discardOldPackets() {
    int old_message_id = -1;
    int last_packet_count = 0;
    int packets_found = 0;
    
    /* This method discards all packets with a message ID smaller than __i_message_id - 10.
     * This eventually limits the size of the incoming packet queue.
     *
     * It also counts the number of lost packets. A packet is considered lost, if all packets with
     * the same message ID have been discarded by this method and the packet in question has not been received.
     * In other words, this method discards all packets in the queue whose message ID is too old and counts
     * the number of discarded packets. The difference between the number of discarded packets and
     * the number of packets that make up the discarded messages is the estimated packet lost.
     * It is an estimation because the packets considered lost might still arrive, though this is
     * rather unlikely. If we store "waterline" of discarded message fragments, we can also count these
     * packets.
     */
    while (__i_packets.begin() != __i_packets.end()) {
      Packet& packet = *__i_packets.begin();
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      int message_id = header.getMessageID();
      int packet_count = header.getPacketCount();
      
      /* Check if we are done */
      if (message_id >= __i_message_id - __max_partial) {
        break;
      }
      
      /* Update statistics */
      if (old_message_id < message_id) {
        __total_discarded_packets += packets_found;
        __total_lost_packets += last_packet_count - packets_found;
        packets_found = 0;
        last_packet_count = packet_count;
        old_message_id = message_id;
      }
      ++packets_found;
      
      __i_packets.pop_front();
    }
  }
  
  void PacketNetStream::setPacketSize(const ssize_t& size) {
    MutexGuard g(__packet_mutex);
    if (size == __max_packet_size) {
      return;
    }
    
    __max_packet_size = size;
    __i_packet.resize(size);
    __o_packet.resize(size);
  }
  
  ssize_t PacketNetStream::getPacketSize() const {
    MutexGuard g(__packet_mutex);
    return __max_packet_size;
  }
  
  void PacketNetStream::setMaxPartialMessages(const unsigned int& max_partial) {
    MutexGuard g(__packet_queue_mutex);
    __max_partial = max_partial;
  }
  
  unsigned int PacketNetStream::getMaxPartialMessages() const {
    MutexGuard g(__packet_queue_mutex);
    return __max_partial;
  }
  
  void PacketNetStream::resetStatistics() {
    __total_sent_bytes = 0;
    __total_received_bytes = 0;
    __total_sent_packets = 0;
    __total_received_packets = 0;
    __total_discarded_packets = 0;
    __total_lost_packets = 0;
  }
  
  ssize_t PacketNetStream::getTotalSentBytes() const {
    return __total_sent_bytes;
  }
  
  ssize_t PacketNetStream::getTotalReceivedBytes() const {
    return __total_received_bytes;
  }
  
  unsigned int PacketNetStream::getTotalSentPackets() const {
    return __total_sent_packets;
  }
  
  unsigned int PacketNetStream::getTotalReceivedPackets() const {
    return __total_received_packets;
  }
  
  unsigned int PacketNetStream::getTotalDiscardedPackets() const {
    return __total_discarded_packets;
  }
  
  unsigned int PacketNetStream::getTotalLostPackets() const {
    return __total_lost_packets;
  }
  
}
