/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_GENERIC_SYNC_SINK_NODE_HPP_
#define _NMM_GENERIC_SYNC_SINK_NODE_HPP_


#include "nmm/base/GenericSinkNode.hpp"
#include "nmm/interfaces/base/sync/ISynchronizedSinkImpl.hpp"
#include "nmm/interfaces/general/progress/ITimeProgressImpl.hpp"
#include "nmm/interfaces/general/progress/IProgressListenerImpl.hpp"
#include "nmm/interfaces/general/ITrackDurationImpl.hpp"
#include "UserTime.hpp"

namespace NMM {

  class SinkController;
    
  //! This class is a superclass for synchronized sink nodes.
  /*!
   *  Synchronization can be switched on and off by the setSynchronized -
   *  method. By default the switch is turned off, so an application 
   *  that don't uses synchronization should not care about it.
   *  In synchronized mode, be careful that incoming buffers are 
   *  timestamped: The GenericSyncSinkNode tries to present them according
   *  to their sync_time in the timestamps. 
   */
  class GenericSyncSinkNode :
    public GenericSinkNode,
    public ISynchronizedSinkImpl,
    public ITimeProgressImpl,
    public IProgressListenerImpl,
    public ITrackDurationImpl	   
  {
  protected:
    bool synchronized;           //!< Internal flag.
                                 //!< (True iff the node should be synchronized.)
    bool have_prepare_method;    //!< Internal flag.
                                 //!< (True iff the prepareBuffer - method is reinplemented.)
    UserTime media_time;         //!< The actual media_time.

    Time first_time;
    Interval last_time;
    bool first_buffer;
    bool send_progress_update;
    Buffer* saved_buffer;
    Interval progress_interval;
    u_int64_t track_duration;
    bool update_seek_time;
    u_int64_t seek_time;    
    
    SinkController* controller;

    // sync statistics
    int total_buffers;
    int total_buffers_discarded;
    Interval total_wait_time;
    Time start_time;
    
  public:
    //! The main constructor.
    GenericSyncSinkNode(const char* name,
			const StreamQueue::Mode queue_mode 
			 = StreamQueue::DEFAULT_MODE,
			const int size = StreamQueue::DEFAULT_SIZE);

    //! The main destructor.
    virtual ~GenericSyncSinkNode();

    /**
     * With this method synchronization of the sync element can be switched on and off.
     * \param toggle true  if the object should be synchronized
     *               false if the object simply presents the buffers as
     *                     fast as possible (default mode).
     */

    virtual Result setSynchronized(const bool& toggle);

    /**
     * Get the status of the object, synchronized, or not synchronized
     */
    virtual bool isSynchronized();

   
    //! Returns a pointer to the controller.
    ISinkController* getController();
    
    //! This method performs necessary preparations for the presentation of 
    //! a buffer.
    /*!
     *  This method is called by the processBuffer-method after 
     *  dequeueing a new buffer. This buffer should be prepared for 
     *  presentation here. 
     *  There are some nodes that don't have to reimplement this method,
     *  because their presentation contains only one step (like the
     *  PlaybackNode...)
     */
    virtual void prepareBuffer(Buffer*);

    //! This method finally presents the data buffer.
    /*!
     *  This method is called by the processBuffer-method when the time
     *  for the presentation of the buffer is reached. So this method 
     *  only should contain the last step of the presentation, not
     *  a lot of time-critical computations. 
     *  This method must be reimplemented!
     */
    virtual Message* presentBuffer(Buffer*) = 0;

    //! The processBuffer - method from GenericProcessorNode is reimplemented here.
    /*!
     *  It calls the prepareBuffer - method at first, if prepareBuffer is 
     *  reimplemented. Then it takes a look at the clock and compares the
     *  actual time with the presentation time of the buffer: If there's 
     *  still some time until the buffer should be presented, the node's
     *  thread is suspended for this time. 
     *  Then the presentBuffer - method is called. 
     *  \return NULL - a sink should not send any buffers.
     */
    virtual Message* processBuffer(Buffer *in_buffer);

    //! Returns the media time of the last presented buffer.
    inline UserTime getMediaTime() 
    {
      return media_time;
    };
    
    inline Result setSyncOffset(const Rational& /*velocity*/, const Interval& /*offset*/) {
      return SUCCESS;
    }
    
    //! Event - handling method for sync_reset.
    /*!
     *  All internal parameters for synchronization are reset.
     *  (The reset - method of the controller is called. This will have an effect on 
     *  the controller. If a synchronizer is connected to the controller, it will also
     *  be reset.)
     */
    virtual Result syncReset();

    virtual void setProgressUpdateInterval(const Interval& update_interval);

    virtual void setProgressUpdateEnabled(const bool& enable_update);

    void setProgress(const u_int64_t& numerator, const u_int64_t& denominator);

    virtual Result trackDuration(const Interval& i);
    
    Result pause();

    Result wakeup();
    
  };
}

#endif // _NMM_GENERIC_SYNC_SINK_NODE_HPP_
