/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Florian Winter <fw@cs.uni-sb.de>
 *                    Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_MULTI_AUDIO_VIDEO_SYNCHRONIZER_HPP_
#define _NMM_MULTI_AUDIO_VIDEO_SYNCHRONIZER_HPP_

#include "Synchronizer.hpp"
#include "nmm/interfaces/base/sync/IMultiAudioVideoSynchronizerImpl.hpp"

namespace NMM {
  class GenericSyncSinkNode;

  /**
   * Synchronizer for multiple audio and video sinks.
   * This class synchronizes an arbitrary number of audio sinks with an arbitrary number of video sinks
   * to a single (master) audio sink
   */
  class MultiAudioVideoSynchronizer :
    public SinkSynchronizer,
    virtual public IMultiAudioVideoSynchronizerImpl
  {
  private:
    
    /**
     * Internal sink.
     * Internal data structure that manages a sink.
     */
    struct InternalSink {
    private:
      ISinkController* controller;
      Interval average_latency;
      Interval skew;
      bool wait_for_reset;
      
    public:
      InternalSink();
      ~InternalSink();
      
      /** Set sink controller */
      void setController(ISinkController* controller);
      
      /** Get sink controller */
      ISinkController* getController() const;
      
      /** Get average latency */
      Interval getAverageLatency() const;
      
      /** Set average latency */
      void setAverageLatency(const Interval& average_latency);
      
      /** Get skew relative to master */
      Interval getSkew() const;
      
      /** Set skew relative to master */
      void setSkew(const Interval& skew);
      
      /** Set wait for reset flag */
      void setWaitForReset(bool value);

      /** Get wait for reset flag */
      bool getWaitForReset() const;
    };
    
    typedef map<int, InternalSink> sinks_t;
    
    
    /** The audio sinks */
    sinks_t audio_sinks;
    
    /** The video sinks */
    sinks_t video_sinks;
    
    /** Master audio sink ID. */
    int master;
    
    /** Specifies whether average latency of master is already known */
    bool master_average_latency_available;
    
    /** Theoretival latency. */
    Interval theo_latency;
    
    /** Security summand for the output nodes. */
    Interval security_interval;
    
    /** The time of the last pause call. */
    Time pause_start;
    
    /** Total number of sinks */
    int sink_count;
    
    /** Number of sinks waiting for reset */
    int reset_sink_count;
    
  protected:
    
    /** The states of the synchronizer. */
    enum State {
      /** The default state. */
      WORKING,   
      
      /**
       * State during a reset.
       * The Synchronizer is in this state from the event sync_reset 
       * at the first sink to the event at the second sink.
       */
      RESET,
      
      /** Paused */
      PAUSE
    };
    
    /** Current state */
    State state;
    
    /** Mutex to gurantee a consistent state. */
    ThreadMutex state_mutex;

    /** Mutex to guarantee consistent sink maps. */
    ThreadMutex sinks_mutex;
    
    /** Thread condition for pause */
    ThreadCondition pause_cond;
    
    
    /**
     * Compute new theoretical latency.
     * This computes the new theoretical latency based on the average latencies of all sinks.
     * \return true if latency changed, false if not
     */
    bool computeTheoLatency();
    
    /**
     * Tell the sink controllers the new theoretical latency.
     * This calls setTheoLatency on all sink controllers.
     */
    void sendTheoLatency();
    
    /**
     * Resolve internal sink ID.
     * The internal ID of a sink is the audio/video sink ID shifted left by 1 and the lowest bit
     * set if the sink is an audio sink. It is needed for the methods from the base interfaces that
     * only accept a sink ID to identify both audio and video sinks.
     * @param internal_id the internal ID that identifies the sink
     * @param id the actual audio or video sink ID
     * @param is_audio set to true if internal_id specifies an audio sink
     */
    void resolveSinkID(int internal_id, int* id, bool* is_audio);
    
  public:
    /**
     * The main constructor.
     * After constructing a new AudioVideoSynchronizer, it has to be 
     * connected to at least one audio and video sink nodes with the methods
     * setAudioSink and setVideoSink.
     */
    MultiAudioVideoSynchronizer(const char* name = "Synchronizer");
    
    virtual ~MultiAudioVideoSynchronizer();
    
  protected:
    /*---- internal methods -----*/
    /**
     * Set average audio latency.
     * This method is used by the sink controllers and should not be called 
     * directly.
     * With this method, the sink controllers tell the
     * synchronizer the average latency of their incoming buffers.
     * @param id sink ID
     * @param average_latency The average_latency of incoming buffers computed
     *                        by the Controller.
     */
    virtual Result setAverageAudioLatency(const int& id, const Interval& average_latency);
    
    /**
     * Set average video latency.
     * This method is used by the sink controllers and should not be called 
     * directly.
     * With this method, the sink controllers tell the
     * synchronizer the average latency of their incoming buffers.
     * @param id sink ID
     * @param average_latency The average_latency of incoming buffers computed
     *                        by the Controller.
     */
    virtual Result setAverageVideoLatency(const int& id, const Interval& average_latency);

    /*---- methods from IMultiAudioVideoSynchronizer -----*/
    /**
     * Set an audio sink controller.
     * This sets the controller for an audio sink.
     * If the sink id doesn't exist, a new sink is added.
     * @param id audio sink id
     * @param controller controller for this sink
     */
    virtual Result setAudioSinkController(const int& id, const ISinkController* controller);
    
    /**
     * Adds an audio sink controller.
     * This adds the controller for an audio sink.
     * A new id is returned.
     * @param controller controller for this sink
     * @return id audio sink id
     */
    virtual int addAudioSinkController(const ISinkController* controller);

    /**
     * Set an video sink controller.
     * This sets the controller for an video sink.
     * If the sink id doesn't exist, a new sink is added.
     * @param id audio sink id
     * @param controller controller for this sink
     */
    virtual Result setVideoSinkController(const int& id, const ISinkController* controller);
    
    /**
     * Adds an video sink controller.
     * This adds the controller for an video sink.
     * A new id is returned.
     * @param controller controller for this sink
     * @return id video sink id
     */
    virtual int addVideoSinkController(const ISinkController* controller);

    /**
     * Remove an audio sink.
     * Deletes the audio sink with the specified id.
     * @param id sink id
     */
    virtual Result removeAudioSink(const int& id);
    
    /**
     * Remove a video sink.
     * Deletes the video sink with the specified id.
     * @param id sink id
     */
    virtual Result removeVideoSink(const int& id);
    
    /**
     * Remove a sink.
     * Deletes the sink with the specified interface.
     * @param controller sink controller
     */
    virtual Result removeSinkController(const ISinkController* controller);

    /**
     * Set master audio sink.
     * Selects an audio sink as master audio sink. If this method is not called, the first
     * added audio sink becomes the master.
     * @param id audio sink id
     */
    virtual Result setMasterAudioSink(const int& id);
    
    /**
     * Set skew of an audio sink.
     * Sets the skew of an audio sink relative to the master audio sink.
     * @param id sink id
     * @param skew skew of this sink relative to master
     */
    virtual Result setAudioSkew(const int& id, const Interval& skew);
    
    /**
     * Set skew of a video sink.
     * Sets the skew of a video sink relative to the master audio sink.
     * @param id sink id
     * @param skew skew of this sink relative to master
     */
    virtual Result setVideoSkew(const int& id, const Interval& skew);
    
    /*---- methods from ISinkSynchronizer -----*/
    virtual void setAverageLatency(const Interval& value, const int& sink);
    
    virtual Result setFirstBufferInTimeAverageLatency(const Interval& value, const int& sink, const Timestamp& first_ts);

    /*---- methods from ISynchronizer -----*/
    virtual Result pause();
    
    virtual Result wakeup();
    
    /**
     * Resets the Synchronizer.
     * This method is used by the sink controllers and should not be called 
     * directly.
     * Only internal variables are reset, you don't need to set the
     * audio and video sink after a reset call.
     * Switches the state from WORKING to RESET.
     * @param sink sink ID
     */
    virtual Result reset(const int& sink);
  };
}

#endif // _NMM_MULTI_AUDIO_VIDEO_SYNCHRONIZER_HPP_







