/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef _NMM_SYNCHRONIZER_HPP_
#define _NMM_SYNCHRONIZER_HPP_

#include "nmm/base/StdNamedObject.hpp"
#include "TimedElement.hpp"
#include "nmm/interfaces/base/sync/ISourceSynchronizerImpl.hpp"
#include "nmm/interfaces/base/sync/ISinkSynchronizerImpl.hpp"
#include "nmm/base/ReferenceCount.hpp"

namespace NMM {
  //! This class is the superclass for all synchronizer objects.
  /*!
   * Synchronizer objects are used to encapsulate the synchronization 
   * strategy. They build the connection between the controllers in 
   * a graph section. The synchronizer is the central point from which
   * the controllers are configured.
   */
  class Synchronizer :
    public StdNamedObject,
    public TimedElement,
    public ReferenceCount,
    virtual public ISynchronizerImpl
  {
  protected:

    virtual Result pause() = 0;

    virtual Result wakeup() = 0;
   /**
     * Resets the Synchronizer.
     * This method is used by the sink controllers and should not be called 
     * directly.
     * Only internal variables are reset, you don't need to set the
     * audio and video sink after a reset call.
     * Switches the state from WORKING to RESET.
     * @param sink sink ID
     */

    virtual Result reset(const int& sink) = 0;
  public:
    //! The main constructor.
    Synchronizer(const char* name = "Synchronizer") : StdNamedObject(name) { }
    
    //! The main destructor.
    virtual ~Synchronizer() { }
  };

  class SinkSynchronizer : 
    public Synchronizer, 
    virtual public ISinkSynchronizerImpl 
  {
  public:
    //! The main constructor.
    SinkSynchronizer(const char* name = "SinkSynchronizer") : Synchronizer(name) { }
    
    //! The main destructor.
    virtual ~SinkSynchronizer() { }

  protected:

    /**
     * Set average latency.
     * This method is used by the sink controllers and should not be called 
     * directly.
     * With this method, the sink controllers tell the
     * synchronizer the average latency of their incoming buffers.
     * @param average_latency The average_latency of incoming buffers computed
     *                        by the Controller.
     * @param sink sink ID
     */
    virtual void setAverageLatency(const Interval& value, const int& sink) = 0;
  };

  class SourceSynchronizer : 
    public Synchronizer, 
    virtual public ISourceSynchronizerImpl 
  {
  public:
    //! The main constructor.
    SourceSynchronizer(const char* name = "SourceSynchronizer") : Synchronizer(name) { }
    
    //! The main destructor.
    virtual ~SourceSynchronizer() { }

  protected:

    /**
     * Sets a skew between audio and video output.
     * The video output is presented later as the audio output.
     */
    virtual Result setMasterSkew(const Interval& skew) = 0;
  };

}

#endif // _NMM_SYNCHRONIZER_HPP_

