/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#ifndef _NMM_TYPES_HPP_
#define _NMM_TYPES_HPP_

#include "nmm/utils/NamedObject.hpp"
#include "Rational.hpp"
#include <sys/time.h> // for gettimeofday()

namespace NMM {

  //! One billion nanoseconds per second.
  const long int nano_per_sec = 1000000000L;  

  class Rational;
  class Interval;
  class IValueStream;
  class OValueStream;

  //-------------------------------- Time -----------------------------------

  //! This class represents a point in time.
  /*! 
   *  Time is represented with a precision of nanoseconds.
   *  A time value can range over about 68 years in the past and 
   *  in the future.
   *  Time is internally stored in two long int's (one for the seconds and
   *  one for the nanoseconds.) It is important to notice that the
   *  nanoseconds are always between 0 and 10^9 - 1, and that the value of the Interval
   *  must be calculated as seconds + nanoseconds/nano_per_sec. So, e.g. -0.2 sec
   *  is stored as -1 second, 800,000,000 nanoseconds.
   */
  class Time {
  public:
    long int sec;     //!< The seconds.
    long int nsec;    //!< The nanoseconds.

    //! The main constructor.
    /*! 
     *  \return A time object that is initialized to zero.
     */
    explicit Time();

    //! This constructor initializes the time object to a given value.
    /*! 
     *  \param nanoseconds The time value represented as nanoseconds.
     */
    explicit Time(const long int& nanoseconds);

    //! This constructor initializes the time object to a given value.
    /*!
     *  \param seconds The seconds of the new time object.
     *  \param nanoseconds The nanoseconds of the new time object.
     */
    explicit Time(const long int& seconds, const long int& nanoseconds);

    //! This constructor initializes the time object to a given value.
    /*!
     *  \param time The time value represented as a struct timeval.
     */
    Time(const struct timeval& time);

    //! This constructor initializes the time object to a given value.
    /*!
     *  \param time The time value represented as a rational number.
     *         (Where 1 represents one second.)
     */
    explicit Time(const Rational& time);

    //! Adds an interval to a time.
    Time& operator+=(const Interval& interval);

    //! Subtracts an interval from a time.
    Time& operator-=(const Interval& interval);

    //! Get Time in format: "sec.nsec", where nsec are properly filled with
    //! '0' up to 9 digits
    std::string getPrettyString() const;
  };

  //! Adds an interval to a time and returns the result as new time 
  //! object.
  Time operator+(const Time& a, const Interval& b);

  //! Subtracts an interval from a time and returns the result as 
  //! new time object.
  Time operator-(const Time& a, const Interval& b);

  //! Returns true iff the two times are equal.
  inline bool operator==(const Time& a, const Time& b)
  {
    return a.sec == b.sec && a.nsec == b.nsec;
  };

  //! Returns true iff the two times are not equal.
  inline bool operator!=(const Time& a, const Time& b)
  {
    return a.sec != b.sec || a.nsec != b.nsec;
  };

  //! Returns true iff a is less than b.
  inline bool operator<(const Time& a, const Time& b)
  {
    return (a.sec < b.sec || (a.sec == b.sec && a.nsec < b.nsec));
  };

  //! Returns true iff a is less than or equal to b.
  inline bool operator<=(const Time& a, const Time& b)
  {
    return (a.sec < b.sec || (a.sec == b.sec && a.nsec <= b.nsec));
  };

  //! Returns true iff a is greater than b.
  inline bool operator>(const Time& a, const Time& b)
  {
    return (a.sec > b.sec || (a.sec == b.sec && a.nsec > b.nsec));
  };

  //! Returns true iff a is greater than or equal to b.
  inline bool operator>=(const Time& a, const Time& b)
  {
    return (a.sec > b.sec || (a.sec == b.sec && a.nsec >= b.nsec));
  };

  //! Returns an output stream that represents the time value.
  /*! The output should be human readable, and the internal representation can be
   *  confusing when having negative values. E.g. -0.2 sec
   *  is stored as -1 second, 800,000,000 nanoseconds. The output operator
   *  calculates a human readable output from the internal representation.
   */
  std::ostream& operator<<(std::ostream& s, const Time& time);

  /*!
   *  This method must be implemented if Events with an Time parameter
   *  should be transported over a net connection.
   */
  OValueStream& operator<<(OValueStream& s, const Time&) ;

  /*!
   *  This method must be implemented if Events with a Time parameter
   *  should be transported over a net connection.
   */
  IValueStream& operator>>(IValueStream& s, Time&);

  //------------------------------ Interval ---------------------------------

  //! The DivisionByZero - Exception for the class Interval.
  class IntervalDivisionByZeroException {
  public:
    
    //! The main constructor.
    /*!
     *  Writes a short hint to GLOBAL_ERROR_STREAM2.
     */
    IntervalDivisionByZeroException();
  };

  //! This class represents a time interval.
  /*!
   *  A time interval is defined as the difference between two points in 
   *  time. 
   *  Like a time value, an interval is stored in two long ints, one for the 
   *  seconds and one for the nanoseconds. It is important to notice that the
   *  nanoseconds are always between 0 and 10^9 - 1, and that the value of the Interval
   *  must be calculated as seconds + nanoseconds/nano_per_sec. So, e.g. -0.2 sec
   *  is stored as -1 second, 800,000,000 nanoseconds.
   */
  class Interval {
  public:
    int64_t sec;      //!< The seconds.
    int64_t nsec;     //!< The nanoseconds.

    //! The main constructor.
    /*!
     *  \return An interval object that is initialized to zero.
     */
    Interval();

    //! This constructor initializes the interval object to a given value.
    /*!
     *  \param nanoseconds The interval value represented as nanoseconds.
     */
    Interval(const long int& nanoseconds);

    //! This constructor initializes the interval object to a given value.
    /*!
     *  \param seconds The seconds of the interval.
     *  \param nanoseconds The nanoseconds of the interval.
     */
    Interval(const long int& seconds, const long int& nanoseconds);

    //! This constructor initializes the interval object to a given value.
    /*!
     *  \param time The interval value represented as a struct timeval.
     */
    Interval(const struct timeval& time);

    //! This constructor initializes the interval object to a given value.
    /*!
     *  \param time The time value represented as a rational number.
     *         (Where 1 represents one second.)
     */
    Interval(const Rational& time);

    //! Adds an interval to another.
    Interval& operator+=(const Interval& interval);

    //! Subtracts an interval from another.
    Interval& operator-=(const Interval& interval);

    //! Multiplies an Interval by a float
    /*!
     *  \param a The multiplier.
     */
    Interval& operator*=(const float& a);

    //! Divides an Interval by an integer.
    /*!
     *  \param a The divisor. If a is zero, an IntervalDivisionByZeroException
     *  is thrown.
     */
    Interval& operator/=(const int& a)
      throw (const IntervalDivisionByZeroException&);

    operator float() const {
      float r = sec;
      r += ((double)nsec / (double) nano_per_sec);
      return r;
    }

    //! Get Interval in format: "sec.nsec", where nsec are properly filled with
    //! '0' up to 9 digits
    std::string getPrettyString() const;
  };

  //! Returns the negative interval value.
  Interval operator-(const Interval& interval);

  //! Adds an interval to another and returns the result as
  //! a new object.
  Interval operator+(const Interval& a, const Interval& b);

  //! Subtracts an interval from another and returns the result
  //! as a new object.
  Interval operator-(const Interval& a, const Interval& b);

  //! Returns true iff the two intervals are equal.
  inline bool operator==(const Interval& a, const Interval& b)
  {
    return a.sec == b.sec && a.nsec == b.nsec;
  };

  //! Returns true iff the two intervals are not equal.
  inline bool operator!=(const Interval& a, const Interval& b)
  {
    return a.sec != b.sec || a.nsec != b.nsec;
  };

  //! Returns true iff a is less than b.
  inline bool operator<(const Interval& a, const Interval& b)
  {
    return (a.sec < b.sec || (a.sec == b.sec && a.nsec < b.nsec));
  };

  //! Returns true iff a is less than ore equal to b.
  inline bool operator<=(const Interval& a, const Interval& b)
  {
    return (a.sec < b.sec || (a.sec == b.sec && a.nsec <= b.nsec));
  };

  //! Returns true iff a is greater than b.
  inline bool operator>(const Interval& a, const Interval& b)
  {
    return (a.sec > b.sec || (a.sec == b.sec && a.nsec > b.nsec));
  };

  //! Returns true iff a is greater than ore equal to b.
  inline bool operator>=(const Interval& a, const Interval& b)
  {
    return (a.sec > b.sec || (a.sec == b.sec && a.nsec >= b.nsec));
  };

  //! Returns an output stream that represents the interval value.
  /*! The output should be human readable, and the internal representation can be
   *  confusing when having negative values. E.g. -0.2 sec
   *  is stored as -1 second, 800,000,000 nanoseconds. The output operator
   *  calculates a human readable output from the internal representation.
   */
  std::ostream& operator<<(std::ostream& s, const Interval& interval);

  //! Returns an interval that represents the difference between the
  //! two points in time a and b.
  Interval operator-(const Time& a, const Time& b);

  //! Suspends a thread for at least the period which is represented by
  //! interval.
  /*!
   *  Note: The thread may be suspended longer than interval!
   */
  void suspendThread(const Interval& interval);

  /*!
   *  This method must be implemented if Events with an Interval parameter
   *  should be transported over a net connection.
   */
  OValueStream& operator<<(OValueStream& s, const Interval& interval);
 
  /*!
   *  This method must be implemented if Events with an Interval parameter
   *  should be transported over a net connection.
   */
  IValueStream& operator>>(IValueStream& s,Interval& interval);
}

#endif // _NMM_TYPES_HPP_
