/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_GRAPHURL_HPP
#define NMM_GRAPHURL_HPP NMM_GRAPHURL_HPP

#include <map>
#include <string>
#include <set>

using namespace std;

namespace NMM {
  class Format;
  class NodeDescription;
  class INode;
  class GraphURLObject;

  class GraphURL {
  public:
    //! enum of the different supported protocols
    enum URLProtocol { 
      //! if no protocol can be identified
      UNKNOWN = 0,
      //! Digitial Versatile Disc as special source
      DVD,
      //! AudioCD as special source
      AUDIOCD,
      //! Digital Video Broadcast Television as special source
      DVBTV,
      //! TVCard as special source
      TV,
      //! IVTV TVCard as special source
      IVTV,
      //! MPEG TVCard as special source
      MPEGTV,
      //! File source as protocol
      FILE,
      //! HyperText Transfer Protocol
      HTTP,    
      //! File Transfer Protocol
      FTP,     
      //! Server Message Block protocol, used for Windows(TM) file servers
      SMB,     
      //! File In SHell, ssh-encrypted remote file access, like kio_fish (http://ich.bin.kein.hoschi.de/fish/)
      FISH,    
      //! VNC, a remote Desktop access protocol
      VNC
    };
  private:

    typedef map<string, string> URLOptions;
    typedef URLOptions::const_iterator URLOptionsIt;

    typedef map<string, GraphURLObject*> GraphURLObjects;
    typedef GraphURLObjects::const_iterator GraphURLObjectsIt;

    string url;

    bool is_valid;

    URLProtocol protocol;
    string protocol_name;
    string user;
    string password;
    string host;
    unsigned int port;
    string uri;
    string uri_path;
    string uri_options;
    string uri_extension;

    URLOptions options;

    GraphURLObjects url_objects;

    void registerGraphURLObject(GraphURLObject* object);
    void registerGraphURLObjects();
    GraphURLObject* getGraphURLObject() const;

    bool parse();

    void addInterfacesFromProtocol(NodeDescription& n) const;

  public:

    GraphURL();

    GraphURL(const string);

    ~GraphURL();

    Format* getFormat() const;

    NodeDescription getNodeDescription() const;

    void configureConstructed(INode* node);
    void configureInitialized(INode* node);
    void configureOutputInitialized(INode* node);

    inline const bool isValid() const { return is_valid; }

    inline const string getURL() const { return url; }

    inline const URLProtocol getProtocol() const { return protocol; }

    inline const string getProtocolName() const { return protocol_name; }

    inline const string getUser() const { return user; }

    inline const string getPassword() const { return password; }

    inline const string getHost() const { return host; }

    inline const unsigned int getPort() const { return port; }

    inline const string getURI() const { return uri; }

    inline const string getURIPath() const { return uri_path; }

    inline const string getURIOptions() const { return uri_options; }

    inline const string getURIExtension() const { return uri_extension; }

    const string getOption(const string) const;

    const set<string> getOptionList() const;
    
    void setOption(const string& key, const string& value);
    
    void clearOptions();
    
    static string urldecode(const string);

    static string urlencode(const string);
    
    static string getMIMETypeFromFile(const string& filename, bool mime_only = true );
 
  };

  ostream& operator<<(ostream&, const NMM::GraphURL&);
}
#endif //NMM_GRAPHURL_HPP
