
#ifndef NMM_IConnectionManager_HPP
#define NMM_IConnectionManager_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "ITransportStrategy.hpp"

namespace NMM
{
  /**
   *  @brief Connection Manager interface
   *
   *  Connection managers are used to initiate and configure a new connection using an existing
   *  reliable connection.
   */

  class IConnectionManager : virtual public Interface { 
  public:
    IConnectionManager() {  setTypeInfo(TypeInfo(IConnectionManager::IConnectionManager_type, "NMM")); };

    IConnectionManager(NMMObject& parent) : Interface(parent) {  setTypeInfo(TypeInfo(IConnectionManager::IConnectionManager_type, "NMM")); setEventReceiver(parent); };

    virtual ~IConnectionManager() {};

    static const char* IConnectionManager_type;

    virtual  string getInterfaceType() const { return IConnectionManager_type; }

    Interface* copy() const { return new IConnectionManager(*this); }
    static const char* handleConnect_event;
    static const char* sendReliableEvent_event;
    static const char* sendReliableConfigEvent_event;
    static const char* sendReliableInstreamMessage_event;
    static const char* getSupportedStrategies_event;
    /**
     *  @brief Handle a connection request from a remote host
     *  @param remote_addr Fully specified address of the remote transport strategy
     *  @param local_addr Returns the fully specified address of the local transport strategy
     *  @exception BindingException if no %ConnectorAcceptor was found or the %ConenctorAcceptor throws a BindingException
     *
     *  This method checks if a %ConnectorAcceptor object exists for the requested transport strategy type
     *  and forwards the request to this Connector.
     */

    virtual void handleConnect(const TSAddress* remote_addr, TSAddress*& local_addr) const
      throw(const Exception&, const BindingException&);
    /**
     *  @brief Reliably send event to partner object
     *  @param ts_reference NMM reference of transport strategy which should receive the event
     *  @param e The event
     *  @return The result of the event
     *
     *  This method is used to reliably send a configuration event to a partner object via the %ConnectionManager.
     *  The connection manager pretends that the event was sent via the transport strategy.
     */

    virtual Result sendReliableEvent(const NMMReference& ts_reference, Event& e) const
      throw(const Exception&, const BindingException&);
    /**
     *  @brief Reliably send event to partner strategy
     *  @param receiver NMM Reference of transport strategy which should receive the event
     *  @param e The event
     *  @return The result of the event
     *
     *  This method is used to reliably send a configuration event to a partner strategy via the %ConnectionManager.
     */

    virtual Result sendReliableConfigEvent(const NMMReference& receiver, Event& e) const
      throw(const Exception&, const BindingException&);
    /**
     *  @brief Reliably send instream message to partner object
     *  @param receiver NMM Reference of transport strategy which should receive the event
     *  @param message The message
     *  @return The result of the event
     *
     *  This method is used to reliably send an instream to a partner object via the ConnectionManager.
     *  The connection manager pretends that the message was sent via the transport strategy.
     *
     *  This method is only for debugging and should not be used.
     */

    virtual Result sendReliableInstreamMessage(const NMMReference& receiver, CEvent& e) const
      throw(const Exception&, const BindingException&);
    /**
     *  @brief Get strategies that support format
     *  @param format Format to check
     *  @return List of transport strategies that support the format
     *
     *  This method returns the list of transport strategies that support the given format.
     */

    virtual list<string >  getSupportedStrategies(const Format& format) const
      throw(const Exception&);


  };

  typedef auto_ptr<IConnectionManager> IConnectionManager_var;




}

#endif
