
#ifndef NMM_IConnectionManagerImpl_HPP
#define NMM_IConnectionManagerImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IConnectionManager.hpp"

#include "ITransportStrategyImpl.hpp"

namespace NMM
{
  /**
   *  @brief Connection Manager interface
   *
   *  Connection managers are used to initiate and configure a new connection using an existing
   *  reliable connection.
   */

  class IConnectionManagerImpl  : public virtual NMMObject { 
  public:
    IConnectionManagerImpl() { addInterface(new IConnectionManager(*this)); registerEvents(); };

    virtual ~IConnectionManagerImpl() {};

  protected:
    /**
     *  @brief Handle a connection request from a remote host
     *  @param remote_addr Fully specified address of the remote transport strategy
     *  @param local_addr Returns the fully specified address of the local transport strategy
     *  @exception BindingException if no %ConnectorAcceptor was found or the %ConenctorAcceptor throws a BindingException
     *
     *  This method checks if a %ConnectorAcceptor object exists for the requested transport strategy type
     *  and forwards the request to this Connector.
     */

    virtual void handleConnect(const TSAddress* remote_addr, TSAddress*& local_addr) const 
      throw(const BindingException&) = 0;
    /**
     *  @brief Reliably send event to partner object
     *  @param ts_reference NMM reference of transport strategy which should receive the event
     *  @param e The event
     *  @return The result of the event
     *
     *  This method is used to reliably send a configuration event to a partner object via the %ConnectionManager.
     *  The connection manager pretends that the event was sent via the transport strategy.
     */

    virtual Result sendReliableEvent(const NMMReference& ts_reference, Event& e) const 
      throw(const BindingException&) = 0;
    /**
     *  @brief Reliably send event to partner strategy
     *  @param receiver NMM Reference of transport strategy which should receive the event
     *  @param e The event
     *  @return The result of the event
     *
     *  This method is used to reliably send a configuration event to a partner strategy via the %ConnectionManager.
     */

    virtual Result sendReliableConfigEvent(const NMMReference& receiver, Event& e) const 
      throw(const BindingException&) = 0;
    /**
     *  @brief Reliably send instream message to partner object
     *  @param receiver NMM Reference of transport strategy which should receive the event
     *  @param message The message
     *  @return The result of the event
     *
     *  This method is used to reliably send an instream to a partner object via the ConnectionManager.
     *  The connection manager pretends that the message was sent via the transport strategy.
     *
     *  This method is only for debugging and should not be used.
     */

    virtual Result sendReliableInstreamMessage(const NMMReference& receiver, CEvent& e) const 
      throw(const BindingException&) = 0;
    /**
     *  @brief Get strategies that support format
     *  @param format Format to check
     *  @return List of transport strategies that support the format
     *
     *  This method returns the list of transport strategies that support the given format.
     */

    virtual list<string >  getSupportedStrategies(const Format& format) const = 0;


  private:
    void registerEvents();
  };

}

#endif
