
#ifndef NMM_IMasterSlaveSourceSynchronizer_HPP
#define NMM_IMasterSlaveSourceSynchronizer_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/Interface.hpp"
#include "nmm/base/Event.hpp"
#include <memory>

#include "ISourceSynchronizer.hpp"

namespace NMM
{
  /**
   * Master-slave source synchronizer.
   * Interface for a synchronizer which performs inter-stream synchronization between
   * one master and multiple slave sources.
   */

  class IMasterSlaveSourceSynchronizer : virtual public ISourceSynchronizer { 
  public:
    IMasterSlaveSourceSynchronizer() {  setTypeInfo(TypeInfo(IMasterSlaveSourceSynchronizer::IMasterSlaveSourceSynchronizer_type, "NMM")); };

    IMasterSlaveSourceSynchronizer(NMMObject& parent) : ISourceSynchronizer(parent) {  setTypeInfo(TypeInfo(IMasterSlaveSourceSynchronizer::IMasterSlaveSourceSynchronizer_type, "NMM")); setEventReceiver(parent); };

    virtual ~IMasterSlaveSourceSynchronizer() {};

    static const char* IMasterSlaveSourceSynchronizer_type;

    virtual  string getInterfaceType() const { return IMasterSlaveSourceSynchronizer_type; }

    Interface* copy() const { return new IMasterSlaveSourceSynchronizer(*this); }
    static const char* setStartDelay_event;
    static const char* setMasterController_event;
    static const char* addSlaveController_event;
    static const char* removeSlaveController_event;
    static const char* resetAll_event;
    /**
     * Sets the time interval to wait before starting after start() is called.
     * This allows the MasterSlaveSourceSynchronizer to tell its controllers the
     * delay time.
     * With the help of a start delay, the nodes
     * are told to start at the current time + delay to avoid scheduling
     * problems (as one node would always be started some time after the other).
     * This is done by setting calling the setStartTime-Method of the controllers
     * with (current_time + delay).
     */

    virtual Result setStartDelay(const Interval& interval)
      throw(const Exception&);
    /**
     * Sets the master (audio) source node.
     * This allows the MasterSlaveSourceSynchronizer to connect to its master controller.
     */

    virtual Result setMasterController(const ISourceController* value)
      throw(const Exception&);
    /**
     * Adds a slave source node. (e.g. video)
     * This allows the MasterSlaveSourceSynchronizer to connect to its slave controller(s).
     */

    virtual Result addSlaveController(const ISourceController* value)
      throw(const Exception&);
    /**
     * Removes a slave source node. (e.g. video)
     */

    virtual Result removeSlaveController(const ISourceController* value)
      throw(const Exception&);
    /**
     * Resets the MasterSlaveSourceSynchronizer.
     * Only internal variables are resetted, you don't need to set the
     * audio and video source after a reset - call.
     * The reset-method is normally not called directly, but by the
     * ISourceControllers.
     * It switches the state between WORKING and RESET.
     */

    virtual Result resetAll()
      throw(const Exception&);


  };

  typedef auto_ptr<IMasterSlaveSourceSynchronizer> IMasterSlaveSourceSynchronizer_var;




}

#endif
