
#ifndef NMM_ISinkControllerImpl_HPP
#define NMM_ISinkControllerImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "ISinkController.hpp"

#include "IControllerImpl.hpp"

namespace NMM
{

  class ISinkControllerImpl  : virtual public IControllerImpl { 
  public:
    ISinkControllerImpl() { addInterface(new ISinkController(*this)); registerEvents(); };

    virtual ~ISinkControllerImpl() {};

  protected:
    /** Tells the controller the maximal delay.
     *
     * With this method you can tell the controller which delay of an
     * incoming buffer should be accepted.
     * In the isBufferValid - method, the controller computes a 
     * theoretical presentation time for this buffer. If the real time
     * in this moment is greater than this theoretical presentation time
     * plus the maximal delay, the buffer will be recognized as invalid.
     * @param max_delay The maximal delay that will be allowed.
     * @return SUCCESS if the delay could be set (if it is greater than
     *                 or equal zero),
     *         FAILURE otherwise.
     */

    virtual Result setMaxDelay(const Interval& value) = 0;
    /**
     * If the average latency exceeds this value, the synchronizer gets to know 
     * that.
     */

    virtual Result setMaxAverageDelay(const Interval& value) = 0;
    /**
     * Used for the communication with the synchronizer.
     *
     * This method is called by the synchronizer to tell the controllers the
     * theoretical latency of the pipeline.
     */

    virtual Result setTheoLatency(const Interval& value) = 0;
    /**
     * Sets the default theoretical latency.
     *
     * This method is called by the synchronizer to tell the controllers the
     * default theoretical latency of the pipeline.
     */

    virtual Result setDefaultTheoLatency(const Interval& value) = 0;
    /**
     * Sets the synchronizer which refers to the stream.
     * @param sync synchronizer
     * @param sinkid id of this controller at the synchronizer
     */

    virtual Result setSinkSynchronizer(const ISinkSynchronizer* value, const int& id) = 0;
    /**
     * Sets the number of buffers used to compute the average latency.
     * A small average_count means that the synchronizer is told the actual 
     * average_latency of incoming buffers very often. That means the
     * synchronization reacts very quickly (and this reaction may be visible)
     * For this reason higher values (such as 50 for audio and 20 for video) 
     * are recommended.
     */

    virtual Result setAverageCount(const int& value) = 0;
    /**
     * Determines the length of possible gaps if buffers arrive too late. 
     *
     * Sets the maximal number of buffers that can be discarded before at least
     * one buffer has to be presented.
     */

    virtual Result setMaxDiscardingCount(const int& value) = 0;
    /**
     * This method is called by the AudioVideoSynchronizer.
     *
     *  It changes the state 
     *          from WORKING to WAIT_FOR_RESET or
     *          from RESET   to WORKING.
     */

    virtual Result waitForReset() = 0;
    /**
     * Restarts the SinkController.
     * The state is changed from RESET to WORKING.
     */

    virtual Result restart() = 0;
    /**
     * Tells the node to wait for the time stamp to become valid
     * or to simply present the buffer when it arrives
     * @param enable true to enable waiting, false to disable
     */

    virtual Result enableWaitToPresent(const bool& enable) = 0;
    /**
     * Sets the minimum interval that a sink controller will
     * return when it is asked how long the node should wait
     * until it presents the current buffer (usually by the
     * method waitToPresent()). If the interval to wait is 
     * smaller than the interval set with this method, the
     * sink controller should return a zero interval.
     */

    virtual Result setMinWaitInterval(const Interval& min_wait) = 0;
    /**
     * Set whether the sink controller should notify its
     * sink synchronizer if the first buffer after a syncReset
     * arrived in time (in regard to the theoretical latency)
     * The notification is done by calling ISinkSynchronizer::setFirstBufferInTimeAverageLatency
     * @param enable true if the controller should notify 
     *               its synchronizer
     */

    virtual void enableFirstBufferInTime(const bool& enable) = 0;
    /**
     * Enables/disables the blocking of the controller (and
     * therefore the blocking of its parent node) when waiting
     * for restart() after having received a syncReset.
     * Blocking should be enabled per default.
     * @param enable true to enable blocking; else false
     */

    virtual void enableBlocking(const bool& enable) = 0;


  private:
    void registerEvents();
  };

}

#endif
