/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        chwellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef PROGRAMDESCRIPTION_HPP
#define PROGRAMDESCRIPTION_HPP PROGRAMDESCRIPTION_HPP

#include <iostream>
#include <list>
#include <string>

#include "nmm/base/serialize/ComplexValue.hpp"
#include "nmm/interfaces/misc/tvguide/ChannelDescription.hpp"
namespace NMM{

  /** \class ProgramDescription
   * This class represents a Programmeinformation for TVGuide. 
   */


  class ProgramDescription : public SerializedValue{
    friend ostream& operator << (ostream& os, const ProgramDescription& pd);
  public:

    enum sourceType_t {UNKNOWN=0, DVB, XMLTV, NXTVEPG};
    
  private:
    NMM::TValue<string> startTime;
    NMM::TValue<string> endTime;
    
    NMM::TValue<ChannelDescription> channel;
    
    NMM::TValue<int> duration; //in minutes

    NMM::TValue<list<string> > title;
    NMM::TValue<string> desc;
    NMM::TValue<list<string> >category;

    NMM::TValue<unsigned int> sourceType;

  public:
    ProgramDescription();
    ProgramDescription(sourceType_t source);
    ProgramDescription (const ProgramDescription&);
    ~ProgramDescription();

    /**
     *Test, if the ProgramDescirption is empty.
     *@return true if it is empty
     *@return false otherwise
     */
    bool empty();

    /**
     *Set the start-time of the program
     *@param string starttime
     */
    void setStartTime(string st);
    /**
     *Get the start-time of the program
     *@resturn string starttime
     */
    string getStartTime()const;

    /**
     *Set the end-time of the program
     *@param sintr endtime
     */
    void setEndTime(string et);
    /**
     *Get the end-time of the program
     *@resturn string endtime
     */
    string getEndTime() const;

    /**
     *Set the ChannelID 
     *@param channel-ID
     */
    void setChannelID(string id);
    /**
     *Get the ChannelID 
     *@return string  channel-ID
     */
    string getChannelID() const;

    /**
     *Set the ChannelDescription 
     *@param channel-description
     */
    void setChannelDescription(ChannelDescription cd);
    /**
     *Get the ChannelDescription 
     *@return ChannelDescription
     */
    ChannelDescription getChannelDescription() const;

    /**
     *Set the Duration of the program 
     *@param duration
     */
    void setDuration(unsigned int d);
    /**
     *Get the duration 
     *@return unsigned int duration
     */
    unsigned int getDuration() const;

    /**
     *Sets a list of titles for this program
     *@param list of titles
     */
    void setTitles(list<string> titles);
    /**
     *Add a single title to the titles-list
     *@param title
     */
    void addTitle(string t);
    /**
     *Gets the titles of this program
     *@return list<string> titles
     */
    list<string> getTitles() const;

    /**
     *Set the description of this program
     *@param description
     */
    void setDescription(string d);
    /**
     *Get the description of this program
     *@return description
     */
    string getDescription() const;

    /**
     *Adds a category for this program
     *@param category
     */
    void addCategory(string c);
    /**
     *Get the list of categories set for this program
     *@return list<string> categories
     */
    list<string> getCategory() const;

    /**
     *returns the SourceType specified on creation
     *@return string source-type
     */
    unsigned int getSourceType() const { return sourceType.getRValue();}

    /**
     *sets the SourceType
     *@param source-type
     */
    void setSourceType(unsigned int t) {sourceType = t;}

    /**
     *transforms the Start- and End-Time to localtime, if a timezone is given in the representation
     */
    void localizeTimes();

    //Interface SerialzedValue
    ProgramDescription* copy();

    bool operator<(const ProgramDescription &pd) const;
    bool operator==(const ProgramDescription &pd) const;
    
  private:
    string transformTime(string s);
    void registerValues();
  };

  OValueStream& operator<<(OValueStream& os, const list<ProgramDescription>& pd);
  IValueStream& operator>>(IValueStream& is, list<ProgramDescription>& pd);




  class ProgramDescriptionList : public SerializedValue{
  private:
    NMM::TValue<list<ProgramDescription> > internal_list;

  public:
    ProgramDescriptionList() : SerializedValue("ProgramDescriptionList", "NMM"){registerAttributes();}
    ProgramDescriptionList(const ProgramDescriptionList &pl): SerializedValue("ProgramDescriptionList", "NMM"){*this=pl; registerAttributes();}
    ProgramDescriptionList(list<ProgramDescription> l): SerializedValue("ProgramDescriptionList", "NMM"){internal_list = l;registerAttributes();}
  private:
    void registerAttributes(){insertArgument(&internal_list);}

  public:
    list<ProgramDescription> getList() const {return internal_list.getRValue();}
    unsigned int size() const {return internal_list.getRValue().size();}
    void pus_back(ProgramDescription pd){internal_list.getRValue().push_back(pd);}
    ProgramDescription back() const {return internal_list.getRValue().back();}
    void pop_back() {internal_list.getRValue().pop_back();}
  };

  OValueStream& operator<<(OValueStream& os, const list<ProgramDescriptionList>& pl);
  IValueStream& operator>>(IValueStream& is, list<ProgramDescriptionList>&pl);





}
#endif
