
#ifndef NMM_IRTPControlImpl_HPP
#define NMM_IRTPControlImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "IRTPControl.hpp"


namespace NMM
{
  /**
   *  @brief RTP Control interface.
   *
   *  This interface provides method for getting and setting options for the RTP protocol.
   */

  class IRTPControlImpl  : public virtual NMMObject { 
  public:
    IRTPControlImpl() { addInterface(new IRTPControl(*this)); registerEvents(); };

    virtual ~IRTPControlImpl() {};

  protected:
    /**
     *  @name Get/Set SDES Items
     */
//@{
    /**
     *  @brief Set canonical name
     *  @param cname The canonical name
     *
     *  This method sets the canonical name of this RTP session. This is a free-form string
     *  which identifies an RTP session. The default is "NMM-RTP".
     */

    virtual void setCNAME(const string& cname) = 0;
    /**
     *  @brief Get canonical name
     *  @return The canonical name
     */

    virtual string getCNAME() const = 0;
//@}
    /**
     *  @name Connection Parameters
     */
//@{
    /**
     *  @brief Set multicast TTL
     *  @param ttl The multicast TTL
     *
     *  This method sets the time to live (TTL) for multicast packets. For unicast sessions, this
     *  parameter has no effect. The integer number is the maximum number of hops until a packet
     *  is deleted.
     */

    virtual void setTTL(const unsigned int& ttl) = 0;
    /**
     *  @brief Get multicast TTL
     *  @return The multicast TTL
     */

    virtual unsigned int getTTL() const = 0;
    /**
     *  @brief Set buffer size
     *  @param buffer_size The buffer size
     *
     *  This method sets the size for NMM buffers allocated by the RTP receiver.
     */

    virtual void setBufferSize(const size_t& buffer_size) = 0;
    /**
     *  @brief Get buffer size
     *  @return The buffer size (in bytes)
     */

    virtual size_t getBufferSize() const = 0;
    /**
     *  @brief Set RTP socket receive buffer size
     *  @param recv_buffer_size Receive buffer size
     *
     *  This method sets the receive buffer size for receiving RTP sockets
     */

    virtual void setReceiveBufferSize(const size_t& recv_buffer_size) = 0;
    /**
     *  @brief Get RTP socket receive buffer size
     *  @return The receive buffer size
     */

    virtual size_t getReceiveBufferSize() const = 0;
    /**
     *  @brief Set maximum frame size
     *  @param max_frame_size The maximum frame size (in bytes)
     *
     *  This method sets the maximum size that a received RTP frame may have. If an incoming frame
     *  is larger than this size, it may be truncated.
     */

    virtual void setMaxFrameSize(const size_t& max_frame_size) = 0;
    /**
     *  @brief Get maximum frame size
     *  @return The maximum frame size (in bytes)
     */

    virtual size_t getMaxFrameSize() const = 0;
    /**
     *  @brief Set session bandwidth
     *  @param bandwidth The session bandwidth (in kb/s)
     *  @todo Check if this parameter is really in kb/s
     *
     *  This method sets the session bandwidth. Note that this is only an estimate which
     *  serves as a hint to the RTP library. The actual bandwidth is determined by the
     *  data which is transferred (for example the video bitrate).
     */

    virtual void setSessionBandwidth(const unsigned int& buffer_size) = 0;
    /**
     *  @brief Get session bandwidth
     *  @retun The session bandwidth (in kb/s)
     */

    virtual unsigned int getSessionBandwidth() const = 0;

//@}

  private:
    void registerEvents();
  };

}

#endif
