
#ifndef NMM_ISPUDecoderImpl_HPP
#define NMM_ISPUDecoderImpl_HPP

#include "nmm/NMMTypes.hpp"
#include "nmm/base/NMMObject.hpp"
#include "ISPUDecoder.hpp"


namespace NMM
{
  /**
   * Subtitle Picture Unit (SPU) decoder.
   * Interface that defines the events that control a SPU decoder. These events
   * are sent for example by the DVD reader.
   */

  class ISPUDecoderImpl  : public virtual NMMObject { 
  public:
    ISPUDecoderImpl() { addInterface(new ISPUDecoder(*this)); registerEvents(); };

    virtual ~ISPUDecoderImpl() {};

  protected:
    /**
     * Instream event to set colorpalette item.
     * Assigns an RGBA color value to a color palette index.
     * @param index color palette index
     * @param color_y Y color value
     * @param color_u U color value
     * @param color_v V color value
     * @param alpha alpha value
     */

    virtual Result setPaletteItem(const int& index, const int& color_y, const int& color_u, const int& color_v, const int& alpha) /* in-stream */  = 0;
    /**
     * Instream event to init color sheme.
     * Initialize the colors of a color scheme. Assigns the colors (palette entries) for the
     * selection and activation state of a button that has this scheme.
     * @param scheme_nr index of color scheme
     * @param selection palette index of selection color
     * @param activation color palette index of activation color
     */

    virtual Result initColorScheme(const int& scheme_nr, const int& selection, const int& activation) /* in-stream */  = 0;
    /**
     * Instream event to set selected button.
     * Sent by the DVD reader if a new menu button was selected.
     * @param button the button number
     */

    virtual Result setDVDSelectedButton(const int& button) /* in-stream */  = 0;
    /**
     * Instream event to set the number of buttons.
     * Sent by the DVD reader to inform the SPU decoder about the number of buttons.
     * @param buttons number of buttons
     */

    virtual Result setDVDButtonCount(const int& buttons) /* in-stream */  = 0;
    /**
     * Instream event to set button coordinates.
     * Sent by the DVD reader to set the coordinates of the current button.
     * @param start_x left X coordinate of button
     * @param start_y upper Y coordinate of button
     * @param end_x right X coordinate of button
     * @param end_y lower Y coordinate of button
     */

    virtual Result setDVDButtonCoords(const int& start_x, const int& start_y, const int& end_x, const int& end_y) /* in-stream */  = 0;
    /**
     * Instream event to set button color scheme.
     * Sent by the DVD reader to set the color scheme of the current button.
     * @param scheme color scheme index
     */

    virtual Result setDVDButtonScheme(const int& scheme) /* in-stream */  = 0;


  private:
    void registerEvents();
  };

}

#endif
