/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *                     
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_TVGUIDEPEERCONNECTOR_HPP
#define NMM_TVGUIDEPEERCONNECTOR_HPP NMM_TVGUIDEPEERCONNECTOR_HPP

#include "nmm/utils/thread/ThreadedObject.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"

#include "nmm/interfaces/misc/tvguide/ITVGuidePeerObjectImpl.hpp"

#include "nmm/base/proxy/tcp/TCPStrategy.hpp"
#include "nmm/base/ServiceProxy.hpp"

#include <list>

namespace NMM{

  //define variables for max number of handlers
#define MAX_INSERT_HANDLERS 5
#define MAX_REQUEST_HANDLERS 5

  class TVGuidePeerInsertHandler;
  class TVGuidePeerRequestHandler;

  /**
   *This class handles the connection to remote TVGuides
   */
  class TVGuidePeerConnector : 
    public ThreadedObject,
    public ITVGuidePeerObjectImpl
  {
  private:
    ITVGuide* localTVGuide;

    list<TCPAddress> hosts;

    list<TCPAddress> requestHostsForChannels;
    bool running;

    //Handler-Mutexes and -Conditions
    ThreadMutex insertHandlerMutex;
    ThreadCondition insertHandlerCondition;
    ThreadMutex requestHandlerMutex;
    ThreadCondition requestHandlerCondition;
    
    ThreadMutex distMutex;

    map<unsigned int, TVGuidePeerInsertHandler*> insertHandlers;
    map<unsigned int, TVGuidePeerRequestHandler*> requestHandlers;
    unsigned int insertHandlerIdCounter;
    unsigned int requestHandlerIdCounter;

  public:
    TVGuidePeerConnector(ITVGuide* itvg);
    ~TVGuidePeerConnector();

    /** query remote TVGuides for available channels
     */
    bool getChannels();
    bool getChannels(list<Interface*> peers);
    
    /** verify that all Event in the given ProgramDescription-list are next to each other
     */
    void verify(string day, string channel, list<ProgramDescription> verifyList);
    void verify(string start, string end, string channel, list<ProgramDescription> verifyList);
    /** distrubute the search-request
     */
    void search(ProgramFilter pf, list<Interface*> peer);
    /** search remote TVGuides for channels with the given id
     */
    string getChannelById(string id);
    /** distribute the request for current events on the given channel
     */
    void getCurrentEventOnChannel(const string& channel);
    /** distribute the request for next events on the given channel
     */
    void getFollowingEventOnChannel(const string& channel);

    /** add the address of a remote TVGuide
     */
    void addHost(TCPAddress adr);
    /** clear the list of remote TVGuides
     */
    void clearHosts();
    
    /** thismethod is called by an insert-handler to indicate, that the handler can be deleted
     */
    void insertHandlerDone(unsigned int id);
    /** thismethod is called by an request-handler to indicate, that the handler can be deleted
     */
    void requestHandlerDone(unsigned int id);

    /** get a address of a peer. the peer will be removed form the internal list
     */
    TCPAddress getLastChannelHost();
  protected:
    Result peerResult(const ProgramDescriptionList& pdlist, 
                      const int& id, 
                      const list<Interface*>& history);
     
    Result peerChannelResult (const ChannelDescriptionList& cdlist, 
                              const list<Interface*>& history);

  private:    
    void run();
    
    void insertEvents(list<ProgramDescription> pdlist);

  };


  /** 
   *This class inserts search-result into the database using its own thread that the function is not blocked
   */
  class TVGuidePeerInsertHandler: public ThreadedObject{
  private:
    enum insert_t{PROGRAM=0, CHANNEL};

    insert_t mode;

    ITVGuide *localtvg;
    list<ProgramDescription> insertPDList;
    list<ChannelDescription> insertCDList;
    unsigned int myID;
    TVGuidePeerConnector *parentPeer;

  public:
    /**
     *Constructor for inserting ProgramDescriptions
     *@param id the insert-handlers id that the TVGuidePeerConnector knows which handler to delete
     *@param peerConnector pointer to the Peer-Connector to notify it when work is done
     *@param localTVGuide poimter to the lokal TVGuide to insert the events 
     *@param programme list of ProgramDescription to be inserted
     */
    TVGuidePeerInsertHandler(unsigned int id, 
                             TVGuidePeerConnector *peerConnector, 
                             ITVGuide *localTVGuide, 
                             list<ProgramDescription> programme);
    /**
     *Constructor for inserting CahnnelDescriptions
     *@param id the insert-handlers id that the TVGuidePeerConnector knows which handler to delete
     *@param peerConnector pointer to the Peer-Connector to notify it when work is done
     *@param localTVGuide poimter to the lokal TVGuide to insert the events 
     *@param channels list of ChannelDescriptions to be inserted
     */
    TVGuidePeerInsertHandler(unsigned int id, 
                             TVGuidePeerConnector *peerConnector, 
                             ITVGuide *localTVGuide, 
                             list<ChannelDescription> channels);
    ~TVGuidePeerInsertHandler();
  private:
    void run();
  };

  /** this class connects to remote TVGuides and requests them for events using its own thread
   */
  class TVGuidePeerRequestHandler : public ThreadedObject{
  private:
    enum request_t{PROGRAM=0, CHANNEL};

    request_t mode;

    unsigned int myID;
    TVGuidePeerConnector *parentPeer;
    ITVGuide* localTVGuide;
    list<TCPAddress>& requestHosts;
    ProgramFilter currentRequest;
    list<Interface*>  lastHistory;
  public:
    /**
     *constructor to distribute search for programme
     *@param id the request-handlers id that the TVGuidePeerConnector knows which handler to delete
     *@param peerConnector pointer to the Peer-Connector to notify it when work is done
     *@param tvguide pointer to the ITVGuide-interface to set the ITVGuidePeerObject-Interface. This is used to notify the TVGuide about events found.
     *@param hosts list of remote TVGuide to query
     *@param request the Request
     *@param history list of Interfaces of already visited TVGuides
     */
    TVGuidePeerRequestHandler(unsigned int id, 
                              TVGuidePeerConnector *peerConnector, 
                              ITVGuide* tvguide, 
                              list<TCPAddress>& hosts, 
                              ProgramFilter request, 
                              list<Interface*> history);

    /**
     *Constructor to distribute search for channels
     *@param id the request-handlers id that the TVGuidePeerConnector knows which handler to delete
     *@param peerConnector pointer to the Peer-Connector to notify it when work is done
     *@param tvguide pointer to the ITVGuide-interface to set the ITVGuidePeerObject-Interface. This is used to notify the TVGuide about events found.
     *@param hosts list of remote TVGuide to query
     *@param history list of Interfaces of already visited TVGuides
     */
    TVGuidePeerRequestHandler(unsigned int id, 
                              TVGuidePeerConnector *peerConnector, 
                              ITVGuide* tvguide, 
                              list<TCPAddress>& hosts, 
                              list<Interface*> history);
    ~TVGuidePeerRequestHandler();
  private:
    void run();
  };

}

#endif
