/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "Scheduler.hpp"
#include "nmm/base/proxy/tcp/TCPStrategy.hpp"

#include "nmm/interfaces/video/tv/ITV.hpp"
#include "nmm/interfaces/file/IFileHandler.hpp"
#include "nmm/interfaces/video/tv/ITVCard.hpp"
#include "nmm/interfaces/video/tv/IChannelInfo.hpp"

namespace NMM{

  TVTimerScheduler::TVTimerScheduler()
  {
    conflictChecker = new TVTimerConflictChecker();
    channelConfig = new TVChannelList((string(getenv("HOME")) + "/.nmm/tv_channels"));

  }
  TVTimerScheduler::~TVTimerScheduler(){
    delete conflictChecker;
    delete channelConfig;
  }

  Result TVTimerScheduler::schedule(TVTimerEntry &entry){

    if(entry.getEndTime() != 0 && entry.getEndTime() < time(NULL) ){
      entry.setStatus(TVTimerEntry::RECORDED);
      STVTimerList::getInstance().addTimer(entry);

      throw TimerRecordedException(entry.getID());
    }


    TVChannel channel = channelConfig->getChannelByName(entry.getChannelName());
    TVChannel::cardList_t adrlist = channel.getCards();
    for(TVChannel::cardList_t::iterator i = adrlist.begin(); i != adrlist.end(); i++){
      TVCardDescription tmpTVCard = urlToNodeDescription(i->first);
      //list<TVTimerEntry> l = STVTimerList::getInstance().getTimers(tmpTVCard);
      list<TVTimerEntry> l = STVTimerList::getInstance().getTimers();
      if(conflictChecker->checkForConflicts(entry, l) == SUCCESS){
        entry.setChannelNumber(i->second);
        entry.addTVCard(tmpTVCard);
        STVTimerList::getInstance().addTimer(entry);
        return SUCCESS;
      }
    }

    //Exception thrown are catched in TVTimerControl

    entry.setStatus(TVTimerEntry::CONFLICT);
    STVTimerList::getInstance().addTimer(entry);
    //update the conflicting timers 
    list<int> il = entry.getConflictingTimers();
    for(list<int>::iterator i = il.begin(); i != il.end(); i++){
      TVTimerEntry e = STVTimerList::getInstance().getTimer(*i);
      e.addConflict(entry.getID());
      STVTimerList::getInstance().changeTimer(e);
    }

    return FAILURE;
  }

  Result TVTimerScheduler::reSchedule(TVTimerEntry &entry){
    if(entry.getEndTime() < time(NULL) ){
      return FAILURE;
    }

    try{
      TVChannel channel = channelConfig->getChannelByName(entry.getChannelName());
      TVChannel::cardList_t adrlist = channel.getCards();
      for(TVChannel::cardList_t::iterator i = adrlist.begin(); i != adrlist.end(); i++){        
        TVCardDescription tmpTVCard = urlToNodeDescription(i->first);
        
        if(contains(entry.getTVCards(), tmpTVCard)){
          continue;
        }

        list<TVTimerEntry> l = STVTimerList::getInstance().getTimers(tmpTVCard);
        if(conflictChecker->checkForConflicts(entry, l) == SUCCESS){
          entry.setChannelNumber(i->second);
          entry.addTVCard(tmpTVCard);
          STVTimerList::getInstance().addTimer(entry);
          return SUCCESS;
        }
      }

    }catch(const Exception& e){
      //If we can't find the channel, we can't record the show
      GLOBAL_ERROR_STREAM(e << endl);
      return FAILURE;
    }

    return FAILURE;
    
  }

  NodeDescription TVTimerScheduler::urlToNodeDescription(GraphURL url){
    NodeDescription returnValue;
    if(url.getProtocol() == GraphURL::FILE){

      returnValue = url.getNodeDescription();
      returnValue.setNodeName("DVBSimulatorNode");
      returnValue.addNodeEvent(Event(IFileHandler::setFilename_event, 
                                     new TInValue<string>(url.getURIPath())), INode::CONSTRUCTED);

    }else if(url.getProtocol() == GraphURL::IVTV){

      returnValue = url.getNodeDescription();

      if(!url.getOption("input").empty()){
        returnValue.addNodeEvent(Event(ITV::setCurrentInput_event, 
                                       new TInValue<unsigned int>(atoi(url.getOption("input").c_str()))), 
                                       INode::INITIALIZED);
      }
      if(!url.getOption("freqtable").empty()){
        returnValue.addNodeEvent(Event(ITV::setFrequencyTable_event, 
                                       new TInValue<string>(url.getOption("freqtable"))), INode::INITIALIZED);      
      }
    }else{
      returnValue = url.getNodeDescription();
    }

    if(!url.getOption("channelConfig").empty()){
      returnValue.addNodeEvent(Event(IChannelInfo::setChannelConfigFile_event, 
                                     new TInValue<string>(url.getOption("channelConfig"))), INode::CONSTRUCTED);      
    }

    return returnValue;
  }

  TVChannelList* TVTimerScheduler::getChannelConfig(){
    return channelConfig;
  }

  bool TVTimerScheduler::contains(list<TVCardDescription> &tvcardlist, TVCardDescription &tvcard){
    
    for(list<TVCardDescription>::iterator i = tvcardlist.begin(); i != tvcardlist.end(); i++){
      if(*i == tvcard) return true;
    }
    return false;
  }

}
