/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef TVTIMERCONTROL_HPP
#define TVTIMERCONTROL_HPP TVTIMERCONTROL_HPP

#include "TVTimerEntry.hpp"

#include "nmm/interfaces/misc/tvtimer/ITVTimerImpl.hpp"
#include "nmm/interfaces/misc/tvtimer/ITVTimerEventsImpl.hpp"
#include "nmm/interfaces/misc/tvguide/ITVGuide.hpp"

#include "nmm/misc/tvtimer/TVTimerList.hpp"
#include "nmm/misc/tvtimer/TVTimerExceptions.hpp"
#include "nmm/misc/tvtimer/Scheduler.hpp"
#include "nmm/misc/tvtimer/TVRecorderManager.hpp"

#include "nmm/utils/thread/Thread.hpp"

namespace NMM{

  class TVTimerControl : public ITVTimerImpl,
                         public ITVTimerEventsImpl
  {
    friend class CreateUsingNew<TVTimerControl>;

  private:

    ThreadMutex timerMutex;

    TVTimerScheduler *scheduler;
    TVRecorderManager *recorder;

    ITVGuide *itvguide;

  private:
    TVTimerControl();
    TVTimerControl(const TVTimerControl&);
    ~TVTimerControl();

  protected:
    /** Program Timer using TvTimerEntry */
    Result programTimer(const TVTimerEntry &entry) throw (const TimerAlreadyProgrammedException&);

    /** ChangeTimer using TvTimerEntry */
    Result changeTimer(const TVTimerEntry &newEntry) throw (const ActionNotAllowedException&);

    /** Delete Timer with ID i */
    Result deleteTimer(const int &i, const string &user = "") throw (const ActionNotAllowedException&);

    /**
     * get the Timers available filtered by day and channel
     */
    list<TVTimerEntry> getTimers(const string &day, const string &channel);

    /**
     * Get all spleeping Timers
     */
    list<TVTimerEntry> getSleepingTimers();
    /**
     * Get all timers currently recording
     */
    list<TVTimerEntry> getRecordingTimers();
    /**
     *Get all timers already recorded
     */
    list<TVTimerEntry> getRecordedTimers();
    /**
     *Get all timers witch Conflicts
     */
    list<TVTimerEntry> getConflictTimers();
    /**
     *Get all timers witch Conflicts with the given timer
     */
    list<TVTimerEntry> getConflictsWithTimer(const TVTimerEntry &entry);

    /**
     * Get the timer with the given ID
     */
    TVTimerEntry getTimer(const unsigned int &id);

    /**
     * Load timer from file
     */
    Result loadTimer(const string &filename);
    /**
     * Store timers to file
     */
    Result saveTimer(const string &filename);

    /**
     * Set the TVGuide-Interface to request timer-information if necessary
     */
    void setTVGuide(const ITVGuide *itvg);
    /**
     * Get the TVGuide-Interface
     */
    ITVGuide* getTVGuide(){return itvguide;}
    /**
     * Set the Directory where the reodings should be stored
     */
    void setDirectory(const string &dir);
    /**
     * Get the directory where the timers are stored
     */
    string getDirectory();

    TVChannelList getChannelConfig();

    /** ITVTimerEvents-Interface */
    virtual Result timerUpdate();
    virtual Result recordingStopped(const unsigned int &/*id*/){return SUCCESS;}
    virtual Result tvcardUnavailable(const unsigned int &id);

  public: 

    void setApplication(NMMApplication* app);
    void setChannelConfig(TVChannelList *cc);

  private:

  };//end class

  /* Typdef to access the TVTimerControl */
  typedef Singleton<TVTimerControl, CreateUsingNew> STVTimerControl;


} //end namespace NMM

#endif
