/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christoph Wellner <chwellner@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "TVTimerEntry.hpp"

namespace NMM{

  TVCardDescription::TVCardDescription(string t, TSAddress a)    
    : NMM::NodeDescription(t)
  {
    setLocation(a.getHostname());
    setPort(a.getPort());
    registerValues();
  }
  TVCardDescription::TVCardDescription()
    : NMM::NodeDescription()
  {
    registerValues();
  }
  TVCardDescription::TVCardDescription(const TVCardDescription &t)
    : NMM::NodeDescription(t)
  {
    registerValues();
    *this = t;   
  }
  
  TVCardDescription::TVCardDescription(const NodeDescription &nd)
    : NMM::NodeDescription(nd)
  {
    registerValues();
  }
  
  TVCardDescription::~TVCardDescription(){}

  void TVCardDescription::setAddress(TSAddress a){
    setLocation(a.getHostname()); 
    setPort(a.getPort());
  }

  TSAddress TVCardDescription::getAddress(){
    TSAddress a; 
    a.setHostname(getLocation()); 
    a.setPort(getPort()); 
    return a;
  }

  void TVCardDescription::registerValues(){
  }

  bool TVCardDescription::operator==(const TVCardDescription &t) const{
    TVCardDescription tvcard = t;
    if(getNodeName() == tvcard.getType()
       && getLocation() == tvcard.getAddress().getHostname()
       && getPort() == tvcard.getAddress().getPort()){
      return true;
    }else{
      return false;
    }
  }


  TVTimerEntry::TVTimerEntry()
    : NMM::SerializedValue("TVTimerEntry", "NMM"),
      start_time(0),
      end_time(0),
      channelNumber(0),
      channelName(""),
      status(SLEEPING),
      conflicts(false),
      period(ONCE)
  {
    registerValues();
  }

  TVTimerEntry::TVTimerEntry(string u)
    : NMM::SerializedValue("TVTimerEntry", "NMM"),
      start_time(0),
      end_time(0),
      channelNumber(0),
      channelName(""),
      status(SLEEPING),
      conflicts(false),
      period(ONCE),
      user(u)
  {
    registerValues();
  }
  
  TVTimerEntry::TVTimerEntry(const TVTimerEntry& te)
    : NMM::SerializedValue("TVTimerEntry", "NMM"),
      start_time(0),
      end_time(0),
      channelNumber(0),
      channelName(""),
      status(SLEEPING),
      conflicts(false),
      period(ONCE)
  {
    registerValues();
    *this = te;
  }
  
  TVTimerEntry::~TVTimerEntry(){}
  
  void TVTimerEntry::addTVCard(TVCardDescription tvd){
    tvcard.getRValue().push_front(tvd);
  }
  TVCardDescription TVTimerEntry::getTVCard() const{ 
    if(!tvcard.getRValue().empty()){
      return tvcard.getRValue().front();
    }
    throw NoCardFoundException();
    
  }

  void TVTimerEntry::addNextTimer(ProgramDescription pd) {
    nextTimers.getRValue().push_back(pd);
    nextTimers.getRValue().sort();
  }

  Result TVTimerEntry::verify(){
    /* if end_time < start_time, set end time to next day */
    if(end_time.getRValue() != 0){
      if(end_time.getRValue() < start_time.getRValue()) {
        struct tm* t = localtime(&(end_time.getRValue()));
        t->tm_mday++;
        end_time = mktime(t);
      }
    }

    if(period.getRValue() != ONCE){      
    }

    switch(period.getRValue()){
    case WEEKDAYS:
      {
        time_t start = start_time.getRValue();
        struct tm * start_tm = localtime(&start);
        //if period is weekdays but starttime is sunday
        if(start_tm->tm_wday == 0){
          //set start- and endtime to monday
          start_time =  start_time.getRValue() + (60 * 60 * 24);
          end_time = end_time.getRValue() + (60 * 60 * 24);
        }
        //if period is weekdays but starttime is saturday
        if(start_tm->tm_wday == 6){
          start_time = start_time.getRValue() + (60 * 60 * 24 * 2);
          end_time = end_time.getRValue() + (60 * 60 * 24 * 2);
        }
      }
      break;
    case WEEKENDS:
      {
        time_t start = start_time.getRValue();
        struct tm * start_tm = localtime(&start);
        //if period is weekends but starttime is a weekday
        if(1 <= start_tm->tm_wday <= 5){
          //set start- and endtime to monday
          start_time =  start_time.getRValue() + ((6 - start_tm->tm_wday) * 60 * 60 * 24);
          end_time = end_time.getRValue() + ((6 - start_tm->tm_wday) * 60 * 60 * 24);
        }
      }
      break;
    default:
      break;
    }



    return SUCCESS;
  }


  TVTimerEntry* TVTimerEntry::copy(){
    return new TVTimerEntry(*this);
  }

  void TVTimerEntry::registerValues(){
    insertArgument(&id);
    insertArgument(&start_time);
    insertArgument(&end_time);
    insertArgument(&channelName);
    insertArgument(&channelNumber);
    insertArgument(&status);
    insertArgument(&conflicts);
    insertArgument(&period);
    insertArgument(&info);
    insertArgument(&nextTimers);
    insertArgument(&tvcard);
    insertArgument(&user);
    insertArgument(&fileURL);
    insertArgument(&conflictingTimers);
  }


  void TVTimerEntry::addConflict(unsigned int id){
    conflictingTimers.getRValue().push_back(id);
    conflictingTimers.getRValue().sort();
    conflictingTimers.getRValue().unique();
  }

  void TVTimerEntry::removeConflict(unsigned int id){
    conflictingTimers.getRValue().remove(id);
  }

  bool TVTimerEntry::hasConflictWith(unsigned int id){

    if(conflictingTimers.getRValue().empty()) return false;

    for(list<int>::iterator i = conflictingTimers.getRValue().begin(); 
        i != conflictingTimers.getRValue().end(); 
        i++)
      {
        if(*i == (int)id){
          return true;
        }
      }
    return false;
  }

  bool TVTimerEntry::operator<(const TVTimerEntry &t) const{

    if(start_time.getRValue() < t.getStartTime()){
      return true;
    }else if(start_time.getRValue() == t.getStartTime() 
             && end_time.getRValue() < t.getEndTime()){
      return true;
    }else if(id.getRValue() < t.getID()){
      return true;
    }

    return false;
  }

  OValueStream& operator<<(OValueStream& os, const list<TVCardDescription>& td){
    return writeList<TVCardDescription>(os, td, "TVCardDescription");
  }
  IValueStream& operator>>(IValueStream& is, list<TVCardDescription>& td){
    return readList<TVCardDescription>(is, td, "TVCardDescription");
  }
  TConstructor<TVTimerEntry> nmm_tvcarddescpription("TVCardDescription", "NMM");
  TValueConstructor<list<TVCardDescription> > nmm_tvcarddescpription_list("list_TVCardDescription", "NMM");


  OValueStream& operator<<(OValueStream& os, const list<TVTimerEntry>& te){
    return writeList<TVTimerEntry>(os, te, "TVTimerEntry");
  }
  IValueStream& operator>>(IValueStream& is, list<TVTimerEntry>& te){
    return readList<TVTimerEntry>(is, te, "TVTimerEntry");
  }
  TConstructor<TVTimerEntry> nmm_timerentry("TVTimerEntry", "NMM");
  TValueConstructor<list<TVTimerEntry> > nmm_timerentry_list("list_TVTimerEntry", "NMM");

}//end namespace NMM
