/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "AudioConverterNode.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::AudioConverterNode, "AudioConverterNode")

namespace NMM {

  AudioConverterNode::AudioConverterNode(const char* name /*= "AudioConverterNode"*/,
                                         const StreamQueue::Mode queue_mode /*= StreamQueue::MODE_SUSPEND*/,
                                         const int size /*= StreamQueue::DEFAULT_SIZE*/)
    : GenericProcessorNode(name, queue_mode, size),
      in_channels(0),
      in_sample_rate(0),
      stereo_to_mono(false),
      downsample(1) {
  }

  AudioConverterNode::~AudioConverterNode() {
  }

  Result AudioConverterNode::doInit() {
    Format* in_format(getOwnInputProperty()->addNewFormat(Format::audio_raw_type));
    in_format->addWildcard(Format::sample_rate_param);
    in_format->addWildcard(Format::channels_param);
    in_format->addIntParamValue(Format::bitpersample_param, 16);
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::encoding_param, "linear");
    in_format->addStringParamValue(Format::sign_param, "signed");
    getOwnInputProperty()->setDefaultFormat(in_format);

    Format* out_format(getOwnOutputProperty()->addNewFormat(Format::audio_raw_type));
    out_format->addWildcard(Format::sample_rate_param);
    out_format->addWildcard(Format::channels_param);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    getOwnOutputProperty()->setDefaultFormat(out_format);

    in_format->setIOPartner(out_format);
    out_format->setIOPartner(in_format);

    return SUCCESS;
  }

  Result AudioConverterNode::doInitOutput() {
    const Format* cnt_format(getInputFormat());

    if (!cnt_format) {
      ERROR_STREAM("Could not get input Format!" << endl);
      return FAILURE;
    }

    Format* in_format(new Format(*cnt_format));
    getOwnWorkingInputProperty()->addNewFormat(in_format);

    try {
      in_sample_rate = in_format->getIntValue(Format::sample_rate_param);
      in_channels = in_format->getIntValue(Format::channels_param);

      Format* out_format(getOwnWorkingOutputProperty()->addNewFormat(Format::audio_raw_type));
      unsigned int divisor(1);
      while (in_sample_rate/divisor >= 8000) {
        out_format->addIntParamValue(Format::sample_rate_param, in_sample_rate/divisor);
        divisor++;
      }
      out_format->addIntParamValue(Format::channels_param, 1);
      if (in_channels > 1) {
        out_format->addIntParamValue(Format::channels_param, in_channels);
      }
      out_format->addIntParamValue(Format::bitpersample_param, 16);
      out_format->addStringParamValue(Format::endian_param, "little");
      out_format->addStringParamValue(Format::encoding_param, "linear");
      out_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
      
      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

    } catch (const FormatException & e ) {
      ERROR_STREAM("Problems reading input format: " << e << endl);
      getOwnInputProperty()->erase();
      return FAILURE;
    }

    return SUCCESS;
  }

  Result AudioConverterNode::doActivate() {
    const Format* cnt_format(getOutputFormat());

    if (!cnt_format) {
      ERROR_STREAM("Could not get output Format!" << endl);
      return FAILURE;
    }

    try {
      unsigned int out_sample_rate = cnt_format->getIntValue(Format::sample_rate_param);
      unsigned int out_channels = cnt_format->getIntValue(Format::channels_param);

      stereo_to_mono = (out_channels == 1);
      downsample = in_sample_rate / out_sample_rate;
    } catch (const FormatException & e ) {
      ERROR_STREAM("Problems reading output format: " << e << endl);
      return FAILURE;
    }

    return SUCCESS;
  }

  Message* AudioConverterNode::processBuffer(Buffer* in_buffer) {
    if ((downsample>1) || stereo_to_mono) {
      in_buffer = in_buffer->getWriteableInstance();
    }

    if (downsample>1) {
      char* data(reinterpret_cast<char*>(in_buffer->getData()));
      for (unsigned int sample = 0;
           sample < in_buffer->getUsed() / downsample;
           sample += 2 * in_channels) {
        for (unsigned int i = 0; i < 2 * in_channels; i++) {
          data[sample + i] = data[sample*downsample + i];
        }
      }
      in_buffer->setUsed(in_buffer->getUsed() / downsample);
    }

    if (stereo_to_mono) {
      char* data(reinterpret_cast<char*>(in_buffer->getData()));
      // this routine only works for signed data with 2 channels and 16 bitpersample
      for (unsigned int sample = 0;
           sample < in_buffer->getUsed() / 2;
           sample += 2) {
        int lu(static_cast<int>(data[sample*2 + 0 + 0]));
        int ll(static_cast<int>(data[sample*2 + 0 + 1]));
        int ru(static_cast<int>(data[sample*2 + 2 + 0]));
        int rl(static_cast<int>(data[sample*2 + 2 + 1]));

        int left((ll  * 256) + lu);
        int right((rl * 256) + ru);

        int mono((left / 2 + right / 2));

        data[sample + 0] = static_cast<char>(mono % 256);
        data[sample + 1] = static_cast<char>(mono / 256);
      }
      in_buffer->setUsed(in_buffer->getUsed()/2);
    }

    return in_buffer;
  }

} // end namespace NMM
