/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:	      
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "AVDemuxNode.hpp"
#include "nmm/interfaces/av/IAVMux.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::AVDemuxNode, "AVDemuxNode")

namespace NMM {

  AVDemuxNode::AVDemuxNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericDemultiplexerNode(name, qmode, qsize) {
    
    buf_count = 0;
    labelnum_video = IAVMux::LABEL_VIDEO;
    labelnum_audio = IAVMux::LABEL_AUDIO;
  }
  
  Result AVDemuxNode::setBufferLabel(const string& label_name, const int& label_number) {
    DEBUG_STREAM("setBufferLabel: label: " << label_name << " num: " << label_number);
    if ("video" == label_name) {
      labelnum_video = label_number;
    }
    else if ("audio" == label_name) {
      labelnum_audio = label_number;
    }
    else {
      ERROR_STREAM("setBufferLabel: FOUND UNKNOWN LABEL ON BUFFER!!!");
      return FAILURE;
    }
    // delete the event because following nodes should recognize their buffers
    // from the input format
    return DELETE;
  }
  
  Result AVDemuxNode::doInit() {
    Format *format;
    vector<Format*> a_formats;
    vector<Format*> v_formats;
    
    // supported video formats
    format = new Format(Format::video_mpeg4_type);
    format->addWildcard(Format::x_resolution_param);
    format->addWildcard(Format::y_resolution_param);
    format->addWildcard(Format::ratio_x_param);
    format->addWildcard(Format::ratio_y_param);
    format->addWildcard(Format::framerate_param);
    format->addWildcard(Format::compressor_param);
    v_formats.push_back(format);
    
    format = new Format(Format::video_mpeg4_type);
    format->addWildcard(Format::x_resolution_param);
    format->addWildcard(Format::y_resolution_param);
    format->addWildcard(Format::ratio_x_param);
    format->addWildcard(Format::ratio_y_param);
    format->addWildcard(Format::framerate_param);
    format->addWildcard(Format::compressor_param);
    format->addWildcard(Format::bitrate_param);
    v_formats.push_back(format);
    
    // supported audio formats

    // raw audio
    format = new Format(Format::audio_raw_type);
    format->addWildcard(Format::channels_param);
    format->addWildcard(Format::sample_rate_param);
    format->addWildcard(Format::bitpersample_param);
    format->addStringParamValue(Format::endian_param, "little");
    format->addStringParamValue(Format::encoding_param, "linear");
    format->addStringParamValue(Format::sign_param, "unsigned");
    format->addStringParamValue(Format::sign_param, "signed");
    a_formats.push_back(format);

    // need another audio raw format without endian
    format = new Format(Format::audio_raw_type);
    format->addWildcard(Format::channels_param);
    format->addWildcard(Format::sample_rate_param);
    format->addWildcard(Format::bitpersample_param);
    format->addStringParamValue(Format::encoding_param, "linear");
    format->addStringParamValue(Format::sign_param, "unsigned");
    format->addStringParamValue(Format::sign_param, "signed");
    a_formats.push_back(format);

    format = new Format(Format::audio_ac3_type);
    a_formats.push_back(format);
    
    format = new Format(Format::audio_ac3_type);
    format->addWildcard(Format::bitrate_param);
    a_formats.push_back(format);
    
    format = new Format(Format::audio_ac3_type);
    format->addWildcard(Format::channels_param);
    format->addWildcard(Format::sample_rate_param);
    format->addWildcard(Format::bitrate_param);
    format->addWildcard(Format::speaker_setup_param);
    format->addIntParamValue(Format::lfe_param, 0);
    format->addIntParamValue(Format::lfe_param, 1);
    a_formats.push_back(format);
    
    format = new Format(Format::audio_mpeg_type);
    format->addWildcard(Format::layer_param);
    format->addIntParamValue(Format::audio_codec_param, 0x55);
    format->addWildcard(Format::sample_rate_param);
    format->addWildcard(Format::channels_param);
    format->addWildcard(Format::bitpersample_param);
    a_formats.push_back(format);
    
    format = new Format(Format::audio_mpeg_type);
    format->addWildcard(Format::layer_param);
    format->addIntParamValue(Format::audio_codec_param, 0x55);
    format->addWildcard(Format::sample_rate_param);
    format->addWildcard(Format::channels_param);
    format->addWildcard(Format::bitpersample_param);
    format->addWildcard(Format::bitrate_param);
    a_formats.push_back(format);
    
    format = new Format(Format::audio_mpeg_type);
    format->addWildcard(Format::layer_param);
    format->addIntParamValue(Format::audio_codec_param, 0x55);
    a_formats.push_back(format);
    
    format = new Format(Format::audio_mpeg_type);
    format->addWildcard(Format::layer_param);
    format->addIntParamValue(Format::audio_codec_param, 0x55);
    format->addWildcard(Format::bitrate_param);
    a_formats.push_back(format);
    
    // output streams
    addOutputStream("video");
    addOutputStream("audio");
    
    // input format
    format = getOwnInputProperty()->addNewFormat("av/avi");
    for (unsigned int i=0; i<a_formats.size(); ++i) {
      format->addFPointerParamValue(Format::audio_format_param, a_formats[i]);
    }
    for (unsigned int i=0; i<v_formats.size(); ++i) {
      format->addFPointerParamValue(Format::video_format_param, v_formats[i]);
    }
    getOwnInputProperty()->setDefaultFormat(format);
    
    // output formats
    Property* property = getOwnOutputProperty();
    for (unsigned int i=0; i<a_formats.size(); ++i) {
      Format* format = new Format( *a_formats[i] );
      property->addNewFormat(format);
      format->setJackTag("audio");
      if (i == 0) { property->setDefaultFormat(format, "audio"); }
    }
    for (unsigned int i=0; i<v_formats.size(); ++i) {
      Format* format = new Format( *v_formats[i] );
      property->addNewFormat(format);
      format->setJackTag("video");
      if (i == 0) { property->setDefaultFormat(format, "video"); }
    }
    
    return SUCCESS;
  }
  
  Result AVDemuxNode::doDeinit() {
    getOwnOutputProperty()->erase();
    getOwnInputProperty()->erase();
    return SUCCESS;
  }
  
  Result AVDemuxNode::doInitOutput() {
    
    // get input format
    const Format *in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    // init working input property
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // init working output property
    Format* in_v_format = in_format->getFPointerValue(Format::video_format_param);
    Format* in_a_format = in_format->getFPointerValue(Format::audio_format_param);
    Format* out_v_format = new Format(*in_v_format);
    Format* out_a_format = (in_a_format != NULL) ? new Format(*in_a_format) : NULL;
    getOwnWorkingOutputProperty()->addNewFormat(out_v_format);
    out_v_format->setJackTag("video");
    getOwnWorkingOutputProperty()->setDefaultFormat(out_v_format, "video");
    if (out_a_format != NULL) {
      getOwnWorkingOutputProperty()->addNewFormat(out_a_format);
      out_a_format->setJackTag("audio");
      getOwnWorkingOutputProperty()->setDefaultFormat(out_a_format, "audio");
    }
    
    return SUCCESS;
  }
  
  Result AVDemuxNode::doDeinitOutput() {
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  }
  
  Message* AVDemuxNode::processBuffer(Buffer *in_buffer) {
    // if setBufferLabel was never received, assume video to be 0
    // and therefore all buffers to be video
    unsigned int label = in_buffer->getLabel();
    if (label == labelnum_audio) {
      DEBUG_STREAM("Sending Audio Data" << endl);
      setCurrentSendOutputStream("audio");
    } else if (label == labelnum_video) {
      DEBUG_STREAM("Sending Video Frame" << endl);
      setCurrentSendOutputStream("video");
    } else {
      ERROR_STREAM("processBuffer: FOUND UNKNOWN LABEL ON BUFFER!!!");
      return NULL;
    }
    buf_count++;
    
    return (in_buffer);
  }
  
  Result AVDemuxNode::processCEvent (CEvent *event) {
    unsigned int label = event->getLabel();
    if (label == labelnum_audio) {
      DEBUG_STREAM("Sending Audio event" << endl);
      setCurrentSendOutputStream("audio");
    } else if (label == labelnum_video) {
      DEBUG_STREAM("Sending Video event" << endl);
      setCurrentSendOutputStream("video");
    } else if (label == 0) {
      DEBUG_STREAM("Sending Audio/Video event" << endl);
    } else {
      ERROR_STREAM("Unknown label on event: " << label << endl);
    }
    
    return SUCCESS;
  }
}
