/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Marco Lohse <mlohse@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "RawReadNode.hpp"

#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/base/format/Property.hpp"

#include "nmm/interfaces/file/ITrack.hpp"
#include "nmm/interfaces/base/sync/ISyncReset.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::RawReadNode, "RawReadNode")

namespace NMM {
  RawReadNode::RawReadNode(const char* name /* = "RawReadNode" */,
			   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size) {
    read_mode   = IFileModeHandler::FILE_PER_STREAM;
    file_format = IFileModeHandler::DATA_ONLY;
    file_name   = "";
    file_des    = 0;
    loop        = false;
    send_event  = false;
  };

  RawReadNode::~RawReadNode() {
  }
  
  void RawReadNode::setFilename(const string& name) {
    if(read_mode == IFileModeHandler::FILE_PER_STREAM) {
      if(isConstructed()) {
	file_name = name;
      }      
      else {
	ERROR_STREAM("Node not in CONSTRUCTED state, setFileName illegal in current state and with FILE_PER_STREAM mode.");
      }
    } else {
      MutexGuard m(file_name_mutex);
      file_name = name;
    }
  }
  
  void RawReadNode::setMode(const IFileModeHandler::FileMode& mode) {
    read_mode = mode;
  }
  
  IFileModeHandler::FileMode RawReadNode::getMode() {
    return read_mode;
  }
  
  void RawReadNode::setFormat(const IFileModeHandler::FileFormat& format) {
    file_format = format;
  }
  
  IFileModeHandler::FileFormat RawReadNode::getFormat() {
    return file_format;
  }
  
  void RawReadNode::setFilenameMode(const IFileModeHandler::FilenameMode&) {
    WARNING_STREAM("Not implemented yet");
  }
  
  IFileModeHandler::FilenameMode RawReadNode::getFilenameMode() {
    WARNING_STREAM("Not implemented yet");
    return IFileModeHandler::FILENAME_ONLY;
  }
  
  void RawReadNode::setLoop(const bool& d) {
    loop = d;
  }
  
  Result RawReadNode::doInit() {
    Format* out_format = getOwnOutputProperty()->addNewFormat("*/*");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }
  
  Result RawReadNode::doInitOutput() {
    if(isOutputConnected("default")) {
      send_format = true;
    }

    Format* out_format = getOwnWorkingOutputProperty()->addNewFormat("*/*");
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    
    if(read_mode == IFileModeHandler::FILE_PER_STREAM) {
      if(!SSecurityCheck::getInstance().checkFilenameForReading(file_name)) {
        ERROR_STREAM("Security check not passed for file " << file_name << ", check .nmmrc for options" << endl);
        return FAILURE;
      }
      if((file_des = open(file_name.c_str(), O_RDONLY)) < 0) {
	DEBUG_STREAM( "Error Number " << errno << " while opening file " << file_name << endl);
	return FAILURE;
      } else {
	DEBUG_STREAM( "Opened file " << file_name << endl);
        setProducingFlag(true);
        send_event = true;
	return SUCCESS;
      }
    }
    else {
      setProducingFlag(true);
      return SUCCESS;
    }
  }
  
  Message* RawReadNode::processBuffer(Buffer *) 
  {
    MutexGuard m(file_name_mutex);
    Buffer *buffer;
    int read_status;
    Timestamp ts;
    
    // send start_track event
    if (send_event) {
      CEvent* ce = new CEvent();
      send_event = false;
      ce->insertEvent(ITrack::create_startTrack(file_name)); 
      DEBUG_STREAM("Sending syncReset" << endl);
      ce->insertEvent(ISyncReset::create_syncReset());     

      if(send_format) {
	send_format = false;

        //set new outputformat
        setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

	DEBUG_STREAM("Sending setFormat" << endl);
	ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));

      }

      return ce;

    }
    
    // IFileModeHandler::FILE_PER_FRAME open new file
    if(read_mode == IFileModeHandler::FILE_PER_FRAME) { // create new file
      
      if(!SSecurityCheck::getInstance().checkFilenameForReading(file_name)) {
        ERROR_STREAM("Security check not passed for file " << file_name << ", check .nmmrc for options" << endl);
        setProducingFlag(false);
        return NULL;
      }
      file_des = open(file_name.c_str(), O_WRONLY | O_CREAT | O_TRUNC, 00644);

      if(file_des < 0) {
	DEBUG_STREAM( "Error Number " << errno << " while opening file " << file_name << endl);
        setProducingFlag(false);
	return 0;
      }
      else {
	DEBUG_STREAM( "Opened file " << file_name << endl);      }
    }
    // IFileModeHandler::FILE_PER_FRAME end
    
    // IFileModeHandler::SIZE_DATA or IFileModeHandler::SIZE_TIMESTAMP_DATA, first read size
    if(file_format == IFileModeHandler::SIZE_DATA || file_format == IFileModeHandler::SIZE_TIMESTAMP_DATA) {
      read_status = read(file_des, &frame_size, sizeof(size_t));
      if(read_status != sizeof(size_t)) {
	if(read_status == 0) {
	  if(loop) {
	    lseek(file_des, 0, SEEK_SET);
	    DEBUG_STREAM( "Done with file but looping ..." << endl);
	    return 0;
	  }
	  else {
	    DEBUG_STREAM( "Done with file." << endl);
            setProducingFlag(false);
            Event e(ITrack::endTrack_event);
            return new CEvent(e);
	  }
	}
	ERROR_STREAM("Could not read size from file " << endl);
  	return 0;
      }
    }

    // IFileModeHandler::SIZE_TIMESTAMP_DATA, read timestamp
    if(file_format == IFileModeHandler::SIZE_TIMESTAMP_DATA) {
      read_status = read(file_des, &ts, sizeof(Timestamp));
      if(read_status != sizeof(Timestamp)) {
	if(read_status == 0) {
	  if(loop) {
	    lseek(file_des, 0, SEEK_SET);
	    DEBUG_STREAM( "Done with file but looping ..." << endl);
	    return 0;
	  }
	  else {
	    DEBUG_STREAM( "Done with file." << endl);
	    setProducingFlag(false);
            Event e(ITrack::endTrack_event);
            return new CEvent(e);
	  }
	}
	else {
	  ERROR_STREAM("Could not read timestamp from file" << endl);
	  return 0;
	}
      }

      DEBUG_STREAM("Timestamp " << ts << endl);
    }
    
    // Get new Buffer, and read data
    if ((buffer = getNewBuffer(frame_size)) != 0) {
      read_status = read(file_des, buffer->getData(), frame_size);
      if(file_format == IFileModeHandler::SIZE_TIMESTAMP_DATA) {
	buffer->setTimestamp(ts);
      }
      buffer->setUsed(read_status);
    } else {
      ERROR_STREAM("Could not get new buffer" << endl);
      return 0;
    }

    // IFileModeHandler::FILE_PER_STREAM
    if(read_mode == IFileModeHandler::FILE_PER_STREAM) {
      if(static_cast<unsigned int>(read_status) == frame_size) {
	DEBUG_STREAM( "Ok. Read " << buffer->getUsed() << " bytes. " << endl);
	return buffer;
      }
      if(read_status == 0) {
	if(loop) {
	  lseek(file_des, 0, SEEK_SET);
	  buffer->release();
	  DEBUG_STREAM( "Done with file but looping ..." << endl);
	  return 0;
	}
	else {
	  buffer->release();
	  DEBUG_STREAM( "Done with file." << endl);
          setProducingFlag(false);
          Event e(ITrack::endTrack_event);
          return new CEvent(e);
	}
      }
      if(read_status > 0) {
	ERROR_STREAM("Could only read " << read_status << " instead of " << frame_size << " bytes. Discarding buffer."<< endl);
	close(file_des);	 
	buffer->release();
	return 0;
      }
      else {
	ERROR_STREAM( "Error Number " << errno << " while reading file " << endl);
	close(file_des);	 
	buffer->release();
	return 0;
      }
    }
    // IFileModeHandler::FILE_PER_FRAME
    else {
      if(static_cast<unsigned int>(read_status) == frame_size) {
	DEBUG_STREAM( "Ok. Read " << buffer->getUsed() << " bytes. " << endl);
	return buffer;
      }
      if(read_status == 0) {
	if(loop) {
	  lseek(file_des, 0, SEEK_SET);
	  buffer->release();
	  DEBUG_STREAM( "Done with file but looping ..." << endl);
	  return 0;
	}
	else {
	  buffer->release();
	  DEBUG_STREAM( "Done with file." << endl);
          setProducingFlag(false);
          Event e(ITrack::endTrack_event);
          return new CEvent(e);
	  return 0;
	}
      }
      if(read_status > 0) {
	ERROR_STREAM("Could only read " << read_status << " instead of " << frame_size << " bytes. Discarding buffer."<< endl);
	close(file_des);	 
	buffer->release();
	return 0;
      }
      else {
	ERROR_STREAM( "Error Number " << errno << " while reading file " << endl);
	close(file_des);	 
	buffer->release();
	return 0;
      }
    }
  }

  Result RawReadNode::doDeinitOutput() 
  {
    if(read_mode == IFileModeHandler::FILE_PER_STREAM) {
      close(file_des);
    }
    return SUCCESS;
  }
}
