/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *                    Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "VNCSourceNode.hpp"
#include "nmm/base/EventDispatcher.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/sync/StreamTimer.hpp"
#include "nmm/base/registry/Plugin.hpp"

NMM_REGISTER_PLUGIN(NMM::VNCSourceNode, "VNCSourceNode")

namespace NMM {

  VNCSourceNode::VNCSourceNode(const char* name) : GenericSyncSourceNode(name) {
     
    // this node shall not be the master for source synchronization
    setSyncMaster(false);
    // this node does not produce at a constant rate
    setConstantRateProducer(false);

    // private-Variablen initialisieren!!
    
    buf_count = 0;
    buffer_size = 0;
    width = 0;
    height = 0;
    colordepth = 0;
    bits_per_pixel = 0;
    
        
    timer = new StreamTimer(this, StreamTimer::REAL_TIME);
    proto = new RFBProtoController(IVNC::REAL_TIME);
    mode = proto->getMode();
    mode_changed = false;	   
    
    DEBUG_STREAM("End > Create VNCSourceNode" << endl);
  }

  VNCSourceNode::~VNCSourceNode() {
    if (timer) delete timer;
    if (proto) delete proto;
  }

  Result VNCSourceNode::doInit() {
    Format *out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addWildcard(Format::colorspace_param);
    out_format->addWildcard(Format::bitperpixel_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::format_param, "packet");
    getOwnOutputProperty()->setDefaultFormat(out_format);

    return SUCCESS;
  }
  

  Result VNCSourceNode::doInitOutput() {
    
    if (proto->connectToRFBServer() == FAILURE) {
      ERROR_STREAM("Could not connect to RFBServer" << endl);
      return FAILURE;
    }
    
    if (proto->isConnected()) {
      DEBUG_STREAM("Connected!" << endl);
    } else {
      DEBUG_STREAM("Not connected!" << endl);
      return FAILURE;
    }
    
    if (proto->initialiseRFBConnection() == FAILURE) {
      ERROR_STREAM("An error occured while initializing the rfb-connection" << endl);
      return FAILURE;
    }
    
    /* Find the best pixel format to use */
    proto->initDesktopAndColor();
	
    /* Tell the VNC server which pixel format and encodings we want to use */
    proto->setFormatAndEncodings();
    
    width = proto->getFramebufferWidth();
    height = proto->getFramebufferHeight();
    bits_per_pixel = proto->getBitsPerPixel();
    buffer_size = width * height * bits_per_pixel/8;
    colordepth = proto->getColordepth();

    DEBUG_STREAM("setting output format parameters: width = " <<
		 width << ", height = " << height <<
		 ", depth = " << colordepth << ", bitsperpixel = " << bits_per_pixel << endl);

    //Request for initial complete picture 
    proto->sendIncrementalFramebufferUpdateRequest();

    // reset the stream timer
    timer->reset();

    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    out_format->addIntParamValue(Format::x_resolution_param, width);
    out_format->addIntParamValue(Format::y_resolution_param, height);
    out_format->addIntParamValue(Format::ratio_x_param, width);
    out_format->addIntParamValue(Format::ratio_y_param, height);
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addIntParamValue(Format::bitperpixel_param, bits_per_pixel);
    
    switch (proto->getChannelorder()) {
    case RFBProtoController::rgb:
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      break;
    case RFBProtoController::bgr:
      out_format->addStringParamValue(Format::channelorder_param,"bgr");
      break;
    default:
      out_format->addStringParamValue(Format::channelorder_param,"000");
      WARNING_STREAM("No valid channelorder in format" << endl);
    }

    switch (proto->getColorspace()) {
    case RFBProtoController::rgb565:
      out_format->addStringParamValue(Format::colorspace_param, "rgb565");
      break;
    case RFBProtoController::rgb24:
      out_format->addStringParamValue(Format::colorspace_param, "rgb24");
      break;
    case RFBProtoController::rgb32:
      out_format->addStringParamValue(Format::colorspace_param, "rgb32");
      break;
    default:
      out_format->addStringParamValue(Format::colorspace_param, "rgb00");
      WARNING_STREAM("Dummy-Format added!!!" << endl);
      break;
    }

    out_format->addStringParamValue(Format::format_param, "packet");
    out_format->addFloatParamValue(Format::framerate_param, 1.0);

    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    return SUCCESS;
  }

  Result VNCSourceNode::doDeactivate() {
    if (proto->isConnected()) {
      proto->handleRFBServerMessage(0);
      if (proto->disconnectFromRFBServer() == FAILURE)
	return FAILURE;
    }
    return SUCCESS;
  }

  void VNCSourceNode::setTolerance(const float& tolerance){
    proto->setTolerance(tolerance);
  } 
  
  void VNCSourceNode::setInterval(const float& interval){
    proto->setInterval(interval);
  }
  
  void VNCSourceNode::setShared(const bool& shared) {
    proto->setShared(shared);
  }

  float VNCSourceNode::getTolerance() const {
    return proto->getTolerance();
  }

  float VNCSourceNode::getInterval() const {
    return proto->getInterval();
  }

  bool VNCSourceNode::isShared() const {
    return proto->isShared();
  }

  int VNCSourceNode::getServerHeight() const {
    return proto->getFramebufferHeight();
  }

  int VNCSourceNode::getServerWidth() const {
    return proto->getFramebufferWidth();
  }

  void VNCSourceNode::sendKeyEvent(const CARD32& key, const int& down) {
    proto->sendKeyEvent(key, down);
  }

  void VNCSourceNode::sendPointerEvent(const int& x, const int& y, const int& buttonmask) {
    proto->sendPointerEvent(x, y, buttonmask);
  }
    
  void VNCSourceNode::sendFrameBufferUpdateRequest(){
    if(mode != IVNC::BLANK_SHEET){
      proto->sendFramebufferUpdateRequest(0, 0, proto->getFramebufferWidth(), proto->getFramebufferHeight(), false);
    }else{
      mode_changed = true;
    }
  }

  void VNCSourceNode::setHostname(const string& value) {
    proto->setHostname(value);
    hostname = value;
  }

  void VNCSourceNode::setPort(const int& port) {
    proto->setPort(port);
    this->port = port;
  }

  void VNCSourceNode::setProtocol(const string&) {
    // not needed
  }

  void VNCSourceNode::setPasswdPath(const string& passwdpath) {
    proto->setPasswdPath(passwdpath.c_str()); 
  }

  void VNCSourceNode::setPasswd(const string& passwd){
    proto->setPasswd(passwd.c_str());
  }
  
  void VNCSourceNode::setMode(const IVNC::Mode& _mode) {
    proto->setMode(_mode);
    mode = _mode;
    mode_changed = true;
  }

  string VNCSourceNode::getHostname() const {
    return hostname;
  }

  int VNCSourceNode::getPort() const {
    return port;
  }

  string VNCSourceNode::getProtocol() const {
    // not needed
    ERROR_STREAM("Method not implemented here" << endl);
    return string("");
  }

  void VNCSourceNode::connectionLost() {
    // maybe clean up something???
  }

  // virtual functions inherited from GenericSyncSourceNode
  Time VNCSourceNode::getLocalTime(u_int64_t)
  {
    // as VNCSourceNode is not a constant rate producer, getLocalTime should not be called
    // and does not have to deliver a useful value.
    return Time(0,0);
  }

  Message* VNCSourceNode::produceBuffer(Buffer*) {
    Buffer* outBuf;
    VncImage* vncimage;
    
    // test if the connection is still active
    if (!proto->isConnected()) {
      setProducingFlag(false);
      CEvent* ret_event = new CEvent(INetClient::create_connectionLost());
      return ret_event;
    }

    /* Now enter the main loop, processing VNC messages. */
    switch (proto->getMode()) {
     
	case IVNC::REAL_TIME:
	    if (proto->hasNewImage() || mode_changed) {
		mode_changed = false;
		vncimage = proto->getVncImage();
		if ((outBuf = getNewBuffer( buffer_size )) != 0) {
		    memcpy(outBuf->getData(), vncimage->data, buffer_size );
		    proto->setHasNewImage(false);
	  	  
		    // Set the correct time stamp for the buffer
		    timer->setTimestamp(outBuf); 
		    
		    return(outBuf);
		}
	    }else{
		proto->handleRFBServerMessage(0);
	    }
	    break;
	case IVNC::REQUESTED:
	    Buffer *outBuf;
	    if ((proto->hasNewImage() && proto->getTimeIsUp()) || mode_changed) {
		mode_changed = false;
		vncimage = proto->getVncImage();
		if ((outBuf = getNewBuffer( buffer_size )) != 0)
		{
		    memcpy(outBuf->getData(), vncimage->data, buffer_size );
		    proto->setHasNewImage(false);
		    proto->setTimeIsUp(false);	      
	      
		    // Set the correct time stamp for the buffer
		    timer->setTimestamp(outBuf); 
		    DEBUG_STREAM("\n##################\nImage generated!\n##################\n" << endl);
		    
		    return(outBuf);
		}
	    }else{
		proto->handleRFBServerMessage(0);
	    }
	    return 0;
    case IVNC::BLANK_SHEET:
      if(mode_changed){
	if ((outBuf = getNewBuffer( buffer_size )) != 0) { 
	  //initialize Buffer with zeros
	  char* data = outBuf->getData();
	  for(int i=0; i<buffer_size; i++){
	    *data=255;
	    data++;
	  }
	  //Set the correct time stamp for the buffer
	  timer->setTimestamp(outBuf);
	  mode_changed = false;
	  return(outBuf);
	}
      }else{
	proto->handleRFBServerMessage(0);
	return 0;
      }
      break;
    default:
      DEBUG_STREAM("The RFBProtoController is in an unknown mode" << endl);
      return 0;
      break;
    }
    return 0;
  }
  
} //namespace NMM
