/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "DevNullNode.hpp"

#include "nmm/base/format/Property.hpp"
#include "nmm/interfaces/stats/IBuffersPassedEvent.hpp"
#include "nmm/base/memorymgr/Buffer.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include <unistd.h>

NMM_REGISTER_PLUGIN(NMM::DevNullNode, "DevNullNode")

namespace NMM {

  DevNullNode::DevNullNode(const char* name,
			   const StreamQueue::Mode queue_mode,
			   const int size)
    : GenericSinkNode(name, queue_mode, size)
  {
    buffers_passed = 0;
    sleep_time = 0;
    upstream_passed = 0;
    interval = 0;
  }
  
  void DevNullNode::setSleepTime(const int& value) {
    sleep_time = value;
  }
  
  int DevNullNode::getSleepTime() const {
    return sleep_time;
  }
  
  int DevNullNode::getBuffersPassed() const {
    return buffers_passed;
  }
  
  void DevNullNode::resetBufferCounter() {
    buffers_passed = 0;
    upstream_passed = 0;
  }
  
  void DevNullNode::setBufferCountInterval(const int& value) {
    if (interval == 0) {
      /*
       * If the interval was previously zero, we start counting buffers from now.
       * It we just set the interval to a new value, we don't reset the counter.
       */
      upstream_passed = 0;
    }
    
    interval = value;
  }
  
  int DevNullNode::getBufferCountInterval() const {
    return interval;
  }
  
  Result DevNullNode::doInit() {
    Format* format = getOwnInputProperty()->addNewFormat("*/*");
    getOwnInputProperty()->setDefaultFormat(format);
    
    return SUCCESS;
  }
  
  Result DevNullNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    Format* format;
    
    format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);
    
    return SUCCESS;
  }
  
  Message* DevNullNode::processBuffer(Buffer *in_buffer) {
    in_buffer->release();
    buffers_passed++;
    if (interval) { 
      upstream_passed++;
      if(upstream_passed >= interval) {
	Event* e = new Event(IBuffersPassedEventReceiver::buffersPassed_event, new TInValue<int>(upstream_passed));
	CEvent* ce = new CEvent(e);
	// use this line instead 
	// CEvent* ce = new CEvent(); // calls CEvent(bool = true)
	ce->setDirection(Message::UPSTREAM);
	DEBUG_STREAM("Created UPSTREAM event " << IBuffersPassedEventReceiver::buffersPassed_event << " " << upstream_passed << endl);
	upstream_passed = 0;
	return ce;
      }
    }

    if (sleep_time) {
      DEBUG_STREAM("Sleeping " << sleep_time << " usec" << endl);
      usleep(sleep_time);
    }
    return 0;
  }
}

