/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *                    Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */



#ifndef NMM_SVGENCODENODE_HPP
#define NMM_SVGENCODENODE_HPP NMM_SVGENCODENODE_HPP

#include "nmm/base/sync/GenericSyncSourceNode.hpp"

#include "nmm/base/sync/DiffTimer.hpp"
#include "nmm/interfaces/text/ISVGEncodeImpl.hpp"
#include <fstream>
#include <vector>

namespace NMM {
  class StreamTimer;

  class Point {
  public:
    Point(const float&, const float&, DiffTimer*);
    Point(const float&, const float&);
    Point();
    ~Point();
    int getX() const;
    int getY() const;
    float getX_float() const;
    float getY_float() const;
    int distX(const Point&) const;
    int distY(const Point&) const;
    float dist_float(const Point&) const;
    Point add(const Point&) const;
    Point mult_c(const float&) const;
    float getTime() const;
    void setTime(const int) const;
  private:
    float x;
    float y;
    float time;
  };

  class SVGObject : public NamedObject {
  public:
    /**
     * Constructor
     */
    SVGObject(Point&);
    SVGObject(Point&, int, int, int);
    SVGObject(Point&, DiffTimer*, int, int, int, int);

    /**
     * Destructor
     */
    virtual ~SVGObject();
    
    virtual void finish(Point&) = 0 ;
    virtual void addPoint(Point&) = 0;
    virtual string getSVGString() const = 0;
    virtual string getAnimatedSVGString() = 0;
    virtual void setPathNumber(int) = 0;
    void setColor(int red, int green, int blue);
    void send();
    bool getIsFinished();
    bool getIsSent();
        
  protected:
    string getSVGColorString() const;
    int red;
    int green;
    int blue;
    Point startpoint;
    bool isFinished;
    bool isSent;
    DiffTimer* timer;
    int stroke_width;
        
  };

  class SVGLine : public SVGObject {
    
  public:
    SVGLine(Point&);
    SVGLine(Point&, int, int, int);
    SVGLine(Point&, DiffTimer*, int, int, int, int);
    ~SVGLine();
    void finish(Point&);
    void addPoint(Point&);
    string getSVGString() const;
    string getAnimatedSVGString();
    void setPathNumber(int);
    
  private:
    Point endpoint;
  };

  class SVGRectangle : public SVGObject {
    
  public:
    SVGRectangle(Point&);
    SVGRectangle(Point&, int, int, int);
    SVGRectangle(Point&, DiffTimer*, int, int, int, int);
    ~SVGRectangle();
    void finish(Point&);
    void addPoint(Point&);
    string getSVGString() const;
    string getAnimatedSVGString();
    void setPathNumber(int);
    
  private:
    int width;
    int height;
    Point endpoint;
  };

  class SVGFreehand : public SVGObject {
  public:
    SVGFreehand(Point&);
    SVGFreehand(Point&, int, int, int);
    SVGFreehand(Point&, DiffTimer*, int, int, int, int);
    ~SVGFreehand();
    void finish(Point&);
    void addPoint(Point&);
    string getSVGString() const;
    string getAnimatedSVGString();
    void setPathNumber(int);

  private:
    vector<Point> beziers;
    vector<Point> points;
    int pathnumber;
    int bezier1(int, int, int);
    int bezier2(int, int, int);
    void calculateBezier();
    Point mirror(Point, Point);
  };
 
  class SVGPointer : public SVGObject {
  public:
    SVGPointer(Point&);
    SVGPointer(Point&, int, int, int);
    SVGPointer(Point&, DiffTimer*, int, int, int, int);
    ~SVGPointer();
    void finish(Point&);
    void addPoint(Point&);
    string getSVGString() const;
    string getAnimatedSVGString();
    void setPathNumber(int);
    
  private:
    vector<Point> beziers;
    vector<Point> points;
    int bezier1(int, int, int);
    int bezier2(int, int, int);
    void calculateBezier();
    Point mirror(Point, Point);
  };

  class SVGEncodeNode  : public GenericSyncSourceNode,
			 public ISVGEncodeImpl {

  public: 

    /** Constructor
     */
    SVGEncodeNode(string, const char* = "SVGEncodeNode");
    SVGEncodeNode(const char* = "SVGEncodeNode");  
    /** Destructor
     */
    ~SVGEncodeNode();
          
  protected:
      
    Result doInit();
    Result doInitOutput();
    virtual Message* produceBuffer(Buffer*);
    virtual Result doDeactivate();

    // virtual functions inherited from GenericSyncSourceNode
    virtual Time getLocalTime(u_int64_t);

    //ISVGEncode
    virtual void setColor(const int& /*red*/, const int& /*green*/, const int& /*blue*/);
    virtual void setDimension(const int& /*width*/, const int& /*height*/);
    virtual int getPixelTaken();
    virtual void setPixelTaken(const int& /*accuracy*/);
    virtual void setWithFile(const bool& /*withFile=true*/);
    virtual void setStrokeWidth(const int& /*strokeWidth*/);
    virtual void drawSVGLine(const int& /*x*/, const int& /*y*/, const int& /*buttonmask*/);
    virtual void drawSVGRectangle(const int& /*x*/, const int& /*y*/, const int& /*buttonmask*/);
    virtual void drawSVGFreehand(const int& /*x*/, const int& /*y*/, const int& /*buttonmask*/);
    virtual void drawSVGPointer(const int& /*x*/, const int& /*y*/, const int& /*buttonmask*/);
    virtual void newAnnotation();
    virtual void clearAnnotation();
    virtual void undo();
    virtual void redo();
    virtual void repaint();
    
    virtual char* readFile(const char*);
    
  private:
    
    StreamTimer* streamTimer;

    ofstream* svgstream;
    ifstream* svginput;
    bool isMouseButtonPressed;
    bool withFile;
    bool addedObject;
    string filename;
    int pix_count;
    int pixel_taken;
    int red;
    int green;
    int blue;
    int stroke_width;
    int path_counter;
    int width, height;
    SVGObject* actualObject;
    DiffTimer* timer;
    vector<SVGObject*> objects;
    vector<SVGObject*>::iterator object_iter;
    ThreadMutex mutex;

    void newFile();
    void writeFile();
    void closeFile();
    void clearHistory();
    void deleteObjects(vector<SVGObject*>::iterator, vector<SVGObject*>::iterator);
    
  };
}//end namespace
#endif












