/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Christian Gerstner <gerstner@graphics.cs.uni-sb.de>
 *                    Eric Peters <epeters@graphics.cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include "nmm/utils/xml/xml++.hpp"
#include "nmm/utils/utils.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include "SVGPaintNode.hpp"
#include "SVGEncodeNode.hpp"
#include <stdio.h>
#include <string.h>
#include <iostream>

NMM_REGISTER_PLUGIN(NMM::SVGPaintNode, "SVGPaintNode")

using namespace std;

namespace NMM{

  SVGPaintNode::SVGPaintNode(const char* name, const StreamQueue::Mode queue_mode, const int queue_size) : GenericMultiplexerNode(name, queue_mode, queue_size) {
    imagewidth=0;
    imageheight=0;
    bits_per_pixel=32;
    buffer_size=0;
    has_background = false;
    has_annotationbuffer = false;
    showSVGAnnotation = true;
    working_background = 0;
    unchanged_background = 0;
    last_annotationbuffer =0;
    im = 0;
  }
  
  SVGPaintNode::~SVGPaintNode(){
    /* Destroy the image in memory. */
    //Temporary commented out, because this line is causing a segfault
    //if (im) gdImageDestroy(im); 
  }  

    void SVGPaintNode::drawLine(const int& x1, const int& y1, const int& x2, const int& y2, const int& color){
    if(x1<imagewidth && x2<imagewidth && y1<imageheight && y2<imageheight){
      gdImageLine(im,x1,y1,x2,y2,color);
    }else{
      ERROR_STREAM("drawLine-Params too big!" << endl);
    }
  }
  
  
  void SVGPaintNode::drawRectangle(const int& x1, const int& y1, const int& width, const int& height, const int& color){
    if(x1<imagewidth && width<imagewidth && y1<imageheight && height<imageheight){
      gdImageRectangle(im,x1,y1,width,height,color);
    }else{
      ERROR_STREAM("drawRectangle-Params too big!" << endl);
    }
  }
  
  void SVGPaintNode::drawBezier(const vector<Point>& points, const int& color){ 
    int pointssize = points.size();
    if(pointssize==1){
      gdImageSetPixel(im,points[0].getX(),points[0].getY(),color);
    }else{
      for(int i=0; i<(pointssize-1); i+=3){
	drawCurve(points[i],points[i+1],points[i+2],points[i+3],color);
      }
    }    
  }

  void SVGPaintNode::drawCurve(const Point& p1, const Point& p2, const Point& p3, const Point& p4, const int& color){
    if(p2.dist_float(p3)<=1){
      drawLine(p1.getX(),p1.getY(),p2.getX(),p2.getY(),color);
      drawLine(p2.getX(),p2.getY(),p3.getX(),p3.getY(),color);
      drawLine(p3.getX(),p3.getY(),p4.getX(),p4.getY(),color);
    }else{
      drawCurve(p1,
		(p1.add(p2)).mult_c(0.5),
		((p1.add(p2.mult_c(2))).add(p3)).mult_c(0.25),
		(((p1.add(p2.mult_c(3))).add(p3.mult_c(3))).add(p4)).mult_c(0.125),
		color);
      drawCurve((((p1.add(p2.mult_c(3))).add(p3.mult_c(3))).add(p4)).mult_c(0.125),
		(p2.add((p3.mult_c(2)).add(p4))).mult_c(0.25),
		(p3.add(p4)).mult_c(0.5),
		p4,
		color);
    }
  } 

   
  /** Initializes the first imageheight Pointer of the gdImageStructure to point to the imagelines
      in the memoryarea given by the dataptr. */
  
  void SVGPaintNode::initgdImage(gdImagePtr _im, int* dataptr){
    for (int i=0;i<imageheight;i++){
      _im->tpixels[i]= dataptr + i * imagewidth;
    }
  }

  int findAllOccurences(string src, string pattern, int src_length){
    int count = 0;
    int lastpos = 0;
    //make sure, that there are no remains from older svgstrings in the buffer
    while((lastpos=src.find(pattern,lastpos)) != static_cast<int>(string::npos) && src_length > lastpos) {
      count++; 
      lastpos++;
    };
 
    return count;
  }
  
  int SVGPaintNode::parseColor(const string& _svgstring, const int& startpos) const {
    int pos1stcolor;
    if((pos1stcolor = _svgstring.find("stroke:rgb(",startpos)) != static_cast<int>(string::npos)) {
      int pos2ndcolor =_svgstring.find(",",pos1stcolor+11);
      int pos3rdcolor =_svgstring.find(",",pos2ndcolor+1);
      int end3rdcolor =_svgstring.find(")",pos3rdcolor+1);
      int firstcolor = atoi((_svgstring.substr(pos1stcolor+11,pos2ndcolor-pos1stcolor-11)).c_str());
      int secondcolor = atoi((_svgstring.substr(pos2ndcolor+1,pos3rdcolor-pos2ndcolor-1)).c_str());
      int thirdcolor = atoi((_svgstring.substr(pos3rdcolor+1,end3rdcolor-pos3rdcolor-1)).c_str());
      return gdTrueColor(firstcolor,secondcolor,thirdcolor);
    }else{
      throw string("Parseerror: no stroke-color found");      
    }
  }

  int SVGPaintNode::parseStroke(const string& _svgstring, const int& startpos) const {
    int stroke_width_start;
    int stroke_width_end;
    if((stroke_width_start = _svgstring.find("stroke-width:",startpos)) != static_cast<int>(string::npos)) {
      stroke_width_end = _svgstring.find(";",stroke_width_start);
      DEBUG_STREAM("Stroke-width = " << atoi((_svgstring.substr(stroke_width_start+13,stroke_width_end-(stroke_width_start+13))).c_str()) << endl);
      return atoi((_svgstring.substr(stroke_width_start+13,stroke_width_end-(stroke_width_start+13))).c_str());
    }else{
      WARNING_STREAM("Parseerror: no stroke-width found" << endl);
      return 2;
    }
  }

  void SVGPaintNode::processSVGString(Buffer* inBuf){
    string svgstring = string(inBuf->getData());
    int svgstring_length = inBuf->getUsed();
    int svgcolor=0;
    int lastobjectendpos = 0;
    int numberOfRects = findAllOccurences(svgstring,"rect",svgstring_length);
    int numberOfLines = findAllOccurences(svgstring,"line",svgstring_length);
    int numberOfFreehands = findAllOccurences(svgstring,"path",svgstring_length);
    
    //Parsing of the rectobjects    
    for(int i=0; i<numberOfRects; i++){
      lastobjectendpos=svgstring.find("rect",lastobjectendpos);
      svgcolor=parseColor(svgstring,lastobjectendpos);
      int posx1 = svgstring.find("x=\"",lastobjectendpos);
      int posx1end = svgstring.find("\"",posx1+3);
      int x1=atoi((svgstring.substr(posx1+3,posx1end-posx1+3)).c_str());
      
      if(posx1==static_cast<int>(string::npos)||posx1end==static_cast<int>(string::npos)) throw string("Parseerror: x not found");
      int posy1 = svgstring.find("y=\"",posx1end);
      int posy1end = svgstring.find("\"",posy1+3);
      int y1=atoi((svgstring.substr(posy1+3,posy1end-posy1+3)).c_str());
      
      if(posy1==static_cast<int>(string::npos)||posy1end==static_cast<int>(string::npos)) throw string("Parseerror: y not found");
      int posx2 = svgstring.find("width=\"",posy1end);
      int posx2end = svgstring.find("\"",posx2+7);
      int x2=atoi((svgstring.substr(posx2+7,posx2end-posx2+7)).c_str());
      if(posx2==static_cast<int>(string::npos)||posx2end==static_cast<int>(string::npos))throw string("Parseerror: width not found");
      int posy2 = svgstring.find("height=\"",posx2end);
      int posy2end = svgstring.find("\"",posy2+8);
      int y2=atoi((svgstring.substr(posy2+8,posy2end-posy2+7)).c_str());
      if(posy2==static_cast<int>(string::npos)||posy2end==static_cast<int>(string::npos)) throw string("Parseerror: y2 not found");
      gdImageSetThickness(im,parseStroke(svgstring,lastobjectendpos));
      lastobjectendpos = svgstring.find("/>",posy2end);
      if(lastobjectendpos==static_cast<int>(string::npos)) throw string("Parseerror: Endtag of actual object not found");
      MESSAGE_STREAM("Rect-Object with x="<<x1<<", y="<<y1<<", width="<<x2<<", height="<<y2<<endl);
      drawRectangle(x1,y1,x1+x2,y1+y2,svgcolor);
    };

    //Parsing of the lineobjects
    lastobjectendpos=0;
    for(int i=0; i<numberOfLines; i++){
      lastobjectendpos=svgstring.find("line",lastobjectendpos);
      svgcolor=parseColor(svgstring,lastobjectendpos);
      int posx1 = svgstring.find("x1=\"",lastobjectendpos);
      int posx1end = svgstring.find("\"",posx1+4);
      int x1=atoi((svgstring.substr(posx1+4,posx1end-posx1+3)).c_str());
      if(posx1==static_cast<int>(string::npos)||posx1end==static_cast<int>(string::npos)) throw string("Parseerror: x1 not found");
      int posy1 = svgstring.find("y1=\"",posx1end);
      int posy1end = svgstring.find("\"",posy1+4);
      int y1=atoi((svgstring.substr(posy1+4,posy1end-posy1+3)).c_str());
      if(posy1==static_cast<int>(string::npos)||posy1end==static_cast<int>(string::npos)) throw string("Parseerror: y1 not found");
      int posx2 = svgstring.find("x2=\"",posy1end);
      int posx2end = svgstring.find("\"",posx2+4);
      int x2=atoi((svgstring.substr(posx2+4,posx2end-posx2+3)).c_str());
      if(posx2==static_cast<int>(string::npos)||posx2end==static_cast<int>(string::npos))throw string("Parseerror: x2 not found");
      int posy2 = svgstring.find("y2=\"",posx2end);
      int posy2end = svgstring.find("\"",posy2+4);
      int y2=atoi((svgstring.substr(posy2+4,posy2end-posy2+3)).c_str());
      if(posy2==static_cast<int>(string::npos)||posy2end==static_cast<int>(string::npos)) throw string("Parseerror: y2 not found");
      gdImageSetThickness(im,parseStroke(svgstring,lastobjectendpos));
      lastobjectendpos = svgstring.find("/>",posy2end);
      if(lastobjectendpos==static_cast<int>(string::npos)) throw string("Parseerror: Endtag of actual object not found");
      //MESSAGE_STREAM("Line-Object with x1="<<x1<<", y1="<<y1<<", x2="<<x2<<", y2="<<y2<<endl);
      drawLine(x1,y1,x2,y2,svgcolor);
    };
    
    lastobjectendpos=0;
    //Parsing of the pathobjects
    vector<Point> bezierpoints;
    for(int i=0; i<numberOfFreehands; i++){
      lastobjectendpos=svgstring.find("path",lastobjectendpos);
      svgcolor=parseColor(svgstring,lastobjectendpos);
      bezierpoints.clear();
      int thisobjectendpos = svgstring.find("/>",lastobjectendpos);
      string pathstring = svgstring.substr(lastobjectendpos-1,thisobjectendpos-lastobjectendpos+3);
      int numberOfPoints = findAllOccurences(pathstring,"C",svgstring_length);
      int posM1 = pathstring.find("d=\"M",0);
      int posM2 = pathstring.find(",",posM1+3);
      int endM2 = pathstring.find(" ",posM2+1);
      int m1 = atoi((pathstring.substr(posM1+4,posM2-posM1-4)).c_str());
      int m2 = atoi((pathstring.substr(posM2+1,endM2-posM2)).c_str());
      Point m(m1,m2);
      bezierpoints.push_back(m);
      int lastCpos=0;
      for(int i=0;i<numberOfPoints;i++){
	int c1pos=pathstring.find("C",lastCpos);
	lastCpos = c1pos + 1;
	int c2pos=pathstring.find(",",c1pos);
	int c3pos=pathstring.find(" ",c2pos+1);
	int c4pos=pathstring.find(",",c3pos+1);
	int c5pos=pathstring.find(" ",c4pos+1);
	int c6pos=pathstring.find(",",c5pos+1);
	int cendpos=pathstring.find(" ",c6pos+1);
	  
	Point p1= Point(atoi((pathstring.substr(c1pos+1,c2pos-c1pos)).c_str()),atoi((pathstring.substr(c2pos+1,c3pos-c2pos)).c_str()));
	Point p2= Point(atoi((pathstring.substr(c3pos+1,c4pos-c3pos)).c_str()),atoi((pathstring.substr(c4pos+1,c5pos-c4pos)).c_str()));
	Point p3= Point(atoi((pathstring.substr(c5pos+1,c6pos-c5pos)).c_str()),atoi((pathstring.substr(c6pos+1,cendpos-c6pos)).c_str()));
	bezierpoints.push_back(p1);
	bezierpoints.push_back(p2);
	bezierpoints.push_back(p3);
      }
      gdImageSetThickness(im,parseStroke(svgstring,lastobjectendpos));
      drawBezier(bezierpoints,svgcolor);
      lastobjectendpos = thisobjectendpos;
    };	 
  };
    
  Result SVGPaintNode::doInit(){
    //input svgtextstream
    addInputStream("svgtextstream");
    Format *in_svgformat = getOwnInputProperty()->addNewFormat("text/svg");
    in_svgformat->setJackTag("svgtextstream");
    getOwnInputProperty()->setDefaultFormat(in_svgformat,"svgtextstream");
	
    //input imagestream
    addInputStream("imagestream");
    Format *in_vncformat = getOwnInputProperty()->addNewFormat("video/raw");
    in_vncformat->addWildcard(Format::x_resolution_param);
    in_vncformat->addWildcard(Format::y_resolution_param);
    in_vncformat->addWildcard(Format::ratio_x_param);
    in_vncformat->addWildcard(Format::ratio_y_param);
    in_vncformat->addWildcard(Format::framerate_param);
    in_vncformat->addStringParamValue(Format::colorspace_param, "rgb24");
    in_vncformat->addIntParamValue(Format::bitperpixel_param, bits_per_pixel);
    in_vncformat->addStringParamValue(Format::channelorder_param,"rgb");
    in_vncformat->addStringParamValue(Format::endian_param, "little");
    in_vncformat->addStringParamValue(Format::format_param, "packet");
    in_vncformat->setJackTag("imagestream");
    getOwnInputProperty()->setDefaultFormat(in_vncformat,"imagestream");
	
    Format *out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::colorspace_param, "rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param, bits_per_pixel);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::format_param, "packet");
    getOwnOutputProperty()->setDefaultFormat(out_format);
	
    return SUCCESS;
  };

    
  Result SVGPaintNode::doInitOutput(){
    // get input formats and make sure they exist
    const Format *insvg_f = getInputFormat("svgtextstream");
    const Format *invnc_f = getInputFormat("imagestream");
    if (insvg_f == NULL || invnc_f == NULL) { cerr << "FAILURE at input" << endl;return FAILURE; }
	
    // set internal variables
    // check for required parameters
    imagewidth = invnc_f->getIntValue(Format::x_resolution_param);
    imageheight = invnc_f->getIntValue(Format::y_resolution_param);

    buffer_size= imagewidth * imageheight * bits_per_pixel/8; 

    /* Allocate the image: x pixels across by y pixels tall */
    im = gdImageCreateTrueColor(imagewidth-1, imageheight);
    MESSAGE_STREAM("IMAGECREATE with "<< imagewidth << "x" << imageheight << " pixels" << endl);

    // working input property
    Format* isvgf = new Format(*insvg_f);
    Format* ivncf = new Format(*invnc_f);
    getOwnWorkingInputProperty()->addNewFormat(isvgf);
    getOwnWorkingInputProperty()->addNewFormat(ivncf);
    getOwnWorkingInputProperty()->setDefaultFormat(isvgf);
    getOwnWorkingInputProperty()->setDefaultFormat(ivncf);
	

    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    out_format->addIntParamValue(Format::x_resolution_param, imagewidth);
    out_format->addIntParamValue(Format::y_resolution_param, imageheight);
    out_format->addIntParamValue(Format::ratio_x_param, imagewidth);
    out_format->addIntParamValue(Format::ratio_y_param, imageheight);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addIntParamValue(Format::bitperpixel_param, bits_per_pixel);
    out_format->addStringParamValue(Format::colorspace_param, "rgb24");            
    out_format->addStringParamValue(Format::format_param, "packet");
    out_format->addFloatParamValue(Format::framerate_param, 1.0);
            
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    return SUCCESS;
    
  };
  
  
  Message* SVGPaintNode::processBuffer(Buffer* inBuf) {
    if(strcmp(getCurrentRecvInputStream(), "imagestream") == 0) {
      if(unchanged_background) unchanged_background->release();
      has_background = true;
      unchanged_background = inBuf;
      /* Increase Counter of unchanged_background before returning it.
	 Otherwise the receiving node will release the buffer. */
      unchanged_background->increaseCount();
      if(showSVGAnnotation && has_annotationbuffer){
	working_background = unchanged_background->getWriteableInstance();
	/* initialize gdImage-structure to point to the memoryarea of working_background */
	  initgdImage(im,reinterpret_cast<int *>(working_background->getData()));
	  try {
	    processSVGString(last_annotationbuffer);  
	  }catch(string error) {ERROR_STREAM("processSVGString:" << error  << endl);}
	  return working_background;
      }
      return unchanged_background;
    }else if(strcmp(getCurrentRecvInputStream(), "svgtextstream") == 0) {
      if(last_annotationbuffer) last_annotationbuffer->release();
      last_annotationbuffer = inBuf;
      has_annotationbuffer = true;
      if(has_background){
	if(showSVGAnnotation){  //send picture with annotations
	  
	  /* increase Counter of unchanged_buffer so that
	     getWriteableInstance gets a real copy of the buffer */
	  unchanged_background->increaseCount();
	  working_background = unchanged_background->getWriteableInstance();
	  /* initialize gdImage-structure to point to the memoryarea of working_background */
	  initgdImage(im,reinterpret_cast<int *>(working_background->getData()));
	  try {
	    processSVGString(inBuf);  
	  }catch(string error) {ERROR_STREAM("processSVGString:" << error  << endl);}	  
	  /* working_background is no more needed, so may be released by receiving node */
	  
	  return working_background;

	}else{                //send picture without annotations
	  
	  unchanged_background->increaseCount();
	  
	  return unchanged_background; 
	
	}
      }
      WARNING_STREAM("No background available." << endl);
      return 0;
    } 
    ERROR_STREAM("Unknow Buffer received!" << endl);
    return 0;
  }
  
 
  Result SVGPaintNode::doDeactivate(){
    return SUCCESS;
  };

  void SVGPaintNode::setImageHeight(const int& height){
    imageheight=height;
  }

  void SVGPaintNode::setImageWidth(const int& width){
    imagewidth=width;
  }

  void SVGPaintNode::setShowSVGAnnotation(const bool& ssa){
    showSVGAnnotation = ssa;
  }

}
