/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <iostream>
#include "nmm/plugins/video/divx4/DivX4EncodeNode.hpp"
#include "nmm/interfaces/video/IKeyframe.hpp"
#include "nmm/interfaces/stats/IEncodeStats.hpp"
#include "nmm/base/format/Property.hpp"
#include "nmm/base/registry/Plugin.hpp"
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

NMM_REGISTER_PLUGIN(NMM::DivX4EncodeNode, "DivX4EncodeNode")

namespace NMM {

  DivX4EncodeNode::DivX4EncodeNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericProcessorNode(name, qmode, qsize),
    // initialization of variables - values taken as advertised by README-File of lib
    divx(NULL),
    x_dim(0),
    y_dim(0),
    framerate(0),
    bitrate(900),
    rc_period(2000),
    rc_reaction_period(10),
    rc_reaction_ratio(20),
    max_quantizer(31),
    min_quantizer(2),
    max_key_interval(250),
    deinterlace(0),
    quality(5),
    handle(0),
    outframe(NULL),
    length(0),
    colorspace(ENC_CSP_YV12),
    quant(0),
    intra(0),
    mvs(NULL),
    frame_count(0),
    startsec(0),
    neededsecs(0),
    encfps(0),
    encoded_bytes(0),
    released(true),
    compressor("DIVX"),
    enc_buf_size(65536),
    last_buffer(NULL),
    send_frame_count(false),
    last_frame_count(0)
  {    
    divx = new ENC_PARAM;    // create new structure for initialization
    
    outframe = new size_t[2500000];
    
    // get and remember current time
    gettimeofday(&tv, &tz);
    startsec = tv.tv_sec;
  }
  
  DivX4EncodeNode::~DivX4EncodeNode() {
    delete divx;
    delete[] outframe;
  }
  
  Result DivX4EncodeNode::doInit() {
    // general input format
    Format* in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"yvu");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"yv12");
    in_format->addIntParamValue(Format::bitperpixel_param,12);
    in_format->addStringParamValue(Format::format_param, "planar");
    
    // general output format
    Format* out_format = getOwnOutputProperty()->addNewFormat("video/mpeg4");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addWildcard(Format::bitrate_param);
    out_format->addStringParamValue(Format::compressor_param, compressor);
    
    // default formats
    getOwnOutputProperty()->setDefaultFormat(out_format);
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    // I/O partners
    in_format->setIOPartner(out_format);
    out_format->setIOPartner(in_format);
    
    return SUCCESS;
  }
  
  Result DivX4EncodeNode::doDeactivate() {
    if (released) {
      return SUCCESS;
    } else {
      return releaseEncoder();
    }
  }
  
  Result DivX4EncodeNode::doInitOutput() {
    // get input format
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("No input format!");
      return FAILURE;
    }
    
    // set internal variables
    x_dim = in_format->getIntValue(Format::x_resolution_param); 
    y_dim = in_format->getIntValue(Format::y_resolution_param);
    framerate = in_format->getFloatValue(Format::framerate_param);
    int x_ratio = in_format->getIntValue(Format::ratio_x_param);
    int y_ratio = in_format->getIntValue(Format::ratio_y_param);
    
    // setup working input property
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // setup working output property
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/mpeg4");
    o_format->addIntParamValue(Format::x_resolution_param, x_dim);
    o_format->addIntParamValue(Format::y_resolution_param, y_dim);
    o_format->addIntParamValue(Format::ratio_x_param, x_ratio);
    o_format->addIntParamValue(Format::ratio_y_param, y_ratio);
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addWildcard(Format::bitrate_param);
    o_format->addStringParamValue(Format::compressor_param, compressor);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Result DivX4EncodeNode::doActivate() {
    
    /* get output format */
    const Format* o_format = getOutputFormat();
    if (!o_format) {
      ERROR_STREAM("No output format!");
      return FAILURE;
    }
    
    /* init structure */
    memset(divx, 0, sizeof(ENC_PARAM));
    
    /* set divx parameters */
    divx->x_dim = o_format->getIntValue(Format::x_resolution_param);
    divx->y_dim = o_format->getIntValue(Format::y_resolution_param);
    divx->framerate = o_format->getFloatValue(Format::framerate_param);
    divx->bitrate = o_format->getIntValue(Format::bitrate_param);
    
    //recommended
    divx->rc_period = rc_period;
    divx->rc_reaction_period = rc_reaction_period;
    divx->rc_reaction_ratio  = rc_reaction_ratio;
    
    divx->max_quantizer = max_quantizer;
    divx->min_quantizer = min_quantizer;
    
    divx->max_key_interval=max_key_interval;
    divx->deinterlace = deinterlace;
    divx->quality=quality;
    divx->handle=NULL;
    
    //initialize ENC_FRAME
    encframe.image = 0;
    encframe.bitstream = outframe;
    encframe.length = length;
    encframe.colorspace = colorspace;
    encframe.quant = quant;
    encframe.intra = intra;
    encframe.mvs = mvs;
    
    // initialize the library
    int res;
    if((res = encore(divx->handle, ENC_OPT_INIT, divx, NULL)) != ENC_OK) {
      DEBUG_STREAM("divx-library open error; result of init:" << res << endl);
      return(FAILURE);
    }
    
    // now the encoder is not released any more
    released = false;
    
    frame_count = 0;
    last_frame_count = 0;
    neededsecs = 0;
    
    return SUCCESS;
  }

  Result DivX4EncodeNode::releaseEncoder() {
    MutexGuard divx_mg(divx_mutex);
    // release the encoder
    int res;
    if ((res = encore(divx->handle, ENC_OPT_RELEASE, NULL, NULL)) != ENC_OK) {
      DEBUG_STREAM("Release of Encoder failed! result of encore: " << res << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Encoder released successfully!" << endl);
    DEBUG_STREAM("Encoded " << encoded_bytes << " Bytes." << endl);
    DEBUG_STREAM("Average fps was " << encfps << "." << endl);
    // now the encoder is released
    released = true;
    return SUCCESS;
  }
  
  Message* DivX4EncodeNode::processBuffer(Buffer* in_buffer) {
    MutexGuard divx_mg(divx_mutex);
    
    if (getWorkingFlag()) {
      setWorkingFlag(false);
      return last_buffer;
    }

    Buffer *out_buffer;
    
    encframe.image = in_buffer->getData();
    
    // encode the in_buffer (raw-frame) to a divx4-frame
    ENC_RESULT res;
    encore(divx->handle, ENC_OPT_ENCODE, &encframe, &res);

    // check time and calculate encfps (encoding frames per second)
    gettimeofday(&tv, &tz);
    neededsecs = tv.tv_sec - startsec;
    frame_count++;
    encoded_bytes += encframe.length;

    encfps = (float)frame_count/(float)neededsecs;
    
    DEBUG_STREAM("ENCODING RESULTS: " << endl <<
		 "is_key_frame: " << res.is_key_frame << endl <<
		 "quantizer: " << res.quantizer << endl <<
		 "Number of texture bits: " << res.texture_bits << endl <<
		 "motion_bits: " << res.motion_bits << endl <<
		 "total_bits: " << res.total_bits << endl <<
		 "encoded frame size: " << encframe.length << endl);
        
    cerr << "Encoded " << frame_count << " Frames; cur. length: " << encframe.length
	 << " - Framerate: " << encfps
	 << (res.is_key_frame ? " - KeyFrame!\r" : "              \r");
    fflush(stderr);
    
    // optimization for BufferManager: get equally sized buffers if possible
    if (encframe.length > enc_buf_size) {
      out_buffer = getNewBuffer(encframe.length, in_buffer);
    }
    else {
      out_buffer = getNewBuffer(enc_buf_size, in_buffer);
    }

    // copy encoded data to out_buffer
    memcpy(out_buffer->getData(), encframe.bitstream, encframe.length);
    out_buffer->setUsed(encframe.length);
    
    in_buffer->release();

    if (res.is_key_frame) {
      // if we got a key frame, send an Event telling about it first
      setWorkingFlag(true);
      last_buffer = out_buffer;
      return new CEvent(Event(IKeyframe::setKeyframe_event));
    }
    
    if (send_frame_count && frame_count - last_frame_count > 10) {
      setWorkingFlag(true);
      last_buffer = out_buffer;
      
      last_frame_count = frame_count;
      
      return new CEvent( IEncodeStatsListener::create_setFrameCount((unsigned int) frame_count, encfps) );
    }    
    
    return out_buffer;
  }
  
  void DivX4EncodeNode::setQuality(const float& quality) {
    this->quality = (int) quality;
  }
  
  void DivX4EncodeNode::setBitrate(const int& bitrate) {
    MutexGuard divx_mg(divx_mutex);
    this->bitrate = bitrate;
    DEBUG_STREAM("Received Event bitrate; bitrate changed to" << bitrate << endl);
  }
  
  float DivX4EncodeNode::getQuality() 
  {
    // quality not defined
    return 1.0f;
  }
  
  int DivX4EncodeNode::getBitrate() {
    return bitrate;
  }
  
  Result DivX4EncodeNode::startTrack(const string&) {
    return SUCCESS;
  }
  
  Result DivX4EncodeNode::endTrack() {
    DEBUG_STREAM("Received Event: endTrack");
    if (released) { return SUCCESS; }
    else return releaseEncoder();
  }
  
  Result DivX4EncodeNode::finished() {
    return SUCCESS;
  }
  
  void DivX4EncodeNode::sendFrameCount(const bool& value) {
    send_frame_count = value;
  }
}
