/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        NMM work group,
 *                    http://www.networkmultimedia.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_MAGICKMANIPULATIONNODE_HPP
#define NMM_MAGICKMANIPULATIONNODE_HPP

#include "nmm/NMM.hpp"
#include "nmm/interfaces/video/magick/IMagickManipulationNodeImpl.hpp"
#include "Magick++.h"
#include <list>
#include "Manipulators.hpp"
#include "nmm/base/GenericProcessorNode.hpp"

namespace NMM {
  //! Supports most of the image manipulation methods (sixty) from ImageMagick. The Node can perform multiple manipulations by storing a list of manipulators.
  class MagickManipulationNode : public GenericProcessorNode,
				 public IMagickManipulationNodeImpl
  {
  public:
    //! The main constructor.
    MagickManipulationNode(const char* = "MagickManipulationNode");
    
    //! The destructor.
    virtual ~MagickManipulationNode();
    
    //! The doInit method from Node.
    virtual Result doInit();

    //! The doInitOutput method from Node.
    virtual Result doInitOutput();

    //! The processBuffer method from GenericNode.
    virtual Message* processBuffer(Buffer* buffer);

    //! Clears the internal list of manipulators.
    void clearManipulators();

    //! Sets the color that should be used when applying the 'border' manipulator.
    virtual void setBorderColor(const double& r, const double& g, const double& b);

    // The image manipulation methods.
    virtual void adaptiveThreshold(const unsigned int& width, const unsigned int& height, const unsigned int& offset = 0);
    virtual void addNoise(const string& noise_type);
    virtual void annotate(const string& text, const string& bounding_area, const string& gravity, const double& degrees);
    virtual void blur(const double& radius = 1, const double& sigma = 0.5);
    virtual void border(const string& geometry = "6x6+0+0");
    virtual void channel(const string& layer);
    virtual void charcoal(const double& radius = 1, const double& sigma = 0.5);
    virtual void chop(const string& geometry);
    virtual void colorize(const unsigned int& opacity_red, const unsigned int& opacity_green, const unsigned int& opacity_blue, const double& pen_color_r, const double& pen_color_g, const double& pen_color_b);
    virtual void comment(const string& comment);
    virtual void contrast(const unsigned int& sharpen);
    virtual void crop(const string& geometry);
    virtual void cycleColormap(const int& amount);
    virtual void despeckle();
    virtual void edge(const unsigned int& radius = 0.0);
    virtual void emboss(const double& radius = 1, const double& sigma = 0.5);
    virtual void enhance();
    virtual void equalize();
    virtual void erase();
    virtual void flip();
    virtual void floodFillColor(const unsigned int& x, const unsigned int& y, const double& fill_color_r, const double& fill_color_g, const double& fill_color_b);
    virtual void floodFillOpacity(const long& x, const long& y, const unsigned int& opacity, const string& method);
    virtual void flop();
    virtual void frame(const unsigned int& width, const unsigned int& height, const int& x, const int& y, const int& innerBevel = 0, const int& outerBevel = 0);
    virtual void gamma(const double& gamma_red, const double& gamma_green, const double& gamma_blue);
    virtual void gaussianBlur(const double& width, const double& sigma);
    virtual void implode(const double& factor);
    virtual void label(const string& name);
    virtual void magnify();
    virtual void matteFloodfill(const double& target_r, const double& target_g, const double& target_b, const unsigned int& opacity, const int& x, const int& y, const string& method);
    virtual void medianFilter(const double& radius = 0.0);
    virtual void minify();
    virtual void modulate(const double& brightness, const double& saturation, const double& hue);
    virtual void negate();
    virtual void normalize();
    virtual void oilPaint(const unsigned int& radius = 3);
    virtual void opacity(const unsigned int& opacity);
    virtual void opaque(const double& opaque_color_r, const double& opaque_color_g, const double& opaque_color_b, const double& pen_color_r, const double& pen_color_g, const double& pen_color_b);
    virtual void quantize(const bool& measureError = false);
    virtual void raise(const string& geometry, const bool& raisedFlag = false);
    virtual void reduceNoise(const unsigned int& order);
    virtual void roll(const int& columns, const int& rows);
    virtual void rotate(const double& degrees);
    virtual void sample(const string& geometry);
    virtual void scale(const string& geometry);
    virtual void segment(const double& cluster_threshold = 1.0, const double& smoothing_threshold = 1.5);
    virtual void shade(const double& azimuth = 30, const double& elevation = 30, const bool& color_shading = false);
    virtual void sharpen(const double& radius = 1, const double& sigma = 0.5);
    virtual void shave(const string& geometry);
    virtual void shear(const double& x_shear_angle, const double& y_shear_angle);
    virtual void solarize(const double& factor = 50.0);
    virtual void spread(const unsigned int& amount = 3);
    virtual void swirl(const double& degrees);
    virtual void threshold(const double& threshold);
    virtual void transform(const string& image_geometry, const string& crop_geometry);
    virtual void transparent(const double& color_r, const double& color_g, const double& color_b);
    virtual void trim();
    virtual void unsharpmask(const double& radius, const double& sigma, const double& amount, const double& threshold);
    virtual void wave(const double& amplitude = 25.0, const double& wavelength = 150.0);
    virtual void zoom(const string& geometry);

  private:
    // Internal variables
    pair<int, int> image_resolution;       //!< The image resolution.
    pair<int, int> new_image_resolution;   //!< The new image resolution.
    int bit_per_pixel;                     //!< The number of bits per pixel.
    double border_r;                       //!< The red border color component.
    double border_g;                       //!< The green border color component.
    double border_b;                       //!< The blue border color component.
    list<Manipulator*> manipulators;       //!< The manipulator list.
  };
}

#endif
