/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        
 *                    
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


#include "TVCardReadNode.hpp"
#include "nmm/base/sync/SourceController.hpp"
#include "nmm/base/registry/DevicePlugin.hpp"

#include "nmm/interfaces/video/tv/ITVCard.hpp"

#include "nmm/utils/SecurityCheck.hpp"

#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

static NMM::TDevicePlugin<NMM::TVCardReadNode> plugin("/dev/video", false, "TVCardReadNode");
NMM_REGISTER_SPECIAL_PLUGIN(plugin)

namespace NMM {

  TVCardReadNode::TVCardReadNode(const char *name /* = "TVCardReadNode" */,
				 StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
				 int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSyncSourceNode(name, mode, queue_size), frame_rate_base(1000) {

    /* set default values */
    setSyncMaster(false);

    // this node produces at a constant rate
    setConstantRateProducer(true);

    device = "/dev/video0";
    device_fd = -1;
    width = 720;
    height = 576;
    frame_rate = 25.0;
    int_frame_rate = static_cast<int>(frame_rate * frame_rate_base);

    memset(&_capability,  0, sizeof(video_capability));
    memset(&_vid_buffer,  0, sizeof(video_buffer));
    memset(&_vid_mbuf,    0, sizeof(video_mbuf));
    memset(&_vid_mmap,    0, sizeof(video_mmap));
    memset(&_vid_picture, 0, sizeof(video_picture));
    memset(&_vid_win,     0, sizeof(video_window));
    memset(&_vid_channel, 0, sizeof(video_channel));
    memset(&_vid_tuner,   0, sizeof(video_tuner));

    _frame = 0;
    grab_format = VIDEO_PALETTE_YUV420P;

    _vid_mmap = video_mmap();
    _vid_mmap.frame = _frame;
    _vid_mmap.width = width;
    _vid_mmap.height = height;

    timer = new StreamTimer(this, StreamTimer::REAL_TIME);

  } /* constructor */

  TVCardReadNode::~TVCardReadNode() {
//    close( fd );
    delete timer;
  } /* destructor */


  Result TVCardReadNode::setOutputFormat(int &format) {
    switch (format) {
    case 0: grab_format = VIDEO_PALETTE_YUV420P; break;
    case 1: grab_format = VIDEO_PALETTE_RGB32; break;
    default : grab_format = VIDEO_PALETTE_YUV420P;
    }

    return SUCCESS;
  } /* setOutputFormat */
  

  /* Event functions */
  void TVCardReadNode::setDevice(const string& device) throw(const DeviceNotFoundException&) {
    int fd = -1;
    try {
      /* to test if the device can be accessed, we try to open the device */
      if(!SSecurityCheck::getInstance().checkFilenameForWriting(device)) {
        DEBUG_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << endl);
        throw DeviceNotFoundException(this, "Security check failed for device", device);
      }
      if((fd = ::open(device.c_str(), O_RDONLY | O_NONBLOCK)) < 0) {
        /* if the device can not be accessed, then we throw an exception */
        DEBUG_STREAM("Open video4linux device " << device << " failed" << endl);
        fd = -1;
        throw DeviceNotFoundException(this, "Device not available or can not be accessed", device);
      } else {
        /* close the device again */
        ::close(fd);
        fd = -1;
      }
    } catch(...) {
      /* close the device again */
      if(fd >= 0) {
        ::close(fd);
        fd = -1;
      }
      
      /* rethrow the exception */
      throw;
    }
    
    /* now we can set the device */
    this->device = device;
  } /* setDevice */

  string TVCardReadNode::getDevice() const {
    return device;
  } /* getDevice */

  void TVCardReadNode::setResolution(const int& x, const int& y) {
    width = x;
    height = y;
  } /* getSize */



  void TVCardReadNode::getResolution(int& x, int& y) {
    x =  _vid_win.width;
    y =  _vid_win.height;
  } /* getSize */



  Result TVCardReadNode::setChannel(const int &source) {
    MutexGuard mg_v4l(mutex_v4l);

    DEBUG_STREAM("Setting video channel to " << source << endl);

    _vid_channel.channel = source;
    if(-1 == ioctl(device_fd, VIDIOCSCHAN, &_vid_channel)) {
      ERROR_STREAM("ERROR : Can\'t ioctl set video channel to " << _vid_channel.channel << endl);
      return FAILURE;
    }
    return SUCCESS;
  } /* TVCardReadNode::setChannel */



  Result TVCardReadNode::getChannel(int &source, string &name) {
    if(-1 == ioctl(device_fd, VIDIOCGCHAN, &_vid_channel)) {
      ERROR_STREAM("Can\'t get video channel");
      return FAILURE;
    }
    source = _vid_channel.channel;
    name   = string(_vid_channel.name);

    return SUCCESS;
  }


  /**
   * NMM  functions 
   */
  Result TVCardReadNode::doInit() {
    // set the FrameTime at the Controller
    controller->setFrameTime(Interval(Rational(1, frame_rate)));

    Format* o_format;

    /* setting up supported formats */
    o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addFloatParamValue(Format::framerate_param, frame_rate);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param, 12);
    o_format->addStringParamValue(Format::format_param, "planar");
  
    getOwnOutputProperty()->setDefaultFormat(o_format);

    o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addFloatParamValue(Format::framerate_param, frame_rate);
    o_format->addStringParamValue(Format::channelorder_param,"rgba");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"rgba32");
    o_format->addIntParamValue(Format::bitperpixel_param, 32);
    o_format->addStringParamValue(Format::format_param, "planar");
  

    string nmm_v4l_conf = "v4l-conf > /dev/null 2>&1";
#ifdef __linux__
    switch (system(nmm_v4l_conf.c_str())) {
    case -1: /* can't run */
      ERROR_STREAM("could'nt start v4l-conf - make sure you have v4l-conf installed!" << endl);
      break;
    case 0: /* ok */
      DEBUG_STREAM("v4l-conf ok." << endl);
      break;
    default: /* non-zero return */
      ERROR_STREAM("v4l-conf had some trouble, trying to continue anyway" << endl);
      break;
    }
#endif

    /* try to open the device */
    if (-1 != device_fd) {
      DEBUG_STREAM("closing device_fd before opening it..." << endl);
      close(device_fd);
    }
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(device)) {
      DEBUG_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    device_fd = open(device.c_str(), O_RDWR);
    if(device_fd < 0) {
      ERROR_STREAM("Error: Can\'t open " << device << " for reading." << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Open device " << device << endl);


    /**
     * get caps 
     */
    if(-1 == ioctl(device_fd, VIDIOCGCAP, &_capability)) {
      ERROR_STREAM("Error: Can\'t get video capability");
      close (device_fd);
      return FAILURE;
    }
    DEBUG_STREAM("  Capability.name     : " << _capability.name << endl);
    DEBUG_STREAM("  Capability.width    : " << _capability.minwidth  << " <= x <= " << _capability.maxwidth  << endl);
    DEBUG_STREAM("  Capability.height   : " << _capability.minheight << " <= y <= " << _capability.maxheight << endl);
    DEBUG_STREAM("  Capability.channels : " << _capability.channels << endl);


    /** 
     * set channel aka input
     */
    // to set a channel, you first have to do a get-channel ioctl:
    // have to make sure that _vid_channel is lower than _capability.channels here
    // (strange, but see bttv-driver.c in linux-2.4.19)
    if(-1 == ioctl(device_fd, VIDIOCGCHAN, &_vid_channel)) {
      ERROR_STREAM("Can\'t get video channel");
      // you could not set a channel, so return FAILURE
      return FAILURE;
    }
    setChannel(0); 

    if (_capability.type & VID_TYPE_TUNER) {
      if(-1 == ioctl(device_fd, VIDIOCGTUNER, &_vid_tuner)) {
	ERROR_STREAM("Error: Can\'t get video capability");
	close (device_fd);
	return FAILURE;
      }
    }

    return SUCCESS;
  }

  Result TVCardReadNode::doDeinit() {
    DEBUG_STREAM("doDeinit..." << endl);
    if (device_fd != -1) {
      DEBUG_STREAM("Closing device_fd" << endl);
      close (device_fd);
      device_fd = -1;
    }
    return SUCCESS;
  }

  Result TVCardReadNode::doInitOutput() {
    Format* o_format;

    if (grab_format == VIDEO_PALETTE_YUV420P) {
      o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      o_format->addIntParamValue(Format::x_resolution_param, width);
      o_format->addIntParamValue(Format::y_resolution_param, height);
      o_format->addIntParamValue(Format::ratio_x_param, width);
      o_format->addIntParamValue(Format::ratio_y_param, height);
      o_format->addFloatParamValue(Format::framerate_param, frame_rate);
      o_format->addStringParamValue(Format::channelorder_param,"yvu");
      o_format->addStringParamValue(Format::endian_param, "little");
      o_format->addStringParamValue(Format::colorspace_param,"yv12");
      o_format->addIntParamValue(Format::bitperpixel_param, 12);
      o_format->addStringParamValue(Format::format_param, "planar");
      
      getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    } else {
      o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      o_format->addIntParamValue(Format::x_resolution_param,width);
      o_format->addIntParamValue(Format::y_resolution_param,height);
      o_format->addFloatParamValue(Format::framerate_param, frame_rate);
      o_format->addStringParamValue(Format::channelorder_param,"rgba");
      o_format->addStringParamValue(Format::endian_param, "little");
      o_format->addStringParamValue(Format::colorspace_param,"rgba32");
      o_format->addIntParamValue(Format::bitperpixel_param, 32);
      o_format->addStringParamValue(Format::format_param, "planar");
      
      getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    }
    return SUCCESS;
  }
      


  Result TVCardReadNode::doActivate() {
    int    zero   = 0;
    string colorspace = "";

    const Format* o_format = getOutputFormat();
    if (NULL == o_format) {
      ERROR_STREAM("out_format is NULL!!" << endl);
      return FAILURE;
    }

    /**
     * clamp width and height 
     */
    width = (width > _capability.maxwidth ? _capability.maxwidth :
             (width < _capability.minwidth ? _capability.minwidth : width));
    height = (height > _capability.maxheight ? _capability.maxheight :
              (height < _capability.minheight ? _capability.minheight : height));
    DEBUG_STREAM("Try to grab with resolution " << width << "x" << height << endl);


    /**
     * get/set picture properties 
     */
    if(-1 == ioctl(device_fd, VIDIOCGPICT, &_vid_picture)) {
      ERROR_STREAM("Error: Can\'t get picture properties" << endl);
      close (device_fd);
      return FAILURE;
    }
  
    /**
     * choose colorspace for grabbing
     */
    colorspace = o_format->getStringValue(Format::colorspace_param);
    if ("yv12" == colorspace) {
      grab_format = VIDEO_PALETTE_YUV420P;
    } else if ("rgba32" == colorspace) {
      grab_format = VIDEO_PALETTE_RGB32;
    }
    _vid_picture.palette = grab_format;
    if(-1 == ioctl(device_fd, VIDIOCSPICT, &_vid_picture)) { 
      ERROR_STREAM("ERROR: Can'\t set picture properties!" << endl);
      close (device_fd);
      return FAILURE;
    }


    /*
     * get video_mbuf for mmapping
     */
    if(-1 == ioctl(device_fd, VIDIOCGMBUF, &_vid_mbuf)) {
      ERROR_STREAM("ERROR: Can'\t get mbuf!" << endl);
      close (device_fd);
      return FAILURE;
    }
    _mmap_buffer = (unsigned char*) mmap(0, _vid_mbuf.size,
                                         PROT_READ | PROT_WRITE,
                                         MAP_SHARED, device_fd, 0);
    if(_mmap_buffer < 0) {
      ERROR_STREAM("ERROR: Can\'t get mmap!" << endl);
      close (device_fd);
      return FAILURE;
    }
      
    if (-1 == ioctl(device_fd, VIDIOCGWIN, &_vid_win)) {
      ERROR_STREAM("ERROR: Can\'t get overlay window!" << endl);
      close (device_fd);
      return FAILURE;
    }

    _vid_win.width = width;
    _vid_win.height = height;

    if(-1 == ioctl(device_fd, VIDIOCSWIN, &_vid_win)) {
      ERROR_STREAM("ERROR: VIDIOCSWIN failed!" << endl);
      //        close (device_fd);
      //        return FAILURE;
    }

    DEBUG_STREAM("_vid.win : " 
                 << _vid_win.x << " " << _vid_win.y << " " 
                 << _vid_win.width << " " << _vid_win.height << endl);

    if (-1 == ioctl(device_fd, VIDIOCGWIN, &_vid_win)) {
      ERROR_STREAM("ERROR: VIDIOCGWIN (2) failed!" << endl);
      close (device_fd);
      return FAILURE;
    } else {
      DEBUG_STREAM("_vid.win : " 
                   << _vid_win.x << " " << _vid_win.y << " " 
                   << _vid_win.width << " " << _vid_win.height << endl);
      DEBUG_STREAM("Got resolution " << _vid_win.width << "x" << _vid_win.height << endl);
    }
 

    /**
     * pre-capture
     */
    if(-1 == ioctl(device_fd, VIDIOCCAPTURE, &zero)) {
      ERROR_STREAM("ERROR: Can't set capturing!" << endl);
      close (device_fd);
      return FAILURE;
    }
    
    /**
     * init mmapped frames
     */
    _vid_mmap.width  = width;
    _vid_mmap.height = height;
    _vid_mmap.format = _vid_picture.palette;
    
    _vid_win.width = width;
    _vid_win.height = height;

    picture_size = _vid_win.width * _vid_win.height; 
    offset1   = picture_size;
    offset2   = int(offset1 * 1.25);
    uv_size   = offset1 / 4;

    switch (grab_format) {
    case VIDEO_PALETTE_YUV420P:
      framesize = int(offset1 * 1.5);
      break;
    case VIDEO_PALETTE_RGB32: 
      framesize = offset1 * 4;
      break;
    }
    
    // reset the stream timer
    timer->reset();

    return SUCCESS;
  }

  Time TVCardReadNode::getLocalTime(u_int64_t valid_read_bytes) {
    // calcluate Rational exactely as
    // valid_read_bytes * frame_rate_base / int_frame_rate; framesize
    // to avoid overflow (framesize * int_frame_rate would create one)
    return Time(Rational(valid_read_bytes * frame_rate_base / int_frame_rate, framesize));
  }  
  
  Message* TVCardReadNode::produceBuffer(Buffer *) {
    Buffer *out;
    char* target;
    char* source;

    DEBUG_STREAM("Grab with framesize " << framesize << endl);

    out = getNewBuffer(framesize);

    if ( out ) {
      MutexGuard mg_v4l(mutex_v4l);

      DEBUG_STREAM("Got buffer, try to grab " << framesize << endl);

      /* Capture */
      DEBUG_STREAM("Capture frame " << _vid_mmap.frame << " (" << width << "x" << height << ")" << endl);
      if (0 > ioctl(device_fd, VIDIOCMCAPTURE, &_vid_mmap)) {
        ERROR_STREAM("ERRROR: VIDIOCMCAPTURE frame " << _vid_mmap.frame << endl);
        close (device_fd);
        out->release();
        return NULL;
      }
      // switch capture buffer
      _vid_mmap.frame = 1-_frame;
      _frame = 1-_frame;

      /* Sync */
      DEBUG_STREAM("Sync frame" << _frame << endl);
      ioctl(device_fd, VIDIOCSYNC, &_frame);

      /* get new frame */
      source = (char*) _mmap_buffer + _vid_mbuf.offsets[_frame];
      target = out->getData();

      switch (grab_format) {
      case VIDEO_PALETTE_YUV420P:
        DEBUG_STREAM("Copy YUV" << endl);    
        /* in order : copy Y part then V and U values */
        memcpy(target, source, offset1);
        memcpy(target + offset1, source + offset2, uv_size );
        memcpy(target + offset2, source + offset1, uv_size );
        out->setUsed(framesize);        
        break;
      case VIDEO_PALETTE_RGB32: 
        DEBUG_STREAM("Copy RGB" << endl);    
        /* all in right order, only one copy step needed */
        memcpy(out->getData(), source, framesize);
        out->setUsed(framesize);
        break;
      }

      // Set the correct time stamp for the buffer
      timer->setTimestamp(out);

      return out;
    } 

    return NULL;
  } /* produceBuffer */

} /* namespace */
