/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:         
 *                    Michael Repplinger <repplinger@cs.uni-sb.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#ifndef NMM_SOCKACCEPTOR_HPP
#define NMM_SOCKACCEPTOR_HPP NMM_SOCKACCEPTOR_HPP

#include "NetExceptions.hpp"
#include "Sock.hpp"
#include "INetAddr.hpp"

namespace NMM {

  /** This wrapper class can be used to accept and control a network connection.
   * With tcp, this is a wrapper for the socket(2), bind(2), listen(2) and accept(2) functions.
   * With udp, a dummy implementation is provided for binding a socket and finding a free port.
   * No emulation of tcp functonality is done for udp. A udp SockAcceptor can only accept one connection,
   * then it must be opened again. Also the accept method trivially succeeds for udp and does nothing
   * but returning the udp socket.
   */
  class SockAcceptor {
  public:
    /** @name Constructor & Destructor
      */
    //@{
    /** Default constructor
      */
    SockAcceptor();

    /** Destructor
      */
    ~SockAcceptor();
    //@ }


    /** @name Accepting Methods
      */
    //@{
    /** Init a new port and calls bind and listen on it.
     * If the transport protocol is "udp", this method simple creates a udp socket and binds it to the given address.
     * @param addr Local address, address part should be INADDR_ANY, port should be the port to listen on
     * @param proto Transport protocol, must be either "tcp" or "udp"
     * @exception ConnectionException if creating, binding or listening on the socket fails
     */
    void open(const INetAddr& addr, const char* proto = "tcp")
      throw (const ConnectionException&);
    
    /** This method closes the SockAcceptor
     */
    void close()
      throw (const ConnectionException&);
    
    /** Scans for free port between [from,to], bind and listen on it.
     *
     * The used port is stored in the given INetAddr.
     * The INetAddr address value should be INADDR_ANY.
     *
     * @param addr Reference to the addr. The hostname must be the address of the localhost.
     *        The used port is stored in addr
     * @param from Begin of the port range
     * @param to End of the port range
     * @param proto The used protocol, must be either "tcp" or "udp"
     * @exception ConnectionExcepion if listening failed on all ports in the given port range
     */
    void tryOpen(INetAddr& addr, u_int32_t from, u_int32_t to, const char* proto = "tcp")
      throw (const ConnectionException&);
    
    /** This method accepts an incoming method.
     * If the transport protocol is UDP, this method simply stores the internal socket in the Sock object and
     * the SockAcceptor must be opened again to accept a new connection.
     * If the transport protocol is TCP, this method also calls setPeerAddress on the Sock object, based on
     * the address returned by the accept(2) call. With udp, the peer address must be set manually or using a
     * SockConnector.
     * @param sock Stores the new socket in the given sock
     */
    void accept(Sock& sock)
      throw (const ConnectionException&);
    //@}
    
    /** @name Status Methods
     */
    //@{
    /** This method returns the internal socket.
     * @return The current socket which is used to accept connections.
     */
    int getSocket();
    
    /** This method tests if connections are waiting. You should use the NetReactor to be informed for incoming connections
     * @return True if connections are waiting, otherwise false
     */
    bool isAcceptable();
    
    /** This method disables blocking for accepting a connection.
     */
    void setNonBlocking();
    //@}
    
  private: //methods
    void createBindException(int);
    void createListenException(int);
    void createSocketException(int);
    void createShutdownException(int);
    void createAcceptException(int);

  private: //member
    int __socket;          ///<The internal socket for accepting incoming connections
    string __trans_prot;   ///<The Trnaport protocol which can be "TCP" or "UDP"
  };

}

#endif
