/* NMM - Network-Integrated Multimedia Middleware
 *
 * Copyright (C) 2002-2005
 *                    NMM work group,
 *                    Computer Graphics Lab,
 *                    Saarland University, Germany
 *                    http://www.networkmultimedia.org
 *
 * Maintainer:        Michael Repplinger <repplinger@cs.uni-sb.de>
 *                     
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */


/*
 * xml++.hpp based on xml++.h from libxml++.
 * libxml++ is copyright (C) 2000 by Ari Johnson, and
 * are covered by the GNU Lesser General Public License.
 */

#include <string>
#include <list>

#include "nmm/nmm_hash_map.hpp"

#include <libxml/parser.h>
#include <libxml/encoding.h>

#ifndef __XML_H
#define __XML_H

using namespace std;
using namespace HASH_MAP_STL;

namespace HASH_MAP_STL{
  template<> struct hash<string> {
    size_t operator()(const string &s) const {
      return __stl_hash_string(s.c_str());
    };
  };
}

class XMLTree;
class XMLNode;
typedef list<XMLNode *> XMLNodeList;
typedef XMLNodeList::iterator XMLNodeIterator;
typedef XMLNodeList::const_iterator XMLNodeConstIterator;
class XMLProperty;
typedef list<XMLProperty *> XMLPropertyList;
typedef XMLPropertyList::iterator XMLPropertyIterator;
typedef XMLPropertyList::const_iterator XMLPropertyConstIterator;
typedef hash_map<string, XMLProperty *> XMLPropertyHash;

class XMLParserCallback {
public:
  virtual void start_document(void) { };
  virtual void end_document(void) { };
  virtual void start_element(const string &, const XMLPropertyHash &) { };
  virtual void end_element(const string &) { };
  virtual void characters(const string &) { };
  virtual void comment(const string &) { };
  virtual void warning(const string &) { };
  virtual void error(const string &) { };
  virtual void fatal_error(const string &) { };
};

template <class ParserCallback = XMLParserCallback>
class XMLParser {
private:
  XMLParserCallback *_parser_callback;
  xmlParserCtxtPtr _context;

  static xmlEntityPtr _get_entity(void *_parser, const xmlChar *n) {
    return xmlGetPredefinedEntity(n);
  };

  static void _start_document(void *_parser) {
    XMLParser *parser;

    parser = (XMLParser *) _parser;
    parser->start_document();
  };

  static void _end_document(void *_parser) {
    XMLParser *parser;

    parser = (XMLParser *) _parser;
    parser->end_document();
  };

  static void _start_element(void *_parser, const xmlChar *n,
        const xmlChar **p) {
    XMLParser *parser;
    XMLPropertyHash properties;

    if(p) {
      const xmlChar **cur;

      for(cur = p; cur && *cur; cur++) {
        string name, value;

        name = string((const char *) *cur++);
        value = string((const char *) *cur);

        properties[name] = new XMLProperty(name, value);
      }
    }

    parser = (XMLParser *) _parser;
    parser->start_element(string((const char *) n), properties);
  };

  static void _end_element(void *_parser, const xmlChar *n) {
    XMLParser *parser;

    parser = (XMLParser *) _parser;
    parser->end_element(string((const char *) n));
  };

  static void _characters(void *_parser, const xmlChar *s, int len) {
    XMLParser *parser;

    parser = (XMLParser *) _parser;
    parser->characters(string((const char *) s, len));
  };

  static void _comment(void *_parser, const xmlChar *s) {
    XMLParser *parser;

    parser = (XMLParser *) _parser;
    parser->comment(string((const char *) s));
  };

  static void _warning(void *_parser, const char *fmt, ...) {
    XMLParser *parser;

#if (__GNUC__>=3)
    __builtin_va_list arg;
#else
    va_list arg;
#endif
    
    char buff[1024];

    va_start(arg, fmt);
    vsprintf(buff, fmt, arg);
    va_end(arg);

    parser = (XMLParser *) _parser;
    parser->warning(string(buff));
  };

  static void _error(void *_parser, const char *fmt, ...) {
    XMLParser *parser;

#if (__GNUC__>=3)
    __builtin_va_list arg;
#else
    va_list arg;
#endif

    char buff[1024];

    va_start(arg, fmt);
    vsprintf(buff, fmt, arg);
    va_end(arg);

    parser = (XMLParser *) _parser;
    parser->error(string(buff));
  };

  static void _fatal_error(void *_parser, const char *fmt, ...) {
    XMLParser *parser;

#if (__GNUC__>=3)
    __builtin_va_list arg;
#else
    va_list arg;
#endif

    char buff[1024];

    va_start(arg, fmt);
    vsprintf(buff, fmt, arg);
    va_end(arg);

    parser = (XMLParser *) _parser;
    parser->fatal_error(string(buff));
  };

  void start_document(void) {
    if(_parser_callback)
      _parser_callback->start_document();
  };

  void end_document(void) {
    if(_parser_callback)
      _parser_callback->end_document();
  };

  void start_element(const string &n, const XMLPropertyHash &p) {
    if(_parser_callback)
      _parser_callback->start_element(n, p);
  };

  void end_element(const string &n) {
    if(_parser_callback)
      _parser_callback->end_element(n);
  };

  void characters(const string &s) {
    if(_parser_callback)
      _parser_callback->characters(s);
  };

  void comment(const string &s) {
    if(_parser_callback)
      _parser_callback->comment(s);
  };

  void warning(const string &s) {
    if(_parser_callback)
      _parser_callback->warning(s);
  };

  void error(const string &s) {
    if(_parser_callback)
      _parser_callback->error(s);
  };

  void fatal_error(const string &s) {
    if(_parser_callback)
      _parser_callback->fatal_error(s);
  };
public:
  XMLParser() {
    xmlSAXHandler sax_handler = {
      NULL,    // internalSubset
      NULL,    // isStandalone
      NULL,    // hasInternalSubset
      NULL,    // hasExternalSubset
      NULL,    // resolveEntity
      _get_entity,  // getEntity
      NULL,    // entityDecl
      NULL,    // notationDecl
      NULL,    // attributeDecl
      NULL,    // elementDecl
      NULL,    // unparsedEntityDecl
      NULL,    // setDocumentLocator
      _start_document,  // startDocument
      _end_document,  // endDocument
      _start_element,  // startElement
      _end_element,  // endElement
      NULL,    // reference
      _characters,  // characters
      NULL,    // ignorableWhitespace
      NULL,    // processingInstruction
      _comment,    // comment
      _warning,    // warning
      _error,    // error
      _fatal_error,  // fatalError
      NULL,    // getParameterEntity
      NULL,    // cdataBlock
      NULL    // externalSubset
    };

    _parser_callback = new ParserCallback;
    _context = xmlCreatePushParserCtxt(&sax_handler, this, NULL, 0, NULL);
  };

  ~XMLParser() {
    if(_parser_callback)
      delete _parser_callback;

    if(_context)
      xmlFreeParserCtxt(_context);
  };

  void parse_chunk(const string &s) {
    xmlParseChunk(_context, s.c_str(), s.length(), 0);
  };

  void finish(void) {
    xmlParseChunk(_context, NULL, 0, 1);
  };
};

class XMLTree {
public:
  XMLTree() : _initialized(true), _filename(), _root(NULL), _compression(0) { };
  XMLTree(const string &fn)
  : _initialized(true), _filename(fn), _root(NULL), _compression(0) { read(); };

/** Copy constructor
  */
  XMLTree(const XMLTree& );

/** Destructor
  */         
  ~XMLTree();

/**
  */
  inline bool     initialized(void)   const   { return _initialized; };

/** returns the current root node
  */
  inline XMLNode* root(void)          const   { return _root; };

/** sets a new root node. The old one is deleted
  */
  XMLNode *set_root(XMLNode *n);

/** This resets the root node to 0 and returns a  pointer to the old. This pointer is not deleted.
  * The intend of this method is to integrate patial node without copy.
  */
  inline  XMLNode *release_root() { XMLNode* tmp = root(); _root = 0; return tmp;}

/** returns the current filename
  */
  inline const string & filename(void)const   { return _filename; };

/** sets a new filename
  */
  inline const string & set_filename(const string &fn) { return _filename = fn; };

/** returns current compression
  */
  inline int compression(void)        const { return _compression; };

/** sets the compression rate (value betwee 0-9)
  */
  int set_compression(int);

/** reads an XML tree from current file. the old root node is deleted so you must request
  * the root node again if it is stored outside the class
  */
  bool read(void);

/**
  */
  inline bool read(const string &fn) { set_filename(fn); return read(); };
  bool read_buffer(const string &);
  bool write(void) const;
  inline bool write(const string &fn) { set_filename(fn); return write(); };
  string write_buffer(void) const;


  const XMLTree& operator=(const XMLTree&);
  bool operator==(const XMLTree&);
//  bool isSubset(const XMLTree&);

  XMLNodeList children(const string& name) const;
private: //attributes
  bool _initialized;
  string _filename;
  XMLNode *_root;
  int _compression;

};

class XMLNode {
private:
  bool _initialized;
  string _name;
  string _content;
  XMLNodeList _children;
  XMLPropertyList _proplist;
  XMLPropertyHash _prophash;
public:
/** constructor.
  */ 
  XMLNode(const string &);
  XMLNode(const string &, const string &);

/**
  */
  XMLNode(const XMLNode&);

  ~XMLNode();



  inline bool  initialized(void)      const { return _initialized; };
  inline bool  is_content(void)       const { return !_content.empty(); };

  inline const string & name(void)    const { return _name; };
  inline const string & content(void) const { return _content; };

  const string & set_content(const string &);
  string getData();

  XMLNodeList children(const string & ) const;

  inline const XMLNodeList& children() const { return _children; }
  inline       XMLNodeList& children()       { return _children; }
  XMLNode *add_child(const string &);
  XMLNode *add_child(XMLNode *);
  XMLNode *add_child(const string &, int pos);
  XMLNode *add_content(const string & = string());
  void remove_child(XMLNode *);

  inline const XMLPropertyList & properties(void) const { return _proplist; };
  XMLProperty *property(const string &);
  const XMLProperty *property(const string &n) const
  { return ((XMLNode *) this)->property(n); };
  XMLProperty *add_property(const string &, const string & = string());
  void remove_property(const string &);

  const XMLNode& operator=(const XMLNode& );
  bool operator==(const XMLNode& );
private: //methods
  void erase();
};

class XMLProperty {
private:
  string _name;
  string _value;
public:
  XMLProperty(const string &n, const string &v = string())
  : _name(n), _value(v) { };
  const string & name(void) const           { return _name; };
  const string & value(void) const          { return _value; };
  const string & set_value(const string &v) { return _value = v; };
        string & value(void)                { return _value; };

};

/**
 * class to encode Strings to and from UTF-8
 */
class XMLEncoder {
private:
  string _encoding;
public:
  /**
   * Constructor
   *@param encoding-standard as string
   */
  XMLEncoder(const string & e)
    : _encoding(e){};
  /**
   * encodes the given string to UTF-8. The input-string must have a encoding as specified in the Contructor
   */
  string getUTF8(const string &s);
  /**
   * encodes a string to the standard specified in the constructor. The input-string must be in UTF-8
   */
  string getEncodedString(const string &s);
};

#endif /* __XML_H */

